/*
 * Copyright (c) 2011 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef __TIZEN_SYSTEM_DLOG_DOC_H__
#define __TIZEN_SYSTEM_DLOG_DOC_H__

/**
 * @defgroup CAPI_SYSTEM_DLOG Dlog
 * @brief The Dlog API provides functions for sending log output.
 * @ingroup CAPI_SYSTEM_FRAMEWORK
 * @section CAPI_SYSTEM_DLOG_HEADER Required Header
 *   \#include <dlog.h>
 *
 * @section CAPI_SYSTEM_DLOG_OVERVIEW Overview
 *
Sending log message to circular buffer. Dlog APIs include Priority and Tag. By using priority and Tag, we can easily filtered messages what we want to see.
<h2 class="pg">priority</h2>
priority level indicates the urgency of log message

<table>
<tr>
	<th>Priority</th>
	<th>Description</th>
</tr>
<tr>
	<td>DLOG_DEBUG</td>
	<td>Debug message. - Log message which developer want to check.</td>
</tr>
<tr>
	<td>DLOG_INFO</td>
	<td>Information message - Normal operational messages. above of this priority will always be logged.</td>
</tr>
<tr>
	<td>DLOG_WARN</td>
	<td>Warning messages - Not an error, but indication that an error will occur if action is not taken.</td>
</tr>
<tr>
	<td>DLOG_ERROR</td>
	<td>Error message - Indicate error.</td>
</tr>
</table>

<h2 class="pg">Macro example for useful usage</h2>
The Dlog APIs can be used by defining own macros. The macros can be defined like below examples.
Thus, developers can use appropriate method as a matter of convenience.

@code
#undef LOG_TAG
#define LOG_TAG "APP_TAG"

#define LOGI(fmt, arg...) \
	({ do { \
		dlog_print(DLOG_INFO, LOG_TAG, "%s: %s(%d) > " fmt, __MODULE__, __func__, __LINE__, ##arg); \
	} while (0); })
#define LOGW(fmt, arg...) \
	({ do { \
		dlog_print(DLOG_WARN, LOG_TAG, "%s: %s(%d) > " fmt, __MODULE__, __func__, __LINE__, ##arg); \
	} while (0); })
#define LOGE(fmt, arg...) \
	({ do { \
		dlog_print(DLOG_ERROR, LOG_TAG, "%s: %s(%d) > " fmt, __MODULE__, __func__, __LINE__, ##arg); \
	} while (0); })
@endcode


@section CAPI_SYSTEM_DLOG_UTIL dlogutil CLI

dlogutil is a utility that can be utilized to read stored logs.

The most rudimentary way to use it is to run it with no parameters; it then prints all stored logs to the standard output and awaits more.

Parameters exist to change behaviour:

<table>
<tr><td>PARAM          </td><td>DESCRIPTION</td></tr>
<tr><td>-s             </td><td>Set default filter to *:s (see later for details)</td></tr>
<tr><td>-f filename    </td><td>Output to given file instead of stdout</td></tr>
<tr><td>-r kbytes      </td><td>Rotate log file every this many kilobytes.</td></tr>
<tr><td>-n files       </td><td>Keep this many old rotations</td></tr>
<tr><td>-v             </td><td>Set log output format.</td></tr>
<tr><td>-b buffer      </td><td>Select another buffer</td></tr>
<tr><td>-c             </td><td>Clear the selected buffers</td></tr>
<tr><td>-d             </td><td>Exit when no more logs are currently available</td></tr>
<tr><td>-t count       </td><td>Similar to -d, except only print this many logs from the end</td></tr>
<tr><td>-g             </td><td>Get the size of selected buffers instead of printing logs</td></tr>
<tr><td>-h, --help     </td><td>Show help info</td></tr>
</table>

Additionally, you can specify filters. Filters go behind the parameters and are used to decide which logs are shown.
<br/>
Each log line has priority and a tag. Filters are specified in the format TAG[:PRIO]. Only logs with given tags and priorities are shown.
<br/>
You can specify * as the tag to match anything. Beware: only a single asterisk works that way, asterisk is not otherwise expanded. For example, "T*G" filter will NOT match "TAG", only literal "T*G".
<br/>
If you don't specify any filter, it defaults to *:I<br/>
If you specify just * as the tag, without priority, its priority defaults to D<br/>
If you specify any other tag without priority, the priority defaults to V<br/>

dlogutil allows you to specify buffers. These are:
<table>
<tr><td>NAME</td><td>REMARKS</td></tr>
<tr><td>main</td><td>the default, for general platform logging</td></tr>
<tr><td>system</td><td>for system logs</td></tr>
<tr><td>radio</td><td>for radio logs</td></tr>
<tr><td>apps</td><td>for applications</td></tr>
<tr><td>kmsg</td><td>a copy of kernel messages (dmesg)</td></tr>
</table>

 There are also priorities to be specified. These are one letter abbreviations of the following levels:

<table>
<tr><td>LEVEL</td><td>DESCRIPTION</td></tr>
<tr><td>V    </td><td>verbose</td></tr>
<tr><td>D    </td><td>debug</td></tr>
<tr><td>I    </td><td>info</td></tr>
<tr><td>W    </td><td>warning</td></tr>
<tr><td>E    </td><td>error</td></tr>
<tr><td>F    </td><td>fatal</td></tr>
<tr><td>S    </td><td>silent</td></tr>
</table>

 Formats, and what they look like:

<table>
<tr><td>NAME</td>      <td>SPECIFICATION</td>                               <td>EXAMPLE</td></tr>
<tr><td>brief</td>     <td>PRI/TAG(PID): MSG</td>                           <td>E/TEST_APP(123): FOO</td></tr>
<tr><td>process</td>   <td>PRI(PID) MSG (TAG)</td>                          <td>E(123) FOO (TEST_APP)</td></tr>
<tr><td>tag</td>       <td>PRI/TAG: MSG</td>                                <td>E/TEST_APP: FOO</td></tr>
<tr><td>thread</td>    <td>PRI('P'PID, 'T'TID) MSG</td>                     <td>E(P123, T456) FOO</td></tr>
<tr><td>raw</td>        <td>MSG</td>                                         <td>FOO</td></tr>
<tr><td>time</td>      <td>REALTIME PRI/TAG(PID): MSG</td>                  <td>12-31 23:59:59.999 +0200 E/TEST_APP(123): FOO</td></tr>
<tr><td>threadtime</td><td>REALTIME PRI/TAG('P'PID, 'T'TID): MSG</td>       <td>12-31 23:59:59.999 +0200 E/TEST_APP(P123, T456): FOO</td></tr>
<tr><td>kerneltime</td><td>BOOTTIME PRI/TAG('P'PID, 'T'TID): MSG</td>       <td>0.123 E/TEST_APP(P123, T456): FOO</td></tr>
<tr><td>long</td>      <td>[REALTIME PRI/TAG 'P'PID, 'T'TID]<br>MSG<br></td><td>[12-31 23:59:59.999 +0200 E/TEST_APP P123, T456]<br>FOO<br></td></tr>
</table>

*/

#endif /* __TIZEN_SYSTEM_DLOG_DOC_H__ */
