#pragma once

/*
 * Copyright (c) 2018, Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdbool.h>
#include <stdint.h>
#include <logconfig.h>
#include <log_file.h>
#include <logprint.h>
#include <ptrs_list.h>
#include <queued_entry.h>
#include <sort_vector.h>

struct fd_ops;

struct fd_info {
	int fd;
	struct fd_ops *ops;
	const char *name;
	void *priv_data;
};


struct fd_ops {
	/// Lifetime management: constructor and destructor
	int (*create)(struct fd_info *fdi, const struct log_config *conf, const char *name, list_head *used_paths);
	void (*destroy)(struct fd_info *fdi);

	/// Read an entry into an internal buffer; requires `prepare_print` to have been called
	int (*read)(struct fd_info *fdi);

	/// Checks whether a log has been buffered and is ready for extraction
	bool (*has_log)(struct fd_info *fdi);

	/// Extracts a buffered entry and relinquishes its ownership
	struct logger_entry *(*extract_entry)(struct fd_info *fdi);

	/// Returns a buffered entry without relinquishing ownership
	const struct logger_entry *(*peek_entry)(const struct fd_info *fdi);

	/// Transition into a state that allows reading and printing out logs
	/// Return values > 0 mean everything went fine but we don't want to print
	int (*prepare_print)(struct fd_info *fdi, int dump, list_head filter_list, log_filter *filter_object);

	/// Clear the buffer
	int (*clear)(struct fd_info *fdi);

	/// Get the size of the buffer
	int (*getsize)(struct fd_info *fdi, uint32_t *val);
};

struct fd_info *fdi_create(struct fd_ops *ops, const char *name);
void fdi_free(struct fd_info *fdi);
void fdi_array_free(struct fd_info ***arr);
bool fdi_has_log(struct fd_info *fdi);
int fdi_push_log(struct fd_info *fdi, struct sort_vector *logs, write_callback callback, void *userdata, log_filter *filter);
struct logger_entry *fdi_extract_entry(struct fd_info *fdi);
int fdi_read(struct fd_info *fdi);
