/*
 * DLOG
 * Copyright (c) 2005-2008, The Android Open Source Project
 * Copyright (c) 2012-2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _LOGCOMMON_H
#define _LOGCOMMON_H

#define LOG_MAX_PAYLOAD_SIZE 4076 // has to be at least the value in the kernel: drivers/staging/android/logger.h

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/syscall.h>
#include <dlog.h>
#include <fcntl.h>
#include <stdbool.h>

#define __STRINGIFY(x) #x
#define STRINGIFY(x) __STRINGIFY(x)

#ifdef DEBUG_ON
#define DBG(...) printf(__VA_ARGS__)
#else
#define DBG(...) do { } while (0)
#endif
#define ERR(...) fprintf(stderr, __VA_ARGS__)

#define gettid() syscall(SYS_gettid)

#define NELEMS(arr) (sizeof(arr)/sizeof(arr[0]))

/** these values are mostly arbitrary, the default being
    fairly generous so as to be lenient towards the odd
    failure, and the min being 7 because it is a lucky
    number (which also happens to be long enough to allow
    the other side to handle the reconnection attempt) */
#define DEFAULT_WAIT_PIPE_MS 500
#define MIN_WAIT_PIPE_MS 7

#ifdef __cplusplus
extern "C" {
#endif

static inline void free_ptr(const void *address)
{
	/* Attribute cleanup does not pass the pointer itself, but rather the address of the pointer,
	 * which is on the stack so should never be null. Of course there is little point to use this
	 * outside attribute cleanup - might as well call free directly in this case. */
	assert(address);

	/* Ideally, we'd accept a parameter of type `void **` since this is the address of a pointer,
	 * but implicit conversions are only legal to and from `void *` which is why the cast */
	free(*(void *const *)address);
}

static inline void close_FILE(FILE *const *file)
{
	assert(file);

	if (*file)
		fclose(*file);
}

static inline void close_fd(int *fd)
{
	assert(fd);

	if (*fd >= 0)
		close(*fd);
}

char * log_name_by_id(log_id_t id);
log_id_t log_id_by_name(const char *name);

log_priority log_priority_from_char(signed char c);
void syslog_critical_failure(const char *message, ...);
int send_dlog_request(int sockfd, int request, void *data, int datalen);
int send_dlog_reply(int sockfd, int request, int result, void *data, int datalen);
int recv_dlog_reply(int sockfd, int request, void **data, int *datalen);
int send_pipe(int sockfd, int pipefd);
int recv_pipe(int sockfd, int timeout_ms);
int connect_sock(const char *path);


static inline int max_int(int a, int b) { return a > b ? a : b; }
static inline int min_int(int a, int b) { return a < b ? a : b; }
static inline int clamp_int(int i, int min, int max) { return min_int(max, max_int(min, i)); }

static inline void bit_set(int *bitset, int index) { *bitset |= (1 << index); }
static inline void bit_clear(int *bitset, int index) { *bitset &= ~(1 << index); }
static inline int bit_test(int bitset, int index) { return bitset & (1 << index); }
static inline int bit_count(int bitset) { return __builtin_popcount(bitset); }

/**
 * @brief Set FD flags
 * @details Set a file descriptor's flags
 * @param[in] fd The FD whose flags to set
 * @param[in] mask The flags to set
 * @returns 0 on success, -errno on failure
 */
static inline int fd_set_flags(int fd, int mask)
{
	assert(fd >= 0);

	int flags = fcntl(fd, F_GETFL);
	if (flags < 0)
		return -errno;

	return fcntl(fd, F_SETFL, flags | mask) < 0 ? -errno : 0;
}

/**
 * @brief Clear FD flags
 * @details Clear a file descriptor's flags
 * @param[in] fd The FD whose flags to clear
 * @param[in] mask The flags to clear
 * @returns 0 on success, -errno on failure
 */
static inline int fd_clear_flags(int fd, int mask)
{
	assert(fd >= 0);

	int flags = fcntl(fd, F_GETFL);
	if (flags < 0)
		return -errno;

	return fcntl(fd, F_SETFL, flags & ~mask) < 0 ? -errno : 0;
}


int str_ends_with(const char *str, const char *suffix);
char *create_glued_string(char const *const *strings, int n);

#ifdef __cplusplus
}
#endif

#endif /* _LOGCOMMON_H */
