/*
 * DLOG
 * Copyright (c) 2005-2008, The Android Open Source Project
 * Copyright (c) 2012-2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _LOGPRINT_H
#define _LOGPRINT_H

#include <stdbool.h>
#include <time.h>
#include <pthread.h>

#include <logcommon.h>
#include <queued_entry.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef enum {
	FORMAT_OFF = 0,
	FORMAT_BRIEF,
	FORMAT_PROCESS,
	FORMAT_TAG,
	FORMAT_THREAD,
	FORMAT_RAW,
	FORMAT_TIME,
	FORMAT_THREADTIME,
	FORMAT_KERNELTIME,
	FORMAT_RECV_REALTIME,
	FORMAT_RWTIME,
	FORMAT_LONG,
	FORMAT_JSON,
} log_print_format;

struct log_format {
	log_print_format format;
	bool color;
};

typedef struct log_filter_t log_filter;

log_filter *log_filter_new();

void log_filter_free(log_filter *p_filter);

/**
 * Returns a deep copy of the passed object.
 */
log_filter *log_filter_from_filter(const log_filter *p_filter);

/**
 * Returns FORMAT_OFF on invalid string
 */
log_print_format log_format_from_string(const char *s);

/**
 * filterExpression: a single filter expression
 * eg "AT:d"
 *
 * returns 0 on success and -1 on invalid expression
 *
 * Assumes single threaded execution
 *
 */

int log_add_filter_rule(log_filter *p_filter, const char *filterExpression);

/**
 * filterString: a whitespace-separated set of filter expressions
 * eg "AT:d *:i"
 *
 * returns 0 on success and -1 on invalid expression
 *
 * Assumes single threaded execution
 *
 */

int log_add_filter_string(log_filter *p_filter, const char *filterString);


/**
 * pid: an int with a pid of a log source process,
 * can be added multiple times with different values
 * (in that case log_should_print_line() will logically
 * OR all filter pid values to make a decision)
 *
 * returns 0 on success and -1 on invalid expression
 *
 * Assumes single threaded execution
 *
 */

int log_add_filter_pid(log_filter *p_filter, pid_t pid);


/**
 * tid: an int with a tid of a log source thread,
 * can be added multiple times with different values
 * (in that case log_should_print_line() will logically
 * OR all filter tid values to make a decision)
 *
 * returns 0 on success and -1 on invalid expression
 *
 * Assumes single threaded execution
 *
 */

int log_add_filter_tid(log_filter *p_filter, pthread_t tid);

/**
 * returns true if this log line should be printed based on its entry
 * data (priority, tag, pid and tid), and false if it should not
 */
bool log_should_print_line(
		log_filter *p_filter, const struct logger_entry *entry);

/**
 * Returns pointer to log entry tag
 *
 * Finds the beginning of a log tag in entry's payload and returns it
 */
char* log_buffer_get_tag(const struct logger_entry *entry);

/**
 * Returns pointer to log entry message
 *
 * Finds the beginning of a log message in entry's payload and returns it
 */
char* log_buffer_get_message(const struct logger_entry *entry);

/**
 * Formats a log message into a buffer
 *
 * Uses defaultBuffer if it can, otherwise malloc()'s a new buffer
 * If return value != defaultBuffer, caller must call free()
 * Returns NULL on malloc error
 */

char *log_format_log_line(
		const struct log_format p_format,
		char *defaultBuffer,
		size_t defaultBufferSize,
		const struct logger_entry *p_line,
		size_t *p_outLength);

/**
 * Either print or do not print log line, based on filter
 *
 * Assumes single threaded execution
 *
 */
int log_print_log_line(
		struct log_format p_format,
		int fd,
		bool colors_auto,
		const struct logger_entry *entry);

log_priority filter_char_to_pri(char c);

char filter_pri_to_char(log_priority pri);
enum sorting_order get_format_sorting(log_print_format format);

#ifdef __cplusplus
}
#endif

#endif /*_LOGPRINT_H*/
