/*
 * Copyright (c) 2017, Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef PTRS_LIST_INCLUDED
#define PTRS_LIST_INCLUDED

#include <stdbool.h>

struct list_elem;

/**
 * list_head is a pointer to list element
 * and must be initialized by the user to NULL
 * before using following API
 */
typedef struct list_elem *list_head;

/**
 * list element value is a pointer to user data
 * which resources are not managed by the list
 * user must allocate and free data memory outside
 * list functions
 */
typedef void *elem_value;

/**
 * @brief Condition callback for list element
 * @param value list element value
 * @param user_data pointer to user data to be forwarded to callback
 * @return true if condition if fulfilled, false if not
 */
typedef bool (*cond_cb)(elem_value value, void *user_data);

/**
 * @brief Apply callback to list element
 * @param value list element value
 * @param user_data pointer to user data to be forwarded to callback
 */
typedef void (*apply_cb)(elem_value value, void *user_data);

/**
 * @brief Map callback to list element
 * @param value list element value
 * @param user_data pointer to user data to be forwarded to callback
 * @return new element, or NULL in case of a failure
 */
typedef elem_value (*map_cb)(elem_value value, void *user_data);

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief Add an element to the list
 * @details Push front a new element
 * @param[in,out] head pointer to the list head
 * @param[in] value user data pointer
 * @return true on success, false on allocation failure
 */
bool list_add(list_head *head, elem_value value);

/**
 * @brief Remove list elements equal to given parameter
 * @param[in,out] head pointer to the list head
 * @param[in] value list element value
 */
void list_remove(list_head *head, elem_value value);

/**
 * @brief Remove list elements if given condition is fulfilled
 * @param[in,out] head pointer to the list head
 * @param[in] user_data pointer to user data to be forwarded to callback
 * @param[in] cond condition callback to check which elements to remove
 */
void list_remove_if(list_head *head, void *user_data, cond_cb cond);

/**
 * @brief Clear list
 * @details Remove all list elements.
 * @param[in,out] head pointer to the list head
 */
void list_clear(list_head *head);

/**
 * @brief Free list contents and clear the list
 * @details Remove all list elements and free them
 * @param[in] head pointer to the list head
 * @notes Requires the list elements to be valid pointers
 */
void list_clear_free_contents(list_head *head);

/**
 * @brief Free list contents using a custom operation and clear the list
 * @details Remove all list elements and free them using a custom operation
 * @param[in] head pointer to the list head
 * @param[in] clear callback to call for each freed element
 */
void list_clear_custom(list_head *head, void *user_data, apply_cb clear);

/**
 * @brief Call given callback for alle elements in the list
 * @param[in] head list head
 * @param[in] user_data pointer to user data to be forwarded to callback
 * @param[in] apply callback to call for each element in the list
 */
void list_foreach(list_head head, void *user_data, apply_cb apply);

/**
 * @brief Create a new list by looping over a given one
 * @param[in] head list head
 * @param[in] user_data pointer to user data to be forwarded to the callbacks
 * @param[in] map callback to call for each element in the list
 * @param[in] clear callback to call for new elements in case of an allocation failure
 * @return a new list, or NULL on allocation failure
 * @remark returned list has to be deallocated by the caller
 */
list_head list_map(list_head head, void *user_data, map_cb map, apply_cb clear);

/**
 * @brief Return the value at iterator
 * @param[in] iter iterator
 * @returns the value at the iterator
 */
elem_value list_at(list_head iter);

/**
 * @brief Advance iterator
 * @param[in,out] iter iterator to advance
 */
void list_next(list_head *iter);

/**
 * @brief Count list elements
 * @param[in] head list head
 * @return list elements number
 */
int list_count(list_head head);

#ifdef __cplusplus
}
#endif

#endif // PTRS_LIST_INCLUDED
