/*
 * DLOG
 * Copyright (c) 2005-2008, The Android Open Source Project
 * Copyright (c) 2012-2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _QUEUED_ENTRY_H
#define _QUEUED_ENTRY_H

#include <logcommon.h>
#include <stdint.h>

struct log_config;

/**
 * @brief sorting order
 */
enum sorting_order {
	SORT_SENT_MONO = 0,
	SORT_SENT_REAL,
	SORT_RECV_MONO,
	SORT_RECV_REAL,
	SORT_MAX,
	SORT_DEFAULT,
};


enum {
	LOGGER_ENTRY_REALTIME,
	LOGGER_ENTRY_MONOTONIC
};

/*
 * Invalid timestamp marker. Note that
 * 1. the number of seconds can be arbitrary, we only check the nanoseconds value.
 *    However, we still set it to 0 in order to prevent accidental uninitialized reads,
 * 2. nsec = -1 isn't a value that's otherwise possible.
 */
#define INVALID_SEC 0
#define INVALID_NSEC -1

struct logger_entry {
	uint16_t    len;    /* length of the payload */
	uint8_t     priority; /* entry's priority*/
	uint8_t     padding;/* padding for alignment */
	int32_t     pid;    /* generating process's pid */
	int32_t     tid;    /* generating process's tid */
	int32_t     sec_sent_mono;  /* time sent, seconds, monotonic */
	int32_t     nsec_sent_mono; /* time sent, nanoseconds, monotonic */
	int32_t     sec_sent_real;  /* time sent, seconds, realtime */
	int32_t     nsec_sent_real; /* time sent, nanoseconds, realtime */
	int32_t     sec_recv_mono;  /* time received, seconds, monotonic */
	int32_t     nsec_recv_mono; /* time received, nanoseconds, monotonic */
	int32_t     sec_recv_real;  /* time received, seconds, realtime */
	int32_t     nsec_recv_real; /* time received, nanoseconds, realtime*/
	int32_t     tag_len;
	int32_t     default_send_ts_type;  /* default backend-dependant send time type */
	char        msg[];  /* the entry's payload */
};

struct logger_entry_with_msg {
	struct logger_entry header;
	char msg[LOG_MAX_PAYLOAD_SIZE];
};

/* Note! This is the header used by AL protocol version 2, and is backward
 * incompatible with the one used in version 1 due to the addition of some
 * extra members (marked below), one of those being header_size. Meanings
 * of the members are the same as in `logger_entry` unless noted otherwise.
 * The header for protocol version 1 is also provided below for reference. */
struct android_logger_entry {
	uint16_t len;
	uint16_t header_size; // since v2 (was padding in v1)
	int32_t  pid;
	int32_t  tid;
	int32_t  sec_sent;    // always realtime, never monotonic
	int32_t  nsec_sent;
	int32_t  euid;        // since v2 (message started here in v1). EUID of log sender; not actually used by DLog.
	char     msg[];       // had different offset in v1 (started where euid is in v2)
};

/* Unused, for reference only - the header for the version 1 of the protocol.
 * See above for the modern header for version 2. This version is obsolete as
 * unlike version 2 it is not forward-compatible since it lacks the size of
 * the header member. It seems that the majority of existing devices support
 * protocol version 2 so there is little need to pollute the codebase with
 * extra support for version 1. */
struct android_logger_entry_compat {
	uint16_t len;
	uint16_t padding;     // became header_size in v2
	int32_t  pid;
	int32_t  tid;
	int32_t  sec_sent;
	int32_t  nsec_sent;
	char     msg[];       // offset changed in v2
};

#ifdef USE_ANDROID_MONOTONIC

#define ANDROID_LOGGER_FOOTER_MAGIC 0xa1379841

struct android_logger_footer {
	uint32_t sec_sent_mono;  /* time sent, seconds, monotonic */
	uint32_t nsec_sent_mono; /* time sent, nanoseconds, monotonic */
	uint32_t magic;          /* arbitrary magic number, encode other info here (e.g. version number) in future versions, last field of structure no matter real version is */
};

#endif /* USE_ANDROID_MONOTONIC */

struct pipe_logger_entry {
	/* same meanings as above */
	uint16_t len;
	uint16_t padding;
	int32_t  pid;
	int32_t  tid;
	int32_t  sec_sent_mono;  /* time sent, seconds, monotonic */
	int32_t  nsec_sent_mono; /* time sent, nanoseconds, monotonic */
	int32_t  sec_sent_real;  /* time sent, seconds, realtime */
	int32_t  nsec_sent_real; /* time sent, nanoseconds, realtime */
	char     msg[];
};

#ifdef __cplusplus
extern "C" {
#endif

void create_pipe_message(void * buf, int prio, char const * tag, char const * msg);
int parse_kmsg_message(char *buffer, struct logger_entry_with_msg *lem, int buffer_size);
int parse_syslog_datagram(const char * buffer, int buffer_size, struct logger_entry * le);
void parse_androidlogger_message(struct android_logger_entry *ale, struct logger_entry *le, size_t dgram_size);
void parse_pipe_message(struct pipe_logger_entry *ple, struct logger_entry *le, size_t msg_size);

#ifdef __cplusplus
}
#endif

#endif /* _QUEUED_ENTRY_H */
