/* DLOG
 * Copyright (c) 2018 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// C
#include <assert.h>
#include <stdlib.h>
#include <string.h>

// POSIX
#include <limits.h>
#include <pthread.h>
#include <unistd.h>
#include <sys/inotify.h>

// DLog
#include <dynamic_config.h>
#include <libdlog.h>
#include <logcommon.h>
#include <logconfig.h>
#include <loglimiter.h>

static int inotify_fd = -1;
static typeof(((struct inotify_event *)0)->wd) inotify_wd = -1;
static char *inotify_path;

static pthread_mutex_t log_dynamic_config_lock = PTHREAD_MUTEX_INITIALIZER;

static bool __setup_runtime_watch(char const *path)
{
	assert(path);
	assert(inotify_fd == -1);
	assert(inotify_wd == -1);

	inotify_fd = inotify_init1(IN_NONBLOCK | IN_CLOEXEC);
	if (inotify_fd == -1) {
		syslog_critical_failure("DLog: inotify_init failed (%d)", errno);
		return false;
	}

	inotify_wd = inotify_add_watch(inotify_fd, path, IN_CREATE | IN_CLOSE_WRITE | IN_MOVE | IN_DELETE);
	if (inotify_wd == -1) {
		syslog_critical_failure("DLog: inotify_add_watch failed (%d)", errno);
		close(inotify_fd);
		inotify_fd = -1;
		return false;
	}

	return true;
}

static void __apply_update()
{
	assert(inotify_path);

	__attribute__((cleanup(log_config_free))) struct log_config config = {NULL, NULL};
	const int r = log_config_read_file(&config, inotify_path);
	if (r && r != -ENOENT) {
		syslog_critical_failure("DLog: keeping current runtime filters, can't read from \"%s\" (%d)", inotify_path, -r);
		return;
	}

	/* This could be concurrently read by other threads, but it's just a bunch
	 * of bools whose values carry no implications regarding others so it is
	 * fine not to lock. */
	__update_plog(&config);

	if (limiter) {
		if (!pthread_rwlock_wrlock(&log_limiter_lock)) {
			__log_limiter_update(&config);
			pthread_rwlock_unlock(&log_limiter_lock);
		} else {
			syslog_critical_failure("DLog: Failed to pthread_rwlock_wrlock for log_limiter_update.");
		}
	}
}

/// caller has to guarantee exclusive access
bool __dynamic_config_create(struct log_config *config)
{
	assert(config);
	assert(!inotify_path);

	const char *const extra_config_path = log_config_get(config, DYNAMIC_CONFIG_CONF_KEY);
	if (!extra_config_path || extra_config_path[0] != '/')
		return false;

	if (asprintf(&inotify_path, "%s/%s", extra_config_path, DYNAMIC_CONFIG_FILENAME) < 0) {
		inotify_path = NULL;
		syslog_critical_failure("DLog: out of memory");
		return false;
	}

	if (!__setup_runtime_watch(extra_config_path)) {
		free(inotify_path);
		inotify_path = NULL;
		return false;
	}

	/* missing filters config file at startup
	 * is not an error so no need to check
	 * return value here
	 */
	log_config_read_file(config, inotify_path);

	return true;
}

/// caller has to guarantee exclusive access
void __dynamic_config_destroy()
{
	assert(inotify_fd <  0 ||  inotify_path);
	assert(inotify_fd >= 0 || !inotify_path);
	assert(inotify_wd < 0 || inotify_fd >= 0);
	assert(inotify_fd < 0 || inotify_wd >= 0);

	if (!inotify_path)
		return;

	free(inotify_path);
	inotify_path = NULL;

	/* Important: We can not call inotify_rm_watch() here - the
	 * process might have forked, sharing inotify_fd. Removing the
	 * watch would break monitoring capability in all processes
	 * sharing the inotify_fd.
	 */
	inotify_wd = -1;

	assert(inotify_fd >= 0);
	close(inotify_fd);
	inotify_fd = -1;
}

void __dynamic_config_update()
{
	if (inotify_fd < 0)
		return;

	struct ievent_with_filename {
		struct inotify_event ie;
		char filename[NAME_MAX + 1];
	} ievent;
	const int r = read(inotify_fd, &ievent, sizeof ievent);
	if (r < 0)
		return;

	/* read() cannot really return 0 here (technically it can do
	 * that in kernels before 2.6.21 if the buffer is too small,
	 * but that is not really applicable here). The reason for
	 * this check is that some daemons close all file descriptors
	 * including this one (they don't have much choice since dlog
	 * does not expose any way to get its FD numbers), open some
	 * other files (which potentially reuses the same FD number
	 * as inotify_fd) and then try to log something using dlog.
	 * The result is that read(inofify_fd, ...) instead reads from
	 * that other descriptor, which is not necessarily (and in fact,
	 * quite unlikely to be) an inotify descriptor, meaning it can
	 * return 0. Sometimes we're lucky and it happens - the client
	 * shouldn't see any difference afterwards so this situation
	 * can be ignored without much fuss. Other times we "steal"
	 * some data which does not belong to us and is not a correctly
	 * formatted inotify event, in which case WD will contain some
	 * garbage value, triggering the assert. This is not actually
	 * a bad thing; it's a fairly decent way to let the client know
	 * that something is amiss. */
	if (r == 0)
		return;

	/* -1 means an overflow of the the event queue. This is fine
	   (if a bit worrisome) since it still tells us that an event
	   has arrived, which is all we care about at this point. */
	assert(ievent.ie.wd == inotify_wd || ievent.ie.wd == -1);

	/* Somebody else might already be updating, in which case we
	 * needn't bother since we would be reading the same files. */
	int lock_r = pthread_mutex_trylock(&log_dynamic_config_lock);
	if (lock_r < 0)
		return;

	__apply_update();

	pthread_mutex_unlock(&log_dynamic_config_lock);
}
