/*-*- Mode: C; c-basic-offset: 8; indent-tabs-mode: t -*-
 * DLOG
 * Copyright (c) 2005-2008, The Android Open Source Project
 * Copyright (c) 2012-2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// C
#include <assert.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>

// POSIX
#include <pthread.h>
#include <unistd.h>

// DLog
#include <dynamic_config.h>
#include <libdlog.h>
#include <logcommon.h>
#include "logconfig.h"
#include "loglimiter.h"

#define DEFAULT_CONFIG_LIMITER false
#define DEFAULT_CONFIG_PLOG true
#define DEFAULT_CONFIG_DEBUGMODE 0
#define DEFAULT_CONFIG_LIMITER_APPLY_TO_ALL_BUFFERS 0

/**
 * @brief Points to a function which writes a log message
 * @details The function pointed to depends on the backend used
 * @param[in] log_id ID of the buffer to log to. Belongs to (LOG_ID_INVALID, LOG_ID_MAX) non-inclusive
 * @param[in] prio Priority of the message.
 * @param[in] tag The message tag, identifies the sender.
 * @param[in] msg The contents of the message.
 * @return Returns the number of bytes written on success and a negative error value on error.
 * @see __dlog_init_backend
 */
int (*write_to_log)(log_id_t log_id, log_priority prio, const char *tag, const char *msg) = NULL;
void (*destroy_backend)();

pthread_rwlock_t log_limiter_lock = PTHREAD_RWLOCK_INITIALIZER;
static pthread_mutex_t log_construction_lock = PTHREAD_MUTEX_INITIALIZER;
static bool is_initialized = false;

extern void __dlog_init_pipe(const struct log_config *conf);
extern void __dlog_init_android(const struct log_config *conf);

bool limiter;
bool dynamic_config;
bool plog[LOG_ID_MAX];
bool plog_default_values[LOG_ID_MAX];

static int debugmode;
static int fatal_assert;
static int limiter_apply_to_all_buffers;

static void __configure_limiter(struct log_config *config)
{
	assert(config);

	if (!limiter)
		return;

	limiter = __log_limiter_create(config);
}

static int __configure_backend(struct log_config *config)
{
	assert(config);

	const char *const backend = log_config_get(config, "backend");
	if (!backend)
		return 0;

	if (!strcmp(backend, "pipe"))
		__dlog_init_pipe(config);
	else if (!strcmp(backend, "logger"))
		__dlog_init_android(config);
	else
		return 0;

	return 1;
}

static void __set_plog_default_values()
{
	for (int i = 0; i < NELEMS(plog); ++i)
		plog_default_values[i] = plog[i];
}

static void __initialize_plog(const struct log_config *config)
{
	assert(config);

	const bool plog_default = log_config_get_boolean(config, "plog", DEFAULT_CONFIG_PLOG);
	for (int i = 0; i < NELEMS(plog); ++i)
		plog[i] = plog_default;
	plog[LOG_ID_APPS] = true; // the default does not apply here for backward compatibility reasons.
	__set_plog_default_values();
}

static void __configure_parameters(struct log_config *config)
{
	assert(config);

	__initialize_plog(config);
	__update_plog(config);
	__set_plog_default_values();

	debugmode = log_config_get_int(config, "debugmode", DEFAULT_CONFIG_DEBUGMODE);
	fatal_assert = access(DEBUGMODE_FILE, F_OK) != -1;
	limiter = log_config_get_boolean(config, "limiter", DEFAULT_CONFIG_LIMITER);
	limiter_apply_to_all_buffers = log_config_get_int(config,
									"limiter_apply_to_all_buffers",
									DEFAULT_CONFIG_LIMITER_APPLY_TO_ALL_BUFFERS);
}

void __update_plog(const struct log_config *conf)
{
	assert(conf);

	for (int i = 0; i < NELEMS(plog); ++i) {
		char key[MAX_CONF_KEY_LEN];
		const int r = snprintf(key, sizeof key, "enable_%s", log_name_by_id((log_id_t)i));
		if (r < 0)
			continue;
		plog[i] = log_config_get_boolean(conf, key, plog_default_values[i]);
	}
}

/**
 * @brief Configure the library
 * @details Reads relevant config values
 * @remarks This is more or less a constructor, but there are some obstacles
 *          to using it as such (i.e. with attribute constructor):
 *
 *  - some important pieces of the system link to dlog, they start very early
 *    such that dlog can't properly initialize (which lasts for program lifetime)
 *    but don't actually log anything until later on and would be fine under lazy
 *    initialisation. The way to do it "properly" would be to expose this function
 *    into the API so that people can manually call it when they're ready, but
 *    one of the design goals of the current API is that it requires absolutely no
 *    other calls than `dlog_print`. Changing it would require somebody with a
 *    bird's eye view of the system to produce a design so I wouldn't count on it.
 *
 *  - the constructor would need to have as high of a priority as possible (so as
 *    to minimize the risk of another library's constructor using uninitialized data)
 *    but at the same time others might want some room to wrap functions before
 *    dlog uses them (think mprobe/mcheck). This would also require a design pass.
 */
#ifndef UNIT_TEST
static
#endif
bool __configure(void)
{
	__attribute__((cleanup(log_config_free))) struct log_config config;

	if (log_config_read(&config) < 0)
		return false;

	dynamic_config = __dynamic_config_create(&config);

	__configure_parameters(&config);

	if (!__configure_backend(&config)) {
		__dynamic_config_destroy();
		dynamic_config = false;
		return false;
	}

	__configure_limiter(&config);
	return true;
}

static bool first = true;
static bool initialize()
{
	if (is_initialized)
		return true;

	/* The mutex acts as a barrier, but otherwise the C language's
	 * machine abstraction is single-threaded. This means that the
	 * compiler is free to rearrange calls inside the mutex according
	 * to the as-if rule because it doesn't care if another thread can
	 * access it in parallel. In particular, `is_initialized = true`
	 * directly after `__configure()` could be rearranged to go in
	 * front of it because it is not touched inside that function
	 * if the compiler thinks it helps somehow (not unlikely: since
	 * it is checked before the mutex, it is very probable for it to
	 * still be in the CPU register or something like that). On top
	 * of that, some architectures (in particular, armv7l) don't have
	 * strict memory guarantees and can reorder actual memory stores
	 * on their own, even if the compiler didn't do anything fancy
	 * when creating machine code. For more info about the issue,
	 * see https://www.aristeia.com/Papers/DDJ_Jul_Aug_2004_revised.pdf
	 *
	 * Ultimately this means that there needs to be some sort of
	 * barrier between `__configure` and `is_initialized = true`,
	 * and the simplest way to achieve that is to just wait until
	 * the second entry into the mutex. */

	bool ret;
	pthread_mutex_lock(&log_construction_lock);
		if (first)
			first = !__configure();
		else
			is_initialized = true;
		ret = !first;
	pthread_mutex_unlock(&log_construction_lock);
	return ret;
}

/**
 * @brief Fatal assertion
 * @details Conditionally crash the sucka who sent the log
 * @param[in] prio Priority of the log
 */
static void __dlog_fatal_assert(int prio)
{
	assert(!fatal_assert || (prio != DLOG_FATAL));
}

/**
 * @brief Check log validity
 * @details Checks whether the log is valid and eligible for printing
 * @param[in] log_id The target buffer ID
 * @param[in] prio The log's priority
 * @param[in] tag The log's tag
 * @return DLOG_ERROR_NONE on success, else an error code.
 * @retval DLOG_ERROR_INVALID_PARAMETER Invalid parameter
 */
static int dlog_check_validity(log_id_t log_id, int prio, const char *tag)
{
	if (!tag)
		return DLOG_ERROR_INVALID_PARAMETER;

	if (log_id <= LOG_ID_INVALID || LOG_ID_MAX <= log_id)
		return DLOG_ERROR_INVALID_PARAMETER;

	return DLOG_ERROR_NONE;
}

/**
 * @brief Check log against limiter rules
 * @details Checks whether the log passes current limiter rules
 * @param[in] log_id The target buffer ID
 * @param[in] prio The log's priority
 * @param[in] tag The log's tag
 * @return DLOG_ERROR_NONE on success, else an error code.
 * @retval DLOG_ERROR_NOT_PERMITTED Not permitted
 */
static int dlog_check_limiter(log_id_t log_id, int prio, const char *tag)
{
	if (!debugmode && prio <= DLOG_DEBUG)
		return DLOG_ERROR_NOT_PERMITTED;

	if (dynamic_config)
		__dynamic_config_update();

	if (limiter) {
		int should_log = 0;
		if (!pthread_rwlock_rdlock(&log_limiter_lock)) {
			should_log = __log_limiter_pass_log(tag, prio);
			pthread_rwlock_unlock(&log_limiter_lock);
		}

		if (!should_log) {
			return DLOG_ERROR_NOT_PERMITTED;
		} else if (should_log < 0) {
			write_to_log(log_id, prio, tag,
					"Your log has been blocked due to limit of log lines per minute.");
			return DLOG_ERROR_NOT_PERMITTED;
		}
	}

	/* This can change due to __dynamic_config_update(), but is atomic and its
	 * value implies nothing else so does not need to be under a lock. */
	if (!plog[log_id])
		return DLOG_ERROR_NOT_PERMITTED;

	return DLOG_ERROR_NONE;
}

static int __write_to_log_critical_section(log_id_t log_id, int prio, const char *tag, const char *fmt, va_list ap, bool check_should_log)
{
	if ((check_should_log || limiter_apply_to_all_buffers) && (dlog_check_limiter(log_id, prio, tag) < 0))
		return DLOG_ERROR_NONE;

	char buf[LOG_MAX_PAYLOAD_SIZE];
	vsnprintf(buf, sizeof buf, fmt, ap);
	return write_to_log(log_id, prio, tag, buf);
}

static int __write_to_log(log_id_t log_id, int prio, const char *tag, const char *fmt, va_list ap, bool check_should_log)
{
	int ret = dlog_check_validity(log_id, prio, tag);
	if (ret < 0)
		return ret;

	/* Threads can be cancelled before they give up a lock.
	 * Therefore cancellation is temporarily disabled.
	 * This solution is comparatively simple and cheap.
	 * The other solutions (cleanup handlers, robust mutexes)
	 * would be much more complicated and also inflict larger
	 * runtime costs. The downside of disabling cancellation
	 * is not a problem in our case because it is temporary
	 * and very brief so we don't keep an obsolete thread
	 * for much longer than we otherwise would. */
	int old_cancel_state;
	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &old_cancel_state);

	/* The only thing that needs to be protected here is `write_to_log` since
	 * all other resources already have their own specific locks (and even the
	 * pointer could be made to point at a null handler instead of a true NULL)
	 * but giving this guarantee makes everything a lot simpler as it removes
	 * the risk of something suddenly becoming NULL during processing. */
	ret = !initialize() || !write_to_log
		? DLOG_ERROR_NOT_PERMITTED
		: __write_to_log_critical_section(log_id, prio, tag, fmt, ap, check_should_log)
		;

	pthread_setcancelstate(old_cancel_state, NULL);

	return ret;
}

/**
 * @brief Print log
 * @details Print a log line
 * @param[in] log_id The target buffer ID
 * @param[in] prio Priority
 * @param[in] tag tag
 * @param[in] fmt Format (same as printf)
 * @param[in] ap Argument list
 * @return Bytes written, or negative error
 */
int __dlog_vprint(log_id_t log_id, int prio, const char *tag, const char *fmt, va_list ap)
{
	int ret = __write_to_log(log_id, prio, tag, fmt, ap, true);
	__dlog_fatal_assert(prio);

	return ret;
}

/**
 * @brief Print log
 * @details Print a log line
 * @param[in] log_id The target buffer ID
 * @param[in] prio Priority
 * @param[in] tag tag
 * @param[in] fmt Format (same as printf)
 * @return Bytes written, or negative error
 */
int __dlog_print(log_id_t log_id, int prio, const char *tag, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	int ret = __dlog_vprint(log_id, prio, tag, fmt, ap);
	va_end(ap);

	return ret;
}

/**
 * @brief Print log
 * @details Print a log line
 * @param[in] log_id The target buffer ID
 * @param[in] prio Priority
 * @param[in] tag tag
 * @param[in] fmt Format (same as printf)
 * @return Bytes written, or negative error
 */
int __dlog_sec_print(log_id_t log_id, int prio, const char *tag, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	int ret = __dlog_vprint(log_id, prio, tag, fmt, ap);
	va_end(ap);

	return ret;
}

int dlog_vprint(log_priority prio, const char *tag, const char *fmt, va_list ap)
{
	return __write_to_log(LOG_ID_APPS, prio, tag, fmt, ap, false);
}

int dlog_print(log_priority prio, const char *tag, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	int ret = dlog_vprint(prio, tag, fmt, ap);
	va_end(ap);

	return ret;
}

/**
 * @brief Finalize DLog
 * @details Finalizes and deallocates the library
 * @notes Used directly in tests; brings back the pre-init state
 */
void __dlog_fini(void)
{
	if (destroy_backend) {
		destroy_backend();
		destroy_backend = NULL;
	}
	write_to_log = NULL;
	is_initialized = false;
	first = true;

	__log_limiter_destroy();
	__dynamic_config_destroy();
}
