// C
#include <stdio.h>

// POSIX
#include <fcntl.h>
#include <time.h>
#include <sys/socket.h>

// DLog
#include <backend_androidlogger.h>
#include <libdlog.h>
#include <logcommon.h>
#include <logconfig.h>
#ifdef USE_ANDROID_MONOTONIC
#include <queued_entry.h>
#endif /* USE_ANDROID_MONOTONIC */

/* TODO: initialize in for loop to be sure of initializing all items */
static int log_fds[(int)LOG_ID_MAX] = { -1, -1, -1, -1, -1, -1 };

/**
 * @brief Write to log
 * @details Writes a log
 * @param[in] log_id ID of the buffer to log to. Belongs to (LOG_ID_INVALID, LOG_ID_MAX) non-inclusive
 * @param[in] prio Priority of the message.
 * @param[in] tag The message tag, identifies the sender.
 * @param[in] msg The contents of the message.
 * @return Number of bytes written, or negative errno
 */
static int __write_to_log_android(log_id_t log_id, log_priority prio, const char *tag, const char *msg)
{
	ssize_t ret;
	int log_fd;

	if (log_id > LOG_ID_INVALID && log_id < LOG_ID_MAX)
		log_fd = log_fds[log_id];
	else
		return DLOG_ERROR_INVALID_PARAMETER;

	if (prio < DLOG_VERBOSE || prio >= DLOG_PRIO_MAX)
		return DLOG_ERROR_INVALID_PARAMETER;

	if (!tag)
		tag = "";

	if (!msg)
		return DLOG_ERROR_INVALID_PARAMETER;

#ifdef USE_ANDROID_MONOTONIC
	struct iovec vec[4];

	struct timespec ts;
	struct android_logger_footer alf;

	clock_gettime(CLOCK_MONOTONIC, &ts);
	alf.magic = ANDROID_LOGGER_FOOTER_MAGIC;
	alf.sec_sent_mono = ts.tv_sec;
	alf.nsec_sent_mono = ts.tv_nsec;

	vec[3].iov_base = (void *) &alf;
	vec[3].iov_len = sizeof alf;
#else
	struct iovec vec[3];
#endif /* USE_ANDROID_MONOTONIC */

	vec[0].iov_base = (unsigned char *) &prio;
	vec[0].iov_len  = 1;
	vec[1].iov_base = (void *) tag;
	vec[1].iov_len  = strlen(tag) + 1;
	vec[2].iov_base = (void *) msg;
	vec[2].iov_len  = strlen(msg) + 1;

	ret = writev(log_fd, vec, NELEMS(vec));
	if (ret < 0)
		ret = -errno;

	return ret;
}

static void __destroy_androidlogger()
{
	/* No need for an explicit lock here; this should only be called
	 * by the library destructor which has its own lock */

	for (size_t i = 0; i < NELEMS(log_fds); ++i)
		if (log_fds[i] >= 0)
			close(log_fds[i]);
}

/**
 * @brief Initialize the backend
 * @details Prepares the backend for proper and fruitful work
 */
void __dlog_init_android(const struct log_config *conf)
{
	log_id_t buf_id;

	for (buf_id = 0; buf_id < LOG_ID_MAX; ++buf_id) {
		int ret = logger_open_buffer_from_config(buf_id, conf, O_WRONLY | O_CLOEXEC, &log_fds[buf_id]);

		if (ret < 0)
			goto failure;
		else if (ret == 0) /* zero signifies no buffer */
			continue;
	}
	write_to_log = __write_to_log_android;
	destroy_backend = __destroy_androidlogger;
	return;

failure:
	for (buf_id = 0; buf_id < LOG_ID_MAX; buf_id++)
		if (log_fds[buf_id] >= 0)
			close(log_fds[buf_id]);
}
