// C
#include <assert.h>
#include <stdio.h>

// POSIX
#include <signal.h>
#include <linux/limits.h>
#include <sys/socket.h>
#include <sys/un.h>

// DLog
#include <libdlog.h>
#include <logconfig.h>
#include <logpipe.h>
#include <logprint.h>

/* TODO: initialize in for loop to be sure of initializing all items */
static int pipe_fd[LOG_ID_MAX] = {-1, -1, -1, -1, -1, -1};
static char log_pipe_path[LOG_ID_MAX][PATH_MAX];
static int wait_pipe_ms = DEFAULT_WAIT_PIPE_MS;

static int __write_to_log_pipe(log_id_t log_id, log_priority prio, const char *tag, const char *msg);

static pthread_rwlock_t log_pipe_lock = PTHREAD_RWLOCK_INITIALIZER;

/**
 * @brief Connect to a pipe
 * @details Connects to a socket whence it retrieves a pipe fd
 * @param[in] path Path to the logger daemon write socket from which to retrieve the pipe
 * @return Pipe file descriptor, or negative errno on failure
 */
static int connect_pipe(const char * path)
{
	int r;
	int fd;
	struct sockaddr_un sa = { .sun_family = AF_UNIX };
	struct dlog_control_msg ctrl_msg = DLOG_CTRL_REQ_PIPE;

	fd = socket(AF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0);
	if (fd < 0)
		return -errno;

	strncpy(sa.sun_path, path, sizeof(sa.sun_path));

	r = connect(fd, (struct sockaddr *) &sa, sizeof(sa));
	if (r < 0) {
		close(fd);
		return -errno;
	}

	r = write(fd, &ctrl_msg, ctrl_msg.length);
	if (r < 0) {
		close(fd);
		return -errno;
	}

	int ret = recv_pipe(fd, wait_pipe_ms);
	close(fd);

	return ret;
}

/**
 * @brief Reconnect pipe
 * @details Gets another pipe for a buffer
 * @param[in] log_id The buffer whose pipe to replace
 * @pre `log_pipe_lock` taken as read-only
 * @post `log_pipe_lock` taken as read-only
 * @return Boolean whether pipe successfully reconnected
 */
static int __reconnect_pipe(log_id_t log_id)
{
	int new_fd, old_fd, ret = 0;
	int wrlock;

	pthread_rwlock_unlock(&log_pipe_lock);
	if ((wrlock = pthread_rwlock_wrlock(&log_pipe_lock)) != 0)
		goto finish;

	new_fd = connect_pipe(log_pipe_path[log_id]);
	if (new_fd < 0) {
		/* if connnecting fails consequently assume
		   dlog_logger is broken and reduce the timeout to
		   allow dlog clients to run normally */
		wait_pipe_ms /= 2;
		if (wait_pipe_ms < MIN_WAIT_PIPE_MS)
			wait_pipe_ms = MIN_WAIT_PIPE_MS;

		goto finish;
	}

	old_fd = pipe_fd[log_id];
	if (old_fd < 0)
		pipe_fd[log_id] = new_fd;
	else {
		int r;
#ifdef __PARANOIC
		sigset_t all_set, current_set;
		sigfillset(&all_set);
		sigprocmask(SIG_BLOCK, &all_set, &current_set);
#endif
		r = dup2(new_fd, old_fd);
		assert(r >= 0);
#ifdef __PARANOIC
		sigprocmask(SIG_SETMASK, &current_set, NULL);
#endif
		close(new_fd);
	}

	ret = (pipe_fd[log_id] >= 0);

finish:
	if (!wrlock)
		pthread_rwlock_unlock(&log_pipe_lock);

	if (pthread_rwlock_rdlock(&log_pipe_lock))
		return 0;

	return ret;
}

/* @pre `log_pipe_lock` taken as read-only
 * @post `log_pipe_lock` taken as read-only
 */
static int _write_to_log_pipe_critical_section(log_id_t log_id, char *buf)
{
	if (pipe_fd[log_id] < 0 && !__reconnect_pipe(log_id))
		return -1;

	struct pipe_logger_entry *const ple = (struct pipe_logger_entry *)buf;

	int ret = write(pipe_fd[log_id], buf, ple->len);
	if (ret < 0 && (errno == EPIPE || errno == EBADF)) {
		/* Handling EPIPE is obvious, it means other peer disappeared and
		 * we need to reconnect (eg. dlog_logger restarted).
		 *
		 * EBADF-handling is a hack - it's needed due to Tizen's launchpad
		 * (application preloader) closing all file descriptors explicitly
		 * (including libdlog's file descriptors) but still trying to log.
		 */
		int was_ebadf = (errno == EBADF);
		if (!__reconnect_pipe(log_id))
			return -1;
		ret = write(pipe_fd[log_id], buf, ple->len);
		if (was_ebadf) {
			create_pipe_message(buf, DLOG_FATAL, "LIBDLOG", "libdlog's internal state has been destroyed! The user application closed libdlog's file descriptor.");
			ret = write(pipe_fd[log_id], buf, ple->len);
		}
	}
	return ret;
}

/**
 * @brief Write to log
 * @details Writes a log
 * @param[in] log_id ID of the buffer to log to. Belongs to (LOG_ID_INVALID, LOG_ID_MAX) non-inclusive
 * @param[in] prio Priority of the message.
 * @param[in] tag The message tag, identifies the sender.
 * @param[in] msg The contents of the message.
 * @return Number of bytes written, or negative errno
 */
static int __write_to_log_pipe(log_id_t log_id, log_priority prio, const char *tag, const char *msg)
{
	ssize_t ret = 0;
	char buf[LOG_MAX_PAYLOAD_SIZE + sizeof(struct pipe_logger_entry)];

	if (!tag)
		tag = "";

	if (log_id >= LOG_ID_MAX ||
	    log_id < 0 ||
	    prio < DLOG_VERBOSE ||
	    prio >= DLOG_PRIO_MAX ||
	    !msg ||
	    1 /* priority */ + strlen(tag) + 1 /* NULL delimiter */ + strlen(msg) + 1 /* NULL terminator */ > LOG_MAX_PAYLOAD_SIZE)
		return DLOG_ERROR_INVALID_PARAMETER;

	create_pipe_message(buf, prio, tag, msg);

	if (!pthread_rwlock_rdlock(&log_pipe_lock)) {
		ret = _write_to_log_pipe_critical_section(log_id, buf);
		pthread_rwlock_unlock(&log_pipe_lock);
	}

	return ret;
}

static void __destroy_pipe()
{
	/* No need for an explicit lock here; this should only be called
	 * by the library destructor which has its own lock */

	for (size_t i = 0; i < NELEMS(pipe_fd); ++i)
		if (pipe_fd[i] >= 0)
			close(pipe_fd[i]);

	wait_pipe_ms = DEFAULT_WAIT_PIPE_MS;
	signal(SIGPIPE, SIG_DFL);
}

/**
 * @brief Initialize the backend
 * @details Prepares the backend for proper and fruitful work
 */
void __dlog_init_pipe(const struct log_config *conf)
{
	int i;

	/*
	 * We mask SIGPIPE signal because most applications do not install their
	 * own SIGPIPE handler. Default behaviour in SIGPIPE case is to abort the
	 * process. SIGPIPE occurs when e.g. dlog daemon closes read pipe endpoint
	 */
	signal(SIGPIPE, SIG_IGN);

	for (i = 0; i < LOG_ID_MAX; ++i) {
		char conf_key[MAX_CONF_KEY_LEN];
		snprintf(conf_key, sizeof(conf_key), "%s_write_sock", log_name_by_id(i));
		const char * conf_val = log_config_get(conf, conf_key);
		if (!conf_val) {
			syslog_critical_failure("DLOG CRITICAL FAILURE: DLog config lacks the \"%s\" entry!", conf_key);
			return;
		}

		if (conf_val[0] != '/') {
			syslog_critical_failure("DLOG CRITICAL FAILURE: path \"%s\" from \"%s\" is not absolute!\n", conf_val, conf_key);
			return;
		}

		const size_t pathlen = strlen(conf_val);
		static const size_t MAX_PIPE_PATH = sizeof((struct sockaddr_un *)NULL)->sun_path;
		if (strlen(conf_val) >= MAX_PIPE_PATH) {
			syslog_critical_failure("DLOG CRITICAL FAILURE: path \"%s\" from \"%s\" is too long! %zu >= %zu", conf_val, conf_key, pathlen, MAX_PIPE_PATH);
			return;
		}

		snprintf(log_pipe_path[i], PATH_MAX, "%s", conf_val);
	}
	write_to_log = __write_to_log_pipe;
	destroy_backend = __destroy_pipe;
}
