/*
 * DLOG
 * Copyright (c) 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <assert.h>
#include <stddef.h>
#include <limits.h>
#include <sys/types.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <pthread.h>
#include <sys/time.h>

/* Included for priorities level */
#include <dlog.h>
#include "loglimiter.h"
#include "loglimiter-internal.h"
#include <logconfig.h>

/* Some random big odd number to make hash more diverse */
#define HASH_MAGIC_THINGY         5237231

#define TIME_FRAME                60

typedef int (*hash_cmp_func_t)(struct rule*, struct rule*);
typedef int (*hash_match_func_t)(struct rule*, unsigned, const char*, int);

struct hashmap {
	hash_cmp_func_t cmp;
	hash_match_func_t match;
	int size;
	void* bucket[];
};

struct hashmap* rules_hashmap = NULL;

/* Keep rules table as single-linked list */
struct rule* current_rules_table = NULL;
struct rule* original_rules_table = NULL;

#define HASHMAP_MASK(hm)          ((int)(hm->size - 1))
#define HASHMAP_LINEAR_PROBE_LEAP 1

static void rules_destroy(struct rule **rlist)
{
	assert(rlist);

	struct rule* r;

	if (NULL == *rlist)
		return;

	while ((r = *rlist)) {
		*rlist = (*rlist)->prev;
		free(r);
	}
}

static int rule_compare(struct rule* r1, struct rule* r2)
{
	if (r1->hash == r2->hash) {
		if (r1->prio == r2->prio)
			return strncmp(r1->tag, r2->tag, strlen(r2->tag));
		else
			return (r1->prio > r2->prio ? 1 : (-1));
	}

	return (r1->hash > r2->hash ? 1 : (-1));
}

static int rule_match(struct rule* r1, unsigned key, const char* s, int prio)
{
	if (r1->hash == key) {
		if (r1->prio == prio)
			return strncmp(r1->tag, s, strlen(s));
		else
			return (r1->prio > prio ? 1 : (-1));
	}

	return (r1->hash > key ? 1 : (-1));
}

/* Translate fancy priority notation into common denominator */
static int util_prio_to_char(int prio)
{
	static const char pri_table[DLOG_PRIO_MAX] = {
		[DLOG_UNKNOWN] = '?',
		[DLOG_DEFAULT] = '*',
		[DLOG_VERBOSE] = 'V',
		[DLOG_DEBUG] = 'D',
		[DLOG_INFO] = 'I',
		[DLOG_WARN] = 'W',
		[DLOG_ERROR] = 'E',
		[DLOG_FATAL] = 'F',
		[DLOG_SILENT] = 'S',
	};

	switch (prio) {
	case 'V':
	case 'D':
	case 'I':
	case 'W':
	case 'E':
	case 'F':
	case 'S':
	case '*':
		return prio;

	case 'v':
	case 'd':
	case 'i':
	case 'w':
	case 'e':
	case 'f':
	case 's':
		return toupper(prio);

	case '0' ... (DLOG_PRIO_MAX - 1 + '0'):
		return pri_table[prio - '0'];

	case 0 ... (DLOG_PRIO_MAX - 1):
		return pri_table[prio];

	default:
		return '?';
	}
}

/* The key is built from TAG and priority by DJB algorithm (Dan Bernstein).
 * Key is only 31 bits long. Negative values are keep to hold NULL bucket */
unsigned util_hash_key(const char* s, int c)
{
	unsigned hash = (unsigned)c;

	hash = ((hash << 5) + hash) + c;

	if (!s || !s[0])
		goto finish;

	while ('\0' != (c = *s++))
		hash = ((hash << 5) + hash) + c;

finish:
	/* Makes the hash more diverse */
	hash *= HASH_MAGIC_THINGY;

	return hash;
}

/* Create hashmap, it's internal interface. */
static struct hashmap* hashmap_create(int size, hash_cmp_func_t cmp_func,
				      hash_match_func_t match_func)
{
	assert(cmp_func);
	assert(match_func);
	assert(size > 0);

	struct hashmap* hm = NULL;
	/* please keep hashmap fill ratio around 50% */
	int internal_size = size << 1;

	/* Round up the lines counter to next power of two. */
	internal_size--;
	internal_size |= internal_size >> 1;
	internal_size |= internal_size >> 2;
	internal_size |= internal_size >> 4;
	internal_size |= internal_size >> 8;
	internal_size |= internal_size >> 16;
	internal_size++;

	hm = malloc(sizeof(struct hashmap) + internal_size * sizeof(void*));
	if (!hm)
		return NULL;

	/* Initialize hash field to correct value */
	memset((void*)hm, 0, sizeof(struct hashmap) + internal_size * sizeof(void*));

	hm->size = internal_size;
	hm->cmp = cmp_func;
	hm->match = match_func;

	return hm;
}

static void hashmap_destroy(struct hashmap **hm)
{
	assert(hm);

	if (!*hm)
		return;

	free(*hm);
	*hm = NULL;
}

static void hashmap_add(struct hashmap* hm, struct rule* r)
{
	unsigned b = (r->hash & HASHMAP_MASK(hm));

	while (hm->bucket[b]) {
		if (!hm->cmp(r, (struct rule*)hm->bucket[b]))
			break;
		b = (b + 1) & HASHMAP_MASK(hm);
	}

	hm->bucket[b] = r;
}

static struct rule* hashmap_search(struct hashmap* hm,
				   const char* tag, char prio)
{
	const unsigned key = util_hash_key(tag, prio);

	unsigned b = (key & HASHMAP_MASK(hm));
	unsigned b0 = b;

	while (hm->bucket[b]) {
		if (!hm->match(hm->bucket[b], key, tag, prio))
			break;

		b = (b + 1) & HASHMAP_MASK(hm);

		assert(b0 != b);
	}

	if (!hm->bucket[b])
		return NULL;

	return hm->bucket[b];
}

/* Must be always executed after __log_config_read() */
int __log_limiter_initialize(struct rule *rules_table)
{
	int hm_size;
	struct rule* rlist = NULL;

	/* logconfig.c module had to initialize this correctly */
	if (NULL == rules_table)
		return (-1);

	/* Count rules in the table */
	hm_size = 0;
	rlist = rules_table;
	while (rlist) {
		hm_size++;
		rlist = rlist->prev;
	}

	/* Allocate hashmap */
	rules_hashmap = (struct hashmap*) hashmap_create(hm_size,
							&rule_compare,
							&rule_match);
	if (NULL == rules_hashmap || !rules_hashmap->size) {
		hashmap_destroy(&rules_hashmap);
		return -1;
	}

	/* Add rule to hashmap */
	rlist = rules_table;
	while (rlist) {
		hashmap_add(rules_hashmap, rlist);
		rlist = rlist->prev;
	}

	return 0;
}

void __log_limiter_destroy(void)
{
	hashmap_destroy(&rules_hashmap);
	rules_destroy(&original_rules_table);
}

struct rule* __log_limiter_add_rule(struct rule** rules_table, const char* tag, int prio, int limit)
{
	assert(tag);

	struct rule* r;

	for (struct rule *i = *rules_table; i; i = i->prev) {
		if (strcmp(i->tag, tag) || i->prio != util_prio_to_char(prio))
			continue;

		if (i->limit != limit) { //set type here only if the limit is updated
			i->limit = limit;
			i->type = RULE_DEFAULT;
		}
		return i;
	}

	r = (struct rule*) malloc(sizeof(struct rule));
	if (NULL == r)
		return NULL;

	memset(r, 0, sizeof(struct rule));

	snprintf(r->tag, MAX_CONF_KEY_LEN, "%s", tag);
	r->prio = util_prio_to_char(prio);
	r->hash = util_hash_key(tag, r->prio);
	r->limit = limit;
	r->start = time(NULL);
	r->hit = 0;
	r->type = RULE_DEFAULT;

	r->prev = *rules_table;
	*rules_table = r;

	return r;
}

struct limiter_limits __log_limiter_get_limits(const char *tag, int prio)
{
	const struct rule fallback = { .limit = -1 };
	const char prio_c = util_prio_to_char(prio);

	return (struct limiter_limits) {
		.tag_and_prio = (hashmap_search(rules_hashmap, tag, prio_c) ?: &fallback)->limit,
		.tag          = (hashmap_search(rules_hashmap, tag, '*')    ?: &fallback)->limit,
		.prio         = (hashmap_search(rules_hashmap, "*", prio_c) ?: &fallback)->limit,
		.global       = (hashmap_search(rules_hashmap, "*", '*')    ?: &fallback)->limit,
	};
}


/* Function implement logic needed to decide,
   whenever message is written to log or not.

   Possible return values are:
	0  - to indicate that message is deny to write into log.
	(-1) - to indicate that limit of the messages is reached.
	1  - to indicate that message is allowed to write into log.
*/
int __log_limiter_pass_log(const char* tag, int prio)
{
	if (!rules_hashmap)
		return 1;

	/* allow empty-tagged messages and make it easy to catch an application that does that */
	if (!strlen(tag))
		return 1;

	const char prio_c = util_prio_to_char(prio);
	struct rule *r =
		hashmap_search(rules_hashmap, tag, prio_c) ?:
		hashmap_search(rules_hashmap, tag, '*')    ?:
		hashmap_search(rules_hashmap, "*", prio_c) ?:
		hashmap_search(rules_hashmap, "*", '*');

	if (!r)
		return 1;
	if (!r->limit)
		return 0;
	if (__LOG_LIMITER_LIMIT_MAX < r->limit)
		return 1;

	/* Decide, if it should go through or stop */
	time_t now = time(NULL);

	if (0 > now)
		return 1;

	if (now - r->start <= TIME_FRAME) {
		if (r->hit >= 0) {
			if (r->hit < r->limit) {
				r->hit++;
				return 1;
			}
			r->hit = INT_MIN+1;
			return (-1);
		} else {
			r->hit++;
			return 0;
		}
	} else {
		r->start = now;
		r->hit = 0;
		return 1;
	}
}

/**
 * @brief Limiter rule from config
 * @details Adds a limiter rule from a config entry, if one exists inside
 * @param[in] key The entry key
 * @param[in] value The entry value
 * @param[in,out] u Userdata (unused)
 */
static void __config_iteration(const char* key, const char* value, void *userdata)
{
	assert(key);
	assert(value);
	struct rule **rules_table = (struct rule **)userdata;
	assert(rules_table);

	static const int prefix_len = sizeof("limiter|") - 1;
	char * delimiter_pos;
	char limiter_tag[MAX_CONF_KEY_LEN];
	int limit;

	if (strncmp(key, "limiter|", prefix_len))
		return;

	delimiter_pos = strchr(key + prefix_len + 1, '|');
	if (!delimiter_pos || (delimiter_pos + 1 == key + strlen(key)))
		return;

	snprintf(limiter_tag, delimiter_pos - (key + prefix_len) + 1, "%s", key + prefix_len);

	if (!strcmp(value, "allow"))
		limit = __LOG_LIMITER_LIMIT_MAX + 1;
	else if (!strcmp(value, "deny"))
		limit = 0;
	else
		limit = atoi(value);

	__log_limiter_add_rule(rules_table, limiter_tag, *(delimiter_pos + 1), limit);
}

int __log_limiter_create(const struct log_config *config)
{
	assert(config);
	assert(!original_rules_table);
	assert(!rules_hashmap);

	log_config_foreach(config, __config_iteration, &original_rules_table);

	const int r = __log_limiter_initialize(original_rules_table);
	if (r) {
		rules_destroy(&original_rules_table);
		return 0;
	}

	for (struct rule *i = original_rules_table; i; i = i->prev)
		i->type = RULE_STATIC;

	return 1;
}

void __log_limiter_update(const struct log_config *config)
{
	assert(config);
	assert(!original_rules_table || rules_hashmap);

	struct rule *rules_table = NULL;
	for (struct rule *i = original_rules_table; i; i = i->prev) {
		struct rule *r = __log_limiter_add_rule(&rules_table, i->tag, i->prio, i->limit);
		if (!r)
			continue;
		r->type = RULE_STATIC;
	}

	log_config_foreach(config, __config_iteration, &rules_table);

	if (rules_hashmap) {
		for (struct rule *i = rules_table; i; i = i->prev) {
			const char prio_c = util_prio_to_char(i->prio);
			const struct rule *const prev_rule = hashmap_search(rules_hashmap, i->tag, prio_c);
			if (!prev_rule)
				continue;

			i->hit = prev_rule->hit;
			i->start = prev_rule->start;
		}
	}

	hashmap_destroy(&rules_hashmap);
	if (__log_limiter_initialize(rules_table) < 0) {
		__log_limiter_destroy();
		return;
	}

	rules_destroy(&current_rules_table);
	current_rules_table = rules_table;
}

/**
 * @brief Dump a limiter rule to provided buffer
 * @details Formats a string containing information about a rule and inserts it into provided buffer
 * @param[in,out] r The pointer to the rule to be dumped (in), the pointer to the previous rule (out)
 * @param[in,out] buf The buffer to dump the rule into
 * @param[in] size The size of the buffer
 * @returns 0 on success, -1 on failure
 */
int __log_limiter_dump_rule(struct rule **r, char *buf, const size_t size)
{
	struct rule *ruleptr = *r ? : current_rules_table;
	int s = 0;

	if (!ruleptr) {
		*r = NULL;
		return 0;
	}

	if (ruleptr->limit == 0)
		s = snprintf(buf, size, "Denied ");
	else if (ruleptr->limit == (__LOG_LIMITER_LIMIT_MAX + 1))
		s = snprintf(buf, size, "Unlimited ");
	else
		s = snprintf(buf, size, "%d logs/minute ", ruleptr->limit);

	if (s < 0)
		return -1;

	s = snprintf(buf + s, size - s, "for %s:%c (%s)",
			ruleptr->tag,
			util_prio_to_char(ruleptr->prio),
			(ruleptr->type == RULE_STATIC) ? "static" : "dynamic");
	if (s < 0)
		return -1;

	*r = ruleptr->prev;
	return 0;
}
