// C
#include <assert.h>
#include <ctype.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// POSIX
#include <getopt.h>
#include <unistd.h>
#include <sys/stat.h>

// DLog
#include <dynamic_config.h>
#include <logcommon.h>
#include <logconfig.h>
#include <logctl.h>
#include <loglimiter.h>

bool parse_options(int argc, const char **argv, struct parsed_params *params, int *exit_value)
{
	assert(argv);
	assert(params);
	assert(params->action == ACTION_NONE);
	assert(!params->value);
	assert(!params->tag);
	assert(exit_value);

	params->tag = "*";
	params->prio = '\0';
	params->plog_buffers_bitset = 0;

	bool use_global_action = true; // if neither tag not priority is specified

#define SET_ACTION(x) do { \
	if (params->action != x && params->action != ACTION_NONE) \
		goto failure; \
	params->action = x; \
} while (false)

	for (;;) {

		static const struct option long_options[] = {
			{"get"     ,       no_argument, NULL,  'g'},
			{"set"     , required_argument, NULL,  's'},
			{"tag"     , required_argument, NULL,  't'},
			{"priority", required_argument, NULL,  'p'},
			{"help"    ,       no_argument, NULL,  'h'},
			{"buffer"  , required_argument, NULL,  'b'},
			{"enable"  ,       no_argument, NULL, '\0'},
			{"disable" ,       no_argument, NULL, '\0'},
			{"clear"   ,       no_argument, NULL,  'c'},
			{0}
		};

		int long_index;
		int opt = getopt_long(argc, (char **) argv, "chb:gs:t:p:", long_options, &long_index);
		if (opt < 0)
			break;

		switch (opt) {
		case 0:
			SET_ACTION(ACTION_PLOG);
			params->plog_enable = (long_index == 6);
			break;
		case 'b': {
			const int buf_id = log_id_by_name(optarg);
			if (buf_id == LOG_ID_INVALID)
				goto failure;

			params->plog_buffers_bitset |= 1 << buf_id;
			break;
		} case 'h':
			*exit_value = EXIT_SUCCESS;
			goto print_help;
		case 'g':
			SET_ACTION(ACTION_GET);
			break;
		case 'c':
			SET_ACTION(ACTION_CLEAR);
			params->value = NULL;
			break;
		case 's': {
			const int value_num = atoi(optarg);
			if (strcmp(optarg, "allow") && strcmp(optarg, "deny") && (value_num <= 0 || value_num >= __LOG_LIMITER_LIMIT_MAX))
				goto failure;
			params->value = optarg;
			SET_ACTION(ACTION_SET);
			break;
		} case 't': {
			params->tag = optarg;
			use_global_action = false;
			break;
		} case 'p':
			params->prio = toupper(*optarg);
			if (!strchr("DVIWEFS*", params->prio))
				goto failure;
			use_global_action = false;
			break;

		default:
			break;
		}
	}

#undef SET_ACTION

	if (params->plog_buffers_bitset != 0 && params->action != ACTION_PLOG)
		goto failure;

	if (params->action == ACTION_NONE) {
		*exit_value = EXIT_SUCCESS;
		goto print_help;
	}

	switch (params->action) {
	case ACTION_GET:
		if (use_global_action)
			params->action = ACTION_DUMP;
		break;
	case ACTION_CLEAR:
		if (!use_global_action)
			params->action = ACTION_SET;
		break;

	default:
		break;
	}

	if (params->action == ACTION_SET && params->prio == '\0')
		params->prio = '*';

	return true;

failure:
	*exit_value = EXIT_FAILURE;

print_help:
	printf("Usage: %s [-t/--tag tag] [-p/--priority priority] (-g/--get | -s/--set value)\n"
		"\ttag: defaults to *\n"
		"\tpriority: F[atal], E[rror], W[arning], I[nfo], V[erbose], D[ebug] S[ilent] * (all, default)\n"
		"\t-g/--get: get current limit for given tag/priority; if neither given, dump the whole filterset\n"
		"\t-s/--set value: set a new limit to value (\"allow\", \"deny\", or number (0; %d) - allowed logs per minute)\n"
		"\t-c/--clear: clear the limit for given tag/priority; if neither given, clear the whole filterset\n"
		"Note: static filters are shadowed by dynamic ones but cannot be overridden at runtime\n\n"
		"       %s [-b/--buffer bufname] [-b ...] (--enable | --disable)\n"
		"\tbufname: main system radio apps (default: all of them)\n"
		"\t-b/--buffer bufname: pick given buffer for plog control\n"
		"\t--enable/--disable: control platform logging for chosen buffers\n\n"
		"NOTE: to use dlogctl, dynamic dlog config has to be enabled!\n"
		"Define the %s config entry to a *DIRECTORY* path.\n"
		"ALSO: filters won't work for the APPS buffer (whence most spam comes)\n"
		"unless you set `limiter_apply_to_all_buffers` in the config to 1!\n\n"
		, argv[0]
		, __LOG_LIMITER_LIMIT_MAX
		, argv[0]
		, DYNAMIC_CONFIG_CONF_KEY
	);

	return false;
}

static void free_keys_values(struct keys_values *kv)
{
	assert(kv);

	if (kv->keys) {
		for (size_t i = 0; i < kv->n; ++i)
			free(kv->keys[i]);
		free(kv->keys);
	}

	if (kv->values)
		free(kv->values);
}

/**
 * @brief Copy file, except modify some lines
 * @details Copies a file linewise to another, except ensures lines starting with given prefixes have certain follow-up
 * @param[in] in Copy from this file
 * @param[in] out Copy into this file
 * @param[in] kv A set of key-value pairs, where lines which do not match any of the prefixes are copied verbatim, while
 *               lines that do are ignored. For each key where the value is not NULL, a line with the key as the prefix
 *               and the value as the suffix is then appended to the resulting file.
 *
 * @notes Usually the keys will contain a '=' at the end so as to only replace exact matches in a config file,
 *        and keys without '=' at the end should usually have the value NULL to keep the config format valid
 *
 * @example Consider a file with the contents:
 *   FOO=BAR
 *   FOOOO=BAZ
 *   ABC=XYZ
 *   ABCABC=QQQ
 *   ABC=ZYX
 *
 * And a set of kv pairs:
 *   {"FOO", NULL}
 *   {"ABC=", "YYY"}
 *   {"QWER=", "ASDF"}
 *
 * Thus:
 *   FOO matches both FOO=BAR and FOOOO=BAZ. Since the value is NULL, both these lines are removed.
 *   ABC= matches both ABC=XYZ and ABC=ZYX, these lines are not copied. Instead, ABC=YYY is added
 *     behind any lines that were not matched.
 *   ABCABC is not matched by any kv pair, so it is copied verbatim.
 *   The QWER= prefix is not matched by any line, so QWER=ASDF is appended to the end.
 *
 * The resulting file will therefore look like:
 *  ABCABC=QQQ
 *  ABC=YYY
 *  QWER=ASDF
 *
 */
int copy_except(FILE *in, FILE *out, struct keys_values *kv)
{
	__attribute__((cleanup(free_ptr))) size_t *const key_len = calloc(kv->n, sizeof *key_len);
	if (!key_len)
		return -ENOMEM;

	for (size_t i = 0; i < kv->n; ++i)
		key_len[i] = strlen(kv->keys[i]);

	char line[MAX_CONF_KEY_LEN + MAX_CONF_VAL_LEN + 2]; // +2 for delimiter and terminator: "K=V\0"
	while (fgets(line, sizeof line, in)) {
		size_t i = 0;
		while (i < kv->n && strncmp(line, kv->keys[i], key_len[i]))
			++i;
		if (i == kv->n)
			fputs(line, out);
	}

	for (size_t i = 0; i < kv->n; ++i)
		if (kv->values[i])
			fprintf(out, "%s%s\n", kv->keys[i], kv->values[i]);

	return 0;
}

int copy_file_with_exceptions(const char *config_path, struct keys_values *kv)
{
	__attribute__((cleanup(close_FILE))) FILE *const configfile = fopen(config_path, "a+"); // won't actually append anything but this is the only mode that both positions the read steam at the beginning and opens the file if missing
	if (!configfile) {
		ERR("config file %s fopen failed: %m\n", config_path);
		return EXIT_FAILURE;
	}

	__attribute__((cleanup(free_ptr))) char *tempname = NULL;
	int r = asprintf(&tempname, "%sXXXXXX", config_path);
	if (r < 0) {
		tempname = NULL;
		ERR("asprintf tempname failed: %m\n");
		return EXIT_FAILURE;
	}

	/* In older libc versions, umask(077) needed to be called
	 * before mkstemp() to prevent a potential race condition
	 * vulnerability. POSIX-2008 requires mkstemp() to create
	 * the file with rights 600, but our static analysis tools
	 * still complain about this and in this case it was safe
	 * to just do this and silence them. In the general case,
	 * doing this is unsafe as umask() is not thread-aware but
	 * logctl is single-threaded anyway. */
	mode_t old = umask(077);
	const int tempfd = mkstemp(tempname);
	umask(old);
	if (tempfd < 0) {
		ERR("mkstemp failed %m\n");
		return EXIT_FAILURE;
	}

	__attribute__((cleanup(close_FILE))) FILE *const tempfile = fdopen(tempfd, "w");
	if (!tempfile) {
		ERR("tmpfile fdopen failed: %s %m\n", tempname);
		goto remove_temp;
	}

	if (fchmod(tempfd, 0644) < 0) {
		ERR("fchmod failed on %s %m\n", tempname);
		goto remove_temp;
	}

	if (copy_except(configfile, tempfile, kv) < 0) {
		ERR("copy_except failed\n");
		goto remove_temp;
	}

	if (rename(tempname, config_path) < 0) {
		ERR("rename failed: %m\n");
		goto remove_temp;
	}

	return EXIT_SUCCESS;

remove_temp:
	unlink(tempname);

	return EXIT_FAILURE;
}

int handle_set(const struct parsed_params *params, const char *config_path, struct log_config *conf)
{
	__attribute__((cleanup(free_ptr))) char *key;
	if (asprintf(&key, "limiter|%s|%c=", params->tag, params->prio) < 0) {
		key = NULL;
		ERR("asprintf key failed: %m\n");
		return EXIT_FAILURE;
	}
	const char *value = params->value;

	struct keys_values kv = {
		.keys = &key,
		.values = &value,
		.n = 1,
	};

	return copy_file_with_exceptions(config_path, &kv);
}

int handle_clear(const struct parsed_params *params, const char *config_path, struct log_config *conf)
{
	char *key = "limiter|"; // note: no '=' at the end, so as to match all rules
	const char *value = NULL; // removal

	struct keys_values kv = {
		.keys = &key,
		.values = &value,
		.n = 1,
	};

	return copy_file_with_exceptions(config_path, &kv);
}

int handle_plog(const struct parsed_params *params, const char *config_path, struct log_config *conf)
{
	const int buffers_bitset = params->plog_buffers_bitset ?: ((1 << LOG_ID_MAX) - 1);

	int count = 0;
	for (int i = 0; i < LOG_ID_MAX; ++i)
		if (buffers_bitset & (1 << i))
			++count;

	__attribute__((cleanup(free_keys_values))) struct keys_values kv = {
		.keys = calloc(count, sizeof *kv.keys),
		.values = calloc(count, sizeof *kv.values),
		.n = count,
	};
	if (!kv.keys || !kv.values)
		return EXIT_FAILURE;

	count = 0;
	for (int i = 0; i < LOG_ID_MAX; ++i) {
		if (!(buffers_bitset & (1 << i)))
			continue;

		if (asprintf(kv.keys + count, "enable_%s=", log_name_by_id((log_id_t)i)) < 0) {
			kv.keys[count] = NULL;
			ERR("asprintf bufname %d failed: %m\n", i);
			return EXIT_FAILURE;
		}
		kv.values[count] = params->plog_enable ? "1" : "0";
		++count;
	}
	assert(count == kv.n);

	return copy_file_with_exceptions(config_path, &kv);
}

#ifndef UNIT_TEST
static void print_limit(int limit, const char *tag, char prio, bool *shadowed, bool dynamic)
{
	const char *const shadow_str = *shadowed ? "[shadowed] " : "";
	const char *const source_str = dynamic ? " (dynamic)" : " (static)";

	switch (limit) {
	case -1:
		printf("%sNothing set for %s:%c%s\n", shadow_str, tag, prio, source_str);
		break;
	case 0:
		printf("%sDenied for %s:%c%s\n", shadow_str, tag, prio, source_str);
		break;
	case __LOG_LIMITER_LIMIT_MAX + 1:
		printf("%sUnlimited for %s:%c%s\n", shadow_str, tag, prio, source_str);
		break;
	default:
		printf("%s%d logs/minute for %s:%c%s\n", shadow_str, limit, tag, prio, source_str);
		break;
	}

	if (limit != -1)
		*shadowed = true;
}

static void get_limits(const struct parsed_params *params, const char *config_path, struct log_config *conf)
{
	struct limiter_limits lims_static = __log_limiter_get_limits(params->tag, params->prio);

	log_config_read_file(conf, config_path); // not an error on failure - config still valid if missing, static rules apply
	__log_limiter_update(conf);
	struct limiter_limits lims_dynamic = __log_limiter_get_limits(params->tag, params->prio);

	bool shadowed = false;
	if (strcmp(params->tag, "*")) {
		if (params->prio != '*') {
			print_limit(lims_dynamic.tag_and_prio, params->tag, params->prio, &shadowed, true);
			print_limit(lims_static.tag_and_prio,  params->tag, params->prio, &shadowed, false);
		}
		print_limit(lims_dynamic.tag, params->tag, '*', &shadowed, true);
		print_limit(lims_static.tag,  params->tag, '*', &shadowed, false);
	}
	if (params->prio != '*') {
		print_limit(lims_dynamic.prio, "*", params->prio, &shadowed, true);
		print_limit(lims_static.prio,  "*", params->prio, &shadowed, false);
	}
	print_limit(lims_dynamic.global, "*", '*', &shadowed, true);
	print_limit(lims_static.global,  "*", '*', &shadowed, false);
}

struct prio_applies_to {
	char prios[DLOG_PRIO_MAX];
	size_t count;
	struct limiter_limits lims[DLOG_PRIO_MAX];
};

static void print_limits_for_prio(const struct parsed_params *params, const char prio[], struct prio_applies_to *applies, size_t prio_idx, bool *shadowed, int dynamic)
{
	assert(prio_idx < DLOG_PRIO_MAX);

	if (strcmp(params->tag, "*") && prio[prio_idx] != '*') {
		if (applies->lims[prio_idx].tag_and_prio == -1)
			applies->prios[applies->count++] = prio[prio_idx];
		else {
			print_limit(applies->lims[prio_idx].tag_and_prio, params->tag, prio[prio_idx], shadowed, dynamic);
			*shadowed = true;
		}
	} else if (prio[prio_idx] == '*') {
		if (applies->count == 0)
			*shadowed = true;
		else if (applies->count != NELEMS(applies->prios) - 2) {
			printf("[applies to ");
			for (size_t i = 0; i < applies->count; i++)
				printf("%c%c ", applies->prios[i], (i == applies->count-1) ? ']' : ',');
		}
		print_limit(applies->lims[prio_idx].tag, params->tag, '*', shadowed, dynamic);
		*shadowed = true;
	} else if (applies->lims[prio_idx].prio == -1)
		applies->prios[applies->count++] = prio[prio_idx];
	else {
		print_limit(applies->lims[prio_idx].prio, "*", prio[prio_idx], shadowed, dynamic);
		*shadowed = true;
	}
}

static void get_prio_limits(const struct parsed_params *params, const char *config_path, struct log_config *conf)
{
	static const char prio_list[] = {'V', 'D', 'I', 'W', 'E', 'F' , 'S' , '*'};
	struct prio_applies_to applies_dynamic = { .count = 0 };
	struct prio_applies_to applies_static = { .count = 0 };

	for (size_t i = 0; i < NELEMS(prio_list); i++)
		applies_static.lims[i] = __log_limiter_get_limits(params->tag, prio_list[i]);

	log_config_read_file(conf, config_path); // not an error on failure - config still valid if missing, static rules apply
	__log_limiter_update(conf);

	for (size_t i = 0; i < NELEMS(prio_list); i++) {
		applies_dynamic.lims[i] = __log_limiter_get_limits(params->tag, prio_list[i]);
		bool shadowed = false;

		print_limits_for_prio(params, prio_list, &applies_dynamic, i, &shadowed, true);
		print_limits_for_prio(params, prio_list, &applies_static, i, &shadowed, false);
	}
}


int handle_get(const struct parsed_params *params, const char *config_path, struct log_config *conf)
{
	if (!__log_limiter_create(conf)) {
		ERR("error creating limiter\n");
		return EXIT_FAILURE;
	}

	if (params->prio == '\0')
		get_prio_limits(params, config_path, conf);
	else
		get_limits(params, config_path, conf);

	__log_limiter_destroy();
	return EXIT_SUCCESS;
}

int handle_dump(const struct parsed_params *params, const char *config_path, struct log_config *conf)
{
	if (!__log_limiter_create(conf)) {
		ERR("error creating limiter\n");
		return EXIT_FAILURE;
	}
	char buf[1024];

	log_config_read_file(conf, config_path); // not an error on failure - config still valid if missing, static rules apply
	__log_limiter_update(conf);

	struct rule *r = NULL;
	do {
		int ret = __log_limiter_dump_rule(&r, buf, sizeof buf);
		if (ret < 0) {
			ERR("Error dumping rule\n");
			__log_limiter_destroy();
			return EXIT_FAILURE;
		}
		printf("%s\n", buf);
	} while (r);

	printf("\nLogging buffer status:\n");
	const bool plog_default = log_config_get_boolean(conf, "plog", true);
	for (int i = 0; i < LOG_ID_MAX; ++i) {
		char key[MAX_CONF_KEY_LEN];
		const int r = snprintf(key, sizeof key, "enable_%s", log_name_by_id((log_id_t)i));
		if (r < 0)
			continue;
		printf("* %s: %s\n", log_name_by_id((log_id_t)i), log_config_get_boolean(conf, key, i == LOG_ID_APPS ? plog_default : true) ? "ENABLED" : "DISABLED");
	}

	__log_limiter_destroy();
	return EXIT_SUCCESS;
}

int main(int argc, const char **argv)
{
	int exit_code = EXIT_FAILURE;
	struct parsed_params params = {0};
	if (!parse_options(argc, argv, &params, &exit_code))
		return exit_code;

	__attribute__((cleanup(log_config_free))) struct log_config conf = {0};
	const int r = log_config_read(&conf);
	if (r < 0) {
		ERR("error reading config: %d\n", r);
		return EXIT_FAILURE;
	}

	const char *const extra_config_path = log_config_get(&conf, DYNAMIC_CONFIG_CONF_KEY);
	if (!extra_config_path) {
		printf("Dynamic config is disabled (\"%s\" not defined in config)\n", DYNAMIC_CONFIG_CONF_KEY);
		return EXIT_SUCCESS;
	}
	if (extra_config_path[0] != '/') {
		printf("Dynamic config: invalid path in (is \"%s\" but has to be absolute)\n", extra_config_path);
		return EXIT_FAILURE;
	}

	__attribute__ ((cleanup(free_ptr))) char *full_inotify_path = NULL;
	if (asprintf(&full_inotify_path, "%s/%s", extra_config_path, DYNAMIC_CONFIG_FILENAME) < 0) {
		full_inotify_path = NULL;
		ERR("asprintf: no memory");
		return EXIT_FAILURE;
	}

	int (*handles[])(const struct parsed_params *, const char *, struct log_config *) = {
		[ACTION_GET]   = handle_get,
		[ACTION_SET]   = handle_set,
		[ACTION_DUMP]  = handle_dump,
		[ACTION_PLOG]  = handle_plog,
		[ACTION_CLEAR] = handle_clear,
	};
	exit_code = handles[params.action](&params, full_inotify_path, &conf);

	return exit_code;
}
#endif // !UNIT_TEST
