/*
 * Copyright (c) 2018, Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include "log_storage.h"

#include <ptrs_list.h>
#include <stdlib.h>
#include <stdbool.h>
#include <assert.h>
#include <queued_entry_timestamp.h>

/* This is implementation of log storage.
 *
 * log_storage is a structure which represents the log storage.
 * Logs can be added to storage, but can't be selectively removed. The log storage
 * manages its storage size by itself. The maximum size is set at the creation time.
 *
 * Log storage handles following operations:
 * - create
 * - release
 * - getting maximum storage size
 * - adding a new log reader
 * - adding a new log entry
 * - clearing the storage
 *
 *
 * log_storage_entry: a single log.
 *
 * Logs are stored in a bidirectional list. Each entry is stored as a log_storage_entry.
 * Adding a new entry makes the list grow by one element. The entry is pushed into the list
 * up to the place, where all following logs are later, and preceding logs are earlier.
 *
 *
 * log_storage_reader: a client.
 *
 * Readers are entities that want to read logs. Each reader is free to read the logs in its own pace.
 * Each reader knows where it finished reading and may resume reading starting from the next unread log.
 *
 * A user may choose if the reader is a continuous reader or if it should stop after reading all the logs
 * that are available at the moment of creation of the reader.
 *
 * Each reader tracks its current position by remembering last entry it has read.
 *
 * Each entry has its list of readers. It means that it is the last entry those readers have read.
 *
 * Log storage reader handles following operations:
 * - release
 * - checking if there are new log entries available
 * - getting new log entry
 * - getting how much (unread) data (in bytes) is ready to read in the storage.
 *
 *       :----------:    :--------:    :--------:    :--------:
 *       | storage  | -> | reader | -> | reader | -> | reader | (readers that have not started reading yet)
 *       :----------|    :--------:    :--------:    :--------:
 *            |
 *            v
 *    -> :-----------| -> :-----------| -> :-----------| -> :-----------| ->
 * ...   | log entry |    | log entry |    | log entry |    | log entry |   ...
 *    <- :-----------: <- :-----------: <- :-----------: <- :-----------: <-
 *         ^    |           ^    |                            ^    |
 *         |    v           |    v                            |    v
 *         | :--------:     | :--------:                      | :--------:
 *         \-| reader |     \-| reader |                      \-| reader |
 *         | :--------:       :--------:                      | :--------:
 *         |    |                                             |    |
 *         |    v                                             |    v
 *         | :--------:                                       | :--------:
 *         \-| reader |                                       \-| reader |
 *         | :--------:                                         :--------:
 *         |    |                                                   These two readers
 *         |    v                                                   have already read all four logs.
 *         | :--------:
 *         \-| reader |
 *           :--------:
 *
 *
 * When a pushed log passes an entry with active readers, they are notified about the log by a registered
 * callback. The callback is also called when the earliest log entries are freed from the storage, and there
 * are some readers that did not read them yet.
 *
 * Readers support checking how much data is already buffered ahead, ready to read.
 * This is implemented by two 64-bit counters in the storage, and one counter in each reader.
 * Storage counters track total size of the log entries:
 *  - counter_begin - counts how much data was already freed from the storage; in other words it points
 *                    to the beginning of the log entries "size window" (in contrast to time window)
 *  - counter_end - counts how much data was inserted to the storage.
 * Current size of the data in the storage is counter_end - counter_begin.
 * Each reader has assigned a counter which points to a specific log entry. When a reader is created
 * the counter points to counter_begin. The counter advances with each log entry reading.
 * When a reader is asked about how much data is ahead, it just needs to subtract its counter
 * from counter_end.
 * In other words - storage stores some part of a log stream, and reader reads from a specific point.
 *
 * A bit tricky part is if a late log arrives and it is pushed over readers.
 * counter_end has to change, because the log stream is increased. However, it is sufficient
 * to artificially "move" the reader's stream pointer at the moment of log passing by.
 *
 */

typedef struct log_storage_entry {
	struct log_storage_entry *next;
	struct log_storage_entry *prev;
	list_head readers;
	// the field below must be the very last field in the struct, because it has variable size
	struct logger_entry entry;
} log_storage_entry;

struct log_storage {
	unsigned capacity;      // maximum size of stored logs
	enum sorting_order sort_by; // stored logs sorting order
	uint64_t counter_begin; // "pointer" to the first stored log in the log stream, see description above
	uint64_t counter_end;   // "pointer" to the end of the stored log stream
	struct log_storage_entry *entries; // the log entries are stored here
	list_head waiting_readers; // these are the readers that did not read anything from the current entries
};

struct log_storage_reader {
	log_storage *storage;

	// callback which is called when the last chance
	// for handling a specific log occurs
	log_storage_reader_callback callback;

	void *user_data;                      // user data passed to the callback
	log_storage_entry *current;           // pointer to last read log entry

	// 'final' is for dumpers - it is a pointer to the last log that should be handled
	// it is NULL for continuous readers
	const log_storage_entry *final;

	uint64_t counter;                     // current position of the reader in the log stream
};

log_storage *log_storage_create(unsigned capacity, enum sorting_order sort_by)
{
	log_storage *storage = malloc(sizeof(*storage));
	if (NULL != storage) {
		storage->capacity = capacity;
		storage->sort_by = sort_by;
		storage->counter_begin = 0;
		storage->counter_end = 0;
		storage->entries = NULL;
		storage->waiting_readers = NULL;
	}

	return storage;
}

static void log_storage_reader_set_storage(void *reader_, void *storage)
{
	assert(reader_);
	log_storage_reader *reader = reader_;
	reader->storage = storage;
}

void log_storage_free(log_storage *storage)
{
	assert(storage);
	log_storage_clear(storage);
	// disconnect all the remaining readers
	list_foreach(storage->waiting_readers, NULL, log_storage_reader_set_storage);
	free(storage);
}

unsigned log_storage_get_capacity(log_storage *storage)
{
	assert(storage);
	return storage->capacity;
}

log_storage_reader *log_storage_new_reader(log_storage *storage,
		bool dumping, log_storage_reader_callback callback, void *user_data)
{
	assert(storage);

	log_storage_reader *reader = malloc(sizeof(*reader));
	if (NULL != reader) {
		reader->storage = storage;
		reader->callback = callback;
		reader->user_data = user_data;
		reader->current = NULL;
		reader->counter = storage->counter_begin;
		reader->final = NULL;

		if (dumping) {
			// dumping - get all the logs up to this moment
			if (NULL != storage->entries)
				reader->final = storage->entries->prev;
			else
			// but in case there are no logs - we have nothing to do
				reader->storage = NULL;
		}

		if (NULL != reader->storage)
			list_add(&storage->waiting_readers, reader);
	}
	return reader;
}

static void reader_execute_callback(void *reader_, void *lse_)
{
	log_storage_reader *reader = reader_;
	const log_storage_entry *lse = lse_;
	// executing callback means - a lost log just passes by us
	// so we need to move the stream pointer forward
	reader->counter += lse->entry.len;
	reader->callback(&lse->entry, reader->user_data);
}

static void log_storage_entry_readers_notify(list_head readers_list, const log_storage_entry *lse)
{
	// for each reader from input list execute reader's callback with the log entry as argument
	list_foreach(readers_list, (void *)lse, reader_execute_callback);
}

static void reader_move_to_entry(void *reader_, void *lse_)
{
	assert(reader_);
	log_storage_reader *reader = reader_;
	log_storage_entry *lse = lse_;
	if (NULL != lse)
		list_add(&lse->readers, reader); // LCOV_EXCL_LINE : never actually used this way and not exposed
	else
		list_add(&reader->storage->waiting_readers, reader);
	reader->current = lse;
}

static void readers_move_all_to_another_entry(list_head *list_from, log_storage_entry *lse)
{
	assert(list_from);
	list_foreach(*list_from, lse, reader_move_to_entry);
	list_clear(list_from);
}

static bool log_storage_reader_check_final(void *reader_, void *lse_)
{
	assert(reader_);
	assert(lse_);

	log_storage_reader *reader = reader_;
	const log_storage_entry *lse = lse_;

	if (lse == reader->final) {
		// dumping/one-shot reader arrived at the end, disconnect it
		reader->storage = NULL;
	}
	return reader->final == lse;
}

static void log_storage_remove_the_earliest_log(log_storage *storage)
{
	assert(storage);

	struct log_storage_entry *free_me = storage->entries;

	// notify waiting readers about the log we are going to remove
	log_storage_entry_readers_notify(storage->waiting_readers, free_me);
	// we don't notify those in free_me->readers because they have already processed this log
	// move readers from this log to waiting list
	readers_move_all_to_another_entry(&free_me->readers, NULL);
	// for readers with "final" on the removed log - disconnect them at all
	list_remove_if(&storage->waiting_readers, free_me, log_storage_reader_check_final);

	// disconnect the entry from the storage
	storage->entries = storage->entries->next;
	storage->entries->prev = free_me->prev;
	storage->entries->prev->next = storage->entries;
	if (storage->entries == free_me)
		storage->entries = NULL;

	// move start position indicator
	storage->counter_begin += free_me->entry.len;

	free(free_me);
}

bool log_storage_add_new_entry(log_storage *storage, const struct logger_entry *le)
{
	assert(storage);
	assert(le);

	log_storage_entry *log_storage_entry = malloc(sizeof(*log_storage_entry) -
			sizeof(log_storage_entry->entry) + le->len);
	if (NULL == log_storage_entry)
		return false;

	log_storage_entry->readers = NULL;

	memcpy(&log_storage_entry->entry, le, le->len);
	if (NULL == storage->entries) {
		// simple insert to empty storage
		storage->entries = log_storage_entry;
		log_storage_entry->next = log_storage_entry;
		log_storage_entry->prev = log_storage_entry;
	} else {
		struct log_storage_entry *iterator = storage->entries;
		bool front = true;
		// find the log after which we should insert the new log
		do {
			iterator = iterator->prev;
			front = log_entry_is_earlier(storage->sort_by, le, &iterator->entry);
			if (front) {
				// In this moment we put the log before existing logs.
				// It may happen that some readers are already at the iterator,
				// so we need to notify them about late log.
				log_storage_entry_readers_notify(iterator->readers, log_storage_entry);
			}
		} while (front && iterator != storage->entries);
		if (front) {
			// the new log is the earliest log, push it to the front
			log_storage_entry->next = iterator;
			log_storage_entry->prev = iterator->prev;
			iterator->prev->next = log_storage_entry;
			iterator->prev = log_storage_entry;
			storage->entries = log_storage_entry;
		} else {
			// the new log goes after the iterator
			log_storage_entry->next = iterator->next;
			log_storage_entry->prev = iterator;
			iterator->next->prev = log_storage_entry;
			iterator->next = log_storage_entry;
		}
	}

	storage->counter_end += le->len;
	// if we have reached the size limit for the buffer - free some space
	while (storage->counter_end - storage->counter_begin > storage->capacity)
		log_storage_remove_the_earliest_log(storage);
	return true;
}

void log_storage_clear(log_storage *storage)
{
	assert(storage);
	while (NULL != storage->entries)
		log_storage_remove_the_earliest_log(storage);
}

enum sorting_order log_storage_get_sorting_order(const log_storage *storage)
{
	assert(storage);
	return storage->sort_by;
}

void log_storage_release_reader(log_storage_reader *reader)
{
	assert(reader);

	if (reader->current)
		list_remove(&reader->current->readers, reader);
	else if (reader->storage)
		list_remove(&reader->storage->waiting_readers, reader);
	free(reader);
}

static bool log_storage_reader_entry_is_not_last(const log_storage_reader *reader, const log_storage_entry *entry)
{
	assert(reader);
	assert(reader->storage);

	return NULL != entry && entry->next != reader->storage->entries && // next entry is first entry
		!(NULL != reader->final && entry == reader->final); // not (dumping reader on last entry)
}

bool log_storage_reader_is_new_entry_available(const log_storage_reader *reader)
{
	assert(reader);

	// storage is NULL if a dumping/one-shot reader has finished reading
	if (NULL == reader->storage)
		return false;

	return log_storage_reader_entry_is_not_last(reader, reader->current) ||
		(NULL == reader->current && NULL != reader->storage->entries);
}

const struct logger_entry *log_storage_reader_get_new_entry(log_storage_reader *reader)
{
	assert(reader);

	const struct logger_entry *le = NULL;

	// storage is NULL if a dumping/one-shot reader has finished reading
	if (NULL == reader->storage)
		return NULL;

	if (NULL == reader->current) {
		if (NULL != reader->storage->entries) {
			// there are some logs in the buffer, let's get them
			le = &reader->storage->entries->entry;
			reader->current = reader->storage->entries;
			list_remove(&reader->storage->waiting_readers, reader);
			list_add(&reader->current->readers, reader);
		}
	} else if (log_storage_reader_entry_is_not_last(reader, reader->current)) {
		// there are more logs in the buffer, let's get them
		list_remove(&reader->current->readers, reader);
		reader->current = reader->current->next;
		le = &reader->current->entry;
		if (!log_storage_reader_check_final(reader, reader->current))
			list_add(&reader->current->readers, reader);
	}

	if (NULL != le)
		reader->counter += le->len;

	return le;
}

uint64_t log_storage_reader_get_ready_bytes(const log_storage_reader *reader)
{
	assert(reader);

	if (NULL == reader->storage)
		return 0;
	return reader->storage->counter_end - reader->counter;
}

const struct log_storage *log_storage_reader_get_storage(const log_storage_reader *reader)
{
	assert(reader);
	return reader->storage;
}
