#pragma once

/*
 * Copyright (c) 2018, Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <queued_entry.h>

#include <stdbool.h>

/**
 * @file
 * @brief This file is the header file for log storage used by logger.
 * @author Adrian Szyndela <adrian.s@samsung.com>
 */

/**
 * @brief This is a storage.
 *
 * A storage stores log entries. They are added to a storage by a logger server.
 * Log entries can be read from a storage by readers.
 * Specific log entries cannot be removed from a storage, it removes logs when needed.
 * All log entries can be removed at once.
 * Read log_storage.c for the internal structure description.
 */
struct log_storage;
typedef struct log_storage log_storage;

/**
 * @brief Creates a storage instance.
 * @param[in] capacity Maximum capacity of the created instance.
 * @return The created storage instance or NULL in case of lack of memory.
 */
log_storage *log_storage_create(unsigned capacity, enum sorting_order sort_by);

/**
 * @brief Frees the storage and all the resources associated with it.
 * @param[in] storage Storage that will be freed.
 *
 * This function frees all the stored log entries, but does not free readers.
 * However it is safe to call before releasing readers.
 */
void log_storage_free(log_storage *storage);

/**
 * @brief Returns the capacity of the storage, as specified on the storage creation.
 * @return The capacity, in bytes.
 */
unsigned log_storage_get_capacity(log_storage *storage);

/**
 * @brief Returns the sorting order of the storage, as specified on the storage creation.
 * @return Log storage's sorting order.
 */
enum sorting_order log_storage_get_sorting_order(const log_storage *storage);

/**
 * @brief This function adds a new entry to the storage.
 * @param[in] storage The log entry will be added to this storage.
 * @param[in] le The log entry which will be added.
 * @return true on success, false on failure - on lack of memory.
 */
bool log_storage_add_new_entry(log_storage *storage, const struct logger_entry *le);

/**
 * @brief This function makes all the log entries disappear. They vanish. Completely.
 * @param[in] storage The storage to clear aka the Augean stables.
 */
void log_storage_clear(log_storage *storage);

/**
 * @brief This is a reader.
 *
 * A reader reads log entries. It can check if there is an entry available for getting
 * and retrieve available entries. It can also check how much data is already buffered
 * and ready to read.
 */
struct log_storage_reader;
typedef struct log_storage_reader log_storage_reader;

/**
 * @brief A type for callbacks that are called when there is last chance for a reader
 *        to handle a log entry.
 */
typedef void (*log_storage_reader_callback)(const struct logger_entry *le, void *user_data);

/**
 * @brief This function creates a new reader.
 * @param[in] storage A storage to read from.
 * @param[in] dumping If the reader is dumping/one-shot type of reader.
 * @param[in] callback A callback to call when there is last chance for a reader
 *                     to handle a log entry.
 * @param[in] user_data User data which will be passed to the callback.
 *
 * There are two types of readers:
 *  - continuous - it reads all the available logs and stays around for more
 *  - dumping/one-shot - it reads all the logs available at the moment of the reader's creation
 *                       but no more
 */
log_storage_reader *log_storage_new_reader(log_storage *storage,
		bool dumping, log_storage_reader_callback callback, void *user_data);

/**
 * @brief Free all resources allocated for a reader.
 * @param[in] reader The reader which will be freed.
 * @note This is safe to call this after freeing the storage. However, it is not recommended.
 */
void log_storage_release_reader(log_storage_reader *reader);

/**
 * @brief This function checks if there are new, unread log entries available in the storage associated
 *        with a reader.
 * @param[in] reader The reader for which the check is performed.
 * @return true if there are unread entries, false in all the other cases.
 */
bool log_storage_reader_is_new_entry_available(const log_storage_reader *reader);

/**
 * @brief This function is responsible for retrieving next unread log entry from the storage associated
 *        with a reader.
 * @param[in] reader The reader which will be used for reading.
 * @return A new, freshly retrieved, shiny log entry.
 * @note Enjoy the entry, but do not free it. It is not yours. Do not store the pointer (for a long time),
 *       because this log entry may be deleted by the storage without warning. Precisely, when you put
 *       another log entry into the storage.
 */
const struct logger_entry *log_storage_reader_get_new_entry(log_storage_reader *reader);


/**
 * @brief Returns the storage that is the owner of a reader.
 * @param[in] reader The reader which is desired to return it's storage.
 * @return The pointer to a log storage.
 */
const struct log_storage *log_storage_reader_get_storage(const log_storage_reader *reader);

/**
 * @brief This function returns the size of buffered, ready to read log entries for a reader.
 * @param[in] reader The reader which will be used for checking.
 * @return The size of buffered, ready to read log entries in bytes.
 * @note Please do not call this for dumping/one-shot readers. Please. It will still return
 *       size of buffered data instead of size of data that can be eventually read by such reader.
 */
uint64_t log_storage_reader_get_ready_bytes(const log_storage_reader *reader);
