/*
 * Copyright (c) 2016, Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include "logger_internal.h"

/**
 * @addtogroup DLOG_IMPLEMENTATION
 * @{
 * @defgroup DLOG_LOGGER Logger
 * @brief Logger daemon
 * @details The logger is the core component of the logging framework. It is responsible for collecting, processing and exposing logs.
 * @{
 */

// function prototypes
static int service_writer_socket(struct logger* server, struct writer* wr, struct epoll_event* event);
static int service_writer_pipe(struct logger* server, struct writer* wr, struct epoll_event* event);
int service_writer_kmsg(struct logger* server, struct writer* wr, struct epoll_event* event);
int service_writer_syslog(struct logger* server, struct writer* wr, struct epoll_event* event);
static int service_writer_handle_req_ctrl(struct logger *server, struct writer *wr, struct dlog_control_msg *msg);
static int service_writer_handle_req_pipe(struct logger *server, struct writer *wr, struct dlog_control_msg *msg);
void dispatch_event_writer(struct logger *server, struct epoll_event *event);
void dispatch_event_sock(struct logger *server, struct epoll_event *event);
static void dispatch_event_reader(struct logger *server, struct epoll_event *event);
void reader_free(struct reader* reader);
static void logger_free(struct logger* l);
static int parse_args(int argc, char **argv, struct buf_params *b);
int socket_initialize(struct sock_data *sock, struct log_buffer *buffer, service_socket_t service_socket, struct socket_config_data *data);
static void help();

/** global state when logger is not interrupted by any handled signals */
static volatile sig_atomic_t g_logger_run = 1;

struct {
	int (*reader_init)(struct reader *reader, const char *buf_name, struct logger *server);
	char logger_devices[LOG_ID_MAX][MAX_CONF_VAL_LEN];
	int logger_device_throttling[LOG_ID_MAX];
} g_backend = {
	.reader_init = NULL
};

/**
 * @brief Parse permissions
 * @details Parse a string representation of permissions to the internal integral one
 * @param[in] str The string representation
 * @return The integer representation
 */
int parse_permissions(const char * str)
{
	int ret, parsed;
	char * parse_safety;

	if (!str)
		return S_IWUSR | S_IWGRP | S_IWOTH; // 0222: everyone can write

	parsed = strtol(str, & parse_safety, 8); // note, rights are octal
	if (parse_safety != (str + strlen(str)))
		return 0;

	ret = 0;

	// note that R and X are pretty useless, only W makes sense
	if (parsed & 00001) ret |= S_IXOTH;
	if (parsed & 00002) ret |= S_IWOTH;
	if (parsed & 00004) ret |= S_IROTH;
	if (parsed & 00010) ret |= S_IXGRP;
	if (parsed & 00020) ret |= S_IWGRP;
	if (parsed & 00040) ret |= S_IRGRP;
	if (parsed & 00100) ret |= S_IXUSR;
	if (parsed & 00200) ret |= S_IWUSR;
	if (parsed & 00400) ret |= S_IRUSR;
	if (parsed & 01000) ret |= S_ISVTX;
	if (parsed & 02000) ret |= S_ISGID;
	if (parsed & 04000) ret |= S_ISUID;

	if (parsed & ~07777)
		printf("Warning: useless bits in permissions %s!", str);

	return ret;
}

/**
 * @brief Get numerical ids for specified user and group
 * @details Converts user and group strings to proper uid and gid identifiers
 * @param[in] user The new user
 * @param[in] group The new group
 * @param[out] uid The uid for user
 * @param[out] gid The numerical gid for group
 * @return 0 on success, else -errno
 */
int usergr2id(const char * user, const char * group, uid_t *uid, gid_t *gid)
{
	struct passwd pwd, *ppwd;
	struct group  grp, *pgrp;
	int bufsize = sysconf(_SC_GETPW_R_SIZE_MAX);

	assert(user);
	assert(group);

	if (bufsize == -1)
		bufsize = 1024;

	char *const buffer = (char *)alloca(bufsize);

	if (getgrnam_r(user, &grp, buffer, bufsize, &pgrp))
		return -errno;
	if (!pgrp)
		return -ENOENT;

	if (getpwnam_r(group, &pwd, buffer, bufsize, &ppwd))
		return -errno;
	if (!ppwd)
		return -ENOENT;

	*uid = pwd.pw_uid;
	*gid = grp.gr_gid;

	return 0;
}

/**
 * @brief Reset privileges
 * @details Resets privileges to those specified at build-time
 * @return 0 on success, else -errno
 */
int reset_self_privileges()
{
	uid_t uid;
	gid_t gid;
	int r;

	r = usergr2id(DLOG_SERVER_USER, DLOG_SERVER_GROUP, &uid, &gid);
	if (r < 0)
		return r;

	if (getegid() != gid && setgid(gid) < 0)
		return -errno;
	if (geteuid() != uid && setuid(uid) < 0)
		return -errno; // should never happen

	return 0;
}

/**
 * @brief Create a bound socket
 * @details Creates a socket of given type under the given path
 * @param[in] path The path to the socket
 * @param[in] type Socket type (SOCK_DGRAM or SOCK_STREAM) with flags
 * @return The socket FD, or negative errno
 */
int bind_fd_create(const char* path, int type)
{
	struct sockaddr_un server_addr;
	int sd;

	sd = socket(AF_UNIX, type, 0);
	if (sd == -1)
		return -errno;

	memset(&server_addr, 0, sizeof(server_addr));
	server_addr.sun_family = AF_UNIX;
	strncpy(server_addr.sun_path, path, sizeof(server_addr.sun_path)-1);
	unlink(server_addr.sun_path);

	if (bind(sd, (struct sockaddr*)&server_addr, sizeof(server_addr)) == -1)
		goto failure;

	return sd;

failure:
	close(sd);
	return -errno;
}

/**
 * @brief Create a listening socket
 * @details Creates a socket with given permissions under the given path
 * @param[in] path The path to the socket
 * @param[in] permissions File permissions (the internal representation)
 * @return The socket FD, or negative errno
 */
int listen_fd_create(const char* path, int permissions)
{
	int sd = bind_fd_create(path, SOCK_STREAM);
	if (sd < 0)
		return sd;

	if (permissions)
		if (chmod(path, permissions) < 0) // ideally, fchmod would be used, but that does not work
			goto failure;

	if (listen(sd, MAX_CONNECTION_Q) < 0)
		goto failure;

	return sd;

failure:
	close(sd);
	return -errno;
}

/**
 * @brief Add given FD entity to logger event notification
 * @details Uses epoll mechanism for notification of events
 * @param[in] logger server owning epoll main loop
 * @param[in] fd_entity file descriptor entity
 * @return 0 on success, -errno on failure
 */
int add_fd_entity(struct logger *logger, struct fd_entity *fd_entity)
{
	assert(logger);
	assert(fd_entity);

	int r = epoll_ctl(logger->epollfd, EPOLL_CTL_ADD, fd_entity->fd, &fd_entity->event);
	if (r < 0)
		return -errno;

	++logger->epollcnt;
	return 0;
}

/**
 * @brief Modify given FD entity to logger event notification
 * @details Uses epoll mechanism for notification of events
 * @param[in] logger server owning epoll main loop
 * @param[in] fd_entity file descriptor entity
 * @param[in] wanted_mask wanted event mask
 * @return 0 on success, -errno on failure
 */
int modify_fd_entity(const struct logger *logger, struct fd_entity *fd_entity, int wanted_mask)
{
	assert(logger);
	assert(fd_entity);

	if (fd_entity->fd < 0)
		return -EINVAL;

	uint32_t pre = fd_entity->event.events;
	if (pre == wanted_mask)
		return 0;
	fd_entity->event.events = wanted_mask;

	if (epoll_ctl(logger->epollfd, EPOLL_CTL_MOD, fd_entity->fd, &fd_entity->event) < 0) {
		fd_entity->event.events = pre;
		return -errno;
	}
	return 0;
}

/**
 * @brief Remove given FD entity from logger event notification
 * @details Uses epoll mechanism for notification of events
 * @param[in] logger server owning epoll main loop
 * @param[in] fd_entity file descriptor entity
 * @return 0 on success, else -errno
 */
int remove_fd_entity(struct logger *logger, struct fd_entity *fd_entity)
{
	assert(logger);
	assert(logger->epollcnt > 0);
	assert(fd_entity);

	int r = epoll_ctl(logger->epollfd, EPOLL_CTL_DEL, fd_entity->fd, NULL);
	if (r < 0)
		return -errno;

	--logger->epollcnt;
	return 0;
}

/**
 * @brief Init fd entity to call given function with user data on event
 * @details Initializes dispatch_event callback and user_data pointer
 * @param[in] fd_entity file descriptor entity to initialize
 * @param[in] dispatch_event callback function to dispatch an event to
 * @param[in] user_data pointer to data to place in event structure when dispatch event is called
 */
void init_fd_entity(struct fd_entity *fd_entity, dispatch_event_t dispatch_event, void *user_data)
{
	assert(fd_entity);
	*fd_entity = (struct fd_entity) {
		.dispatch_event = dispatch_event,
		.event = (struct epoll_event) {
			.data = (epoll_data_t) { .ptr = user_data },
			.events = 0
		},
		.fd = -1
	};
}

/**
 * @brief Set file descriptor to epoll entity for ready to read events
 * @details Initializes fd and events mask in epoll entity
 * @param[in] fd_entity file descriptor entity to initialize
 * @param[in] fd file descriptor to set the entity with
 */
void set_read_fd_entity(struct fd_entity *fd_entity, int fd)
{
	assert(fd_entity);
	assert(fd >= 0);
	fd_entity->event.events = EPOLLIN;
	fd_entity->fd = fd;
}

/**
 * @brief Set file descriptor to epoll entity for ready to write events
 * @details Initializes fd and events mask in epoll entity
 * @param[in] fd_entity file descriptor entity to initialize
 * @param[in] fd file descriptor to set the entity with
 */
void set_write_fd_entity(struct fd_entity *fd_entity, int fd)
{
	assert(fd_entity);
	assert(fd >= 0);

	fd_entity->event.events = EPOLLHUP; /* not EPOLLOUT - readers start dormant,
	                                     * get written to in atomic-like chunks,
	                                     * and only temporarily use EPOLLOUT if
	                                     * their pipe gets clogged mid-chunk */
	fd_entity->fd = fd;
}

/**
 * @brief Create a writer
 * @details Create a writer structure
 * @param[out] writer The newly-allocated writer
 * @param[in] fd The FD writer initially belongs to
 * @param[in] log_buffer The logging buffer
 * @param[in] service_writer callback function for handling new data on writer descriptor FD
 * @return 0 on success, -ENOMEM when out of memory
 */
int writer_create(struct writer **writer, int fd, struct log_buffer *log_buffer,
						 service_writer_t service_writer, service_socket_t service_socket)
{
	struct writer *w = (struct writer *)calloc(1, sizeof(struct writer));
	if (!w)
		return -ENOMEM;

	init_fd_entity(&w->fd_entity, dispatch_event_writer, w);
	set_read_fd_entity(&w->fd_entity, fd);
	w->buf_ptr = log_buffer;
	w->readed = 0;
	w->service_writer = service_writer;
	w->service_socket = service_socket;

	*writer = w;
	return 0;
}

/**
 * @brief Dummy function placeholder that should not be called at all
 * @details Use this function as a service socket callback for writers that handle other resources than sockets
 *          This function is temporary and only for testing purposes
 * @param[in] server Unused
 * @param[in] wr Unused
 * @param[in] msg Unused
 * @return Will not return - just raises SIGABRT signal
 *
 * @todo Remove this function after unification of evant handling functionality of writers
 */
int service_socket_dummy(struct logger *server, struct writer *wr, struct dlog_control_msg *msg) // LCOV_EXCL_START
{
	assert(0 && "This function should not be called");
	return -1;
} // LCOV_EXCL_STOP

/**
 * @brief Create kmsg writer
 * @details Create the structure responsible for handling writing data from /dev/kmsg to given logging buffer
 * @param[out] writer The newly-allocated writer
 * @param[in] log_buffer The logging buffer
 * @return 0 on success, -errno on failure
 */
int create_kmsg_writer(struct writer ** writer, struct log_buffer *log_buffer)
{
	int fd = open("/dev/kmsg", O_RDONLY | O_NONBLOCK);
	if (fd < 0)
		return -errno;

	int ret = writer_create(writer, fd, log_buffer, service_writer_kmsg, service_socket_dummy);
	if (ret != 0) {
		close(fd);
	}
	return ret;
}

int systemd_sock_get(const char *path, int type)
{
	int sock_fd = -1;
	int n = sd_listen_fds(0);
	if (n > 0) {
		int i;
		for (i = SD_LISTEN_FDS_START; sock_fd < 0 && i < SD_LISTEN_FDS_START + n; i++)
			if (sd_is_socket_unix(i, type, -1, path, 0) > 0)
				sock_fd = i;
	}
	return sock_fd;
}

/* TODO: move the path to the configuration */
static const char dev_log_sock[] = "/run/dlog/dev-log";
static const char dev_log_link[] = "/dev/log";

int dev_log_sock_get()
{
	return systemd_sock_get(dev_log_sock, SOCK_DGRAM);
}

int dev_log_sock_create()
{
	int fd = bind_fd_create(dev_log_sock, SOCK_DGRAM);
	if (fd >= 0) {
		unlink(dev_log_link);
		if (symlink(dev_log_sock, dev_log_link) == -1) {
			close(fd);
			fd = -errno;
		}
	}
	return fd;
}

/**
 * @brief Create syslog writer
 * @details Create the structure responsible for handling writing data from syslog to given logging buffer
 * @param[out] writer The newly-allocated writer
 * @param[in] log_buffer The logging buffer
 * @return 0 on success, -errno on failure
 */
int create_syslog_writer(struct writer ** writer, struct log_buffer *log_buffer)
{
	int fd = dev_log_sock_get();
	if (fd < 0)
		fd = dev_log_sock_create();
	if (fd < 0)
		return -errno;

	int err = fd_set_flags(fd, O_NONBLOCK);
	if (err < 0) {
		close(fd);
		return err;
	}

	int ret = writer_create(writer, fd, log_buffer, service_writer_syslog, service_socket_dummy);
	if (ret != 0) {
		close(fd);
	}
	return ret;
}

/**
 * @brief Close socket that was initialized using socket_initialize function
 * @details Closes socket entity fd
 * @param[in] sock socet data to close
 */
void socket_close(struct sock_data *sock)
{
	assert(sock);
	if (sock->fd_entity.fd >= 0)
		close(sock->fd_entity.fd);
}

/**
 * @brief Create buffer
 * @details Allocate a buffer structure
 * @param[out] lb The newly-allocated buffer
 * @param[in] buf_id The buffer ID
 * @param[in] data Buffer config data
 * @return 0 on success, -errno on failure
 */
static int buffer_create(struct log_buffer **log_buffer, log_id_t buf_id, struct buffer_config_data *data)
{
	assert(data);
	struct log_buffer *lb = calloc(1, sizeof(*lb));

	if (!lb)
		return -ENOMEM;

	lb->id = buf_id;
	lb->log_storage = log_storage_create(data->size, data->sort_by);
	if (!lb->log_storage) {
		free(lb);
		return -ENOMEM;
	}

	int r;
	r = socket_initialize(&lb->sock_ctl, lb, service_writer_handle_req_ctrl, &data->ctl_socket);
	if (r < 0) {
		log_storage_free(lb->log_storage);
		free(lb);
		return r;
	}

	r = socket_initialize(&lb->sock_wr, lb, service_writer_handle_req_pipe, &data->write_socket);
	if (r < 0) {
		socket_close(&lb->sock_ctl);
		log_storage_free(lb->log_storage);
		free(lb);
		return r;
	}

	*log_buffer = lb;
	return 0;
}

/**
 * @brief Free reader given in ptr
 * @details Use this function when deleting all readers in the list
 * @param[in] ptr pointer to the reader
 * @param[in] user_data pointer to the logger to remove readers waiting for an event
 * @return always true (the reader is to be removed)
 */
static bool cond_reader_free(void *ptr, void *user_data)
{
	struct reader *reader = (struct reader *)ptr;
	struct logger *logger = (struct logger *)user_data;
	assert(reader);
	if (reader->fd_entity.fd >= 0)
		remove_fd_entity(logger, &reader->fd_entity);
	reader_free(reader);
	return true;
}

/**
 * @brief Free buffer
 * @details Deallocate a buffer
 * @param[in] buffer The buffer to deallocate
 * @param[in] logger The server to remove buffer readers from fd loop
 */
static void buffer_free(struct log_buffer *buffer, struct logger *logger)
{
	assert(buffer);

	list_remove_if(&buffer->readers, logger, cond_reader_free);

	socket_close(&buffer->sock_ctl);
	socket_close(&buffer->sock_wr);

	log_storage_free(buffer->log_storage);

	free(buffer);
}

uint64_t reader_buffered_space(const struct reader *reader)
{
	assert(reader);
	assert(reader->buf_ptr);
	return log_storage_reader_get_ready_bytes(reader->log_storage_reader);
}

int reader_is_bufferable(const struct reader *reader)
{
	assert(reader);
	return reader->buf_ptr && reader->file.path != NULL;
}

int reader_ms_since(const struct reader *reader, struct timespec *ts)
{
	return (ts->tv_sec - reader->last_read_time.tv_sec) * 1000 + (ts->tv_nsec - reader->last_read_time.tv_nsec) / 1000000;
}

int reader_should_buffer(struct reader *reader, const struct buf_params *buf_params, struct timespec now)
{
	assert(reader);
	assert(buf_params);

	if (!reader_is_bufferable(reader))
		return 0;

	if (reader_buffered_space(reader) < (uint64_t)buf_params->bytes && reader_ms_since(reader, &now) < (buf_params->time * 1000))
		return 1;

	reader->last_read_time = now;
	return 0;
}

/**
 * @brief Append to buffer
 * @details Appends an entry to the buffer
 * @param[in] s The logger server
 * @param[in] entry The entry to append
 * @param[in] b The buffer whither to append
 */
int buffer_append(const struct logger_entry *entry, struct log_buffer *b)
{
	if (!log_storage_add_new_entry(b->log_storage, entry))
		return -ENOMEM;
	return 0;
}

/**
 * @brief FD limit handler
 * @details Checks whether the FD limit was reached and leaves logs about it if so
 * @param[in] server The logger server
 * @param[in] err errno from the call that failed to create an FD
 */
static void check_if_fd_limit_reached(struct logger *server, int err)
{
	assert(server);

	if (err != ENFILE && err != EMFILE)
		return;

	printf("ERROR: dlog_logger fd limit reached!\n");

	/* pick one representative buffer to send the log to;
	 * no point spamming all the buffers, especially since
	 * default dlogutil invocations display 3 buffers so
	 * it would appear multiple times */
	struct log_buffer *const buf = server->buffers[LOG_ID_MAIN];
	if (!buf)
		return;

	struct logger_entry_with_msg entry;
	create_pipe_message(&entry,
		DLOG_FATAL, /* not really FATAL, but since we're at the FD limit
		             * there are thousands of logging programs so this one
		             * would likely get lost in the flood since developers
		             * tend to overuse ERROR (and FATAL is the one above) */
		"DLOG",
		"\x1b[31m DLOG DAEMON FD LIMIT REACHED \x1b[0m" // make it stand out
	);
	if (buffer_append(&entry.header, buf))
		printf("ERROR: not enough memory either, please check platform settings as the daemon is seriously resource-starved!\n");
}

static int reader_print_out_single_log(struct reader *reader, const struct logger_entry *logger_entry)
{
	assert(reader);
	assert(reader->buf_ptr);
	assert(logger_entry);

	if (!log_should_print_line(reader->filter, logger_entry))
		return 0;

	if (reader->file.path) {
		logfile_write_with_rotation(logger_entry, &reader->file, reader->buf_ptr->sort_by);
		return 0;
	}

	const char *tag = logger_entry->msg + 1;
	if (!strlen(tag))
		return 0;

	int r = write(reader->file.path ? reader->file.fd : reader->fd_entity.fd, logger_entry, logger_entry->len);
	if (r < 0) {
		if (errno == EPIPE)
			return 1;
		return -1;
	}

	reader->file.size += r;
	if (r < logger_entry->len) {
		reader->partial_log_size = logger_entry->len - r;
		memcpy(reader->partial_log, ((char *)logger_entry) + r, reader->partial_log_size);
		return -1;
	} else if (logfile_rotate_needed(&reader->file) > 0) {
		logfile_do_rotate(&reader->file);
	}

	return 0;
}

/**
 * @brief Print out logs
 * @details Make sure the reader is up to date on printed logs
 * @param[in] reader The reader to read the data
 * @param[in] server The logger server
 * @return 0 if data remains for the next iteration, 1 if the buffer is to be removed, else -1
 */
int print_out_logs(struct reader *reader)
{
	assert(reader);

	assert(reader->buf_ptr);

	int r, ret = 0;

	if (reader->partial_log_size) {
		r = write(reader->fd_entity.fd, reader->partial_log, reader->partial_log_size);
		if (r <= 0)
			return ret;

		if (r < reader->partial_log_size) {
			reader->partial_log_size -= r;
			memmove(reader->partial_log, reader->partial_log + r, reader->partial_log_size);
			return ret;
		}

		reader->partial_log_size = 0;
	}

	while (log_storage_reader_is_new_entry_available(reader->log_storage_reader)) {
		const struct logger_entry* ple = (const struct logger_entry *)log_storage_reader_get_new_entry(reader->log_storage_reader);

		assert(ple);

		switch (reader_print_out_single_log(reader, ple)) {
		case 0: /* nothing more to do, let's do next loop */
			break;

		case 1: /* error after which we need to end the reader */
			return 1;

		default: /* writing error, bounce out */
			return -1;
		}
	}

	if (reader->dumpcount)
		ret = 1;
	else
		ret = -1;

	return ret;
}

/**
 * @brief Writer close FD
 * @details Close a writer's connections
 * @param[in] server The logger server
 * @param[in] wr The writer whose connections to close
 */
void writer_close_fd(struct logger* server, struct writer* wr)
{
	assert(wr);
	remove_fd_entity(server, &wr->fd_entity);
	close(wr->fd_entity.fd);
}

/**
 * @brief Free writer
 * @details Deallocate a writer
 * @param[in] w The writer to deallocate
 */
static void writer_free(struct writer *w, struct logger *logger)
{
	writer_close_fd(logger, w);
	free(w);
}

/**
 * @brief Free a reader
 * @details Deallocates a reader
 * @param[in] reader The reader to free
 */
void reader_free(struct reader* reader)
{
	if (!reader)
		return;

	if (reader->fd_entity.fd >= 0)
		close(reader->fd_entity.fd);
	if (reader->read_fd >= 0)
		close(reader->read_fd);
	logfile_free(&reader->file);
	if (reader->filter)
		log_filter_free(reader->filter);
	if (reader->log_storage_reader)
		log_storage_release_reader(reader->log_storage_reader);
	free(reader);
}

/**
 * @brief Service reader file
 * @details Handles readers reading directly from file
 * @param[in] reader The reader to service
 * @return 0 on success, else an errno value
 * @return 1 if the reader is to be removed, 0 if kept, -errno on error
 */
int service_reader_file(struct reader* reader)
{
	assert(reader);

	struct logger_entry_with_msg entry;
	static char buffer[sizeof entry + 1];
	buffer[sizeof buffer - 1] = '\0';

	/* The devices for both KMSG and Android Logger only return one log per read().
	   So using an 'if' here would be wasteful and, more importantly, too slow in the case where other logs come in.
	   However, with an unlimited loop, if there are extreme amounts of logs incoming,
	   the loop handles them slower than they come so the program stays in the loop
	   for a very long time, starving all other log sources. Using a limited loop
	   makes sure that the daemon works efficiently in the usual case while preventing
	   starvation under heavy load.
	*/
	int max_loop_iterations = g_backend.logger_device_throttling[reader->buf_id];
	while (max_loop_iterations--) {
		int r = read(reader->read_fd, buffer, sizeof buffer - 1);
		if (r == 0)
			break;
		else if (r == -1) {
			if (errno == EAGAIN) // no data left in the buffer
				break;
			else if (errno != EINTR) // EINTR is safe to ignore; retry
				return -errno;
		} else {
			buffer[r] = '\0';
			parse_androidlogger_message((struct android_logger_entry *) buffer, &entry.header, r);
			add_recv_timestamp(&entry.header);

			if (!log_should_print_line(reader->filter, &entry.header))
				return 0;

			logfile_write_with_rotation(&entry.header, &reader->file, SORT_SENT_REAL);
		}
	}

	return 0;
}

void reader_notify_losing_log(const struct logger_entry *le, void *reader_)
{
	struct reader *reader = reader_;
	assert(le);
	assert(reader);
	reader_print_out_single_log(reader, (struct logger_entry *)le);
}

/**
 * @brief Add reader to log buffer
 * @details Adds a reader to the log buffers reader list
 * @param[in] log_buffer The buffer whither to add the reader
 * @param[in] reader The reader to add to the buffer
 * @return 0 on success, -ENOMEM on memory allocation error
 */
int add_buffer_reader(struct log_buffer *buffer, struct reader *reader)
{
	assert(reader);
	assert(buffer);

	reader->log_storage_reader = log_storage_new_reader(buffer->log_storage,
			reader->dumpcount, reader_notify_losing_log, reader);

	if (!reader->log_storage_reader)
		return -ENOMEM;

	return list_add(&buffer->readers, reader) ? 0 : -ENOMEM;
}

/**
 * @brief Add reader to server
 * @details Adds a reader to the server's reader list
 * @param[in] server The server whither to add the reader
 * @param[in] reader The reader to add to the server
 * @return 0 on success, -EINVAL if related buffer not enabled or invalid reader type given
 *                       -ENOMEM on memory allocation error
 */
static int add_reader(struct logger *server, struct reader *reader)
{
	assert(reader);
	assert(server);

	if (reader->fd_entity.fd >= 0) {
		/* Readers who write to file have no FD entity since their FD
		 * is not eligible to be added to epoll. However, epoll primarily
		 * serves to handle pipes getting clogged mid-write (which cannot
		 * happen for files) and is not involved in starting a write,
		 * which happens periodically for all readers anyway regardless
		 * of whether they've been added to epoll or not. The other use
		 * case for epoll - the one where connection FDs leak for unused
		 * buffers - is not a problem here because the deamon is not
		 * supposed to ever stop writing to files. */

		int r = add_fd_entity(server, &reader->fd_entity);
		if (r < 0)
			return r;
	}

	if (reader->buf_ptr)
		return add_buffer_reader(reader->buf_ptr, reader);

	assert(reader->read_fd >= 0);
	return list_add(&server->extra_readers, reader) ? 0 : -ENOMEM;
}

int reader_init_for_pipe(struct reader *reader, const char *buf_name, struct logger *server)
{
	assert(reader);
	assert(buf_name);
	assert(server);

	reader->service_reader = print_out_logs;

	log_id_t buf_id = log_id_by_name(buf_name);
	if (buf_id == LOG_ID_INVALID)
		return -EINVAL;

	reader->buf_ptr = server->buffers[buf_id];
	reader->buf_id = buf_id;
	return (reader->buf_ptr ? 0 : -EINVAL);
}

int reader_init_for_logger(struct reader *reader, const char *buf_name, struct logger *server)
{
	assert(reader);
	assert(buf_name);
	assert(server);

	if (!strcmp(buf_name, "kmsg") || !strcmp(buf_name, "syslog"))
		return reader_init_for_pipe(reader, buf_name, server);

	reader->service_reader = service_reader_file;

	log_id_t buf_id = log_id_by_name(buf_name);
	if (buf_id == LOG_ID_INVALID)
		return -EINVAL;

	assert(buf_id >= 0 && (unsigned) buf_id < NELEMS(g_backend.logger_devices));
	char const *config_list = g_backend.logger_devices[buf_id];
	if (strlen(config_list) == 0)
		return -ENOENT;
	reader->buf_id = buf_id;

	int ret = logger_open_buffer(buf_id, config_list, O_RDONLY | O_NONBLOCK, &reader->read_fd);

	if (ret <= 0)
		return -ENOENT;

	return 0;
}

static int create_fifo_fds(struct logger *server, int fifo_id, int *write_fd, int *read_fd, bool dump)
{
	assert(write_fd);
	assert(read_fd);

	char fifo_path[64];
	if (snprintf(fifo_path, sizeof fifo_path, "/run/dlog/priv/fifo/%d", fifo_id) < 0)
		return -errno;

	if (mkfifo(fifo_path, 0600) < 0) {
		assert(errno != EEXIST);
		return -errno;
	}

	int ret = 0;

	*read_fd = open(fifo_path, O_RDONLY | O_NONBLOCK);
	if (*read_fd < 0) {
		check_if_fd_limit_reached(server, errno);
		ret = -errno;
		goto finish;
	}

	*write_fd = open(fifo_path, O_WRONLY | O_NONBLOCK);
	if (*write_fd < 0) {
		check_if_fd_limit_reached(server, errno);
		ret = -errno;
		close(*read_fd);
		goto finish;
	}

	/* Speed up dumping dlogutils by increasing their pipe size,
	 * as otherwise this pipe's size becomes a major bottleneck.
	 *
	 * Continuous connections don't really care if the initial
	 * burst of "historic" logs is slow and the following flow
	 * of logs is just fine with a small pipe. Meanwhile they
	 * live for a long time during which they would take away
	 * from the valuable total pipe size. */
	if (dump)
		if (fcntl(*write_fd, F_SETPIPE_SZ, DUMPING_READER_PIPE_SIZE) < 0) {
			/* Ignore; this is just a performance optimisation
			 * and doesn't affect functionality so while errors
			 * are worrisome they not a big deal or something
			 * we can do anything about, at any rate. */
		}

finish:
	unlink(fifo_path);
	return ret;
}

/**
 * @brief Parse a command line
 * @details Creates a reader from a parsed command line
 * @param[in] cmdl The command line string
 * @param[in] wr The writer who sent the line (NULL if internal)
 * @param[in] server The server
 * @param[out] rd The new reader
 * @return 0 on success, else -errno
 */
static int parse_command_line(const char *cmdl, struct writer *wr, struct logger *server, struct reader **rd)
{
	assert(cmdl);
	// wr optional: NULL means created by config
	assert(server);
	assert(rd);

	char cmdline[512];
	int option, argc = 0;
	char *argv[100];
	char *tok;
	char *tok_sv;
	int retval = 0;
	struct reader *reader;
	int silence = 0;

	snprintf(cmdline, sizeof(cmdline), "%s", cmdl);

	reader = (struct reader *)calloc(1, sizeof(struct reader));
	if (!reader) {
		retval = -ENOMEM;
		goto cleanup;
	}

	tok = strtok_r(cmdline, DELIMITER, &tok_sv);
	if (!tok || strcmp(tok, "dlogutil")) {
		retval = -EINVAL;
		goto cleanup;
	}

	while (tok && (argc < NELEMS(argv))) {
		argv[argc++] = tok;
		tok = strtok_r(NULL, DELIMITER, &tok_sv);
	}

	init_fd_entity(&reader->fd_entity, dispatch_event_reader, reader);

	logfile_init(&reader->file);

	reader->filter = log_filter_new();
	if (!reader->filter)
		goto cleanup;

	reader->buf_ptr = NULL;
	reader->buf_id = -1;
	reader->read_fd = -1;
	reader->dumpcount = false;
	reader->partial_log_size = 0;
	reader->service_reader = NULL;
	reader->log_storage_reader = NULL;
	clock_gettime(CLOCK_MONOTONIC, &reader->last_read_time);

	while ((option = getopt(argc, argv, "cdt:gsf:r:n:v:b:u:")) != -1) {
		switch (option) {
			break;
		case 'd':
		case 't':
			reader->dumpcount = true;
			break;
		case 'f':
			if (wr) {
				/* Do not trust writer-based readers (only config-based).
				 * The control socket's privilege checks are fairly lenient
				 * so this prevents people from asking us to overwrite
				 * some potentially important files at logger privilege.
				 *
				 * At some point it would be good to be able to skip the
				 * middleman and become able to write to a file directly
				 * though. The daemon should become able to receive an
				 * opened file descriptor from a writer. */
				retval = -EPERM;
				goto cleanup;
			}

			retval = logfile_set_path(&reader->file, optarg);
			if (retval < 0)
				goto cleanup;
			break;
		case 'b':
			assert(g_backend.reader_init);
			retval = g_backend.reader_init(reader, optarg, server);
			if (retval != 0)
				goto cleanup;
			break;
		case 'r':
			if (sscanf(optarg, "%zu", &reader->file.rotate_size_kbytes) != 1)
				goto cleanup;
			break;
		case 'n':
			if (sscanf(optarg, "%zu", &reader->file.max_rotated) != 1)
				goto cleanup;
			break;
		case 'v':
			reader->file.format.format = log_format_from_string(optarg);
			break;
		case 's':
			silence = 1;
			log_add_filter_string(reader->filter, "*:s");
			break;
		case '?': // invalid option or missing argument
			retval = -EINVAL;
			goto cleanup;
		default:
			// everything else gets handled in util directly
			break;
		}
	}

	if (wr && wr->buf_ptr) {
		reader->service_reader = print_out_logs;
		reader->buf_ptr = wr->buf_ptr;
	}

	if (reader->service_reader == NULL) {
		retval = -EINVAL;
		goto cleanup;
	}

	if (optind < argc)
		while (optind < argc)
			log_add_filter_string(reader->filter, argv[optind++]);
	else if (!silence)
		log_add_filter_string(reader->filter, "*:D");

	if ((reader->read_fd < 0) && !reader->buf_ptr) {
		retval = -EINVAL;
		goto cleanup;
	}

	if (reader->file.path != NULL) {
		retval = logfile_open(&reader->file);
		if (retval < 0)
			goto cleanup;
	} else if (wr) {
		int write_fd = -1, read_fd = -1;
		retval = create_fifo_fds(server, wr->fd_entity.fd, &write_fd, &read_fd, reader->dumpcount);
		if (retval < 0)
			goto cleanup;

		set_write_fd_entity(&reader->fd_entity, write_fd);
		retval = send_pipe(wr->fd_entity.fd, read_fd);
		if (read_fd > 0)
			close(read_fd);
		if (retval)
			goto cleanup;
	} else {
		retval = -EINVAL;
	}
cleanup:
	/* recycle for further usage */
	optarg = NULL;
	optind = 0;
	optopt = 0;

	if (retval) {
		if (wr) {
			int r = send_dlog_reply(wr->fd_entity.fd, DLOG_REQ_HANDLE_LOGUTIL, DLOG_REQ_RESULT_ERR, NULL, 0);
			if (r < 0)
				printf("ERROR: both parse_command_line() and send_dlog_reply() failed\n");
		}
		*rd = NULL;
		reader_free(reader);
	} else
		*rd = reader;
	return retval;
}

/**
 * @brief Service util request
 * @details Handle a request from util
 * @param[in] server The logger server
 * @param[in] wr The writer who sent the request
 * @param[in] msg The message containing the request
 * @return 0 on success, else -errno
 */
static int service_writer_handle_req_util(struct logger* server, struct writer* wr, struct dlog_control_msg* msg)
{
	assert(msg);

	// check request type, that should be always DLOG_REQ_HANDLE_LOGUTIL
	// as dispatched by service_writer_handle_req_ctrl handler
	// don't assert for compatibility with service_writer_handle_req_pipe
	// and possible mistakes in the future that would be hard to track
	if (msg->request != DLOG_REQ_HANDLE_LOGUTIL)
		return -EINVAL;

	if (msg->length <= sizeof(struct dlog_control_msg) ||
	    msg->length > sizeof(struct dlog_control_msg) + MAX_LOGGER_REQUEST_LEN)
		return -EINVAL;

	if (msg->data[msg->length - sizeof(struct dlog_control_msg)] != 0)
		return -EINVAL;

	struct reader *rd;
	int r = parse_command_line(msg->data, wr, server, &rd);
	if (r < 0)
		return r;

	assert(server);
	assert(rd->buf_ptr);

	r = add_reader(server, rd);
	if (r < 0)
		return r;

	assert(wr);
	if (wr->readed > msg->length) {
		wr->readed -= msg->length;
		memmove(wr->buffer, wr->buffer + msg->length, wr->readed);
	} else
		wr->readed = 0;

	return 0;
}

static int service_writer_handle_req_getsize(struct logger *server, struct writer *wr, struct dlog_control_msg *msg)
{
	assert(server);
	assert(wr);
	assert(msg);
	assert(msg->request == DLOG_REQ_GETSIZE);

	uint32_t buf_size = log_storage_get_capacity(wr->buf_ptr->log_storage);

	return send_dlog_reply(wr->fd_entity.fd, DLOG_REQ_GETSIZE, DLOG_REQ_RESULT_OK, &buf_size, sizeof buf_size);
}

/**
 * @brief Service clear request
 * @details Handle a clear-buffer request
 * @param[in] server The logger server
 * @param[in] wr The writer who sent the request
 * @param[in] msg The message containing the request
 * @return 0 on success, else -errno
 */
static int service_writer_handle_req_clear(struct logger* server, struct writer* wr, struct dlog_control_msg* msg)
{
	assert(msg);

	// check request type, that should be always DLOG_REQ_CLEAR
	// as dispatched by service_writer_handle_req_ctrl handler
	// don't assert for compatibility with service_writer_handle_req_pipe
	// and possible mistakes in the future that would be hard to track
	if (msg->request != DLOG_REQ_CLEAR)
		return -EINVAL;

	if (msg->length != (sizeof(struct dlog_control_msg)))
		return -EINVAL;

	if (!wr || !wr->buf_ptr)
		return -EINVAL;

	log_storage_clear(wr->buf_ptr->log_storage);

	if (wr->readed > msg->length) {
		wr->readed -= msg->length;
		memmove(wr->buffer, wr->buffer + msg->length, wr->readed);
	} else
		wr->readed = 0;

	return 0;
}

/**
 * @brief Service control request
 * @details Handle a clear-buffer or util request in respect to msg request type
 * @param[in] server The logger server
 * @param[in] wr The writer who sent the request
 * @param[in] msg The message containing the request
 * @return 0 on success, else -errno
 */
static int service_writer_handle_req_ctrl(struct logger *server, struct writer *wr, struct dlog_control_msg *msg)
{
	assert(msg);

	switch (msg->request) {
	case DLOG_REQ_CLEAR:
		return service_writer_handle_req_clear(server, wr, msg);
	case DLOG_REQ_HANDLE_LOGUTIL:
		return service_writer_handle_req_util(server, wr, msg);
	case DLOG_REQ_GETSIZE:
		return service_writer_handle_req_getsize(server, wr, msg);
	default:
		return -EINVAL;
	}
}

/**
 * @brief Service a pipe acquisition request
 * @details Handle a pipe request
 * @param[in] server The logger server
 * @param[in] wr The writer who sent the request
 * @param[in] msg The message containing the request
 * @return 0 on success, else -errno
 */
static int service_writer_handle_req_pipe(struct logger* server, struct writer* wr, struct dlog_control_msg* msg)
{
	int r;

	assert(msg);

	// check request type given by user
	// don't assert that as the message is not parsed before
	if (msg->request != DLOG_REQ_PIPE)
		return -EINVAL;

	if (msg->length != sizeof(struct dlog_control_msg))
		return -EINVAL;

	int pipe_fd[2];
	if (pipe2(pipe_fd, O_CLOEXEC | O_NONBLOCK) < 0) { // O_NONBLOCK just for pipe_fd[0]; writer removes it for pipe_fd[1] on its own
		check_if_fd_limit_reached(server, errno);
		return -errno;
	}

	if (fcntl(pipe_fd[1], F_SETPIPE_SZ, PIPE_REQUESTED_SIZE) < 0) {
		/* Ignore failures. This call is just a performance optimisation
		 * and doesn't affect functionality; we can't do anything about
		 * an error anyway. */
	}

	assert(server);
	assert(wr);

	struct fd_entity pipe_entity = wr->fd_entity;
	set_read_fd_entity(&pipe_entity, pipe_fd[0]);
	r = add_fd_entity(server, &pipe_entity);
	if (r < 0)
		goto err_close;

	r = send_pipe(wr->fd_entity.fd, pipe_fd[1]);
	if (r < 0)
		goto err_remove;
	close(pipe_fd[1]);

	writer_close_fd(server, wr);
	wr->service_writer = service_writer_pipe;
	wr->fd_entity = pipe_entity;
	wr->readed = 0;
	return 0;

err_remove:
	remove_fd_entity(server, &pipe_entity);

err_close:
	close(pipe_fd[0]);
	close(pipe_fd[1]);

	return r;
}

/**
 * @brief Service a socket request
 * @details Handle a socket request
 * @param[in] server The logger server
 * @param[in] wr The writer who sent the request
 * @param[in] event The event containing the request
 * @return 0 on success, else -errno
 */
static int service_writer_socket(struct logger* server, struct writer* wr, struct epoll_event* event)
{
	int r = 0;

	assert(wr);
	struct dlog_control_msg* const msg = (struct dlog_control_msg*) wr->buffer;

	assert(event);
	if (event->events & EPOLLIN)
		r = read(wr->fd_entity.fd, wr->buffer + wr->readed, sizeof wr->buffer - wr->readed);

	if ((r == 0 || r == -1) && event->events & EPOLLHUP)
		return -EINVAL;

	do {
		if (r > 0)
			wr->readed += r;

		/* The socket is SOCK_STREAM (see `listen_fd_create`), so one message
		 * could be split into chunks returned across multiple read() calls. */
		if (wr->readed < sizeof(msg->length)
		||  wr->readed < msg->length)
			goto dont_process_yet_and_read_more_data;

		assert(wr->service_socket);
		r = wr->service_socket(server, wr, msg);
		if (r <= 0)
			return r;

dont_process_yet_and_read_more_data:
		r = read(wr->fd_entity.fd, wr->buffer + wr->readed, sizeof wr->buffer - wr->readed);
	} while (r > 0 || ((wr->readed >= sizeof(msg->length) && wr->readed >= msg->length)));

	return (r >= 0 || errno == EAGAIN)  ? 0 : r;
}

/**
 * @brief Service pipe log data
 * @details Handle log messages incoming through a pipe
 * @param[in] server The logger server
 * @param[in] wr The writer who sent the logs
 * @param[in] event The event associated with the data
 * @return 0 on success, else -errno
 */
static int service_writer_pipe(struct logger *server, struct writer *wr, struct epoll_event *event)
{
	if (event->events & EPOLLIN) {
		int r = read(wr->fd_entity.fd, wr->buffer + wr->readed, sizeof wr->buffer - wr->readed);

		if (r == -1 && errno == EAGAIN)
			return 0;
		else if ((r == 0 || r == -1) && event->events & EPOLLHUP)
			return -EINVAL;
		else if (r == 0)
			return -EBADF;

		wr->readed += r;

		struct pipe_logger_entry *const ple = (struct pipe_logger_entry *const)wr->buffer;
		while ((wr->readed >= sizeof(ple->len)) && (ple->len <= wr->readed)) {
			const int payload_size = ple->len - sizeof *ple;
			if (payload_size < 0 || payload_size > LOG_MAX_PAYLOAD_SIZE)
				return -EINVAL;

			struct logger_entry_with_msg lem;
			parse_pipe_message(ple, &lem.header, ple->len);
			add_recv_timestamp(&lem.header);
			r = buffer_append(&lem.header, wr->buf_ptr);
			wr->readed -= ple->len;
			memmove(wr->buffer, wr->buffer + ple->len, sizeof wr->buffer - ple->len);

			if (r)
				return r;
		}
	} else if (event->events & EPOLLHUP)
		return -EBADF;

	return 0;
}

/**
 * @brief Service /dev/kmsg
 * @details Read from the /dev/kmsg device
 * @param[in] server The logger server
 * @param[in] wr The writer who sent the request
 * @param[in] event The relevant event
 * @return 0 on success, else -errno
 */
int service_writer_kmsg(struct logger* server, struct writer* wr, struct epoll_event* event)
{
	if (event->events & EPOLLHUP)
		return -EBADF;
	if (!(event->events & EPOLLIN))
		return 0;

	/* The KMSG device returns just 1 log per read() so it is done in a loop.
	 * In theory this could starve everything else out if logs appeared faster
	 * than the daemon could process them, which would then necessitate some
	 * sort of throttling. In practice, KMSG doesn't really get flooded with
	 * logs the same way Android Logger devices are so the throttling is mostly
	 * there because we get it for free anyway and consistency doesn't hurt. */

	int max_loop_iterations = g_backend.logger_device_throttling[LOG_ID_KMSG];
	while (max_loop_iterations--) {
		int r = read(wr->fd_entity.fd, wr->buffer, sizeof wr->buffer - 1);

		if (r == -1 && (errno == EAGAIN || errno == EPIPE))
			return 0;
		else if ((r == 0 || r == -1) && event->events & EPOLLHUP)
			return -EINVAL;
		else if (r == 0)
			return -EBADF;
		else if (r == -1)
			return -errno;

		wr->buffer[r] = '\0';
		struct logger_entry_with_msg lem;
		if (parse_kmsg_message(wr->buffer, &lem, r)) {
			// don't signal an error: KMSG writer is too important to remove; besides, it would not get fixed that way.
			return 0;
		}
		add_recv_timestamp(&lem.header);
		r = buffer_append(&lem.header, wr->buf_ptr);
		if (r < 0)
			return r;
	}

	return 0;
}

/**
 * @brief Service reader given in ptr
 * @details Use this function when servicing all readers in the list
 * @param[in] ptr pointer to the reader
 * @param[in] user_data pointer to the logger to add readers waiting for an event
 * @return true if the reader is to be removed
 */
static bool cond_service_reader(void *ptr, void *user_data)
{
	assert(ptr);
	assert(user_data);
	struct reader *reader = (struct reader *)ptr;
	struct logger *logger = (struct logger *)user_data;

	if (!logger->exiting && reader_should_buffer(reader, &logger->buf_params, logger->now))
		return false;

	assert(reader->service_reader);
	int r = reader->service_reader(reader);

	if (r > 0) {
		if (reader->fd_entity.fd >= 0)
			remove_fd_entity(logger, &reader->fd_entity);
		reader_free(reader);
		return true;
	}

	/* `service_reader()` returns -1 if everything was flushed, or 0 if
	 * a mild error happened that can be recovered from simply by waiting,
	 * the prime example being the pipe getting clogged. As soon as the
	 * reader is available again, we'd like to know about it to ensure
	 * logs are flushed as quickly as possible, which is why the EPOLLOUT.
	 *
	 * On the other hand, we don't want to remove readers from epoll even
	 * if they successfully flushed and have no logs to wait for. Consider
	 * the case where a buffer is unused (for example through libdlog's
	 * buffer disabling feature). If we relied on receiving an error on
	 * calling `write()` to learn that the connection had been closed,
	 * we would never learn about it because there would be no incoming
	 * logs to trigger the flush and so any FDs representing connections
	 * to such buffer would leak until a log finally arrived (which could
	 * be never). This is why waiting is also done on EPOLLHUP. */
	if (modify_fd_entity(logger, &reader->fd_entity, (r == 0) ? EPOLLOUT : EPOLLHUP) < 0) {
		/* ignore, can't really happen and it's not
		 * like we can do anything about it either */
	}

	return false;
}

/**
 * @brief Dispatch reader event
 * @details Receives and handles an event
 * @param[in] server The logger server
 * @param[in] event The received event
 */
static void dispatch_event_reader(struct logger *server, struct epoll_event *event)
{
	assert(event);
	struct reader *reader = (struct reader *)event->data.ptr;

	assert(reader);

	if (event->events & (EPOLLHUP | EPOLLERR)) {
		list_remove(&reader->buf_ptr->readers, reader);
		remove_fd_entity(server, &reader->fd_entity);
		reader_free(reader);
		return;
	}

	assert(reader->service_reader);
	int r = reader->service_reader(reader);
	if (r != 0)
		remove_fd_entity(server, &reader->fd_entity);
}

/**
 * @brief Service syslog
 * @details Read from the syslog socket
 * @param[in] server The logger server
 * @param[in] wr The writer who sent the request
 * @param[in] event The relevant event
 * @return 0 on success, else -errno
 */
int service_writer_syslog(struct logger* server, struct writer* wr, struct epoll_event* event)
{
	if (event->events & EPOLLHUP)
		return -EBADF;
	if (!(event->events & EPOLLIN))
		return 0;

	int r = read(wr->fd_entity.fd, wr->buffer, sizeof wr->buffer - 1);

	if (r == -1 && (errno == EAGAIN || errno == EPIPE))
		return 0;
	else if ((r == 0 || r == -1) && event->events & EPOLLHUP)
		return -EINVAL;
	else if (r == 0)
		return -EBADF;
	else if (r == -1)
		return -errno;

	wr->buffer[r] = '\0';

	struct logger_entry_with_msg lem;
	if (parse_syslog_datagram(wr->buffer, r + 1, &lem.header)) {
		/* don't return parse error as writers are removed then */
		return 0;
	}

	add_recv_timestamp(&lem.header);
	return buffer_append(&lem.header, wr->buf_ptr);
}

/**
 * @brief Service all readers
 * @details Update all readers with latest data
 * @param[in] server The logger server
 * @param[in] force_push Whether to force logs to be pushed to the readers
 */
static void service_all_readers(struct logger *server)
{
	for (int i = 0; i < LOG_ID_MAX; i++) {
		struct log_buffer *const buffer = server->buffers[i];
		if (!buffer)
			continue;

		list_remove_if(&buffer->readers, server, cond_service_reader);
	}
	list_remove_if(&server->extra_readers, server, cond_service_reader);
}

/**
 * @brief Initialize a socket
 * @details Initializes a socket
 * @param[out] sock The socket to initialize
 * @param[in] buffer The buffer to whom the socket belongs
 * @param[in] type Type of the socket
 * @param[in] data Socket config data
 * @return 0 on success, else -errno
 */
int socket_initialize(struct sock_data *sock, struct log_buffer *buffer, service_socket_t service_socket, struct socket_config_data *data)
{
	assert(data);
	int sock_fd = listen_fd_create(data->path, data->permissions);
	if (sock_fd < 0)
		return sock_fd;

	init_fd_entity(&sock->fd_entity, dispatch_event_sock, sock);
	sock->service_socket = service_socket;
	sock->buf_ptr = buffer;
	set_read_fd_entity(&sock->fd_entity, sock_fd);

	return 0;
}

/**
 * @brief Add writer to server
 * @details Adds a writer to the server's witers list and registers its event to epoll loop
 * @param[in] l The server to add the writer to
 * @param[in] writer The writer to add to the server and register its working_fd to epoll loop
 */
static void logger_add_writer(struct logger* l, struct writer* wr)
{
	assert(l);
	assert(wr);

	list_add(&l->writers, wr);
	add_fd_entity(l, &wr->fd_entity);
}

/**
 * @brief Create the logger server
 * @details Allocate the logger server's auxiliary structures (buffers etc.)
 * @param[in] data Initialisation data
 * @param[out] l The logger server
 * @return 0 on success, -errno on failure
 */
static int logger_create(struct logger_config_data *data, struct logger *l)
{
	memset(l, 0, sizeof *l);

	l->epollfd = epoll_create1(0);
	if (l->epollfd == -1)
		return -errno;

	l->buf_params = data->buf_params;

	for (log_id_t id = 0; id < LOG_ID_MAX; id++)
		if (data->is_buffer_enabled[id]) {
			int r = buffer_create(&l->buffers[id], id, data->buffers + id);
			if (r < 0)
				return r;
		}

	for (log_id_t id = 0; id < LOG_ID_MAX; id++)
		if (l->buffers[id]) {
			add_fd_entity(l, &l->buffers[id]->sock_ctl.fd_entity);
			add_fd_entity(l, &l->buffers[id]->sock_wr.fd_entity);
		}

	/* TODO: make writers creation optional/configurable */
	int (*writers_factories[LOG_ID_MAX])(struct writer **writer, struct log_buffer *log_buf) = {
		[LOG_ID_KMSG] = create_kmsg_writer,
		[LOG_ID_SYSLOG] = create_syslog_writer,
	};
	for (unsigned u = 0; u < NELEMS(writers_factories); ++u)
		if (writers_factories[u] && data->is_buffer_enabled[u]) {
			struct writer *wr;
			int r = writers_factories[u](&wr, l->buffers[u]);
			if (r < 0)
				return r;

			logger_add_writer(l, wr);
		}

	return 0;
}

static bool cond_writer_free(void *ptr, void *user_data)
{
	writer_free((struct writer *)ptr, user_data);
	return true;
}

/**
 * @brief Free logger
 * @details Deallocate the logger and its auxiliary structures
 * @param[in] l The logger server
 */
static void logger_free(struct logger *l)
{
	assert(l);

	list_remove_if(&l->writers, l, cond_writer_free);
	list_remove_if(&l->extra_readers, l, cond_reader_free);

	int j;
	for (j = 0; j < LOG_ID_MAX; j++)
		if (l->buffers[j])
			buffer_free(l->buffers[j], l);

	if (l->epollfd >= 0)
		close(l->epollfd);
}

/**
 * @brief Dispatch writer event
 * @details Receives and handles an event
 * @param[in] server The logger server
 * @param[in] event The received event
 */
void dispatch_event_writer(struct logger *server, struct epoll_event *event)
{
	assert(server);
	assert(event);
	struct writer *writer = (struct writer*)event->data.ptr;

	assert(writer->service_writer);
	int r = writer->service_writer(server, writer, event);
	if (r) {
		list_remove(&server->writers, writer);
		writer_free(writer, server);
	}
}

/**
 * @brief Dispatch socket event
 * @details Receives and handles an event
 * @param[in] server The logger server
 * @param[in] event The received event
 */
void dispatch_event_sock(struct logger *server, struct epoll_event *event)
{
	assert(event);
	struct sock_data const * const sock = (struct sock_data const * const) event->data.ptr;

	assert(sock);
	int sock_pipe = accept4(sock->fd_entity.fd, NULL, NULL, SOCK_NONBLOCK);
	if (sock_pipe < 0) {
		check_if_fd_limit_reached(server, errno);
		return;
	}

	struct writer *writer;
	if (writer_create(&writer, sock_pipe, sock->buf_ptr, service_writer_socket, sock->service_socket) == 0)
		logger_add_writer(server, writer);
	else
		close(sock_pipe);
}

/**
 * @brief Handle interrupting/terminating signals
 * @details Clears global flag to stop main loop
 * @param[in] signo signal number
 */
static void handle_signals(int signo)
{
	g_logger_run = 0;
}

static void ensure_epoll_size(struct epoll_event **events, unsigned *size, unsigned wanted_size)
{
	assert(events);
	assert(size);

	if (wanted_size <= *size)
		return;

	typeof(*events) temp = realloc(*events, wanted_size * sizeof **events);
	if (!temp)
		return;

	*events = temp;
	*size = wanted_size;
}

static int initialize_epoll_size(struct epoll_event **events, unsigned *size)
{
	assert(events);
	assert(size);

	static const size_t default_size = 16U;
	typeof(*events) ev = malloc(default_size * sizeof *ev);
	if (!ev)
		return -ENOMEM;

	*events = ev;
	*size = default_size;
	return 0;
}

/**
 * @brief Do logging
 * @details The main logging loop
 * @param[in] server The logger server
 * @return 0 on success, else -errno
 */
static int do_logger(struct logger* server)
{
	struct sigaction action = {
		.sa_handler = handle_signals,
		.sa_flags   = 0
	};
	sigemptyset(&action.sa_mask);

	static const int handled_signals[] = { SIGINT, SIGTERM };
	for (unsigned u = 0; u < NELEMS(handled_signals); ++u)
		sigaction(handled_signals[u], &action, NULL);

	unsigned events_size;
	struct epoll_event *events;
	int r = initialize_epoll_size(&events, &events_size);
	if (r < 0)
		return r;
	while (g_logger_run) {
		clock_gettime(CLOCK_MONOTONIC, &server->now);

		ensure_epoll_size(&events, &events_size, server->epollcnt);

		int nfds = epoll_wait(server->epollfd, events, events_size, 1000);
		if (nfds < 0 && errno == EINTR)
			continue;

		if (nfds < 0) {
			r = -errno;
			break;
		}

		for (int i = 0; i < nfds; i++) {
			struct fd_entity* entity = (struct fd_entity*) events[i].data.ptr;
			assert(entity->dispatch_event);
			entity->dispatch_event(server, &events[i]);
		}
		service_all_readers(server);
	}

	free(events);

	/* ensure all logs are written no matter when the program was interrupted */
	server->exiting = 1;
	service_all_readers(server);

	return 0;
}

/**
 * @brief Prepare socket data
 * @details Extracts initialisation data specific to each socket
 * @param[in] conf Config database
 * @param[out] data Socket init config data
 * @param[in] buf_name The name of the buffer the socket belongs to
 * @param[in] type The name of the buffer type
 * @return 0 on success, -errno on failure
 */
int prepare_socket_data(const struct log_config *conf, struct socket_config_data *data, char *buf_name, const char *type)
{
	char conf_key[MAX_CONF_KEY_LEN];
	int r;

	r = snprintf(conf_key, MAX_CONF_KEY_LEN, "%s_%s_sock", buf_name, type);
	if (r < 0)
		return -errno;
	const char * const path = log_config_get(conf, conf_key);

	r = snprintf(conf_key, MAX_CONF_KEY_LEN, "%s_%s_sock_rights", buf_name, type);
	if (r < 0)
		return -errno;
	const char * const permissions_str = log_config_get(conf, conf_key);

	if (!permissions_str || !path)
		return -ENOENT;

	data->permissions = parse_permissions(permissions_str);
	if (data->permissions <= 0)
		return -EINVAL;

	strncpy(data->path, path, MAX_CONF_VAL_LEN - 1);

	return 0;
}

/**
 * @brief Prepare buffer data
 * @details Extracts data specific for each buffer
 * @param[in] conf Config database
 * @param[out] data Buffer init config data
 * @param[in] buf_id Index of the buffer to work with
 * @return 0 on success, -errno on failure
 */
int prepare_buffer_data(const struct log_config *conf, struct buffer_config_data *data, log_id_t buf_id)
{
	char * const buf_name = log_name_by_id(buf_id);
	char * validity_check_ptr;
	char conf_key[MAX_CONF_KEY_LEN];
	int r;

	r = prepare_socket_data(conf, &data->write_socket, buf_name, "write");
	if (r < 0)
		return r;

	r = prepare_socket_data(conf, &data->ctl_socket, buf_name, "ctl");
	if (r < 0)
		return r;

	r = snprintf(conf_key, MAX_CONF_KEY_LEN, "%s_size", buf_name);
	if (r < 0)
		return -errno;

	const char * const size_str = log_config_get(conf, conf_key);
	if (!size_str)
		return -ENOENT;

	data->size = strtol(size_str, &validity_check_ptr, 10);
	if (*validity_check_ptr)
		return -EINVAL;
	if (data->size <= 0)
		return -EINVAL;

	return 0;
}

/**
 * @brief Save logfile line
 * @detail Saves logfile config line in user_data list
 * @param[in] key Config entry key
 * @param[in] value Config entry value
 * @param[in] userdata Userdata
 */
void save_logfile_config(char const *key, char const *value, void *userdata)
{
	assert(userdata);
	if (strncmp(key, CONF_PREFIX, sizeof CONF_PREFIX - 1))
		return;

	struct logger_config_data *data = (struct logger_config_data *)userdata;
	void *cmd = (void *)strdup(value);
	if (cmd)
		list_add(&data->logfile_configs, cmd);
	//ignore errors
}

/**
 * @brief Prepare config data
 * @details Extracts relevant config data for logger initialisation
 * @param[out] data Parsed configuration
 * @param[in] argc Execution argument count
 * @param[in] argv Execution argument values
 * @return 0 on success, -errno on failure
 */
int prepare_config_data(struct logger_config_data *data, int argc, char **argv)
{
	assert(data);
	assert(argv);
	assert(argc >= 0);

	memset(data, 0, sizeof *data);

	int ret = parse_args(argc, argv, &data->buf_params);
	if (ret != 0) {
		help();
		return ret;
	}

	struct log_config conf;
	ret = log_config_read(&conf);
	if (ret < 0)
		return ret;

	const enum sorting_order sort_by = get_order_from_config(&conf);

	const char * const backend = log_config_get(&conf, "backend");
	if (!backend) {
		ret = -ENOENT;
		goto end;
	}
	int throttling_default = log_config_get_int(&conf, "logger_dev_throttling", LOGGER_DEVICE_THROTTLING_DEFAULT);

	for (int i = 0; i < LOG_ID_MAX; i++) {
		char key[MAX_CONF_KEY_LEN];
		const int r = snprintf(key, sizeof key, "logger_dev_throttling_%s", log_name_by_id((log_id_t)i));
		if (r < 0)
			continue;

		g_backend.logger_device_throttling[i] = max_int(1, log_config_get_int(&conf, key, throttling_default));
	}

	memset(data->is_buffer_enabled, 0, sizeof(data->is_buffer_enabled));
	if (!strcmp(backend, "pipe")) {
		data->is_buffer_enabled[LOG_ID_MAIN] =
			data->is_buffer_enabled[LOG_ID_RADIO] =
			data->is_buffer_enabled[LOG_ID_SYSTEM] =
			data->is_buffer_enabled[LOG_ID_APPS] = 1;
		g_backend.reader_init = reader_init_for_pipe;
	} else if (!strcmp(backend, "logger")) {
		g_backend.reader_init = reader_init_for_logger;
		for (log_id_t buf_id = 0; buf_id < LOG_ID_MAX; ++buf_id) {
			const char *const logger_device = log_config_get(&conf, log_name_by_id(buf_id));
			if (logger_device)
				strncpy(g_backend.logger_devices[buf_id], logger_device,
					NELEMS(g_backend.logger_devices[buf_id]) - 1);
		}
	} else {
		ret = -ENOENT;
		goto end;
	}
	data->is_buffer_enabled[LOG_ID_KMSG] = 1;
	data->is_buffer_enabled[LOG_ID_SYSLOG] =
		dev_log_sock_get() >= 0 || log_config_get_boolean(&conf, "syslog_force", false);

	for (log_id_t buf_id = 0; buf_id < LOG_ID_MAX; ++buf_id) {
		if (data->is_buffer_enabled[buf_id]) {
			ret = prepare_buffer_data(&conf, data->buffers + buf_id, buf_id);
			if (ret < 0)
				goto end;
			data->buffers[buf_id].sort_by = sort_by;
		}
	}
	data->logfile_configs = NULL;
	log_config_foreach(&conf, save_logfile_config, data);

end:
	log_config_free(&conf);
	return ret;
}

static bool cond_string_free(void *ptr, void *user_data)
{
	free(ptr);
	return true;
}

static void free_config_data(struct logger_config_data *data)
{
	list_remove_if(&data->logfile_configs, NULL, cond_string_free);
}

/**
 * @brief Print help
 * @details Prints basic usage tips
 */
static void help()
{
	printf("Usage: %s [options]\n"
	       "\t-h    Show this help\n"
	       "\t-b N  Set the size of the log buffer (in bytes)\n"
	       "\t-t N  Set time between writes to file (in seconds)\n",
	       program_invocation_short_name);
}

/**
 * @brief Parse args
 * @details Parses execution parameters of the program
 * @param[in] argc Argument count
 * @param[in] argv Argument values
 * @param[out] b Buffering parameters
 * @return 0 or 1 on success, else -errno. Nonzero if the program is to close.
 */
int parse_args(int argc, char **argv, struct buf_params *b)
{
	int option;

	b->time = BUF_PARAM_TIME_DEFAULT;
	b->bytes = BUF_PARAM_BYTES_DEFAULT;

	while ((option = getopt(argc, argv, "hb:t:")) != -1) {
		switch (option) {
		case 't':
			if (!isdigit(optarg[0]))
				return -EINVAL;
			b->time = clamp_int(atoi(optarg), BUF_PARAM_TIME_MIN, BUF_PARAM_TIME_MAX);
			break;
		case 'b':
			if (!isdigit(optarg[0]))
				return -EINVAL;
			b->bytes = clamp_int(atoi(optarg), BUF_PARAM_BYTES_MIN, BUF_PARAM_BYTES_MAX);
			break;
		case 'h':
			return 1;
		default:
			return -EINVAL;
		}
	}
	optind = 0;
	optopt = 0;
	optarg = NULL;
	return 0;
}

/**
 * @brief Parse logfile line
 * @detail Parses a logfile config line
 * @param[in] key Config entry key
 * @param[in] value Config entry value
 * @param[in] userdata Userdata
 */
void parse_logfile_config(void *value, void *userdata)
{
	assert(value);
	assert(userdata);

	struct logger *server = (struct logger *) userdata;
	struct reader *rd = NULL;
	char *configline = (char *) value;

	int r = parse_command_line(configline, NULL, server, &rd);
	if (r == 0)
		add_reader(server, rd);
	else {
		errno = -r;
		printf("Warning: unable to add logutil reader for provided configuration. Ignoring.\n"
		       "  Config line: %s\n"
		       "  Reason given: %m\n",
		       configline);
	}
}

/**
 * @brief Finalize initialisation
 * @details Do misc stuff needed at the end of the initialisation
 * @param[in] data configuration dat to read logfiles config lines from
 * @param[in] server logger instance to configure
 * @return 0 on success, -errno on failure
 */
static int finalize_init(struct logger_config_data *data, struct logger *server)
{
	int r = sd_notify(0, "READY=1");
	if (r < 0)
		return r;

	//create files after resetting self privileges
	list_foreach(data->logfile_configs, server, parse_logfile_config);

	return 0;
}

#ifndef UNIT_TEST
/**
 * @brief The logger
 * @return 0 on success, nonzero on failure
 * @retval 1 Configuration error
 * @retval 2 Runtime error
 */
int main(int argc, char** argv)
{
	int r, ret;

	signal(SIGPIPE, SIG_IGN);

	r = reset_self_privileges();
	if (r < 0) {
		errno = -r;
		printf("Unable to drop privileges to build-time defaults (%m). Exiting.\n");
		return DLOG_EXIT_ERR_RUNTIME;
	}

	struct logger_config_data data;
	if ((r = prepare_config_data(&data, argc, argv)) != 0) {
		if (r < 0) {
			errno = -r;
			printf("Unable to prepare config (%m). Exiting.\n");
			return DLOG_EXIT_ERR_CONFIG;
		} else
			return DLOG_EXIT_SUCCESS;  // handle -h option
	}

	struct logger server;
	if ((r = logger_create(&data, &server)) < 0) {
		errno = -r;
		printf("Unable to initialize logger with provided configuration (%m). Exiting.\n");
		ret = DLOG_EXIT_ERR_CONFIG;
		goto cleanup;
	}

	if ((r = finalize_init(&data, &server)) < 0) {
		errno = -r;
		printf("Unable to finalize initialisation (%m). Exiting.\n");
		ret = DLOG_EXIT_ERR_CONFIG;
		goto cleanup;
	}

	if ((r = do_logger(&server)) < 0) {
		errno = -r;
		printf("Runtime failure (%m). Exiting.\n");
		ret = DLOG_EXIT_ERR_RUNTIME;
		goto cleanup;
	}

	ret = DLOG_EXIT_SUCCESS;

cleanup:
	free_config_data(&data);
	logger_free(&server);
	return ret;
}
#endif

/**
 * @}
 * @}
 */
