#pragma once

#include "log_storage.h"

#include <syslog.h>

#include <ctype.h>
#include <fcntl.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <pwd.h>
#include <grp.h>
#include <assert.h>
#include <linux/limits.h>
#include <sys/epoll.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/un.h>
#include <logpipe.h>
#include <log_file.h>
#include <logconfig.h>
#include <ptrs_list.h>
#include <sd-daemon.h>
#include <backend_androidlogger.h>
#include <queued_entry_timestamp.h>

/* supported main() return codes */
enum {
	DLOG_EXIT_SUCCESS = 0,
	DLOG_EXIT_ERR_CONFIG,
	DLOG_EXIT_ERR_RUNTIME
};

/* Size in bytes of the pipe given to libdlog clients.
 * The default size (1 page, usually 4 kiB) is fairly small
 * which leads programs that log a lot to easily clog the
 * pipe. On the other hand, the amount of memory allocated
 * to pipes specifically can be limited and on most relevant
 * targets seems to be set to 16ki pages (64 MiB raw memory)
 * so pipes should not be too large so as not to waste it. */
#define PIPE_REQUESTED_SIZE (64 * 1024) // 16 pages at 4 kiB per page

/* Dumping readers should ideally get a larger pipe than clients
 * so that it doesn't get easily clogged by the odd spammer.
 * We can afford more generosity here because these are invoked
 * quite rarely yet require handling priority, either because
 * they are called by a developer manually or because a crashing
 * program caused a system state dump. */
#define DUMPING_READER_PIPE_SIZE (4 * PIPE_REQUESTED_SIZE)

#define FILE_PATH_SIZE (256)
#define MAX_CONNECTION_Q 100
#define DELIMITER " "
#define CONF_PREFIX "dlog_logger_conf_"

enum {
	BUF_PARAM_TIME_MIN = 0,
	BUF_PARAM_TIME_MAX = 3600,
	BUF_PARAM_TIME_DEFAULT = 1,
};

enum {
	BUF_PARAM_BYTES_MIN = 0,
	BUF_PARAM_BYTES_MAX = 65536,
	BUF_PARAM_BYTES_DEFAULT = 100
};

enum {
	LOGGER_DEVICE_THROTTLING_DEFAULT = 100,
};

struct logger;
struct writer;
struct reader;
struct log_buffer;

/**
 * @brief Event handler
 * @details Receives and handles an event
 * @param[in] server The logger server
 * @param[in] event The received event
 */
typedef void (*dispatch_event_t)(struct logger *server, struct epoll_event *event);

struct fd_entity {
	dispatch_event_t dispatch_event;
	struct epoll_event event;
	int fd;
};

/**
 * @brief Service a writer
 * @details Specific handler that event is delegated to, according to writers type
 * @param[in] server The logger server
 * @param[in] wr The writer to handle the request
 * @param[in] event The relevant event on file descriptor
 * @return 0 on success, else -errno
 */
typedef int (*service_writer_t)(struct logger* server, struct writer* wr, struct epoll_event* event);

/**
 * @brief Service socket request
 * @details Handle request on socket in respect to msg request type
 * @param[in] server The logger server
 * @param[in] wr The writer to handle the request on the socket
 * @param[in] msg The message containing the request
 * @return 0 on success, else -errno
 */
typedef int (*service_socket_t)(struct logger *server, struct writer *wr, struct dlog_control_msg *msg);

#define LARGEST_STRUCT(a, b) (sizeof(struct a) > sizeof(struct b) ? sizeof(struct a) : sizeof(struct b))
struct writer {
	struct fd_entity   fd_entity;
	struct log_buffer* buf_ptr;
	unsigned           readed;
	service_writer_t   service_writer;
	service_socket_t   service_socket;
	char               buffer[LOG_MAX_PAYLOAD_SIZE + LARGEST_STRUCT(android_logger_entry, pipe_logger_entry)];
};
#undef LARGEST_STRUCT

/**
 * @brief Service reader
 * @details Passes the reader to a more specific servicing handler
 * @param[in] server The logger server
 * @param[in] reader The reader being serviced
 * @return Positive integer if the reader is to be removed, else a nonpositive one
 */
typedef int (*service_reader_t)(struct reader *reader);

struct reader {
	struct fd_entity   fd_entity;
	struct log_file    file;
	log_filter*        filter;
	struct log_buffer* buf_ptr;
	int                read_fd;
	bool               dumpcount;
	struct timespec    last_read_time;
	int                partial_log_size;
	char               partial_log[sizeof(struct logger_entry_with_msg)];
	service_reader_t   service_reader;
	log_storage_reader *log_storage_reader;
	log_id_t           buf_id;
};

struct sock_data {
	struct fd_entity   fd_entity;
	struct log_buffer* buf_ptr;
	struct epoll_event event;
	service_socket_t   service_socket;
};

struct log_buffer {
	struct sock_data   sock_wr;
	struct sock_data   sock_ctl;
	list_head          readers;
	log_id_t           id;
	enum sorting_order sort_by;
	log_storage       *log_storage;
};

struct buf_params {
	int bytes;
	int time;
};

struct logger {
	int                 epollfd;
	unsigned            epollcnt;
	list_head           writers;
	list_head           extra_readers;
	struct log_buffer*  buffers[LOG_ID_MAX];
	struct buf_params   buf_params;
	int                 exiting;
	struct timespec     now;
};

struct socket_config_data {
	int permissions;
	char path[MAX_CONF_VAL_LEN];
};

struct buffer_config_data {
	int size;
	struct socket_config_data write_socket;
	struct socket_config_data ctl_socket;
	enum sorting_order sort_by;
};

struct logger_config_data {
	struct buf_params buf_params;
	list_head logfile_configs;
	int is_buffer_enabled[LOG_ID_MAX];
	struct buffer_config_data buffers[LOG_ID_MAX];
};
