#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>

#include <dlog.h>
#include <logcommon.h>
#include <logprint.h>

struct parsed_params {
	log_id_t      buffer;
	log_priority  priority;
	int           help;
	const char   *tag;
	char         *msg;
	unsigned      count;
	bool          print_stats;
};

void parse_options(int argc, const char **argv, struct parsed_params *params)
{
	assert(argv);
	assert(params);
	assert(!params->msg);

	for (;;) {
		int opt = getopt(argc, (char **) argv, "b:p:t:hc:s");
		if (opt < 0)
			break;

		switch (opt) {
		case 'b':
			params->buffer = log_id_by_name(optarg);
			break;
		case 'p':
			params->priority = filter_char_to_pri(optarg[0]);
			break;
		case 't':
			params->tag = optarg;
			break;
		case 'h':
			params->help = 1;
			return;
		case 'c':
			params->count = atoi(optarg);
			break;
		case 's':
			params->print_stats = true;
			break;
		default:
			break;
		}
	}

	if (optind == argc)
		return;

	params->msg = create_glued_string(argv + optind, argc - optind);
}

int send_log(struct parsed_params *params)
{
	unsigned i;
	int r = EXIT_SUCCESS;
	struct timespec start, end;

	assert(params);

	if (params->print_stats)
		clock_gettime(CLOCK_MONOTONIC, &start);

	for (i = 0; i < params->count; ++i) {
		dlog_error_e ret = __dlog_print(params->buffer, params->priority, params->tag, "%s", params->msg);
		if (ret < 0) {
			r = EXIT_FAILURE;
			fprintf(stderr, "Log #%d: error %d\n", i, ret);
		}
	}

	if (params->print_stats) {
		clock_gettime(CLOCK_MONOTONIC, &end);
		printf("Sending log took %ld us\n", (end.tv_sec - start.tv_sec)*1000000 + (end.tv_nsec - start.tv_nsec) / 1000);
	}
	return r;
}

int send_from_stdin(struct parsed_params *params)
{
	int r = EXIT_SUCCESS;
	params->msg = malloc(LOG_MAX_PAYLOAD_SIZE);
	if (!params->msg)
		return EXIT_FAILURE;
	while (fgets(params->msg, LOG_MAX_PAYLOAD_SIZE, stdin) != NULL) {
		if (send_log(params) == EXIT_FAILURE)
			r = EXIT_FAILURE;
	}

	free(params->msg);
	return r;
}

void print_help(const char *progname)
{
	assert(progname);

	printf("Usage: %s [-p priority] [-b buffer] [-t tag] [-c number] [-s] [message]\n"
	       "\t-p priority \tone of {Debug, Info, Warning, Error, Verbose, Fatal}\n"
	       "\t            \tfirst letter is enough; defaults to Info\n"
	       "\t-b buffer   \tone of {main, system, radio, apps, kmsg, syslog}\n"
	       "\t            \tdefaults to main\n"
	       "\t-t tag      \ta short identification string for identification and filtering purposes\n"
	       "\t            \tcan be anything, defaults to DLOG_SEND\n"
	       "\t-c number   \tspecify how many times each message should be logged\n"
	       "\t-s          \tcollect and print statistics\n"
	       "\t-h          \tshow this help\n"
	       "\tmessage     \tmessage for logging; if not specified, each line from standard input\n"
	       "\t            \tis logged separately until EOF\n",
	       progname);
}

int main(int argc, const char **argv)
{
	struct parsed_params params = {
		.buffer   = LOG_ID_APPS,
		.priority = DLOG_INFO,
		.help     = 0,
		.tag      = "DLOG_SEND",
		.msg      = NULL,
		.count    = 1,
		.print_stats = false
	};
	parse_options(argc, argv, &params);
	if (!params.msg) {
		if (params.help) {
			print_help(argv[0]);
			return EXIT_FAILURE;
		}
		return send_from_stdin(&params);
	}

	int ret = send_log(&params);
	free(params.msg);
	return ret;
}
