/*
 * Copyright (c) 2016, Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include <fcntl.h>
#include <getopt.h>
#include <sys/epoll.h>

#include <fd_info.h>
#include <fdi_pipe.h>
#include <fdi_logger.h>
#include <log_file.h>
#include <logconfig.h>
#include <logretrieve.h>
#include <queued_entry_timestamp.h>
#include <ptrs_list.h>

#include "logutil_doc.h"

/**
 * @addtogroup DLOG_IMPLEMENTATION
 * @{
 * @defgroup DLOG_UTIL Util
 * @brief The log retrieval utility
 * @details The utility responsible for reading and processing logs for the user.
 * @{
 */

typedef enum action_e {
	ACTION_PRINT = 0,
	ACTION_GETSIZE,
	ACTION_CLEAR,
} action_e;

static void show_version(const char *name)
{
	printf("%s version: %s-%s\n", name, __DLOG_VERSION, __DLOG_RELEASE);
}

int parse_options(int argc, char **argv, struct log_file *l_file, int *enabled_buffers, action_e *action, struct additional_options *opt)
{
	assert(argv);
	assert(l_file);
	assert(enabled_buffers);
	assert(action);
	assert(opt);

	int silence = 0;

	while (1) {
		static const struct option long_options[] = {
			{"tid"    , required_argument, NULL,   0},
			{"pid"    , required_argument, NULL,   1},
			{"version",       no_argument, NULL,   2},
			{"color"  , required_argument, NULL,   3},
			{"help"   ,       no_argument, NULL, 'h'},
			{0}
		};
		int err_arg_nondigit = 0;
		int option = getopt_long(argc, argv, "cdt:gsf:r:n:v:b:u:h", long_options, NULL);

		if (option < 0)
			break;

		switch (option) {
		case 0: { /* tid filter */
			int tid;
			if (sscanf(optarg, "%d", &tid) != 1)
				err_arg_nondigit = 1;
			else if (log_add_filter_tid(opt->filter_object, tid))
				goto enomem;
			break;
		}
		case 1: { /* pid filter */
			int pid;
			if (sscanf(optarg, "%d", &pid) != 1)
				err_arg_nondigit = 1;
			else if (log_add_filter_pid(opt->filter_object, pid))
				goto enomem;
			break;
		}
		case 2: { /* version */
			show_version(argv[0]);
			return 1;
		}
		case 3: /* colored headers */
			l_file->colors_auto = false;
			if (!strcmp(optarg, "always"))
				l_file->format.color = true;
			else if (!strcmp(optarg, "auto"))
				l_file->colors_auto = true;
			else if (!strcmp(optarg, "never"))
				l_file->format.color = false;
			else {
				ERR("Error: invalid color setting\n");
				return -EINVAL;
			}
			break;
		case 'd':
			opt->logs_dump = DUMP_INFINITE;
			break;
		case 't':
			if (sscanf(optarg, "%zu", &opt->logs_dump) != 1)
				err_arg_nondigit = 1;
			break;
		case 'c':
			*action = ACTION_CLEAR;
			break;
		case 'g':
			*action = ACTION_GETSIZE;
			break;
		case 'b': {
			log_id_t id = log_id_by_name(optarg);
			if (id == LOG_ID_INVALID) {
				ERR("Error: there is no buffer \"%s\"\n", optarg);
				return -ENOENT;
			}
			bit_set(enabled_buffers, id);
			break;
		}
		case 'u':
			if (sscanf(optarg, "%zu", &opt->logs_size) != 1)
				err_arg_nondigit = 1;
			break;
		case 'f':
			if (logfile_set_path(l_file, optarg) < 0)
				goto enomem;
			break;
		case 'v': {
			const log_print_format format = log_format_from_string(optarg);
			if (format == FORMAT_OFF) {
				ERR("Error: invalid format\n");
				return -EINVAL;
			}
			l_file->format.format = log_format_from_string(optarg);
			break;
		}
		case 's':
			silence = 1;
			list_add(&opt->filter_list, "*:s");
			break;
		case 'r':
			if (sscanf(optarg, "%zu", &l_file->rotate_size_kbytes) != 1)
				err_arg_nondigit = 1;
			break;
		case 'n':
			if (sscanf(optarg, "%zu", &l_file->max_rotated) != 1)
				err_arg_nondigit = 1;
			break;
		case 'h':
			show_help(argv[0], true);
			return 1;
		default: // invalid option or missing mandatory parameter
			show_help(argv[0], false);
			return -EINVAL;
		}

		if (err_arg_nondigit) {
			ERR("Error: -%c requires a numerical parameter\n", option);
			return -EINVAL;
		}
	}

	if (optind < argc)
		while (optind < argc)
			list_add(&opt->filter_list, argv[optind++]);
	else if (!silence)
		list_add(&opt->filter_list, "*:D");

	return 0;

enomem:
	ERR("Error: out of memory\n");
	return -ENOMEM;
}

#ifndef UNIT_TEST
int main(int argc, char **argv)
{
	int enabled_buffers = 0; // bitset
	__attribute__ ((cleanup(log_config_free))) struct log_config conf = {.begin = NULL, .last = NULL};
	action_e action = ACTION_PRINT;
	__attribute__ ((cleanup(fdi_array_free))) struct fd_info **fdi_ptrs = NULL;
	int fdi_cnt = 0;
	int i;
	__attribute__ ((cleanup(logfile_free))) struct log_file l_file;
	int r;
	__attribute__ ((cleanup(additional_options_cleanup))) struct additional_options opt;

	logfile_init(&l_file);
	if (additional_options_init(&opt)) {
		ERR("Error: couldn't allocate memory");
		return 1;
	}

	logfile_set_fd(&l_file, fileno(stdout), 0);

	r = log_config_read(&conf);
	if (r < 0) {
		errno = -r;
		ERR("Warning: unable to read config file: %m\n");
	}

	r = parse_options(argc, argv, &l_file, &enabled_buffers, &action, &opt);
	if (r)
		return r < 0 ? 1 : 0;

	/** Optimisation for short-lived (i.e. dumping) instances.

	   We can assume that such instances will not see a timezone
	   change happen during their lifetime, and even if they do,
	   it is probably best if they ignore it (a dump would ideally
	   behave as if it was atomic). This means we can optimize the
	   behaviour where printing the timestamp requires us to check
	   whether the timezone changed, which involves calling stat()
	   on /etc/localtime, which is not cheap on its own and made
	   even worse when it has to go through multiple symlinks.

	   glibc has a feature where setting the TZ environmental var
	   will cache the timezone, achieving precisely what we want.
	 */
	if (opt.logs_dump != DUMP_CONTINUOUS && !getenv("TZ"))
		putenv("TZ=:/etc/localtime");

	const char *const backend = log_config_get(&conf, "backend");
	if (!backend) {
		ERR("Error: backend not defined\n");
		return 1;
	}
	const bool is_pipe = !strcmp(backend, "pipe");

	enum sorting_order sort_by = get_format_sorting(l_file.format.format);
	bool sorting_needed;
	enum sorting_order proper_sort_by = get_proper_sort_by(sort_by, is_pipe, &conf, &sorting_needed);

	r = l_file.path ? logfile_open(&l_file) : 0;
	if (r < 0) {
		errno = -r;
		ERR("Error while creating the output file: %m\n");
		return 1;
	}

	fdi_cnt = create_initial_fdis(&fdi_ptrs, enabled_buffers, is_pipe, &conf);
	if (fdi_cnt < 0) {
		errno = -fdi_cnt;
		ERR("Error while creating fd info: %m\n");
		return 1;
	}

	for (i = 0; i < fdi_cnt; ++i) {
		struct fd_info *const fdi = fdi_ptrs[i];

		int r = 0;
		switch (action) {
		case ACTION_CLEAR:
			r = fdi->ops->clear(fdi);
			break;
		case ACTION_PRINT:
			// No action here. We have to keep the case because otherwise we get a warning.
			break;
		case ACTION_GETSIZE: {
			uint32_t size;
			r = fdi->ops->getsize(fdi, &size);
			if (!r)
				printf("%s: %u KiB\n", fdi->name, BtoKiB(size));
			break;
		} }

		if (r < 0) {
			static const char *action_names[] = {
				[ACTION_PRINT] = "preparing for printing",
				[ACTION_GETSIZE] = "size getting",
				[ACTION_CLEAR] = "buffer clearing",
			};
			errno = -r;
			ERR("Error while %s: %m\n", action_names[action]);
			return 1;
		}
	}

	if (action == ACTION_PRINT) {
		int r = do_print(fdi_ptrs, fdi_cnt, logfile_callback, &(struct logfile_callback_data) {
			.file = &l_file,
			.sort_by = proper_sort_by,
		}, sorting_needed, proper_sort_by, &conf, &opt);
		if (r < 0) {
			errno = -r;
			ERR("Error while printing logs: %m\n");
			return r;
		}
	}

	return 0;
}
#endif

/**
 * @}
 * @}
 */
