/**
 @defgroup CLI_DLOGUTIL dlogutil usage
 @ingroup DLOG_COMMAND_LINE_UTILITIES
 @brief Log reading utility
 @details dlogutil is a utility that can be utilised to read stored logs.

The most rudimentary way to use it is to run it with no parameters; it then prints all stored logs to the standard output and awaits more.

Parameters exist to change this behaviour:

Parameter       | Description
----------------|---------------------------------
-s              | Set default filter to *:s (see later for details)
-f filename     | Output to given file instead of stdout
-r kbytes       | Rotate log file every this many kilobytes.
-n files        | Keep this many old rotations
-v              | Set log output format.
-b buffer       | Select another buffer
-c              | Clear the selected buffers
-d              | Exit when no more logs are currently available
-t count        | Similar to -d, except only print this many logs from the end
-g              | Get the size of selected buffers instead of printing logs
-h, --help      | Show help info

Additionally, you can specify filters. Filters go behind the parameters and are used to decide which logs are shown.

Each log line has priority and a tag. Filters are specified in the format TAG[:PRIO]. Only logs with given tags and priorities are shown.

You can specify * as the tag to match anything. Beware: only a single asterisk works that way, asterisk is not otherwise expanded. For example, "T*G" filter will NOT match "TAG", only literal "T*G".

If you don't specify any filter, it defaults to *:I
If you specify just * as the tag, without priority, its priority defaults to D
If you specify any other tag without priority, the priority defaults to V

*/

#include <stdbool.h>

static void show_help(const char *cmd, bool requested_by_user)
{
	static const char *format_str = "Usage: %s [options] [filterspecs]"
		"\noptions include:\n"
		"  -s                Set default filter to silent.\n"
		"                    Like specifying filterspec '*:s'\n"
		"  -f <filename>     Log to file. Default to stdout\n"
		"  -r [<kbytes>]     Rotate log every kB (defaults to: " STRINGIFY(DEFAULT_ROTATE_SIZE_KB) "kB). Requires -f and -n > 0, else no rotation\n"
		"  -n <count>        Sets max number of rotated logs to <count> (defaults to: " STRINGIFY(DEFAULT_ROTATE_NUM_FILES) "). Requires -f and -r > 0, else no rotation\n"
		"  -c                clear (flush) the entire log and exit without printing logs\n"
		"  -d                dump the log and then exit (don't block)\n"
		"  -t <count>        print only the most recent <count> lines (implies -d)\n"
		"  -g                get the size of the log's ring buffer and exit\n"
		"  -b <buffer>       request alternate ring buffers (can use multiple)\n"
		"                    ('main', 'radio', 'system', 'apps', 'kmsg', 'syslog')\n"
		"                    the default set is main + system + apps\n"
		"  -u <size>         Sets the size of sort buffer (0 to disable sorting)\n"
		"                    Smaller is faster but lowers sorting quality\n"
		"  --pid <pid>       Filter messages by process id\n"
		"  --tid <tid>       Filter messages by thread id\n"
		"  --color <when>    Toggle color coding of headers (when can be 'always', 'auto' or 'never').\n"
		"                    'auto' will only enable colors when writing to a TTY and is the default\n"
		"  --version         Print DLog version\n"
		"  -h or --help      show this help\n"
		"  -v <format>       Sets the log print format. Formats and how they look:\n"
		"                        brief (default): PRI/TAG(PID): MSG\n"
		"                        long: [TIME_SENT_DATE PRI/TAG PID TID]\n"
		"                              MSG\n"
		"                        rwtime: TIME_RECV_DATE [TIME_SENT_NUMERIC_PRECISE] PRI/TAG(PID, TID): MSG\n"
		"                        recv_realtime: TIME_RECV_DATE_PRECISE PRI/TAG(PID, TID): MSG\n"
		"                        kerneltime: TIME_SENT_NUMERIC_PRECISE PRI/TAG(PID, TID): MSG\n"
		"                        threadtime: TIME_SENT_DATE_PRECISE_TIMEZONE PRI/TAG(PID, TID): MSG\n"
		"                        time: TIME_SENT_DATE_PRECISE_TIMEZONE PRI/TAG(PID): MSG\n"
		"                        raw: MSG\n"
		"                        thread: PRI(PID, TID) MSG\n"
		"                        process: PRI(PID) MSG (TAG)\n"
		"                        tag: PRI/TAG: MSG\n"
		"                    And the meanings of the symbols:\n"
		"                        PRI: priority character (VDIWEFS, see below)\n"
		"                        TAG: log tag, usually denotes sending program\n"
		"                        PID: process ID, prefix 'P' when alongside TID\n"
		"                        TID: thread ID, prefix 'T' when alongside PID\n"
		"                        TIME: timestamp, with the following variants:\n"
		"                          DATE: shown as date (eg 12-31 23:59:59)\n"
		"                          NUMERIC: shown as raw integer\n"
		"                          SENT: when the log was sent by app\n"
		"                          RECV: when the log was received by dlog\n"
		"                          PRECISE: extra precision to milliseconds\n"
		"                          TIMZEONE: also shows timezone\n"
		"                    IMPORTANT: logs are sorted by their SENT timestamp\n"
		"                    so choosing a format that shows the RECV timestamp\n"
		"                    means they can appear out of order. Also, some logs\n"
		"                    supply their timestamp in seconds since system boot,\n"
		"                    in which case using a DATE timestamp does not make\n"
		"                    much sense (usually shows dates in January 1970).\n"
		"                    The most reliable choices are 'rwtime' and 'recv_realtime',\n"
		"                    but the default is 'brief' for legacy reasons.\n"
		"                    Additionally, a 'json' format is supported, in which\n"
		"                    each log becomes a line with a JSON object containing\n"
		"                    the following fields: 'priority' (string), 'tid' (number),\n"
		"                    'pid' (number), 'recv_real' (string), 'recv_mono' (number),\n"
		"                    'sent_real' (string), 'sent_mono' (number), 'tag' (string)\n"
		"                    and 'msg' (string). Note that some of the timestamps\n"
		"                    can be missing, but at least one will be available.\n"
		"filterspecs are a series of\n<tag>[*][:priority]\n and \n<prefix>*[:=priority]\n"
		"where <tag> is a log component tag\n"
		"<prefix>* matches all log component tags starting with the prefix\n"
		"and * sets global default policy for all tags not matched by other filterspecs \n"
		"priority is:\n"
		"  V    Verbose - matches all priorities\n"
		"  D    Debug   - matches priorities higher or equal to Debug\n"
		"  I    Info    - matches priorities higher or equal to Info\n"
		"  W    Warn    - matches priorities higher or equal to Warn\n"
		"  E    Error   - matches priorities higher or equal to Error\n"
		"  F    Fatal   - matches priorities equal to Fatal\n"
		"  S    Silent  - does not match any priority - supress output\n"
		"\n =<priority> - matches exactly specified policy, for example:\n"
		" =W matches only Warn priority\n"
		"\n'*' means '*:D' and <tag> by itself means <tag>:V\n"
		"If no filterspec is found, filter defaults to '*:D'\n\n";

	if (requested_by_user)
		printf(format_str, cmd);
	else
		ERR(format_str, cmd);
}
