// C
#include <assert.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>

// POSIX
#include <unistd.h>
#include <sys/ioctl.h>

// DLog
#include <backend_androidlogger.h>
#include <dlog.h>
#include <dlog_ioctl.h>
#include <logcommon.h>
#include <logconfig.h>

/**
 * @brief Set Android Logger protocol
 * @details Try to set the protocol version to 2
 * @param[in] fd file desciptor of the logger device
 * @return -errno on failure; 1 or 2 on success depending on which version is set
 * @remarks Only useful under the Android Logger backend.
 *
 * The reason using protocol version 2 is needed is because devices which support
 * it display behaviour which is not very useful (and borderline incorrect):
 * the GET_LOG_LEN ioctl returns how much the contents of the buffer actually
 * take up in the format that they are stored in, not how much they would take
 * when converted to the protocol which is actually used. This means that while
 * using protocol version 1 (i.e. the default) on devices which support version 2
 * (i.e. the vast majority) the sum of read() calls doable on the buffer will
 * not add up to the total returned by GET_LOG_LEN (since v2 has a larger header).
 */
static int logger_set_protocol_version(int fd)
{
	assert(fd >= 0);

	int ver = 2;
	int r = ioctl(fd, LOGGER_SET_VERSION, (char *)&ver);
	if (r >= 0)
		return 2;

	/* EINVAL only happens in version 1 of the protocol, since
	 * it stands for EINs Version Android Logger. Also, somewhat
	 * interestingly, this is not because it's not yet version 2
	 * but because v1 doesn't actually support the SET_VERSION call.
	 * DLog doesn't actually support version 1 devices but it's
	 * good to have feedback on which version is in place. */
	if (errno == EINVAL)
		return 1;

	return -errno;
}

/*
 * @brief Sets proper, existing logger device file descriptor
 * @details Checks which path from list exists in filesystem, opens it and returns fd passed as a parameter
 * @param[in] buf_id Log buffer id
 * @param[in] config_list List of possible buffer device paths, delimited with :
 * @param[in] open_flags Specifies flags passed to open()
 * @param[out] fd Address of the opened fd to return
 * @param[out] actual_path The path that was actually used
 * @param[in] path_size Size of the `actual_path` buffer
 * @return 1 on success, 0 on empty path, otherwise error code
 * @notes fd parameter is modified only on success
 */
int logger_open_buffer_get_path(int buf_id, const char *const config_list, int open_flags, int *fd, char *actual_path, size_t path_size)
{
	assert(fd);
	assert(config_list);
	assert(buf_id >= 0 && buf_id < LOG_ID_MAX);
	assert(path_size > 0 || !actual_path);

	char list[MAX_CONF_VAL_LEN];
	char *saveptr = list;
	char *dev_path;
	int ret;

	strncpy(list, config_list, sizeof list); /* config_list variable may be modified by strtok_r() so we need a local copy */
	list[sizeof list - 1] = '\0';

	dev_path = strtok_r(list, ":", &saveptr);

	while (dev_path != NULL) {
		if ((ret = open(dev_path, open_flags)) >= 0) {

			/* DLog can only read logs if the connection is working under
			 * protocol version 2 (see <queued_entry.h> for more info).
			 * As of version 2, it is not possible to change the protocol
			 * on write-only connections, but that does not matter since
			 * the write interface is the same as in version 1 anyway. */
			if ((open_flags & O_ACCMODE) != O_WRONLY) {
				const int proto_ver = logger_set_protocol_version(ret);
				if (proto_ver != 2) {
					close(ret);

					/* Don't bother iterating further - if this device couldn't
					 * handle this request, any others won't be able to either.
					 * In theory it might be fine for optional buffers, but in
					 * practice that would just hide a deeper problem with the
					 * platform (obsolete kernel, not using pipe backend, etc.) */
					return proto_ver < 0 ? proto_ver : -ENOTSUP;
				}
			}

			*fd = ret;
			if (actual_path) {
				*actual_path = '\0';
				strncat(actual_path, dev_path, path_size - 1);
			}
			return 1;
		} else if (errno != ENOENT) {
			return -errno;
		}

		dev_path = strtok_r(NULL, ":", &saveptr);
	}

	if (config_list[strlen(config_list)-1] == ':') {
		/* no "real" path found but there's the empty string at the end that strok wouldn't return */
		return 0;
	}

	char err_message[MAX_CONF_VAL_LEN + 64];
	snprintf(err_message, sizeof err_message, "LOG BUFFER #%d %s HAS NO PATH SET IN CONFIG", buf_id, log_name_by_id(buf_id));
	syslog_critical_failure(err_message);

	return -ENOENT;
}

int logger_open_buffer(int buf_id, const char *const config_list, int open_flags, int *fd)
{
	return logger_open_buffer_get_path(buf_id, config_list, open_flags, fd, NULL, 0U);
}

/*
 * @brief Sets proper, existing logger device file descriptor
 * @details Checks which path from config list exists in filesystem, opens it and returns fd passed as a parameter
 * @param[in] buf_id Log buffer id
 * @param[in] conf The configuration
 * @param[in] open_flags Specifies flags passed to open()
 * @param[out] fd Address of the opened fd to return
 * @param[out] actual_path The path that was actually used
 * @param[in] path_size Size of the `actual_path` buffer
 * @return 1 on success, 0 on empty path, otherwise error code
 * @notes fd parameter is modified only on success
 */
int logger_open_buffer_from_config_get_path(int buf_id, const struct log_config *conf, int open_flags, int *fd, char *actual_path, size_t path_size)
{
	assert(conf);
	assert(buf_id >= 0 && buf_id < LOG_ID_MAX);
	assert(fd);
	assert(path_size > 0 || !actual_path);

	const char *const config_list = log_config_get(conf, log_name_by_id(buf_id));

	if (!config_list)
		return -ENOENT;

	return logger_open_buffer_get_path(buf_id, config_list, open_flags, fd, actual_path, path_size);
}

int logger_open_buffer_from_config(int buf_id, const struct log_config *conf, int open_flags, int *fd)
{
	return logger_open_buffer_from_config_get_path(buf_id, conf, open_flags, fd, NULL, 0U);
}
