/*
 * Copyright (c) 2018, Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>

#include <log_file.h>
#include <queued_entry_timestamp.h>


#include "fd_info.h"
#include "sort_vector.h"

/**
 * @brief Create struct fd_info
 * @details Allocates memory for the struct and initializes it with given arguments
 * @param[in] ops source-dependent operation set
 * @return FD created struct or NULL on memory allocation failure
 */
struct fd_info *fdi_create(struct fd_ops *ops, const char *name)
{
	assert(ops);

	struct fd_info *fdi = malloc(sizeof(struct fd_info));
	if (!fdi)
		return NULL;

	fdi->fd = -1;
	fdi->ops = ops;
	fdi->name = name;
	fdi->priv_data = NULL;

	return fdi;
}

void fdi_free(struct fd_info *fdi)
{
	if (!fdi)
		return;

	if (fdi->ops->destroy)
		fdi->ops->destroy(fdi);

	if (fdi->fd >= 0)
		close(fdi->fd);

	free(fdi);
}

void fdi_array_free(struct fd_info ***arr)
{
	assert(arr);

	if (!*arr)
		return;

	for (int i = 0; (*arr)[i]; ++i)
		fdi_free((*arr)[i]);

	free(*arr);
}

int fdi_push_log(struct fd_info *fdi, struct sort_vector *logs, write_callback callback, void *userdata, log_filter *filter)
{
	assert(fdi);
	assert(userdata);

	struct logger_entry *temp = fdi->ops->extract_entry(fdi);
	if (!temp)
		return -ENOMEM;

	if (!log_should_print_line(filter, temp)) {
		free(temp);
		return 0;
	}

	int r;

	if (IS_VECTOR_SIZE_SORTABLE(logs->size)) {
		struct timespec ts;
		extract_timestamp(temp, logs->sort_by, &ts);

		/* check if new entry is newer than util's start moment */
		if (ts.tv_sec > logs->start.tv_sec || (ts.tv_sec == logs->start.tv_sec && ts.tv_nsec > logs->start.tv_nsec))
			logs->old_logs_dumped = true;

		r = sort_vector_push(logs, temp, callback, userdata);
	} else {
		r = callback(temp, userdata);
		free(temp);
	}

	return r;
}
