/*
 * Copyright (c) 2018, Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// C
#include <errno.h>

// POSIX
#include <fcntl.h>
#include <unistd.h>

// DLog
#include <backend_androidlogger.h>
#include <dlog_ioctl.h>
#include <logconfig.h>
#include <queued_entry_timestamp.h>

// dlogutil
#include "fd_info.h"
#include "fdi_logger.h"

#define UNLIMITED_LOG_LEN -1

struct logger_priv_data {
	int log_len;
	struct logger_entry *entry;
};

static int logger_ioctl(int fd, int ioctl_id)
{
	assert(fd >= 0);

	int ret = ioctl(fd, ioctl_id);
	if (ret < 0)
		return -errno;
	return ret;
}

static int logger_getsize(struct fd_info *fdi, uint32_t *val)
{
	assert(fdi);

	int r = logger_ioctl(fdi->fd, LOGGER_GET_LOG_BUF_SIZE);
	if (r < 0)
		return r;

	*val = (uint32_t)r;
	return 0;
}

/**
 * @brief Get buffer filled size
 * @details Sends a get log len ioctl to given fd
 * @param[in] fd File descriptor of the buffer
 * @return -errno on failure, else >= 0 length in bytes
 * @remarks ANDROID LOGGER version
 */
static int logger_get_log_len(int fd)
{
	return logger_ioctl(fd, LOGGER_GET_LOG_LEN);
}

/**
 * @brief Clear the buffer
 * @details Sends a clear ioctl to given fd
 * @param[in] fdi File descriptor info
 * @return 0 on success, -errno on failure
 */
static int logger_clear(struct fd_info *fdi)
{
	assert(fdi);
	const int ret = logger_ioctl(fdi->fd, LOGGER_FLUSH_LOG);
	return ret < 0 ? ret : 0;
}

static bool path_already_opened(char *path, list_head used_paths)
{
	for (list_head i = used_paths; i; list_next(&i))
		if (!strcmp(path, (char *)list_at(i)))
			return true;

	return false;
}

/**
 * @brief Open an Android Logger device
 * @param[in] fdi File descriptor info
 * @param[in] conf The configuration
 * @return 0 on success, -errno on failure
 */
static int logger_create(struct fd_info *fdi, const struct log_config *conf, const char *name, list_head *used_paths)
{
	assert(fdi);
	assert(conf);
	assert(name);
	assert(used_paths);

	int ret = 0;
	int const PATH_SIZE = MAX_CONF_VAL_LEN;

	char *const actual_path = malloc(PATH_SIZE);
	struct logger_priv_data *const lpd = malloc(sizeof *lpd);
	if (!actual_path || !lpd) {
		ret = -ENOMEM;
		goto failure;
	}

	int opened = logger_open_buffer_from_config_get_path(log_id_by_name(name), conf, O_RDWR, &fdi->fd, actual_path, PATH_SIZE);
	if (opened <= 0) {
		ret = opened ?: -EINVAL;
		goto failure;
	}

	if (path_already_opened(actual_path, *used_paths)) {
		ret = -EALREADY;
		goto free_fdi;
	} else if (!list_add(used_paths, actual_path)) {
		ret = -ENOMEM;
		goto free_fdi;
	}

	lpd->entry = NULL;
	fdi->priv_data = lpd;

	return 0;

free_fdi:
	close(fdi->fd);
	fdi->fd = -1;
failure:
	free(lpd);
	free(actual_path);

	return ret;
}

static void logger_destroy(struct fd_info *fdi)
{
	assert(fdi);
	struct logger_priv_data *lpd = (struct logger_priv_data *)fdi->priv_data;

	if (!lpd)
		return;

	free(lpd->entry);
	free(lpd);
	fdi->priv_data = NULL;
}

static int logger_prepare_print(struct fd_info *fdi, int dump, list_head filter_list, log_filter *filter_object)
{
	assert(fdi);
	struct logger_priv_data *const lpd = (struct logger_priv_data *)fdi->priv_data;
	assert(lpd);
	assert(filter_object);

	for (list_head iter = filter_list; iter; list_next(&iter))
		log_add_filter_string(filter_object, (const char *)list_at(iter));

	if (dump == DUMP_CONTINUOUS) {
		lpd->log_len = UNLIMITED_LOG_LEN;
		return 0;
	}

	lpd->log_len = logger_get_log_len(fdi->fd);
	if (lpd->log_len < 0)
		return lpd->log_len;

	if (lpd->log_len == 0) {
		/* signify we don't want to get printed;
		   we can't rely on the regular removal
		   path because there won't be a read()
		   to trigger it */
		return 1;
	}

	return 0;
}

static int logger_read(struct fd_info *fdi)
{
	assert(fdi);
	struct logger_priv_data *lpd = (struct logger_priv_data *)fdi->priv_data;
	assert(lpd);

	if (lpd->entry)
		return -EAGAIN;

	char buff[LOG_MAX_PAYLOAD_SIZE + sizeof(struct android_logger_entry)];
	int r = read(fdi->fd, buff, sizeof buff);
	if (r < 0)
		return -errno;

	struct android_logger_entry *const ale = (struct android_logger_entry *)buff;
	if (r <= sizeof *ale)
		return -EINVAL;

	lpd->entry = malloc(r - sizeof *ale + sizeof *lpd->entry);
	if (!lpd->entry)
		return -ENOMEM;

	parse_androidlogger_message(ale, lpd->entry, r);

	lpd->entry->sec_recv_mono = INVALID_SEC;
	lpd->entry->nsec_recv_mono = INVALID_NSEC;
	lpd->entry->sec_recv_real = INVALID_SEC;
	lpd->entry->nsec_recv_real = INVALID_NSEC;

	if (lpd->log_len != UNLIMITED_LOG_LEN) {
		lpd->log_len -= r;

		if (lpd->log_len <= 0)
			return 0; // emulate EOF
	}

	return r;
}

static struct logger_entry *logger_extract_entry(struct fd_info *fdi)
{
	assert(fdi);
	struct logger_priv_data *lpd = (struct logger_priv_data *)fdi->priv_data;
	assert(lpd);

	struct logger_entry *const ret = lpd->entry;
	lpd->entry = NULL;
	return ret;
}

static const struct logger_entry *logger_peek_entry(const struct fd_info *fdi)
{
	assert(fdi);
	const struct logger_priv_data *const lpd = (const struct logger_priv_data *)fdi->priv_data;
	assert(lpd);

	return lpd->entry;
}

static bool logger_has_log(struct fd_info *fdi)
{
	assert(fdi);
	struct logger_priv_data *lpd = (struct logger_priv_data *)fdi->priv_data;
	assert(lpd);

	return lpd->entry != NULL;
}

struct fd_ops ops_logger = {
	.create = logger_create,
	.destroy = logger_destroy,
	.read = logger_read,
	.has_log = logger_has_log,
	.extract_entry = logger_extract_entry,
	.peek_entry = logger_peek_entry,
	.prepare_print = logger_prepare_print,
	.clear = logger_clear,
	.getsize = logger_getsize,
};
