/*
 * Copyright (c) 2018, Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <assert.h>
#include <stdio.h>
#include <string.h>

#include <logconfig.h>
#include <logcommon.h>
#include <logpipe.h>

#include "fd_info.h"
#include "fdi_pipe.h"
#include "fdi_pipe_internal.h"

static int pipe_getsize(struct fd_info *fdi, uint32_t *val)
{
	assert(fdi);
	struct pipe_priv_data *ppd = (struct pipe_priv_data *)fdi->priv_data;
	assert(ppd);
	assert(ppd->sock_fd >= 0);

	int r = send_dlog_request(ppd->sock_fd, DLOG_REQ_GETSIZE, NULL, 0);
	if (r < 0)
		return r;

	uint32_t *data;
	int datalen;
	r = recv_dlog_reply(ppd->sock_fd, DLOG_REQ_GETSIZE, (void **)&data, &datalen);
	if (r < 0)
		goto cleanup;
	if (!data || datalen != sizeof *data) {
		r = -EINVAL;
		goto cleanup;
	}
	*val = *data;

cleanup:
	free(data);
	return r;
}

/**
 * @brief Clear request
 * @details Send a CLEAR request to the logger daemon
 * @param[in] fdi The fd info to work on
 * @return 0 on success, -errno on failure
 */
static int pipe_clear(struct fd_info *fdi)
{
	assert(fdi);
	struct pipe_priv_data *ppd = (struct pipe_priv_data *)fdi->priv_data;
	assert(ppd);
	assert(ppd->sock_fd >= 0);

	return send_dlog_request(ppd->sock_fd, DLOG_REQ_CLEAR, NULL, 0);
}

/**
 * @brief Send a logger request
 * @details Send a logging request to the daemon
 * @param[in] filters Log filters to use
 * @param[in] dump Whether dumping or not
 * @param[in] sock_fd Socket file descriptor
 * @return 0 on success, -errno on failure
 */
static int send_logger_request(list_head filters, int dump, int sock_fd)
{
	assert(filters);
	assert(sock_fd >= 0);

	char request_string[MAX_LOGGER_REQUEST_LEN] = "dlogutil";
	int len = strlen(request_string) + 1 /* null byte */;

	if (dump) {
		static const char dump_str[] = " -d";
		strncat(request_string, dump_str, sizeof request_string - len);
		len += sizeof dump_str - 1;
	}

	for (list_head iter = filters; iter; list_next(&iter)) {
		const char *filter = (const char *)list_at(iter);
		int arglen = strnlen(filter, MAX_LOGGER_REQUEST_LEN);
		int needed = arglen + 1 /* space */;
		if (len + needed > sizeof(request_string))
			return -E2BIG;

		strncat(request_string, " ", 1);
		strncat(request_string, filter, arglen + 1);
		len += needed;
	}

	return send_dlog_request(sock_fd, DLOG_REQ_HANDLE_LOGUTIL, request_string, len);
}

/**
 * @brief Open a connection to the dlog daemon
 * @param[in] fdi File descriptor info
 * @param[in] conf The configuration
 * @return 0 on success, -errno on failure
 */
static int pipe_create(struct fd_info *fdi, const struct log_config *conf, const char *name, list_head *used_paths)
{
	assert(fdi);
	assert(conf);
	assert(name);
	assert(used_paths);

	char conf_key[MAX_CONF_KEY_LEN];
	snprintf(conf_key, sizeof(conf_key), "%s_ctl_sock", name);

	const char *sock_path = log_config_get(conf, conf_key);
	if (!sock_path)
		return -ENOENT;

	int sock_fd = connect_sock(sock_path);
	if (sock_fd < 0)
		return sock_fd;

	struct pipe_priv_data *ppd = malloc(sizeof *ppd);
	if (!ppd) {
		close(sock_fd);
		return -ENOMEM;
	}
	ppd->data_len = 0;
	ppd->sock_fd = sock_fd;
	ppd->offset = 0;

	fdi->priv_data = ppd;
	return 0;
}

static void pipe_destroy(struct fd_info *fdi)
{
	assert(fdi);
	struct pipe_priv_data *ppd = (struct pipe_priv_data *)fdi->priv_data;
	if (!ppd)
		return;

	if (ppd->sock_fd >= 0)
		close(ppd->sock_fd);
	free(ppd);

	fdi->priv_data = NULL;
}

static int pipe_prepare_print(struct fd_info *fdi, int dump, list_head filter_list, log_filter *filter_object)
{
	assert(filter_list);
	assert(filter_object);
	assert(fdi);
	struct pipe_priv_data *ppd = (struct pipe_priv_data *)fdi->priv_data;
	assert(ppd);
	assert(ppd->sock_fd >= 0);

	log_add_filter_string(filter_object, "*:V"); // lets everything through (actual filtering done by the daemon)

	int r = send_logger_request(filter_list, dump, ppd->sock_fd);
	if (r < 0)
		return r;

	int pipe_fd = recv_pipe(ppd->sock_fd, DEFAULT_WAIT_PIPE_MS);
	if (pipe_fd < 0)
		return pipe_fd;

	fdi->fd = pipe_fd;

	return 0;
}

static int pipe_read(struct fd_info *fdi)
{
	assert(fdi);
	struct pipe_priv_data *ppd = (struct pipe_priv_data *)fdi->priv_data;
	assert(ppd);

	if (ppd->data_len == sizeof ppd->buff)
		return -EAGAIN;

	if (ppd->offset > 0 && ppd->data_len > 0)
		memmove(ppd->buff, ppd->buff + ppd->offset, ppd->data_len);
	ppd->offset = 0;

	int r = read(fdi->fd, ppd->buff + ppd->offset + ppd->data_len, RECEIVE_BUFFER_SIZE - ppd->data_len - ppd->offset);
	if (r < 0)
		return -errno;

	ppd->data_len += r;
	return r;
}

static bool pipe_has_log(struct fd_info *fdi)
{
	assert(fdi);
	struct pipe_priv_data *ppd = (struct pipe_priv_data *)fdi->priv_data;
	assert(ppd);

	const struct logger_entry *const le = (const struct logger_entry *)(ppd->buff + ppd->offset);
	return ppd->data_len >= sizeof le->len && ppd->data_len >= le->len;
}

static struct logger_entry *pipe_extract_entry(struct fd_info *fdi)
{
	assert(fdi);
	struct pipe_priv_data *ppd = (struct pipe_priv_data *)fdi->priv_data;
	assert(ppd);

	struct logger_entry *const buf_le = (struct logger_entry *)(ppd->buff + ppd->offset);
	if (!buf_le->len)
		return NULL;

	struct logger_entry *const new_le = malloc(buf_le->len);
	if (!new_le)
		return NULL;

	memcpy(new_le, buf_le, buf_le->len);
	ppd->data_len -= buf_le->len;
	ppd->offset += buf_le->len;

	return new_le;
}

static const struct logger_entry *pipe_peek_entry(const struct fd_info *fdi)
{
	assert(fdi);
	const struct pipe_priv_data *const ppd = (const struct pipe_priv_data *)fdi->priv_data;
	assert(ppd);

	return (const struct logger_entry *)(ppd->buff + ppd->offset);
}

struct fd_ops ops_pipe = {
	.create = pipe_create,
	.destroy = pipe_destroy,
	.read = pipe_read,
	.has_log = pipe_has_log,
	.extract_entry = pipe_extract_entry,
	.peek_entry = pipe_peek_entry,
	.prepare_print = pipe_prepare_print,
	.clear = pipe_clear,
	.getsize = pipe_getsize,
};
