/*
 * DLOG
 * Copyright (c) 2005-2008, The Android Open Source Project
 * Copyright (c) 2012-2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <log_file.h>
#include <queued_entry_timestamp.h>

#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <assert.h>

#define DEFAULT_LOGFILE_PERM (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP)

/**
 * @addtogroup SHARED_FUNCTIONS
 * @{
 */

void logfile_init(struct log_file *l_file)
{
	assert(l_file);

	memset(l_file, 0, sizeof(struct log_file));

	l_file->fd = -1;
	l_file->rotate_size_kbytes = DEFAULT_ROTATE_SIZE_KB;
	l_file->max_rotated = DEFAULT_ROTATE_NUM_FILES;
	l_file->format.format = FORMAT_BRIEF;
	l_file->colors_auto = true;
	l_file->prev_sec = INT_MIN;
	l_file->prev_nsec = INT_MIN;
}

static void logfile_set_should_close(struct log_file *l_file, int on_off)
{
	assert(l_file);
	if (on_off)
		assert(l_file->fd >= 0);

	l_file->should_close = !!on_off;
}

void logfile_set_fd(struct log_file *l_file, int fd, int should_close)
{
	assert(l_file);

	l_file->fd = fd;
	logfile_set_should_close(l_file, should_close);
}

int logfile_set_path(struct log_file *l_file, const char *path)
{
	assert(l_file);

	char *p = strdup(path);
	if (!p)
		return -errno;

	if (l_file->path)
		free(l_file->path);
	l_file->path = p;

	return 0;
}

static void logfile_close_if_needed(struct log_file *l_file)
{
	assert(l_file);

	if (l_file->fd >= 0 && l_file->should_close) {
		close(l_file->fd);
		logfile_set_should_close(l_file, 0);
		l_file->fd = -1;

	}
}

void logfile_free(struct log_file *l_file)
{
	assert(l_file);

	if (l_file->path) {
		free(l_file->path);
		l_file->path = NULL;
	}
	logfile_close_if_needed(l_file);
}

static int logfile_update_fsize(struct log_file *l_file)
{
	assert(l_file);
	assert(l_file->fd >= 0);

	struct stat st;

	if (fstat(l_file->fd, &st) < 0)
		return -errno;

	l_file->size = st.st_size;
	return 0;
}

static int logfile_open_internal(struct log_file *l_file)
{
	assert(l_file);
	assert(l_file->path);

	logfile_close_if_needed(l_file);

	int fd = open(l_file->path, O_CREAT | O_WRONLY | O_APPEND, DEFAULT_LOGFILE_PERM);
	if (fd < 0)
		return -errno;

	logfile_set_fd(l_file, fd, 1);

	l_file->fd = fd;
	logfile_set_should_close(l_file, 1);

	logfile_update_fsize(l_file);

	return 0;
}

/**
 * @brief Open a log file
 * @details Open a log file into the passed structure.
 * @param[in] l_file The file structure to contain the opened file descriptor
 * @remarks Creates the file if it does not exist, else appends to it.
 */
int logfile_open(struct log_file *l_file)
{
	assert(l_file);

	int r = logfile_open_internal(l_file);
	if (r < 0)
		return r;

	r = logfile_rotate_needed(l_file);
	if (r > 0)
		logfile_do_rotate(l_file);

	return r;
}

/**
 * @brief Checks if log file should be rotated
 * @param[in] l_file The file structure to contain the currently opened file
 * @remarks Updates l_file->size with on-disk file size
 * @returns 1 if rotate is needed, 0 if not, -errno on error
 */
int logfile_rotate_needed(struct log_file *l_file)
{
	assert(l_file);

	if (l_file->path == NULL || l_file->rotate_size_kbytes <= 0 || l_file->max_rotated <= 0)
		return 0;

	logfile_update_fsize(l_file);
	return BtoKiB(l_file->size) > l_file->rotate_size_kbytes;
}

/**
 * @brief Rotate log files
 * @details Opens a new file and moves existing files further back
 * @param[in] file The file structure to contain the currently opened file
 */
void logfile_do_rotate(struct log_file *file)
{
	assert(file);
	assert(file->path);

	int i;
	char path0[PATH_MAX];
	char path1[PATH_MAX];

	for (i = file->max_rotated ; i > 0 ; i--) {
		snprintf(path1, PATH_MAX, "%s.%d", file->path, i);
		if (i - 1 == 0)
			snprintf(path0, PATH_MAX, "%s",  file->path);
		else
			snprintf(path0, PATH_MAX, "%s.%d", file->path, i - 1);
		if (rename(path0, path1) < 0 && errno != ENOENT)
			ERR("while rotating log file: %s", file->path);
	}

	/* open log file again */
	logfile_open_internal(file);
}

static void logfile_add_timestamp(struct log_file *file, struct timespec ts)
{
	file->prev_sec = ts.tv_sec;
	file->prev_nsec = ts.tv_nsec;
}

/**
 * @brief Write with rotation
 * @details Writes the entry to given file, automatically handling file rotation
 * @param[in] e The entry to write
 * @param[in] file The file to write to
 * @returns 0 if log was successfully written, else 1
 */
int logfile_write_with_rotation(const struct logger_entry *e, struct log_file *file, enum sorting_order sort_by)
{
	int written_bytes = 0;

	struct timespec ts;
	extract_timestamp(e, sort_by, &ts);

	if (ts.tv_sec < file->prev_sec || (ts.tv_sec == file->prev_sec && ts.tv_nsec < file->prev_nsec)) {
		struct logger_entry_with_msg msg;
		memcpy(&msg, e, sizeof *e);
		int r = snprintf(msg.msg, sizeof msg.msg, "%s%c%s", log_buffer_get_tag(e), '\0', "INFO: Following log entry could not be sorted and is out of order.");

		if (r < 0) {
			ERR("unable to format out-of-order message %m");
			written_bytes = 0;
		} else {
			msg.header.len = r + 1 + sizeof *e;

			written_bytes += log_print_log_line(file->format, file->fd, file->colors_auto, &msg.header);

			if (written_bytes < 0) {
				ERR("unable to write out-of-order message %m");
				written_bytes = 0;
			}
		}
	}

	written_bytes += log_print_log_line(file->format, file->fd, file->colors_auto, e);
	if (written_bytes <= 0)
		return 1;
	file->size += written_bytes;

	logfile_add_timestamp(file, ts);

	if (logfile_rotate_needed(file))
		logfile_do_rotate(file);
	return 0;
}

int logfile_callback(const struct logger_entry *e, void *userdata)
{
	struct logfile_callback_data *data = userdata;
	return logfile_write_with_rotation(e, data->file, data->sort_by);
}

/**
 * @}
 */
