/*
 * DLOG
 * Copyright (c) 2005-2008, The Android Open Source Project
 * Copyright (c) 2012-2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <syslog.h>

#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/un.h>
#include <unistd.h>
#include <fcntl.h>
#include <poll.h>
#include <logcommon.h>
#include <logpipe.h>
#include <assert.h>
#include <ctype.h>
#include <dlog_ioctl.h>

/**
 * @addtogroup SHARED_FUNCTIONS
 * @{
 */

#define MAX_PREFIX_SIZE 32
#define LINE_MAX        (MAX_PREFIX_SIZE+PATH_MAX)

static struct {
	log_id_t id;
	char * name;
} logid_map[LOG_ID_MAX] = {
	{ .id = LOG_ID_MAIN,   .name = "main" },
	{ .id = LOG_ID_RADIO,  .name = "radio" },
	{ .id = LOG_ID_SYSTEM, .name = "system" },
	{ .id = LOG_ID_APPS,   .name = "apps" },
	{ .id = LOG_ID_KMSG,   .name = "kmsg" },
	{ .id = LOG_ID_SYSLOG, .name = "syslog" },
};

// message header structure to send and receive pipe file descriptor
// contains header data required for building message
// to send it using sendmsg(2) and receive using recvmsg(2)
// all header data may be placed on stack while sending or receiving
struct pipe_msg_header {
	struct msghdr hdr;
	struct iovec iov;
	char ctl[CMSG_SPACE(sizeof(int))];
};

// message control header containing file descriptor
// to initialize pipe fd sending message with
// and to compare incoming header on receiver side with
static struct cmsghdr g_pipe_msg_ctl = {
	.cmsg_len = CMSG_LEN(sizeof(int)),
	.cmsg_level = SOL_SOCKET,
	.cmsg_type = SCM_RIGHTS
};

/**
 * @brief Initialize message header structure for sending or receiving pipe fd
 * @details A file descriptor is placed in control header
 * @param[in] msg message header structure to initialize with
 * @param[in] data dlog control message to be send with the message or placeholder
 *                 for receiving the data to be verified on receiver side
 */
static void init_pipe_msg_header(struct pipe_msg_header *msg, struct dlog_control_msg *data)
{
	assert(msg);
	assert(data);

	msg->iov = (struct iovec) {
		.iov_base = data,
		.iov_len = sizeof(*data)
	};
	memset(msg->ctl, 0, sizeof(msg->ctl));
	msg->hdr = (struct msghdr) {
		.msg_name = NULL,
		.msg_namelen = 0,
		.msg_iov = &msg->iov,
		.msg_iovlen = 1,
		.msg_control = msg->ctl,
		.msg_controllen = sizeof(msg->ctl)
	};
}

/**
 * @brief Initialize message data for sending a control message
 * @param[in] data dlog control message to be filled
 * @param[in] request The request type
 * @param[in] result The result of the request
 */
static void init_pipe_msg_data(struct dlog_control_msg *data, int request, int result)
{
	assert(data);

	*data = (struct dlog_control_msg) {
		.length = sizeof(*data),
		.request = request,
		.result = result,
	};
}

/**
 * @brief verify dlog control message data on receiver side
 * @param[in] data dlog control message received
 * @return non zero if positively verified
 */
static int verify_pipe_msg_data(struct dlog_control_msg *data)
{
	assert(data);

	return data->length == sizeof(*data)
		&& data->request == (DLOG_REQ_PIPE | DLOG_REQ_REPLY)
		&& data->result == DLOG_REQ_RESULT_OK;
}

/**
 * @brief Initialize message control for sending pipe fd
 * @param[in] msg_hdr message control header to be populated with pipe fd to send
 * @param[in] pipefd the pipe file descriptor to send with the message
 */
static void init_pipe_msg_control(struct msghdr *msg_hdr, int pipefd)
{
	assert(msg_hdr);

	struct cmsghdr *cmsg = CMSG_FIRSTHDR(msg_hdr);
	*cmsg = g_pipe_msg_ctl;
	memcpy(CMSG_DATA(cmsg), &pipefd, sizeof(pipefd));
}

/**
 * @brief Retrieve a file descriptor from a received control message header
 * @param[in] msg_hdr message control header to parse
 * @return The file descriptor received in the control header, -errno on failure
 * @retval -EIO No descriptor was received from the socket
 */
static int parse_pipe_msg_control(struct msghdr *msg_hdr)
{
	assert(msg_hdr);

	int pipefd = -EIO;
	struct cmsghdr *cmsg;
	for (cmsg = CMSG_FIRSTHDR(msg_hdr); cmsg != NULL; cmsg = CMSG_NXTHDR(msg_hdr, cmsg))
		if (memcmp(cmsg, &g_pipe_msg_ctl, sizeof(g_pipe_msg_ctl)) == 0) {
			memcpy(&pipefd, CMSG_DATA(cmsg), sizeof(pipefd));
			break;
		}
	return pipefd;
}

/**
 * @brief Get buffer ID by name
 * @details Returns the ID of the buffer with the given name
 * @param[in] name The name of the buffer whose ID we seek
 * @return If such buffer exists, its ID.
 *         Else, LOG_ID_INVALID.
 * @see log_name_by_id
 */
log_id_t log_id_by_name(const char *name)
{
	log_id_t i;

	for (i = 0; i < LOG_ID_MAX; ++i)
		if (!strcmp(name, logid_map[i].name))
			return logid_map[i].id;

	return LOG_ID_INVALID;
}

/**
 * @brief Get buffer name by ID
 * @details Returns the name of the buffer with the given ID
 * @param[in] id The ID of the buffer whose name we seek
 * @return If such buffer exists, its name.
 *         Else, empty string.
 * @see log_id_by_name
 */
char * log_name_by_id(log_id_t id)
{
	log_id_t i;

	for (i = 0; i < LOG_ID_MAX; ++i)
		if (id == logid_map[i].id)
			return logid_map[i].name;

	return "";
}

/**
 * @brief Log crtical failure
 * @details Writes the given message to syslog with a critical failure note
 * @param[in] message The message to write to syslog
 */
void syslog_critical_failure(const char *message, ...)
{
	openlog(NULL, LOG_PID | LOG_NDELAY, 0);
	va_list va;
	va_start(va, message);
	vsyslog(LOG_CRIT, message, va);
	va_end(va);
	closelog();
}

static int __send_dlog_msg(int sockfd, int result, int request, void *data, int datalen)
{
	struct dlog_control_msg msg;

	init_pipe_msg_data(&msg, request, result);
	msg.length += datalen;

	struct iovec iov[2] = {
		{ .iov_base = &msg, .iov_len = sizeof(msg) },
		{ .iov_base = data, .iov_len = datalen },
	};

	return (writev(sockfd, iov, data ? 2 : 1) < 0) ? -errno : 0;
}

/**
 * @brief Receive dlog protocol reply
 * @param[in] sockfd Socket file descriptor to read from
 * @param[in] request The expected request type
 * @param[out] data Pointer to extra data from the message
 * @param[out] datalen Length of the extra data
 * @notes If data is not NULL, it has to be deallocated manually after use.
 * @return 0 on success, -errno on failure
 */
int recv_dlog_reply(int sockfd, int request, void **data, int *datalen)
{
	assert(sockfd >= 0);
	assert(data);
	assert(datalen);

	*data = NULL;

	struct dlog_control_msg header;
	int r = read(sockfd, &header, sizeof header);
	if (r < 0)
		return -errno;
	if (r < sizeof header)
		return -EINVAL;
	if (header.request != (request | DLOG_REQ_REPLY)
	||  header.result != DLOG_REQ_RESULT_OK)
		return -EINVAL;

	*datalen = header.length - sizeof header;
	if (*datalen < 0)
		return -EINVAL;
	if (*datalen == 0) {
		*data = NULL;
		return 0;
	}

	*data = malloc(*datalen);
	if (!*data)
		return -ENOMEM;

	r = read(sockfd, *data, *datalen);
	if (r < *datalen) {
		free(*data);
		*data = NULL;
		return r < 0 ? -errno : -EINVAL;
	}

	return 0;
}

int send_dlog_request(int sockfd, int request, void *data, int datalen)
{
	return __send_dlog_msg(sockfd, 0, request, data, datalen);
}

int send_dlog_reply(int sockfd, int request, int result, void *data, int datalen)
{
	return __send_dlog_msg(sockfd, result, DLOG_REQ_REPLY | request, data, datalen);
}

/**
 * @brief Send a pipe over socket
 * @details Sends a pipe over socket
 * @param[in] sockfd The FD of the socket
 * @param[in] pipefd The pipe to send over the socket
 * @return 0 on success, else -errno
 */
int send_pipe(int sockfd, int pipefd)
{
	struct pipe_msg_header msg;
	struct dlog_control_msg data;
	init_pipe_msg_data(&data, DLOG_REQ_REPLY | DLOG_REQ_PIPE, DLOG_REQ_RESULT_OK);
	init_pipe_msg_header(&msg, &data);
	init_pipe_msg_control(&msg.hdr, pipefd);

	if (sendmsg(sockfd, &msg.hdr, 0) == -1)
		return -errno;

	return 0;
}

/**
 * @brief Receive a file descriptor from a socket
 * @details Receives a message from given socket fd and retrieves another fd from inside
 * @param[in] sockfd The socket's file descriptor
 * @param[in] timeout_ms How long to wait for input on sockfd, forever if 0
 * @return The file descriptor received from the socket, -errno on failure
 * @retval -EIO No descriptor was received from the socket
 */
int recv_pipe(int sockfd, int timeout_ms)
{
	struct pipe_msg_header msg;
	struct dlog_control_msg data;
	init_pipe_msg_header(&msg, &data);

	if (timeout_ms > 0) {
		struct pollfd pfd[] = { { .fd = sockfd, .events = POLLIN }, };
		int r = poll(pfd, NELEMS(pfd), timeout_ms);
		if (r != 1)
			return r == -1 ? -errno : -EWOULDBLOCK;
	}

	int res = recvmsg(sockfd, &msg.hdr, 0);
	if (res < 0)
		return -errno;
	if (res == 0)
		return -EIO;

	if (!verify_pipe_msg_data(&data))
		return -EIO;

	int pipefd = parse_pipe_msg_control(&msg.hdr);
	if (pipefd < 0)
		return pipefd;

	int err = fd_clear_flags(pipefd, O_NONBLOCK);
	if (err < 0) {
		close(pipefd);
		return err;
	}

	return pipefd;
}

/**
 * @brief Connect to socket
 * @details Connects to the socket at the given path
 * @param[in] path The path to the socket
 * @return The FD of the socket connection on success; -errno on failure.
 */
int connect_sock(const char *path)
{
	struct sockaddr_un sa = { .sun_family = AF_UNIX };
	if (strlen(path) >= sizeof sa.sun_path)
		return -ENAMETOOLONG;
	strncpy(sa.sun_path, path, sizeof sa.sun_path - 1);

	int fd = socket(sa.sun_family, SOCK_STREAM | SOCK_CLOEXEC, 0);
	if (fd < 0)
		return -errno;

	int r = connect(fd, (struct sockaddr *)&sa, sizeof sa);
	if (r < 0) {
		close(fd);
		return -errno;
	}

	return fd;
}

/**
 * @brief String suffix comparison
 * @details Checks whether a string ends with given suffix
 * @param[in] str The string to check
 * @param[in] suffix The suffix
 * @return 1 if yes, 0 if no
 */
int str_ends_with(const char *str, const char *suffix)
{
	assert(str);
	assert(suffix);

	const int str_len = strlen(str);
	const int suffix_len = strlen(suffix);

	if (str_len < suffix_len)
		return 0;

	return !strcmp(str + str_len - suffix_len, suffix);
}

/*
 * @brief Create glued string
 * @details Glues multiple strings together
 * @param[in] strings A table of strings to glue
 * @param[in] n The number of strings in the table
 * @return The glued string
 * @notes The caller needs to free the returned string
 */
char *create_glued_string(char const *const *strings, int n)
{
	int len = 0;
	for (int i = 0; i < n; ++i)
		len += strlen(strings[i]) + 1;

	char *ret = (char *) calloc(1, len);
	if (!ret)
		return NULL;

	for (int i = 0; i < n; ++i) {
		strncat(ret, strings[i], len - strlen(ret));
		if (i + 1 != n)
			strncat(ret, " ", len - strlen(ret));
	}

	return ret;
}

log_priority log_priority_from_char(signed char c)
{
	if (c < 0 || c >= DLOG_PRIO_MAX)
		return DLOG_UNKNOWN;

	return (log_priority)c;
}

/**
 * @}
 */
