#include <assert.h>
#include <logconfig.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <dirent.h>
#include <unistd.h>
#include <fcntl.h>
#include <logcommon.h>
#include <limits.h>
#include <stdbool.h>

#include "parsers.h"
#include <queued_entry.h>

/**
 * @addtogroup SHARED_FUNCTIONS
 * @{
 */

/**
 * @brief A singular config entry list element.
 */
struct log_conf_entry {
	char key[MAX_CONF_KEY_LEN];
	char value[MAX_CONF_VAL_LEN];
	struct log_conf_entry *next;
};

/**
 * @brief Get a config value
 * @details Returns a value with the given key from the given config
 * @param[in] config The config to work with
 * @param[in] key The key of the value entry to get
 * @return The value of the entry if such exists, else NULL
 * @see log_config_set
 */
const char *log_config_get(const struct log_config *config, const char *key)
{
	assert(config);
	assert(key);

	struct log_conf_entry *e;

	for (e = config->begin; e; e = e->next) {
		if (!strcmp(e->key, key))
			return e->value;
	}

	return NULL;
}

/**
 * @brief Get a config value converted to int
 * @details Returns a value with the given key from the given config
 * @param[in] c The config to work with
 * @param[in] key The key of the value entry to get
 * @param[in] default_val Default int value to return when proper value not given in config
 * @return The value of the entry if such exists, else default value
 * @see log_config_get
 */
int log_config_get_int(const struct log_config *c, const char *key, int default_val)
{
	int val = default_val;
	const char *str = log_config_get(c, key);
	return (str && parse_number(&str, &val) == 0) ? val : default_val;
}

/**
 * @brief Get a config value converted to boolean
 * @details Returns true if config string is a non-zero number
 * @param[in] c The config to work with
 * @param[in] key The key of the value entry to get
 * @param[in] default_val Default boolean value to return when proper value not given in config
 * @return If the entry exists and is a number, its value cast to boolean; else default value
 * @see log_config_get_int
 */
bool log_config_get_boolean(const struct log_config *c, const char *key, bool default_val)
{
	return !!log_config_get_int(c, key, default_val);
}

/**
 * @brief Modify a config entry
 * @details Sets the entry with the given key in the given config to the given value
 * @param[in] config The config to work with
 * @param[in] key The key of the value entry to set
 * @param[in] value The value to set
 * @remarks The entry must already exist
 * @return 0 if the entry was not added, else 1
 * @see log_config_set, log_config_get
 */
static int log_config_modify(struct log_config *config, const char *key, const char *value)
{
	assert(config);
	assert(key);
	assert(value);

	struct log_conf_entry *e;

	for (e = config->begin; e; e = e->next) {
		if (!strcmp(e->key, key))
			return snprintf(e->value, sizeof e->value, "%s", value) == (strlen(value));
	}

	return 0;
}

/**
 * @brief Append a config entry
 * @details Create an entry composed of the given key and value and add it to the given config
 * @remarks Does not check for duplicates
 * @param[in] config The config to work with
 * @param[in] key The key of the entry to create
 * @param[in] value The value of the new entry
 * @see log_config_set, log_config_get
 */
static void log_config_add(struct log_config *config, const char *key, const char *value)
{
	assert(config);
	assert(key);
	assert(value);

	struct log_conf_entry *e = calloc(1, sizeof(struct log_conf_entry));
	if (!e)
		return;

	snprintf(e->key, sizeof e->key, "%s", key);
	snprintf(e->value, sizeof e->value, "%s", value);

	if (!config->last) {
		config->begin = config->last = e;
		return;
	}

	config->last->next = e;
	config->last = e;
}

/**
 * @brief Set a config entry
 * @details Sets an entry to given value
 * @remarks Adds new key if needed
 * @param[in] config The config to work with
 * @param[in] key The key of the entry
 * @param[in] value The value of the entry
 * @see log_config_get
 */
void log_config_set(struct log_config *config, const char *key, const char *value)
{
	if (!log_config_modify(config, key, value))
		log_config_add(config, key, value);
}

/**
 * @brief Remove a config entry
 * @details Removes entry with given key
 * @param[in] config The config whence to remove the entry
 * @param[in] key The key of the entry to be removed
 * @return true if an entry was removed, false if not
 */
bool log_config_remove(struct log_config *config, const char *key)
{
	assert(config);
	assert(key);

	if (!config->begin)
		return false;

	struct log_conf_entry *prev_entry = config->begin;
	if (!strcmp(prev_entry->key, key)) {
		if (config->last == config->begin)
			config->last = NULL;
		config->begin = prev_entry->next;
		free(prev_entry);

		return true;
	}

	for (struct log_conf_entry *entry = prev_entry->next; entry; prev_entry = entry, entry = entry->next) {
		if (strcmp(entry->key, key))
			continue;

		if (entry == config->last)
			config->last = prev_entry;
		prev_entry->next = entry->next;
		free(entry);

		return true;
	}

	return false;
}

static void log_config_read_file_by_handle(struct log_config *config, FILE *fp)
{
	assert(config);
	assert(fp);

	char line[MAX_CONF_ENTRY_LEN];
	while (fgets(line, sizeof line, fp)) {
		int len = strlen(line);
		char key[MAX_CONF_KEY_LEN];

		if (len <= 1 || line[0] == '#')
			continue;

		if (line[len - 1] == '\n')
			line[len - 1] = '\0';

		char *tok = strchr(line, '=');
		if (!tok || (tok - line >= sizeof key))
			continue;
		++tok;

		snprintf(key, tok - line, "%s", line);
		log_config_set(config, key, tok);
	}
}

static int config_entry_filter(const struct dirent *de)
{
	assert(de);

	if (de->d_type != DT_REG)
		return 0;

	if (!str_ends_with(de->d_name, DEFAULT_CONFIG_SUFFIX))
		return 0;

	return 1;
}

static void log_config_read_file_by_dirfd_and_name(struct log_config *const config, const int dir_fd, const char *const filename)
{
	assert(config);
	assert(filename);
	assert(dir_fd >= 0);

	const int file_fd = openat(dir_fd, filename, O_RDONLY);
	if (file_fd < 0)
		return;

	FILE *fp = fdopen(file_fd, "r");
	if (!fp) {
		close(file_fd);
		return;
	}

	log_config_read_file_by_handle(config, fp);
	fclose(fp); // also closes file_fd
}

static void log_config_read_dir(struct log_config *config, const char *dir_path)
{
	assert(config);
	assert(dir_path);

	const int dir_fd = open(dir_path, O_DIRECTORY | O_RDONLY);
	if (dir_fd < 0)
		return;

	struct dirent **entries;
	int entry_count = scandirat(dir_fd, ".", &entries, config_entry_filter, alphasort);
	if (entry_count < 0)
		goto err_close;

	int i;
	for (i = 0; i < entry_count; ++i) {
		log_config_read_file_by_dirfd_and_name(config, dir_fd, entries[i]->d_name);
		free(entries[i]);
	}
	free(entries);

err_close:
	close(dir_fd);
}

/**
 * @brief Read one file to config
 * @details Appends one file's contents to a config.
 * @remarks Comments are ignored and there is no ordering guarantee
 * @param[in] config The config to append to
 * @param[in] filename The filename of the config file
 * @return 0 on success, -errno on failure
 * @see log_config_read
 */
int log_config_read_file(struct log_config *config, const char *filename)
{
	assert(config);
	assert(filename);

	FILE *fp = fopen(filename, "r");
	if (!fp)
		return -errno;

	log_config_read_file_by_handle(config, fp);
	fclose(fp);

	return 0;
}

/**
 * @brief Read the config
 * @details Read the entire configuration.
 * @param[in] config The config to fill
 * @remarks The path defaults can be overridden with the DLOG_CONFIG_PATH environment variable.
 *          Additionally, configuration snippets will be read from files ending with ".conf"
 *          from DLOG_CONFIG_PATH.d directory
 * @return 0 on success, -errno on failure
 * @see log_config_free
 */
int log_config_read(struct log_config *config)
{
	assert(config);

	config->begin = config->last = NULL;

	const char *main_path = getenv("DLOG_CONFIG_PATH") ?: DLOG_CONFIG_PATH;
	assert(main_path);

	int r = log_config_read_file(config, main_path);
	if (r < 0)
		return r;

	char dir_path[PATH_MAX];
	snprintf(dir_path, sizeof dir_path, "%s.d", main_path);
	log_config_read_dir(config, dir_path);

	return 0;
}

/**
 * @brief Free a config
 * @details Releases the resources owned by given config
 * @remarks Does not deallocate the superstructure itself
 * @param[in] config The config whose contents to erase
 * @see log_config_read
 */
void log_config_free(struct log_config* config)
{
	struct log_conf_entry * current = config->begin;

	while (current) {
		struct log_conf_entry * prev = current;
		current = current->next;
		free(prev);
	}

	config->begin = config->last = NULL;
}

/**
 * @brief Copy a config
 * @details Copies all entries from source to destination
 * @param[in] dst Destination config ("to")
 * @param[in] src Source config ("from")
 */
void log_config_copy(struct log_config *dst, const struct log_config *src)
{
	assert(dst);
	assert(src);

	log_config_free(dst);
	for (const struct log_conf_entry *entry = src->begin; entry; entry = entry->next)
		log_config_set(dst, entry->key, entry->value);
}

/**
 * @brief Call a function on each entry of a config
 * @details A simple helper function that implements a for-each loop that calls a function on all entries of a config.
 * @param[in] config The config to work with
 * @param[in] iter The iteration function called over each entry
 * @param[in,out] userdata User data
 */
void log_config_foreach(const struct log_config *config, configIter iter, void *userdata)
{
	struct log_conf_entry* e = config->begin;
	while (e) {
		iter(e->key, e->value, userdata);
		e = e->next;
	}
}

enum sorting_order get_order_from_config(const struct log_config *conf)
{
	enum sorting_order ret = SORT_RECV_MONO;

	const char *const sort_str = log_config_get(conf, "sort_by");
	if (sort_str) {
		if (!strcmp(sort_str, "sent_real"))
			ret = SORT_SENT_REAL;
		else if (!strcmp(sort_str, "recv_real"))
			ret = SORT_RECV_REAL;
		else if (!strcmp(sort_str, "sent_mono"))
			ret = SORT_SENT_MONO;
	}

	return ret;
}

enum sorting_order get_proper_sort_by(enum sorting_order sort_by, bool is_pipe, const struct log_config *conf, bool *sorting_needed)
{
	enum sorting_order source_sorting = get_order_from_config(conf);

	*sorting_needed = false;

	if (!is_pipe)
		return SORT_SENT_REAL; // the Android logger only provides this one timestamp and sorts on it, so let's use it

	if (sort_by == SORT_DEFAULT)
		return source_sorting; // pick the compatible sorting to make things easier

	*sorting_needed = (sort_by != source_sorting); // if everything is sorted by the source, we don't have to
	return sort_by;
}

/**
 * @}
 */
