/*
 * DLOG
 * Copyright (c) 2005-2008, The Android Open Source Project
 * Copyright (c) 2012-2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <assert.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include <logprint.h>
#include <ptrs_list.h>

/**
 * @addtogroup SHARED_FUNCTIONS
 * @{
 */

#define FILTERINFO_PID_NONE -1
#define FILTERINFO_TID_NONE -1

struct tag_and_prio {
	char *tag;
	int tagLength;
	log_priority pri;
	bool exactPri;
	bool prefix;
};

typedef struct FilterInfo_t {
	enum {
		FILTER_TAG_AND_PRIO,
		FILTER_PID,
		FILTER_TID,
	} type;
	union {
		struct tag_and_prio tnp;
		pid_t pid;
		pthread_t tid;
	};
} FilterInfo;

struct log_filter_t {
	list_head filters;
	log_priority global_pri;
	bool exact_global_pri;
};

/**
 * @brief Allocate filter info
 * @details Allocates filter info with given parameters
 * @param[in] tag The filtering tag
 * @param[in] pri The filtering priority
 * @param[in] exactPri The filtering priority is exact
 * @param[in] prefix The filtering tag is a prefix
 * @return The new structure (or NULL if allocation failed).
 * @see filterinfo_free
 */
static FilterInfo *filterinfo_new(const char *tag, log_priority pri, bool prefix, bool exactPri, pid_t pid, pthread_t tid)
{
	FilterInfo *p_ret = (FilterInfo *)calloc(1, sizeof(FilterInfo));
	if (!p_ret)
		return NULL;

	if (tag) {
		p_ret->type = FILTER_TAG_AND_PRIO;
		p_ret->tnp.tag = strdup(tag);
		if (!p_ret->tnp.tag) {
			free(p_ret);
			return NULL;
		}
		p_ret->tnp.tagLength = strlen(p_ret->tnp.tag);
		p_ret->tnp.pri = pri;
		p_ret->tnp.exactPri = exactPri;
		p_ret->tnp.prefix = prefix;
	} else if (pid != FILTERINFO_PID_NONE) {
		p_ret->type = FILTER_PID;
		p_ret->pid = pid;
	} else if (tid != FILTERINFO_TID_NONE) {
		p_ret->type = FILTER_TID;
		p_ret->tid = tid;
	} else {
		assert(false); // LCOV_EXCL_LINE
	}

	return p_ret;
}

static FilterInfo *filterinfo_clone(FilterInfo *p_info)
{
	assert(p_info);

	switch (p_info->type) {
	case FILTER_TAG_AND_PRIO:
		return filterinfo_new(p_info->tnp.tag, p_info->tnp.pri, p_info->tnp.prefix, p_info->tnp.exactPri, FILTERINFO_PID_NONE, FILTERINFO_TID_NONE);
	case FILTER_PID:
		return filterinfo_new(NULL, 0, false, false, p_info->pid, FILTERINFO_TID_NONE);
	case FILTER_TID:
		return filterinfo_new(NULL, 0, false, false, FILTERINFO_PID_NONE, p_info->tid);

	default: assert(false); // LCOV_EXCL_LINE
	}
}

/**
 * @brief Deallocate filter info
 * @details Deallocates the entire filter info structure
 * @param[in] p_info The structure to deallocate
 * @see filterinfo_new
 */
static void filterinfo_free(FilterInfo *p_info)
{
	assert(p_info);

	if (p_info->type == FILTER_TAG_AND_PRIO) {
		free(p_info->tnp.tag);
		p_info->tnp.tag = NULL;
	}

	free(p_info);
}

// apply_cb versions of two previous functions
elem_value filterinfo_clone_cb(elem_value p_info, void *_)
{
	return filterinfo_clone(p_info);
}
static void filterinfo_free_cb(elem_value p_info, void *_)
{
	filterinfo_free(p_info);
}

/**
 * @brief Char to priority
 * @details Converts a character priority representation to integer
 * @param[in] c The character to convert
 * @remarks Accepts both letters and numbers
 * @return Integral log priority
 * @see filter_pri_to_char
 */
log_priority filter_char_to_pri(char c)
{
	log_priority pri;

	c = tolower(c);

	if (c >= '0' && c <= '9') {
		if (c >= ('0'+DLOG_SILENT))
			pri = DLOG_VERBOSE;
		else
			pri = (log_priority)(c - '0');
	} else if (c == 'v') {
		pri = DLOG_VERBOSE;
	} else if (c == 'd') {
		pri = DLOG_DEBUG;
	} else if (c == 'i') {
		pri = DLOG_INFO;
	} else if (c == 'w') {
		pri = DLOG_WARN;
	} else if (c == 'e') {
		pri = DLOG_ERROR;
	} else if (c == 'f') {
		pri = DLOG_FATAL;
	} else if (c == 's') {
		pri = DLOG_SILENT;
	} else if (c == '*') {
		pri = DLOG_DEFAULT;
	} else {
		pri = DLOG_UNKNOWN;
	}

	return pri;
}
/**
 * @brief Priority to character
 * @details Converts an integer representation of priority to character
 * @param[in] pri The integer to convert
 * @return Character log priority
 * @see filter_char_to_pri
 */
char filter_pri_to_char(log_priority pri)
{
	switch (pri) {
	case DLOG_VERBOSE:
		return 'V';
	case DLOG_DEBUG:
		return 'D';
	case DLOG_INFO:
		return 'I';
	case DLOG_WARN:
		return 'W';
	case DLOG_ERROR:
		return 'E';
	case DLOG_FATAL:
		return 'F';
	case DLOG_SILENT:
		return 'S';
	case DLOG_DEFAULT:
	case DLOG_UNKNOWN:
	default:
		return '?';
	}
}

/**
 * @brief Filter a line
 * @details Passes a line through filters to discover whether it should be logged or not
 * @param[in] p_format The format to work with
 * @param[in] entry The log entry to filter
 * @return True if the line should be printed, else false
 */
bool log_should_print_line(log_filter *p_filter, const struct logger_entry *entry)
{
	assert(p_filter);
	assert(entry);

	bool prefix;
	bool matched = false;
	const char *tag;
	log_priority prio;

	tag = log_buffer_get_tag(entry);
	/* mark empty-tagged messages and make it easy to catch an application that does that */
	if (!tag || !strlen(tag))
		tag = "DLOG_ERROR_NOTAG";

	prio = (log_priority)entry->priority;

	// Sadly, we can't use list_foreach, because it doesn't let us exit early
	for (list_head p_curFilter_iter = p_filter->filters; p_curFilter_iter != NULL; list_next(&p_curFilter_iter)) {
		FilterInfo *p_curFilter = list_at(p_curFilter_iter);

		switch (p_curFilter->type) {
		case FILTER_TAG_AND_PRIO:
			prefix = p_curFilter->tnp.prefix;
			if ((prefix && (0 == strncmp(tag, p_curFilter->tnp.tag, p_curFilter->tnp.tagLength))) ||
			    (!prefix && (0 == strcmp(tag, p_curFilter->tnp.tag)))) {
				matched = true;
				if (p_curFilter->tnp.exactPri) {
					if (prio == p_curFilter->tnp.pri)
						return true;
				} else {
					if (prio >= p_curFilter->tnp.pri)
						return true;
				}
			}
			break;
		case FILTER_PID:
			matched = true;
			if (entry->pid == p_curFilter->pid)
				return true;
			break;
		case FILTER_TID:
			matched = true;
			if (entry->tid == p_curFilter->tid)
				return true;
			break;

		default: assert(false); // LCOV_EXCL_LINE
		}

	}
	if (matched)
		return false;

	if (p_filter->exact_global_pri)
		return (prio == p_filter->global_pri);
	else
		return (prio >= p_filter->global_pri);
}

/**
 * @brief Allocate log format
 * @details Allocates a log format structure
 * @return The new structure (or NULL if allocation failed).
 * @see log_filter_free
 */
log_filter *log_filter_new(void)
{
	log_filter *p_ret;

	p_ret = calloc(1, sizeof(log_filter));

	if (!p_ret)
		return NULL;
	p_ret->global_pri = DLOG_SILENT;
	p_ret->exact_global_pri = false;

	return p_ret;
}

/**
 * @brief Copy a log format
 * @details Allocates a copy of the log format
 * @param[in] p_format The log format to copy
 * @return The new structure (or NULL if allocation failed)
 * @see log_filter_free
 */
log_filter *log_filter_from_filter(const log_filter *p_filter)
{
	log_filter *p_ret;

	if (!(p_ret = log_filter_new()))
		return NULL;

	p_ret->global_pri = p_filter->global_pri;
	p_ret->exact_global_pri = p_filter->exact_global_pri;
	p_ret->filters = list_map(p_filter->filters, NULL, filterinfo_clone_cb, filterinfo_free_cb);

	if (!p_ret->filters)
	{
		log_filter_free(p_ret);
		return NULL;
	}

	return p_ret;
}

/**
 * @brief Deallocate log format
 * @details Deallocates the entire log format structure
 * @param[in] p_format The structure to deallocate
 */
void log_filter_free(log_filter *p_filter)
{
	if (p_filter) {
		list_clear_custom(&p_filter->filters, NULL, filterinfo_free_cb);

		free(p_filter);
	}
}

/**
 * @brief Print format from string
 * @details Converts a string to print format
 * @param[in] formatString The string representation to convert
 * @return The integer print format (on invalid string: FORMAT_OFF)
 */
log_print_format log_format_from_string(const char *formatString)
{
	static const struct {
		char const *string;
		log_print_format format;
	} formats[] = {
		{"brief"        , FORMAT_BRIEF        },
		{"process"      , FORMAT_PROCESS      },
		{"tag"          , FORMAT_TAG          },
		{"thread"       , FORMAT_THREAD       },
		{"raw"          , FORMAT_RAW          },
		{"time"         , FORMAT_TIME         },
		{"threadtime"   , FORMAT_THREADTIME   },
		{"kerneltime"   , FORMAT_KERNELTIME   },
		{"recv_realtime", FORMAT_RECV_REALTIME},
		{"rwtime"       , FORMAT_RWTIME       },
		{"long"         , FORMAT_LONG         },
		{"json"         , FORMAT_JSON         },
	};

	for (size_t i = 0; i < NELEMS(formats); ++i)
		if (!strcmp(formatString, formats[i].string))
			return formats[i].format;

	return FORMAT_OFF;
}

/**
 * @brief Add a filter rule
 * @details Adds a tag/priority filtering rule
 * @param[in] p_format The log format to add the filter rule to
 * @param[in] filterExpression A filter expression ("tag:prio")
 * @return 0 on success, else -1
 * @remarks Assumes single threaded execution
 */
int log_add_filter_rule(log_filter *p_filter,
		const char *filterExpression)
{
	size_t tagNameLength, priPosition;
	bool exact = false;
	bool pattern = false;
	log_priority pri = DLOG_DEFAULT;

	tagNameLength = strcspn(filterExpression, ":");

	if (tagNameLength == 0)
		goto error;

	if (filterExpression[tagNameLength] == ':') {
		if ((strlen(filterExpression + (tagNameLength + 1)) > 1)
			&& (filterExpression[tagNameLength + 1] == '=')) {
			priPosition = tagNameLength + 2;
			exact = true;
		} else {
			priPosition = tagNameLength + 1;
			exact = false;
		}
		pri = filter_char_to_pri(filterExpression[priPosition]);

		if (pri == DLOG_UNKNOWN)
			goto error;
	}

	if (0 == strncmp("*", filterExpression, tagNameLength)) {
		/* This filter expression refers to the global filter
		 * The default level for this is DEBUG if the priority
		 * is unspecified
		 */
		if (pri == DLOG_DEFAULT)
			pri = DLOG_DEBUG;

		p_filter->global_pri = pri;
		p_filter->exact_global_pri = exact;
	} else {
		/* for filter expressions that don't refer to the global
		 * filter, the default is verbose if the priority is unspecified
		 */
		if (pri == DLOG_DEFAULT)
			pri = DLOG_VERBOSE;

		if (filterExpression[tagNameLength - 1] == '*') {
			pattern = true;
			tagNameLength = tagNameLength - 1;
		} else
			pattern = false;

		char *tagName = strndup(filterExpression, tagNameLength);
		if (!tagName)
			goto error;

		FilterInfo *p_fi = filterinfo_new(tagName, pri, pattern, exact, FILTERINFO_PID_NONE, FILTERINFO_TID_NONE);
		free(tagName);
		if (!p_fi)
			goto error;

		if (!list_add(&p_filter->filters, p_fi))
			goto error;
	}

	return 0;
error:
	return -1;
}

/**
 * @brief Add filter string
 * @details Adds multiple tag/priority filtering rule from a string
 * @param[in] p_format The log format to add the filter rules to
 * @param[in] filterString A filter string containing multiple filter rules delimited by spaces, tabs or commas
 * @return 0 on success, else -1
 * @remarks Assumes single threaded execution
 */
int log_add_filter_string(log_filter *p_filter,
		const char *filterString)
{
	char *filterStringCopy = strdup(filterString);
	if (!filterStringCopy)
		return -1;

	char *p_cur = filterStringCopy;
	char *p_ret;
	int err;

	while (NULL != (p_ret = strsep(&p_cur, " \t,"))) {
		/* ignore whitespace-only entries */
		if (p_ret[0] != '\0') {
			err = log_add_filter_rule(p_filter, p_ret);

			if (err < 0)
				goto error;
		}
	}

	free(filterStringCopy);
	return 0;
error:
	free(filterStringCopy);
	return -1;
}

/**
 * @brief Add pid to filter by
 * @details Adds a single pid filtering rule
 * @param[in] p_format The log format to add the filter rules to
 * @param[in] pid A pid to filter by
 * @return 0 on success, else -ENOMEM
 * @remarks Assumes single threaded execution
 */
int log_add_filter_pid(log_filter *p_filter, pid_t pid)
{
	assert(p_filter);

	FilterInfo *p_fi = filterinfo_new(NULL, 0, false, false, pid, FILTERINFO_TID_NONE);
	if (!p_fi)
		return -ENOMEM;

	return list_add(&p_filter->filters, p_fi) ? 0 : -ENOMEM;
}

/**
 * @brief Add tid to filter by
 * @details Adds a single tid filtering rule
 * @param[in] p_format The log format to add the filter rules to
 * @param[in] pid A tid to filter by
 * @return 0 on success, else -ENOMEM
 * @remarks Assumes single threaded execution
 */
int log_add_filter_tid(log_filter *p_filter, pthread_t tid)
{
	assert(p_filter);

	FilterInfo *p_fi = filterinfo_new(NULL, 0, false, false, FILTERINFO_PID_NONE, tid);
	if (!p_fi)
		return -ENOMEM;

	return list_add(&p_filter->filters, p_fi) ? 0 : -ENOMEM;
}

/**
 * @brief Return pointer to log entry tag
 * @details Finds the beginning of a log tag in entry's payload and returns it
 * @param[in] entry The log entry
 * @return Pointer to log tag
 */
char* log_buffer_get_tag(const struct logger_entry *entry)
{
	if (entry->len - (int) sizeof(struct logger_entry) < 2)
		return NULL;

	return (char*)entry->msg;
}

/**
 * @brief Return pointer to log entry message
 * @details Finds the beginning of a log message in entry's payload and returns it
 * @param[in] entry The log entry
 * @return Pointer to log message
 */
char* log_buffer_get_message(const struct logger_entry *entry)
{
	if (entry->len - (int) sizeof(struct logger_entry) < 2)
		return NULL;

	return (char*)entry->msg + entry->tag_len + 1 /* NULL delimiter */;
}

enum sorting_order get_format_sorting(log_print_format format)
{
	switch (format) {
	case FORMAT_TAG:
	case FORMAT_PROCESS:
	case FORMAT_THREAD:
	case FORMAT_RAW:
	case FORMAT_BRIEF:
	case FORMAT_JSON:
		return SORT_DEFAULT;

	case FORMAT_KERNELTIME:
		return SORT_SENT_MONO;

	case FORMAT_TIME:
	case FORMAT_THREADTIME:
	case FORMAT_LONG:
		return SORT_SENT_REAL;

	case FORMAT_RWTIME:
	case FORMAT_RECV_REALTIME:
		return SORT_RECV_REAL;

	default: assert(false); // LCOV_EXCL_LINE
	}
}

#define COLOR_PURPLE "\033[35;1m"
#define COLOR_RED    "\033[31;1m"
#define COLOR_YELLOW "\033[33;1m"
#define COLOR_RESET  "\033[0m"

static const char *get_pre_color(log_priority pri, bool enabled)
{
	if (!enabled)
		return "";
	switch (pri) {
	case DLOG_FATAL:
		return COLOR_PURPLE;
	case DLOG_ERROR:
		return COLOR_RED;
	case DLOG_WARN:
		return COLOR_YELLOW;
	default:
		return "";
	}
}

static const char *get_post_color(log_priority pri, bool enabled)
{
	if (!enabled)
		return "";
	switch (pri) {
	case DLOG_FATAL:
	case DLOG_ERROR:
	case DLOG_WARN:
		return COLOR_RESET;
	default:
		return "";
	}
}

static char to_hex_digit(int digit)
{
	assert(digit >= 0);
	assert(digit < 16);
	if (digit < 10)
		return digit + '0';
	else
		return digit + 'A' - 10;
}

char *json_escape_string(char *in)
{
	assert(in);
	size_t len = strlen(in);
	/* One input string character will be encoded by max 6 characters in the output string.
	 * We use unsigned chars for easy comparisions. */
	unsigned char *buf = malloc(6 * len + 1);
	if (!buf)
		return NULL;

	/* So, RFC 7159 says that JSON output can be in one of three encodings: UTF-{8,16,32}.
	 * In particular, in UTF-8, the multibyte characters are allowed. This means that, assuming
	 * that our input is valid UTF-8, we can just write them directly, simplifying the implementation.
	 * Correctly encoding arbitrary binary data would require something like base64, this, however,
	 * would make it very difficult to read by humans. Almost all logs are UTF-8 anyway. */
	int targ = 0;
	for (int i = 0; i < len; ++i) {
		unsigned char c = in[i];
		switch (c) {
		/* These cases can get nicer encodings than the usual "\uxxxx", even though we technically
		 * dont have to do so. Also, we don't encode '/', because there is no reason to. */
		case '\"':
			buf[targ++] = '\\';
			buf[targ++] = '\"';
			break;
		case '\\':
			buf[targ++] = '\\';
			buf[targ++] = '\\';
			break;
		case '\b':
			buf[targ++] = '\\';
			buf[targ++] = 'b';
			break;
		case '\f':
			buf[targ++] = '\\';
			buf[targ++] = 'f';
			break;
		case '\n':
			buf[targ++] = '\\';
			buf[targ++] = 'n';
			break;
		case '\r':
			buf[targ++] = '\\';
			buf[targ++] = 'r';
			break;
		case '\t':
			buf[targ++] = '\\';
			buf[targ++] = 't';
			break;
		default:
			/* We don't use a range case, because we would have to remove the already done characters from it.
			 * We would end up with three distinct cases for the {0, ..., 31} interval. */
			if (c <= 31) {
				buf[targ++] = '\\';
				buf[targ++] = 'u';
				buf[targ++] = '0';
				buf[targ++] = '0';
				buf[targ++] = to_hex_digit(c / 16);
				buf[targ++] = to_hex_digit(c % 16);
			}
			else
				buf[targ++] = c;
			break;
		}
	}
	buf[targ++] = '\0';

	char *ret = realloc(buf, targ);
	if (!ret)
		free(buf);
	return ret;
}

const char *json_priority_name(log_priority prio)
{
	static const char *const json_priorities_map[] = {
		[DLOG_VERBOSE] = "verbose",
		[DLOG_DEBUG]   = "debug",
		[DLOG_INFO]    = "info",
		[DLOG_WARN]    = "warning",
		[DLOG_ERROR]   = "error",
		[DLOG_FATAL]   = "fatal",
		[DLOG_SILENT]  = "silent",
	};

	const char *ret = NULL;
	if (prio < NELEMS(json_priorities_map))
		ret = json_priorities_map[prio];
	if (!ret) // Either out of the map, or in between its elements
		ret = "unknown";
	return ret;
}

char *log_format_json(
		char *default_buffer,
		size_t default_buffer_size,
		const struct logger_entry *entry,
		size_t *p_outLength,
		char *tag,
		char *msg)
{
	const char *prio_name = json_priority_name(entry->priority);

	__attribute__((cleanup(free_ptr))) char *tag_esc = json_escape_string(tag);
	if (tag_esc == NULL)
		return NULL;

	__attribute__((cleanup(free_ptr))) char *msg_esc = json_escape_string(msg);
	if (msg_esc == NULL)
		return NULL;

	char recv_real[64], recv_mono[64], sent_real[64], sent_mono[64];
#if defined(HAVE_LOCALTIME_R)
	struct tm tmBuf;
#endif
	struct tm* ptm;
	time_t time_t_temp;

	if (entry->nsec_recv_real == INVALID_NSEC)
		recv_real[0] = '\0';
	else {
		time_t_temp = entry->sec_recv_real;
#if defined(HAVE_LOCALTIME_R)
		ptm = localtime_r(&time_t_temp, &tmBuf);
#else
		ptm = localtime(&time_t_temp);
#endif
		if (!ptm)
			return NULL;
		strftime(recv_real, sizeof recv_real, "\"recv_real\":\"%FT%T%z\",", ptm);
	}

	if (entry->nsec_sent_real == INVALID_NSEC)
		sent_real[0] = '\0';
	else {
		time_t_temp = entry->sec_sent_real;
#if defined(HAVE_LOCALTIME_R)
		ptm = localtime_r(&time_t_temp, &tmBuf);
#else
		ptm = localtime(&time_t_temp);
#endif
		if (!ptm)
			return NULL;
		strftime(sent_real, sizeof sent_real, "\"sent_real\":\"%FT%T%z\",", ptm);
	}

	if (entry->nsec_recv_mono == INVALID_NSEC)
		recv_mono[0] = '\0';
	else
		snprintf(recv_mono, sizeof recv_mono, "\"recv_mono\":%d.%.9d,", entry->sec_recv_mono, entry->nsec_recv_mono);

	if (entry->nsec_sent_mono == INVALID_NSEC)
		sent_mono[0] = '\0';
	else
		snprintf(sent_mono, sizeof sent_mono, "\"sent_mono\":%d.%.9d,", entry->sec_sent_mono, entry->nsec_sent_mono);

#define ARGS "{" \
		"\"priority\":\"%s\"," \
		"\"pid\":%d," \
		"\"tid\":%d," \
		"%s%s%s%s" \
		"\"tag\":\"%s\"," \
		"\"msg\":\"%s\"" \
		"}\n", prio_name, entry->pid, entry->tid, \
		recv_real, recv_mono, sent_real, sent_mono, \
		tag_esc, msg_esc

	int len = snprintf(NULL, 0, ARGS);

	char *ret;
	if (len < 0)
		return NULL;
	else if (default_buffer_size >= len + 1) {
		ret = default_buffer;
	} else {
		ret = (char *)malloc(len + 1);

		if (ret == NULL)
			return NULL;
	}

	int res = snprintf(ret, len + 1, ARGS);

#undef ARGS

	if (res <= 0) {
		if (ret != default_buffer)
			free(ret);
		return NULL;
	}

	*p_outLength = len;
	return ret;
}

enum ts_type {
	TS_SENT,
	TS_RECV,
};

static bool entry_get_ts_inner(const struct logger_entry *entry, bool sent, bool mono, struct timespec *target)
{
	int sec = sent
		? (mono ? entry->sec_sent_mono : entry->sec_sent_real)
		: (mono ? entry->sec_recv_mono : entry->sec_recv_real);
	int nsec = sent
		? (mono ? entry->nsec_sent_mono : entry->nsec_sent_real)
		: (mono ? entry->nsec_recv_mono : entry->nsec_recv_real);

	if (nsec == INVALID_NSEC)
		return false;

	target->tv_sec = sec;
	target->tv_nsec = nsec;
	return true;
}

struct timespec entry_get_ts(const struct logger_entry *entry, const log_print_format format, enum ts_type type)
{
	/* The returned timestamp should match the format (as in, the printf specifier) used below.
	 * In particular, monotonic timestamps should be the ones represented by the 'raw' (%u)
	 * format (currently used solely by the KERNELTIME format) while realtime timestamps should
	 * be represented by strftime date-formatted strings.
	 * However, we decided that some timing data is better than none. Therefore, if the needed timestamp
	 * is missing, we use one of the others, even though it probably is at least slightly wrong.
	 * A better solution will be possible after the planned refactor of the format system.
	 * In particular, we allow using the realtime stamp in the monotonic field, as well as using sending
	 * and receiving timestamps interchangeably. */

	struct timespec ret = {};
	bool sent = type == TS_SENT;
	bool mono = format == FORMAT_KERNELTIME;

#define CHECK_AND_CHANGE_SENT \
		if (entry_get_ts_inner(entry, sent, mono, &ret)) \
			return ret; \
		sent = !sent;

	CHECK_AND_CHANGE_SENT
	CHECK_AND_CHANGE_SENT

	if (mono) {
		mono = false;
		CHECK_AND_CHANGE_SENT
		CHECK_AND_CHANGE_SENT
	}

#undef CHECK_AND_CHANGE_SENT

	return ret;
}

/**
 * @brief Format log line
 * @details Formats a log entry for direct printing
 * @param[in] p_format The log format to use
 * @param[out] defaultBuffer The buffer to write into, if possible
 * @param[in] defaultBufferSize The size of the default buffer
 * @param[in] entry The entry to format
 * @param[out] p_outLength The length of the output
 * @return The resulting buffer (defaultBuffer if possible, else a malloc'd one)
 * @remarks If the returned value is not defaultBuffer, it has to be free'd
 */
char *log_format_log_line(
		const struct log_format p_format,
		char *defaultBuffer,
		size_t defaultBufferSize,
		const struct logger_entry *entry,
		size_t *p_outLength)
{
#if defined(HAVE_LOCALTIME_R)
	struct tm tmBuf;
#endif
	struct tm* ptm;
	time_t time_t_temp;
	char timeBuf[32], tzBuf[16];
	char recvTimeBuf[32];
	char prefixBuf[128], suffixBuf[128];
	char priChar;
	int prefixSuffixIsHeaderFooter = 0;
	char *tag, *msg;
	char * ret = NULL;

	priChar = filter_pri_to_char((log_priority)entry->priority);

	tag = log_buffer_get_tag(entry);
	/* mark empty-tagged messages and make it easy to catch an application that does that */
	if (!tag || !strlen(tag))
		tag = "DLOG_ERROR_NOTAG";

	msg = log_buffer_get_message(entry) ?: "DLOG_ERROR_NOMSG";

	/*
	 * JSON is formatted separately. This is because it's so vastly different
	 * to the other formats, it even contains different data.
	 * Also, currently in other formats the data before and after the message
	 * is limited in length. This would make JSON impossible to implement in a good way.
	 */
	if (p_format.format == FORMAT_JSON)
		return log_format_json(defaultBuffer, defaultBufferSize, entry, p_outLength, tag, msg);

	/*
	 * Get the current date/time in pretty form
	 *
	 * It's often useful when examining a log with "less" to jump to
	 * a specific point in the file by searching for the date/time stamp.
	 * For this reason it's very annoying to have regexp meta characters
	 * in the time stamp.  Don't use forward slashes, parenthesis,
	 * brackets, asterisks, or other special chars here.
	 */

	struct timespec ts;

	long int sec_sent, nsec_sent;
	ts = entry_get_ts(entry, p_format.format, TS_SENT);
	sec_sent = ts.tv_sec;
	nsec_sent = ts.tv_nsec;

	long int sec_recv, nsec_recv;
	ts = entry_get_ts(entry, p_format.format, TS_RECV);
	sec_recv = ts.tv_sec;
	nsec_recv = ts.tv_nsec;

	time_t_temp = sec_sent;
#if defined(HAVE_LOCALTIME_R)
	ptm = localtime_r(&time_t_temp, &tmBuf);
#else
	ptm = localtime(&time_t_temp);
#endif
	if (!ptm)
		return NULL;
	strftime(timeBuf, sizeof(timeBuf), "%m-%d %H:%M:%S", ptm);

	time_t_temp = sec_recv;
#if defined(HAVE_LOCALTIME_R)
	ptm = localtime_r(&time_t_temp, &tmBuf);
#else
	ptm = localtime(&time_t_temp);
#endif
	if (!ptm)
		return NULL;

	strftime(recvTimeBuf, sizeof(recvTimeBuf), "%m-%d %H:%M:%S", ptm);

	strftime(tzBuf, sizeof(tzBuf), "%z", ptm);

	/*
	 * Construct a buffer containing the log header and log message.
	 */
	size_t prefixLen, suffixLen;
	bool colorPrefix = true, colorSuffix = true;

	switch (p_format.format) {
	case FORMAT_TAG:
		prefixLen = snprintf(prefixBuf, sizeof(prefixBuf),
				"%c/%-8s: ", priChar, tag);
		strncpy(suffixBuf, "\n", 2);
		suffixLen = 1;
		colorSuffix = false;
		break;
	case FORMAT_PROCESS:
		prefixLen = snprintf(prefixBuf, sizeof(prefixBuf),
				"%c(%5d) ", priChar, (int)entry->pid);
		suffixLen = snprintf(suffixBuf, sizeof(suffixBuf),
				"  (%s)\n", tag);
		break;
	case FORMAT_THREAD:
		prefixLen = snprintf(prefixBuf, sizeof(prefixBuf),
				"%c(P%5d, T%5d) ", priChar, (int)entry->pid, (int)entry->tid);
		strncpy(suffixBuf, "\n", 2);
		colorSuffix = false;
		suffixLen = 1;
		break;
	case FORMAT_RAW:
		prefixBuf[0] = 0;
		prefixLen = 0;
		colorPrefix = false;
		strncpy(suffixBuf, "\n", 2);
		suffixLen = 1;
		colorSuffix = false;
		break;
	case FORMAT_TIME:
		prefixLen = snprintf(prefixBuf, sizeof(prefixBuf),
				"%s.%03ld%s %c/%-8s(%5d): ", timeBuf, nsec_sent / 1000000,
				tzBuf, priChar, tag, (int)entry->pid);
		strncpy(suffixBuf, "\n", 2);
		suffixLen = 1;
		colorSuffix = false;
		break;
	case FORMAT_THREADTIME:
		prefixLen = snprintf(prefixBuf, sizeof(prefixBuf),
				"%s.%03ld%s %c/%-8s(P%5d, T%5d): ",
				timeBuf, nsec_sent / 1000000,
				tzBuf, priChar, tag, (int)entry->pid, (int)entry->tid);
		strncpy(suffixBuf, "\n", 2);
		suffixLen = 1;
		colorSuffix = false;
		break;
	case FORMAT_KERNELTIME:
		prefixLen = snprintf(prefixBuf, sizeof(prefixBuf),
				"%5lu.%03ld %c/%-8s(P%5d, T%5d): ",
				sec_sent, nsec_sent / 1000000,
				priChar, tag, (int)entry->pid, (int)entry->tid);
		strncpy(suffixBuf, "\n", 2);
		suffixLen = 1;
		colorSuffix = false;
		break;
	case FORMAT_RECV_REALTIME:
		prefixLen = snprintf(prefixBuf, sizeof(prefixBuf),
				"%s.%03ld %c/%-8s(P%5d, T%5d): ",
				recvTimeBuf, nsec_recv / 1000000,
				priChar, tag, (int)entry->pid, (int)entry->tid);
		strncpy(suffixBuf, "\n", 2);
		suffixLen = 1;
		colorSuffix = false;
		break;
	case FORMAT_RWTIME:
		/* This shouldn't be needed. However, sadly, the rwtime format contains two different timestamps.
		 * Such a situation is impossible to describe as is. Therefore, it calls for an ugly hack :P
		 * Again, the planned refactor of the format system will fix this. */
		ts = entry_get_ts(entry, FORMAT_KERNELTIME, TS_RECV);
		prefixLen = snprintf(prefixBuf, sizeof(prefixBuf),
				"%s [%5lu.%03ld] %c/%-8s(P%5d, T%5d): ",
				recvTimeBuf, ts.tv_sec, ts.tv_nsec / 1000000,
				priChar, tag, (int)entry->pid, (int)entry->tid);
		strncpy(suffixBuf, "\n", 2);
		suffixLen = 1;
		colorSuffix = false;
		break;
	case FORMAT_LONG:
		prefixLen = snprintf(prefixBuf, sizeof(prefixBuf),
				"[ %s.%03ld %c/%-8s P%5d, T%5d]\n",
				timeBuf, nsec_sent / 1000000, priChar, tag,
				(int)entry->pid, (int)entry->tid);
		strncpy(suffixBuf, "\n\n", 3);
		suffixLen = 2;
		colorSuffix = false;
		prefixSuffixIsHeaderFooter = 1;
		break;
	case FORMAT_BRIEF:
	default:
		prefixLen = snprintf(prefixBuf, sizeof(prefixBuf),
				"%c/%-8s(%5d): ", priChar, tag, (int)entry->pid);
		strncpy(suffixBuf, "\n", 2);
		suffixLen = 1;
		colorSuffix = false;
		break;
	}
	/* snprintf has a weird return value.   It returns what would have been
	 * written given a large enough buffer.  In the case that the prefix is
	 * longer then our buffer(128), it messes up the calculations below
	 * possibly causing heap corruption.  To avoid this we double check and
	 * set the length at the maximum (size minus null byte)
	 */
	if (prefixLen >= sizeof(prefixBuf))
		prefixLen = sizeof(prefixBuf) - 1;
	if (suffixLen >= sizeof(suffixBuf))
		suffixLen = sizeof(suffixBuf) - 1;

	/* the following code is tragically unreadable */

	size_t numLines;
	char *p;
	size_t bufferSize;
	const char *pm;
	int message_len = entry->len - sizeof(struct logger_entry) - entry->tag_len - 2 /* tag and msg delimiters */;

	const char *const pre_color_prefix = get_pre_color((log_priority)entry->priority, colorPrefix && p_format.color);
	const char *const pre_color_suffix = get_pre_color((log_priority)entry->priority, colorSuffix && p_format.color);
	const char *const post_color_prefix = get_post_color((log_priority)entry->priority, colorPrefix && p_format.color);
	const char *const post_color_suffix = get_post_color((log_priority)entry->priority, colorSuffix && p_format.color);
	const int pre_color_prefix_len = strlen(pre_color_prefix);
	const int post_color_prefix_len = strlen(post_color_prefix);
	const int pre_color_suffix_len = strlen(pre_color_suffix);
	const int post_color_suffix_len = strlen(post_color_suffix);

	if (prefixSuffixIsHeaderFooter) {
		/* we're just wrapping message with a header/footer */
		numLines = 1;
	} else {
		pm = msg;
		numLines = 0;

		/* The line-end finding here must match the line-end finding
		 * in for ( ... numLines...) loop below
		 */
		while (pm < (msg + message_len)) {
			if (*pm++ == '\n') numLines++;
		}
		/* plus one line for anything not newline-terminated at the end */
		if (pm > msg && *(pm-1) != '\n') numLines++;
	}

	/* this is an upper bound--newlines in message may be counted
	 * extraneously
	 */
	bufferSize = (numLines * (pre_color_prefix_len + prefixLen + post_color_prefix_len
			+ pre_color_suffix_len + suffixLen + post_color_suffix_len)) + message_len + 1;

	if (defaultBufferSize >= bufferSize) {
		ret = defaultBuffer;
	} else {
		ret = (char *)malloc(bufferSize);

		if (ret == NULL)
			return ret;
	}

	ret[0] = '\0';       /* to start strcat off */

	p = ret;
	pm = msg;

	if (prefixSuffixIsHeaderFooter) {
		strncat(p, pre_color_prefix, pre_color_prefix_len + 1);
		p += pre_color_prefix_len;
		strncat(p, prefixBuf, prefixLen);
		p += prefixLen;
		strncat(p, post_color_prefix, post_color_prefix_len + 1);
		p += post_color_prefix_len;
		strncat(p, msg, message_len);
		p += message_len;
		strncat(p, pre_color_suffix, pre_color_suffix_len + 1);
		p += pre_color_suffix_len;
		strncat(p, suffixBuf, suffixLen);
		p += suffixLen;
		strncat(p, post_color_suffix, post_color_suffix_len + 1);
		p += post_color_suffix_len;
	} else {
		while (pm < (msg + message_len) && *pm != 0) {
			const char *lineStart;
			size_t lineLen;

			lineStart = pm;

			/* Find the next end-of-line in message */
			while (pm < (msg + message_len)
					&& *pm != '\n' && *pm != 0) pm++;
			lineLen = pm - lineStart;

			strncat(p, pre_color_prefix, pre_color_prefix_len + 1);
			p += pre_color_prefix_len;
			strncat(p, prefixBuf, prefixLen);
			p += prefixLen;
			strncat(p, post_color_prefix, post_color_prefix_len + 1);
			p += post_color_prefix_len;
			strncat(p, lineStart, lineLen);
			p += lineLen;
			strncat(p, pre_color_suffix, pre_color_suffix_len + 1);
			p += pre_color_suffix_len;
			strncat(p, suffixBuf, suffixLen);
			p += suffixLen;
			strncat(p, post_color_suffix, post_color_suffix_len + 1);
			p += post_color_suffix_len;

			if (pm < (msg + message_len) && *pm == '\n' && *pm != 0)
				pm++;
		}
	}

	if (p_outLength != NULL)
		*p_outLength = p - ret;

	return ret;
}

/**
 * @brief Print log line
 * @details Prints the given log entry to the supplied file descriptor
 * @param[in] p_format The format to use
 * @param[in] fd The file descriptor to write to
 * @param[in] colors_auto Whether to switch color coding based on TTY-ness of fd
 * @param[in] entry The entry to print
 * @return Bytes written
 */
int log_print_log_line(
		struct log_format p_format,
		int fd,
		bool colors_auto,
		const struct logger_entry *entry)
{
	int ret;
	char defaultBuffer[512];
	char *outBuffer = NULL;
	size_t totalLen;

	if (colors_auto)
		p_format.color = isatty(fd);

	outBuffer = log_format_log_line(p_format,
			defaultBuffer, sizeof(defaultBuffer), entry, &totalLen);

	if (!outBuffer)
		return -1;

	do {
		ret = write(fd, outBuffer, totalLen);
	} while (ret < 0 && errno == EINTR);

	if (ret < 0) {
		ERR("+++ LOG: write failed (errno=%d)\n", errno);
		ret = 0;
		goto done;
	}

	if (((size_t)ret) < totalLen) {
		ERR("+++ LOG: write partial (%d of %d)\n", ret,
				(int)totalLen);
		goto done;
	}

done:
	if (outBuffer != defaultBuffer)
		free(outBuffer);

	return ret;
}

/**
 * @}
 */
