#include <sys/epoll.h>

#include <fdi_pipe.h>
#include <fdi_logger.h>
#include <logretrieve.h>
#include <queued_entry_timestamp.h>

// buffers to use by default, when nothing specified
static const int default_buffers = (1 << LOG_ID_MAIN) | (1 << LOG_ID_SYSTEM) | (1 << LOG_ID_APPS);

int validate_buffers(int *enabled_buffers)
{
	if (*enabled_buffers &  (1 << LOG_ID_KMSG)
	&&  *enabled_buffers & ~(1 << LOG_ID_KMSG))
		return -EINVAL;

	if (!*enabled_buffers)
		*enabled_buffers = default_buffers;

	return 0;
}

int create_initial_fdis(struct fd_info ***fdis, int enabled_buffers, bool is_pipe, const struct log_config *conf)
{
	assert(fdis);
	assert(conf);

	if (validate_buffers(&enabled_buffers) < 0)
		return -EINVAL;

	struct fd_info **fdi_ptrs = calloc(bit_count(enabled_buffers) + 1 /* NULL terminator */, sizeof *fdi_ptrs);
	if (!fdi_ptrs)
		return -ENOMEM;

	__attribute__ ((cleanup(list_clear_free_contents))) list_head used_paths = NULL;

	int fdi_cnt = 0;
	int ret = 0;
	for (int i = 0; i < LOG_ID_MAX; ++i) {
		if (!bit_test(enabled_buffers, i))
			continue;

		const char *const bufname = log_name_by_id(i);
		struct fd_info *fdi;
		switch (i) {
		case LOG_ID_KMSG:
		case LOG_ID_SYSLOG:
			fdi = fdi_create(&ops_pipe, bufname);
			break;
		default:
			fdi = fdi_create(is_pipe ? &ops_pipe : &ops_logger, bufname);
			break;
		}
		if (!fdi) {
			ret = -ENOMEM;
			goto failure;
		}

		int r = fdi->ops->create(fdi, conf, bufname, &used_paths);
		if (r < 0) {
			// TODO: Inform about that again somewhere else
			fdi_free(fdi);
			continue;
		}
		fdi_ptrs[fdi_cnt++] = fdi;
	}

	if (enabled_buffers && !fdi_cnt) {
		ret = -EINVAL;
		goto failure;
	}

	*fdis = fdi_ptrs;
	return fdi_cnt;

failure:

	fdi_array_free(&fdi_ptrs);
	return ret;
}

int additional_options_init(struct additional_options *opt)
{
	opt->filter_list = NULL;
	opt->filter_object = log_filter_new();
	opt->logs_dump = DUMP_CONTINUOUS;
	opt->logs_size = DEFAULT_SORT_BUFFER_SIZE;
	return opt->filter_object ? 0 : -ENOMEM;
}

void additional_options_cleanup(struct additional_options *opt)
{
	list_clear(&opt->filter_list);
	log_filter_free(opt->filter_object);
}

struct fd_info *find_earliest_log(struct fd_info **data_fds, int fd_count, enum sorting_order sort_by)
{
	assert(data_fds);

	struct fd_info *best_fdi = NULL;
	struct timespec best_ts = {};

	for (int i = 0; i < fd_count; ++i) {
		struct fd_info *const fdi = data_fds[i];
		if (!fdi || !fdi->ops->has_log(fdi))
			continue;

		const struct logger_entry *const le = fdi->ops->peek_entry(fdi);
		struct timespec current_ts;
		extract_timestamp(le, sort_by, &current_ts);

		if (best_fdi && ((current_ts.tv_sec > best_ts.tv_sec) || (current_ts.tv_sec == best_ts.tv_sec && current_ts.tv_nsec > best_ts.tv_nsec)))
			continue;

		best_ts = current_ts;
		best_fdi = fdi;
	}

	return best_fdi;
}

int put_logs_into_vector(struct fd_info **data_fds, int fd_count, int nfds, struct sort_vector *logs, write_callback callback, void *userdata, log_filter *filter)
{
	assert(data_fds);
	assert(logs);
	assert(userdata);

	struct fd_info *best_fdi;
	do {
		best_fdi = find_earliest_log(data_fds, fd_count, logs->sort_by);
		if (!best_fdi)
			break;

		int r = fdi_push_log(best_fdi, logs, callback, userdata, filter);
		if (r)
			return r;
	} while (!nfds || best_fdi->ops->has_log(best_fdi)); // if a buffer got drained and there are still logs incoming, break to give it a chance to refill

	return 0;
}

/**
 * @brief Handle input
 * @details The main loop reading log data
 * @param[in] data_fds An array of FDs to read data from
 * @param[in] fd_count Size of the above array
 * @param[in] logs The log sorting vector
 * @param[in] l_file File output metadata
 * @return int 0 if successful, a negative value on error, callback's return value if it isn't 0
 */
int do_print(struct fd_info **data_fds, int fd_count, write_callback callback, void *userdata, bool sorting_needed, enum sorting_order sort_by, const struct log_config *conf, struct additional_options *opt)
{
	assert(data_fds);
	assert(userdata);

	int epoll_cnt = 0;
	__attribute__((cleanup(close_fd))) int epollfd = epoll_create1(0);
	if (epollfd < 0)
		return -errno;

	__attribute__ ((cleanup(sort_vector_free))) struct sort_vector logs;
	sort_vector_init(&logs);
	sort_vector_apply_config(&logs, conf);
	logs.dump = opt->logs_dump;
	logs.size = sorting_needed ? opt->logs_size : 0;
	logs.sort_by = sort_by;
	clock_gettime(get_proper_clock(sort_by), &logs.start);
	if (!sort_vector_finalize(&logs))
		return -ENOMEM;

	for (int nfds = 0; nfds < fd_count; ++nfds) {
		int r = data_fds[nfds]->ops->prepare_print(data_fds[nfds], logs.dump, opt->filter_list, opt->filter_object);
		if (r < 0)
			return r;
		if (r > 0) {
			// everything went fine, but we're not printing this one
			fdi_free(data_fds[nfds]);
			data_fds[nfds] = NULL;
			continue;
		}
		int fd = data_fds[nfds]->fd;
		r = fd_set_flags(fd, O_NONBLOCK);
		if (r < 0)
			return r;
		struct epoll_event ev = {
			.data.ptr = data_fds[nfds],
			.events = EPOLLIN,
		};
		epoll_ctl(epollfd, EPOLL_CTL_ADD, fd, &ev);
		++epoll_cnt;
	}

	__attribute__((cleanup(free_ptr))) struct epoll_event *evs = calloc(epoll_cnt, sizeof *evs);
	if (!evs)
		return -ENOMEM;

	while (epoll_cnt > 0) {
		int nfds;
		do
			nfds = epoll_wait(epollfd, evs, epoll_cnt, 100);
		while (nfds < 0 && errno == EINTR);
		if (nfds < 0)
			return -errno;

		for (int i = 0; i < nfds; ++i) {
			struct fd_info *fdi = evs[i].data.ptr;
			int r = fdi->ops->read(fdi);
			if (r < 0) {
				if (r == -EAGAIN)
					continue;
				else
					break;
			}

			if (r == 0) {
				epoll_ctl(epollfd, EPOLL_CTL_DEL, fdi->fd, NULL);
				--epoll_cnt;
			}
		}

		int r = put_logs_into_vector(data_fds, fd_count, nfds, &logs, callback, userdata, opt->filter_object);
		if (r != 0)
			return r;

		/* The oldest log can be so fresh as to be from the future
		 * (i.e. has a negative age). This can happen when somebody
		 * changes the realtime clock backwards. In that case let
		 * such logs be force-flushed immediately since the realtime
		 * timestamp has lost its meaning and the monotonic timestamp
		 * might not be present. Otherwise the log would wait until
		 * the clock reaches its old value again. */
		const long last_log_age = sort_vector_time_span(&logs);
		const bool is_from_future = (last_log_age < 0);
		if (logs.dump == DUMP_CONTINUOUS && (!nfds || (logs.old_logs_dumped == 1 && (last_log_age > logs.timeout || is_from_future)))) {
			r = sort_vector_flush(&logs, callback, userdata, is_from_future);
			if (r != 0)
				return r;
		}
	}

	int r = put_logs_into_vector(data_fds, fd_count, 0, &logs, callback, userdata, opt->filter_object);
	if (r != 0)
		return r;

	if (logs.dump != DUMP_CONTINUOUS && logs.dump != DUMP_INFINITE)
		while (sort_vector_used_size(&logs) > logs.dump)
			sort_vector_pop(&logs);

	r = sort_vector_flush(&logs, callback, userdata, true);
	if (r != 0)
		return r;

	return 0;
}
