/*
 * DLOG
 * Copyright (c) 2005-2008, The Android Open Source Project
 * Copyright (c) 2012-2018 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>

#include "parsers.h"

/**
 * @addtogroup SHARED_FUNCTIONS
 * @{
 */

int parse_number(const char **cursor, void *user_data)
{
	char *endptr;
	errno = 0;
	int val = strtol(*cursor, &endptr, 10);
	if (errno != 0 || endptr == *cursor)
		return -EINVAL;

	*cursor = endptr;
	int *pid = (int *)user_data;
	if (pid)
		*pid = val;
	return 0;
}

int parse_const_char(const char **cursor, void *user_data)
{
	if (strlen(*cursor) <= 0 || **cursor != *((char *)user_data))
		return -EINVAL;
	++(*cursor);
	return 0;
}

/**
 * @brief Parse syslog date
 * @details Parses timestamp from a syslog message
 * @param[in,out] cursor The beginning of the section, will contain the beginning of the next section
 * @param[out] timestamp The timestamp
 * @returns 0 on success, -errno on failure
 */
int parse_syslog_date(char **cursor, struct tm *timestamp)
{
	memset(timestamp, 0, sizeof *timestamp);

	char * const begin = *cursor;
	char * const end = strptime(begin, "%h %e %T", timestamp);
	if (!end)
		return -EINVAL;

	*cursor = end;
	return 0;
}

int parse_timestamp(const char **cursor, void *user_data)
{
	if (strlen(*cursor) <= 0)
		return -EINVAL;
	return parse_syslog_date((char **)cursor, (struct tm *)user_data);
}

int skip_name(const char **cursor, void *user_data)
{
	if (strlen(*cursor) <= 0)
		return -EINVAL;
	if (!isalpha(**cursor))
		return -EINVAL;
	do
		++(*cursor);
	while (isalpha(**cursor));
	return 0;
}

int parse_message(const char **cursor, void *user_data)
{
	if (strlen(*cursor) <= 0)
		return -EINVAL;
	const char **msg = (const char **)user_data;
	*msg = *cursor;
	return 0;
}

int parse(const char **cursor, struct parsers *parsers)
{
	assert(parsers);
	int i;
	for (i = 0; i < parsers->count; ++i) {
		int err = parsers->parsers[i].parse(cursor, parsers->parsers[i].user_data);
		if (err != 0)
			return err;
	}
	return 0;
}

int parse_optional(const char **cursor, void *user_data)
{
	struct parsers *optional_parsers = (struct parsers *)user_data;
	assert(optional_parsers->count > 0);
	struct parser *first_parser = optional_parsers->parsers;
	assert(first_parser);
	int err = first_parser->parse(cursor, first_parser->user_data);
	if (err != 0)
		return 0;

	struct parsers mandatory_parsers = (struct parsers) {
		.parsers = first_parser + 1,
		.count = optional_parsers->count - 1
	};
	return parse(cursor, &mandatory_parsers);
}

/**
 * @}
 */
