/*
 * Copyright (c) 2017, Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <assert.h>
#include <stdlib.h>
#include <ptrs_list.h>

struct list_elem {
	struct list_elem *next;
	elem_value value;
};

/**
 * @brief Remove element at given pointer
 * @details Removes given element from the list
 * @param[in] elem pointer to the element to remove from the list
 * @return pointer to the next element after removed one
 */
static struct list_elem *list_remove_at(struct list_elem *elem)
{
	assert(elem);  //valid elem pointer

	if (elem->next) {
		struct list_elem *next = elem->next;
		*elem = *next;
		free(next);
	} else {
		free(elem);
		elem = NULL;
	}
	return elem;
}

/**
 * @brief Check value equal to data given in user data
 * @details elem value is passed by pointer
 *          for list elements to be easily exchanged
 *          with some primitive values like integers
 * @param[in] value list value to compare
 * @param[in] user_data pointer to value to compare with
 * @return true if user data points on the same value
 */
static bool cond_value_equal(elem_value value, void *user_data)
{
	return (value == *((elem_value *)user_data));
}

/**
 * @brief Increment given counter
 * @details Used for counting list elements
 * @param[in] value list value to count (not used)
 * @param[in] user_data pointer to integer counter
 */
static void apply_increment_counter(elem_value value, void *user_data)
{
	int *counter = (int *)user_data;
	(*counter)++;
}

bool list_add(list_head *head, elem_value value)
{
	assert(head); //valid pointer

	struct list_elem *elem = (struct list_elem *)calloc(1, sizeof(struct list_elem));
	if (elem == NULL)
		return false;

	*elem = (struct list_elem) {
		.next = *head,
		.value = value
	};

	(*head) = elem;
	return true;
}

void list_remove(list_head *head, elem_value value)
{
	list_remove_if(head, &value, cond_value_equal);
}

void list_remove_if(list_head *head, void *user_data, cond_cb cond)
{
	assert(cond);

	while (*head && cond((*head)->value, user_data))
		*head = list_remove_at(*head);

	struct list_elem *prev = *head;
	while (prev && prev->next) {
		if (cond(prev->next->value, user_data))
			prev->next = list_remove_at(prev->next);
		else
			prev = prev->next;
	}
}

void list_clear(list_head *head)
{
	while (*head)
		*head = list_remove_at(*head);
}

void list_clear_free_contents(list_head *head)
{
	while (*head) {
		free(list_at(*head));
		*head = list_remove_at(*head);
	}
}

void list_clear_custom(list_head *head, void *user_data, apply_cb clear)
{
	while (*head) {
		clear(list_at(*head), user_data);
		*head = list_remove_at(*head);
	}
}

void list_foreach(list_head head, void *user_data, apply_cb apply)
{
	assert(apply);

	struct list_elem *elem;
	for (elem = head; elem; elem = elem->next)
		apply(elem->value, user_data);
}

list_head list_map(list_head head, void *user_data, map_cb map, apply_cb clear)
{
	assert(map);

	list_head ret = NULL, *target = &ret;
	for (list_head p = head; p != NULL; list_next(&p)) {
		elem_value new = map(list_at(p), user_data);
		if (!new) {
			if (clear)
				list_clear_custom(&ret, user_data, clear);
			else
				list_clear(&ret);
			return NULL;
		}

		if (!list_add(target, new)) {
			if (clear) {
				clear(new, user_data);
				list_clear_custom(&ret, user_data, clear);
			} else {
				list_clear(&ret);
			}
			return NULL;
		}

		target = &(*target)->next;
	}

	return ret;
}

elem_value list_at(list_head iter)
{
	return iter->value;
}

void list_next(list_head *iter)
{
	if (!*iter)
		return;

	*iter = (*iter)->next;
}

int list_count(list_head head)
{
	int counter = 0;
	list_foreach(head, &counter, apply_increment_counter);
	return counter;
}
