/*
 * DLOG
 * Copyright (c) 2005-2008, The Android Open Source Project
 * Copyright (c) 2012-2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// C
#include <assert.h>
#include <stdio.h>
#include <string.h>

// DLog
#include <dlog.h>
#include <logconfig.h>
#include "parsers.h"
#include <queued_entry.h>
#include "translate_syslog.h"

/**
 * @addtogroup SHARED_FUNCTIONS
 * @{
 */

/**
 * @brief Create pipe message
 * @details Prepares a pipe message structure by filling its members appropriately
 * @param[out] buf The pointer to (the buffer containing) the struct
 * @param[in] prio Priority level
 * @param[in] tag Log tag
 * @param[in] msg Log message contents
 */
void create_pipe_message(void *buf, int prio, char const *tag, char const *msg)
{
	assert(buf);
	assert(msg);
	assert(tag);
	assert(prio >= DLOG_VERBOSE);
	assert(prio < DLOG_PRIO_MAX);
	assert(1 /* priority */ + strlen(tag) + 1 /* NULL delimiter */ + strlen(msg) + 1 /* NULL terminator */ <= LOG_MAX_PAYLOAD_SIZE);

	struct pipe_logger_entry *ple = (struct pipe_logger_entry *) buf;
	struct timespec ts;

	int tag_l = strlen(tag) + 1;
	int msg_l = strlen(msg) + 1;
	int len = 1 + msg_l + tag_l + sizeof *ple;

	clock_gettime(CLOCK_MONOTONIC, &ts);
	ple->sec_sent_mono = ts.tv_sec;
	ple->nsec_sent_mono = ts.tv_nsec;

	clock_gettime(CLOCK_REALTIME, &ts);
	ple->sec_sent_real = ts.tv_sec;
	ple->nsec_sent_real = ts.tv_nsec;

	ple->padding = 0;
	ple->len = len;
	ple->pid = getpid();
	ple->tid = gettid();
	ple->msg[0] = prio;
	memcpy(ple->msg + 1, tag, tag_l);
	memcpy(ple->msg + 1 + tag_l, msg, msg_l);
}

/**
 * @brief Parse KMSG message
 * @details Parses a raw KMSG message into the internal format
 *
 * See https://www.kernel.org/doc/Documentation/ABI/testing/dev-kmsg
 * for a detailed description of the format (or see below for TL;DR)
 *
 * Here's what our parser does with the message:
 * 1,234,56789,-,FOO,BAR;Some message.\nSOME_TAG=MONTAG\n
 *
 * ^ ^   ^     ^ ^       ^              ^
 * │ │   │     │ │       │              └─ extra tags: ignored.
 * │ │   │     │ │       └──────────────── message: taken.
 * │ │   │     │ └──────────────────────── currently there are no fields past the flag. If such are introduced in the future they will be ignored.
 * │ │   │     └────────────────────────── flag: ignored.
 * │ │   └──────────────────────────────── timestamp: converted from milliseconds and taken.
 * │ └──────────────────────────────────── sequential identifier: ignored.
 * └────────────────────────────────────── syslog priority: taken.
 *
 * @param[in,out] buffer The buffer which contains the message and will contain the entry afterwards
 * @param[in] buffer_size The size of the input buffer
 * @return 0 on success, -errno on failure
 */
int parse_kmsg_message(char *buffer, struct logger_entry_with_msg *lem, int buffer_size)
{
	char *current = buffer;

	// syslog-style priority
	int fac_prio;
	if (parse_number((const char **)&current, &fac_prio) < 0)
		return -EINVAL;

	// sequential ID; skipped
	current = strchr(current + 1, ',');
	if (current == NULL)
		return -EINVAL;

	// timestamp (in usec)
	char *const timestamp_str = current + 1;
	current = strchr(timestamp_str, ',');
	if (current == NULL)
		return -EINVAL;
	*current = '\0';

	errno = 0;
	unsigned long long timestamp = strtoull(timestamp_str, NULL, 10);
	if (errno)
		return -errno;
	lem->header.sec_sent_mono  = timestamp / (1000*1000);
	lem->header.nsec_sent_mono = 1000 * (timestamp % (1000 * 1000));
	/* only monotonic sendtime is supported by kmsg backend */
	lem->header.sec_sent_real = INVALID_SEC;
	lem->header.nsec_sent_real = INVALID_NSEC;
	lem->header.default_send_ts_type = LOGGER_ENTRY_MONOTONIC;

	// afterwards there's the flag (and others, in case the format gets extended); skipped
	current = strchr(current + 1, ';');
	if (current == NULL)
		return -EINVAL;
	++current;

	// the message; always ends with '\n'
	char *const msg_begin = current;
	current = strchr(msg_begin, '\n');
	if (current == NULL)
		return -EINVAL;
	*current = '\0';

	lem->header.pid = 0;
	lem->header.tid = 0;

	lem->header.priority = translate_syslog_priority(fac_prio);
	lem->header.tag_len = snprintf(lem->msg, sizeof lem->msg - 1, "%s", translate_syslog_facility(fac_prio));
	lem->header.len = lem->header.tag_len + 1; // plus NULL delimiter
	lem->header.len += 1 + snprintf(lem->msg + lem->header.len, sizeof lem->msg - 1 - lem->header.len, "%s", msg_begin);

	lem->header.len += sizeof(struct logger_entry);
	lem->header.padding = 0;

	return 0;
}

/* Most of the logic between Android Logger and Pipe messages is similar and they
   happen to share field names, though this can't be a function because they are
   fundamentally different types (and casting pointers would just result in hidden
   bugs when the fields change). Ideally this would be a template but we're no C++.
*/
#define SHARED_PROCESS_MESSAGE(xle) { \
	le->tag_len = -1; \
	le->priority = xle->msg[0]; \
	for (int i = 1 /* 0 is prio */; i < payload_size; ++i) { \
		if (xle->msg[i] != '\0') \
			continue; \
		le->tag_len = i - 1; \
		break; \
	} \
	if (le->tag_len == -1) { \
		/* ignore for now */ \
	} \
	payload_size--; /* there's priority at first byte which is not a part of a message */ \
	le->len = sizeof *le + payload_size; \
	le->pid = xle->pid; \
	le->tid = xle->tid; \
	memmove(le->msg, xle->msg + 1, payload_size); \
}

#ifdef USE_ANDROID_MONOTONIC
static void parse_android_logger_footer(struct android_logger_entry *ale, struct logger_entry *le, size_t *payload_size)
{
	assert(ale);
	assert(le);
	assert(payload_size);

	struct android_logger_footer *alf;
	if (*payload_size <= sizeof *alf)
		return;

	alf = (struct android_logger_footer*)(ale->msg + *payload_size - sizeof *alf);
	if (alf->magic != ANDROID_LOGGER_FOOTER_MAGIC)
		return;

	*payload_size -= sizeof *alf;
	le->sec_sent_mono = alf->sec_sent_mono;
	le->nsec_sent_mono = alf->nsec_sent_mono;
}
#endif

/**
 * @brief Parse AL message
 * @details Parses an Android Logger message into DLog's internal format
 * @param[in] ale The buffer containing the android logger datagram message
 * @param[out] le Target structure to contain the log
 * @param[in] size The size of the datagram
 */
void parse_androidlogger_message(struct android_logger_entry *ale, struct logger_entry *le, size_t dgram_size)
{
	size_t payload_size = dgram_size - sizeof *ale;
	assert(payload_size > 0);

	/* default to using the realtime timestamp
	 * to at least have some ordering information
	 * if the monotonic timestamp is missing */
	le->sec_sent_mono = INVALID_SEC;
	le->nsec_sent_mono = INVALID_NSEC;

#ifdef USE_ANDROID_MONOTONIC
	/* there may be android_logger_footer at the end of payload, right after log message,
	 * but only if the message has been prepared and sent by libdlog,
	 * not directly from shell script via /dev file */
	parse_android_logger_footer(ale, le, &payload_size);
#endif

	SHARED_PROCESS_MESSAGE(ale);
	le->sec_sent_real = ale->sec_sent;
	le->nsec_sent_real = ale->nsec_sent;
	le->default_send_ts_type = LOGGER_ENTRY_REALTIME;
}

void parse_pipe_message(struct pipe_logger_entry *ple, struct logger_entry *le, size_t msg_size)
{
	size_t payload_size = msg_size - sizeof *ple;
	assert(payload_size <= LOG_MAX_PAYLOAD_SIZE);

	SHARED_PROCESS_MESSAGE(ple);
	le->sec_sent_mono = ple->sec_sent_mono;
	le->nsec_sent_mono = ple->nsec_sent_mono;
	le->sec_sent_real = ple->sec_sent_real;
	le->nsec_sent_real = ple->nsec_sent_real;
	le->default_send_ts_type = LOGGER_ENTRY_MONOTONIC;
}

/**
 * @brief Construct log from syslog datagram
 * @details Constructs a logger entry from data contained in a syslog datagram.
 * @param[out] le The logger entry structure to construct
 * @param[in] priority Priority
 * @param[in] tag Tag
 * @param[in] timestamp Timestamp
 * @param[in] pid Process ID
 * @param msg Message
 */
static void construct_logger_entry_from_syslog(struct logger_entry *le, int priority, const char *tag, struct tm *timestamp, int pid, const char *msg)
{
	assert(le);
	assert(tag);
	assert(timestamp);
	assert(msg);

	le->tag_len = strlen(tag);
	const int msg_len = strlen(msg) + 1;

	le->len = sizeof(struct logger_entry) + le->tag_len + 1 + msg_len;
	le->pid = pid;
	le->tid = 0; // could be set to PID but this way it's explicit the TID is not known
	le->sec_sent_real = mktime(timestamp);
	le->nsec_sent_real = 0; // the timestamp does not have such precision
	/* only realtime sendtime is supported by syslog backend */
	le->sec_sent_mono = INVALID_SEC;
	le->nsec_sent_mono = INVALID_NSEC;
	le->default_send_ts_type = LOGGER_ENTRY_REALTIME;
	le->priority = priority;

	char *const msg_begin = le->msg + le->tag_len + 1;
	memcpy(le->msg, tag, le->tag_len + 1);
	memcpy(msg_begin, msg, msg_len);
}

/**
 * @brief Parse syslog message
 * @details Interprets a message received from syslog
 * @param[in] buffer The buffer which contains the message to parse
 * @param[in] buffer_size The size of the received datagram
 * @param[out] le The logger entry to fill
 * @return 0 on failure, 1 on success
 */
int parse_syslog_datagram(const char *buffer, int buffer_size, struct logger_entry * le)
{
	int fac_prio = -1, pid = 0;
	struct tm timestamp;
	const char *msg = NULL;

	struct parser parsePid[] = {
		{ parse_const_char, "[" },
		{ parse_number, &pid },
		{ parse_const_char, "]" },
	};
	struct parsers optionalPid = (struct parsers) {
		.parsers = parsePid,
		.count = NELEMS(parsePid)
	};
	struct parser parseSyslog[] = {
		{ parse_const_char, "<" },
		{ parse_number, &fac_prio },
		{ parse_const_char, ">" },
		{ parse_timestamp, &timestamp },
		{ parse_const_char, " " },
		{ skip_name, NULL },
		{ parse_optional, &optionalPid },
		{ parse_const_char, ":" },
		{ parse_const_char, " " },
		{ parse_message, &msg }
	};
	struct parsers syslogParsers = (struct parsers) {
		.parsers = parseSyslog,
		.count = NELEMS(parseSyslog)
	};
	const char *cursor = buffer;
	int err = parse(&cursor, &syslogParsers);
	if (err != 0)
		return err;

	construct_logger_entry_from_syslog(le, translate_syslog_priority(fac_prio), translate_syslog_facility(fac_prio), &timestamp, pid, msg);

	return 0;
}

/**
 * @}
 */
