/*
 * DLOG
 * Copyright (c) 2005-2008, The Android Open Source Project
 * Copyright (c) 2012-2015 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

// C
#include <assert.h>

// DLog
#include <queued_entry.h>
#include <queued_entry_timestamp.h>

/**
 * @brief Reception timestamp
 * @details Add a timestamp describing reception time.
 * @param[in/out] le The logger entry to stamp time on.
 */
void add_recv_timestamp(struct logger_entry *le)
{
	struct timespec ts;

	clock_gettime(CLOCK_MONOTONIC, &ts);
	le->sec_recv_mono = ts.tv_sec;
	le->nsec_recv_mono = ts.tv_nsec;

	clock_gettime(CLOCK_REALTIME, &ts);
	le->sec_recv_real = ts.tv_sec;
	le->nsec_recv_real = ts.tv_nsec;
}

/**
 * @brief Copy reception timestamp
 * @details Replicates recv timestamp from sent timestamp where adding a new one would not make sense
 * @param[in/out] le The logger entry to replicate the time stamp for.
 */
void copy_recv_timestamp(struct logger_entry *le)
{
	le->sec_recv_mono  = le->sec_sent_mono;
	le->nsec_recv_mono = le->nsec_sent_mono;
	le->sec_recv_real  = le->sec_sent_real;
	le->nsec_recv_real = le->nsec_sent_real;
}

bool log_entry_is_earlier(const enum sorting_order sort_by, const struct logger_entry *lhs, const struct logger_entry *rhs)
{
	assert(lhs);
	assert(rhs);

	int32_t r_sec, r_nsec, l_sec, l_nsec;
	switch (sort_by) {
	case SORT_RECV_REAL:
		r_sec = rhs->sec_recv_real;
		r_nsec = rhs->nsec_recv_real;
		l_sec = lhs->sec_recv_real;
		l_nsec = lhs->nsec_recv_real;
		break;
	case SORT_RECV_MONO:
		r_sec = rhs->sec_recv_mono;
		r_nsec = rhs->nsec_recv_mono;
		l_sec = lhs->sec_recv_mono;
		l_nsec = lhs->nsec_recv_mono;
		break;
	case SORT_SENT_REAL:
		r_sec = rhs->sec_sent_real;
		r_nsec = rhs->nsec_sent_real;
		l_sec = lhs->sec_sent_real;
		l_nsec = lhs->nsec_sent_real;
		break;
	default:
		r_sec = rhs->sec_sent_mono;
		r_nsec = rhs->nsec_sent_mono;
		l_sec = lhs->sec_sent_mono;
		l_nsec = lhs->nsec_sent_mono;
		break;
	};
	/* Returning false for both the left- and the right-handed operand
	 * (as opposed to returning true for the other) makes it so that logs
	 * are left in the order they are received. See sort_vector.c */
	if (l_nsec == INVALID_NSEC || r_nsec == INVALID_NSEC)
		return false;

	return l_sec < r_sec || (l_sec == r_sec && l_nsec < r_nsec);
}

void extract_timestamp(const struct logger_entry *le, enum sorting_order stamp_type, struct timespec *ts)
{
	assert(le);
	assert(ts);

	switch (stamp_type) {
	case SORT_SENT_MONO:
		ts->tv_sec = le->sec_sent_mono;
		ts->tv_nsec = le->nsec_sent_mono;
		break;
	case SORT_SENT_REAL:
		ts->tv_sec = le->sec_sent_real;
		ts->tv_nsec = le->nsec_sent_real;
		break;
	case SORT_RECV_MONO:
		ts->tv_sec = le->sec_recv_mono;
		ts->tv_nsec = le->nsec_recv_mono;
		break;
	case SORT_RECV_REAL:
		ts->tv_sec = le->sec_recv_real;
		ts->tv_nsec = le->nsec_recv_real;
		break;

	default: assert(false); // LCOV_EXCL_LINE
	}
}

clockid_t get_proper_clock(enum sorting_order sort_by)
{
	switch (sort_by) {
	case SORT_SENT_MONO:
	case SORT_RECV_MONO:
		return CLOCK_MONOTONIC;

	case SORT_SENT_REAL:
	case SORT_RECV_REAL:
		return CLOCK_REALTIME;

	default: assert(false); // LCOV_EXCL_LINE
	}
}
