/*
 * Copyright (c) 2018, Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include <log_file.h>
#include <logconfig.h>
#include <queued_entry_timestamp.h>

#include "sort_vector.h"

#define MAX_SORT_BUFFER_SIZE 1048576 // fairly arbitrary, mostly to prevent accidents

#define DEFAULT_SORT_TIMEOUT 1000 // ms
#define MAX_SORT_TIMEOUT 60000

void sort_vector_init(struct sort_vector *logs)
{
	assert(logs);

	logs->data = NULL;
	logs->begin = 0;
	logs->end = 0;

	logs->dump = DUMP_CONTINUOUS;
	logs->size = DEFAULT_SORT_BUFFER_SIZE;
	logs->timeout = DEFAULT_SORT_TIMEOUT;
	logs->sort_by = SORT_SENT_MONO;
	logs->old_logs_dumped = 0;
}

int sort_vector_finalize(struct sort_vector *logs)
{
	assert(logs);

	if (logs->dump != DUMP_INFINITE && logs->dump != DUMP_CONTINUOUS)
		logs->size = logs->dump + 1; // +1 so that `-t 1` works (buffer must have capacity at least 2)

	logs->size = clamp_int(logs->size, 0, MAX_SORT_BUFFER_SIZE);
	logs->timeout = clamp_int(logs->timeout, 0, MAX_SORT_TIMEOUT);

	if (!IS_VECTOR_SIZE_SORTABLE(logs->size))
		return 1;

	logs->data = (struct logger_entry **)calloc(logs->size, sizeof(struct logger_entry *));
	return logs->data != NULL;
}

void sort_vector_free(struct sort_vector *logs)
{
	while (!sort_vector_empty(logs))
		sort_vector_pop(logs);
	free(logs->data);
}

void sort_vector_apply_config(struct sort_vector *logs, const struct log_config *config)
{
	assert(config);
	assert(logs);

	logs->size = log_config_get_int(config, "util_sorting_buffer_size", logs->size);
	logs->timeout = log_config_get_int(config, "util_sorting_time_window", logs->timeout);
}

struct logger_entry *sort_vector_back(struct sort_vector *logs)
{
	return logs->data[logs->begin];
}

int sort_vector_empty(struct sort_vector *logs)
{
	return logs->begin == logs->end;
}

int sort_vector_full(struct sort_vector *logs)
{
	return (logs->end + 1) % logs->size == logs->begin;
}

void sort_vector_pop(struct sort_vector *logs)
{
	free(logs->data[logs->begin]);
	logs->data[logs->begin] = NULL;
	logs->begin = (logs->begin + 1) % logs->size;
}

int sort_vector_used_size(struct sort_vector *logs)
{
	if (logs->end >= logs->begin)
		return logs->end - logs->begin;
	else
		return logs->size - logs->begin + logs->end;
}

/**
 * @brief Push log
 * @details Push a log to the sorting container
 * @param[in] p Pushee (taken ownership of)
 * @param[in] logs Log sorting vector
 * @return 0 on success, -errno on failure
 */
int sort_vector_push(struct sort_vector *logs, struct logger_entry *p, write_callback callback, void *userdata)
{
	assert(logs);
	assert(logs->data);
	assert(IS_VECTOR_SIZE_SORTABLE(logs->size));
	assert(p);
	assert(userdata);

	if (sort_vector_full(logs)) {
		if (logs->dump == DUMP_CONTINUOUS || logs->dump == DUMP_INFINITE) {
			const int r = callback(sort_vector_back(logs), userdata);
			if (r != 0)
				return r;
		}
		sort_vector_pop(logs);
	}

	/* Insertion sort. Logs usually appear in order, so it makes a lot of sense to just iterate
	 * linearly from the end to try and find the correct index. The algorithm starts working
	 * from the empty slot behind the last index and populates it with the greater of the log
	 * in the last index and the one we're trying to insert, which usually ends up being the
	 * latter, but if the choice was the former, it decrements the index and repeats the process.
	 *
	 * In general, this means that during iteration some entries will be duplicated. This looks
	 * a bit like a memory leak (especially to static analysis tools) but is fine since the
	 * duplicated value is always in the spot we're currently replacing and it is never itself
	 * eligible to be placed there, ensuring that all values are unique and sorted afterward.
	 *
	 * Let's illustrate this with a visual example. Suppose `p` has value 6, and that in the
	 * table there are already four values at indices 0, 1, 2, and 3 (thus logs->end == 4) with
	 * values 2, 3, 7, and 9 respectively:
	 * ┌─┬─┬─┬─┐   ┌─┐
	 * │2│3│7│9│  p│6│
	 * └─┴─┴─┴─┘   └─┘
	 * After insertion, we'd like the table to look like this:
	 * ┌─┬─┬─┬─┬─┐
	 * │2│3│6│7│9│
	 * └─┴─┴─┴─┴─┘
	 * Initially, i = 4 (the empty spot, marked by 'v' in the graphic below) is
	 * due replacement and the two candidates (marked by '?') are 6 (from `p`)
	 * and 9 (from the slot at i-1, represented in the code by `e`):
	 * ┌─┬─┬─┬?┬v┐   ┌?┐
	 * │2│3│7│9│ │  p│6│
	 * └─┴─┴─┴─┴─┘   └─┘
	 * 9 > 6, so it gets picked:
	 * ┌─┬─┬?┬v┬─┐   ┌?┐
	 * │2│3│7│9│9│  p│6│
	 * └─┴─┴─┴─┴─┘   └─┘
	 * It looks as if this duplicated this value, but now i = 3, and one of them is due replacement
	 * with either 7 (the value from slot i-1) or 6 (from `p`). Again, the earlier slot gets picked:
	 * ┌─┬?┬v┬─┬─┐   ┌?┐
	 * │2│3│7│7│9│  p│6│
	 * └─┴─┴─┴─┴─┘   └─┘
	 * This time, at i = 2, the choice is won by the value from `p`, ending iteration:
	 * ┌─┬─┬─┬─┬─┐
	 * │2│3│6│7│9│
	 * └─┴─┴─┴─┴─┘
	 */
	int i;
	for (i = logs->end; i > logs->begin || (logs->end < logs->begin && i <= logs->end); i = (i ?: logs->size) - 1) {
		/* Note that in the case in which neither of p and e is earlier than the other, we pretend
		 * e is earlier. This way, in the case of the invalid timestamps, we output the entries
		 * in the order we receive them, because this best reflects the "true" order of those logs.
		 * This breaks if only some of the entries have invalid timestamps (for example,
		 * we receive 2.0, invalid, 1.0 in this order), but this situation is rare since almost
		 * always all of the entries we receive have the same timestamps. Also, in such cases
		 * correct sorting is impossible anyway.
		 */
		struct logger_entry *const e = logs->data[(i ?: logs->size) - 1];
		if (!log_entry_is_earlier(logs->sort_by, p, e))
			break;
		logs->data[i] = e;
	}

	logs->data[i] = p;
	logs->end = (logs->end + 1) % logs->size;
	return 0;
}

long sort_vector_time_span(struct sort_vector *logs)
{
	struct timespec now, oldest;
	long diff_s, diff_ns;
	if (sort_vector_empty(logs))
		return 0;

	clock_gettime(get_proper_clock(logs->sort_by), &now);
	extract_timestamp(sort_vector_back(logs), logs->sort_by, &oldest);

	diff_s = now.tv_sec - oldest.tv_sec;
	diff_ns = now.tv_nsec - oldest.tv_nsec;
	if (diff_ns < 0) {
		diff_ns += 1000000000;
		--diff_s;
	}

	return (diff_s * 1000) + (diff_ns/1000000);
}

/**
 * @brief Process log
 * @details Processes a raw log and, if it's old enough, prints it
 * @param[in] e The entry to process
 * @param[in] reference_ts Reference timestamp
 * @param[in] stamp_type Type of the reference time stamp
 * @param[in] callback Callback to call in order to print the log
 * @param[in] userdata Data to pass to the callback
 * @param[in] timeout Sort timeout in ms
 * @param[out] callback_ret Return value of the callback if it has been called
 * @return bool whether the log was old enough
 */
bool process_log(const struct logger_entry *e, struct timespec reference_ts, enum sorting_order stamp_type, write_callback callback, void *userdata, long timeout, int *callback_ret)
{
	assert(e);
	assert(userdata);
	assert(callback_ret);

	if (timeout != -1) {

		struct timespec log_ts;
		extract_timestamp(e, stamp_type, &log_ts);

		long  s = reference_ts.tv_sec  - log_ts.tv_sec;
		long ns = reference_ts.tv_nsec - log_ts.tv_nsec;
		if (ns < 0) {
			ns += 1000000000;
			--s;
		}

		if (timeout > (s * 1000) + (ns / 1000000))
			return false;
	}

	*callback_ret = callback(e, userdata);
	return true;
}

int sort_vector_flush(struct sort_vector *logs, write_callback callback, void *userdata, bool force)
{
	struct timespec ts;
	clock_gettime(get_proper_clock(logs->sort_by), &ts);

	int callback_ret = 0;
	while (!sort_vector_empty(logs) && process_log(sort_vector_back(logs), ts, logs->sort_by, callback, userdata, force ? -1 : logs->timeout, &callback_ret) && callback_ret == 0)
		sort_vector_pop(logs);

	return callback_ret;
}
