/*
 * DLOG
 * Copyright (c) 2005-2008, The Android Open Source Project
 * Copyright (c) 2012-2018 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <assert.h>
#include <syslog.h>

#include <dlog.h>
#include <logcommon.h>

#include "translate_syslog.h"

/**
 * @addtogroup SHARED_FUNCTIONS
 * @{
 */

enum syslog_priority {
	SYSLOG_EMERGENCY = 0,
	SYSLOG_ALERT,
	SYSLOG_CRITICAL,
	SYSLOG_ERROR,
	SYSLOG_WARNING,
	SYSLOG_NOTICE,
	SYSLOG_INFO,
	SYSLOG_DEBUG,
};

static const int syslog_priority_mapping[] = {
	[SYSLOG_EMERGENCY] = DLOG_FATAL,
	[SYSLOG_ALERT]     = DLOG_FATAL,
	[SYSLOG_CRITICAL]  = DLOG_FATAL,
	[SYSLOG_ERROR]     = DLOG_ERROR,
	[SYSLOG_WARNING]   = DLOG_WARN,
	[SYSLOG_NOTICE]    = DLOG_INFO,
	[SYSLOG_INFO]      = DLOG_INFO,
	[SYSLOG_DEBUG]     = DLOG_DEBUG,
};

#define STR_FAC(x)[LOG_FAC(x)] = "SYS" #x // [LOG_FOO] = "SYSLOG_FOO"
static const char *syslog_facilities[LOG_NFACILITIES] = {
	STR_FAC(LOG_KERN),
	STR_FAC(LOG_USER),
	STR_FAC(LOG_MAIL),
	STR_FAC(LOG_DAEMON),
	STR_FAC(LOG_AUTH),
	STR_FAC(LOG_SYSLOG),
	STR_FAC(LOG_LPR),
	STR_FAC(LOG_NEWS),
	STR_FAC(LOG_UUCP),
	STR_FAC(LOG_CRON),
	STR_FAC(LOG_AUTHPRIV),
	[11] = "SYSLOG_SYSTEM_11", // 11-15 are system reserved without any LOG_SYSTEM constants
	[12] = "SYSLOG_SYSTEM_12",
	[13] = "SYSLOG_SYSTEM_13",
	[14] = "SYSLOG_SYSTEM_14",
	[15] = "SYSLOG_SYSTEM_15",
	STR_FAC(LOG_LOCAL0),
	STR_FAC(LOG_LOCAL1),
	STR_FAC(LOG_LOCAL2),
	STR_FAC(LOG_LOCAL3),
	STR_FAC(LOG_LOCAL4),
	STR_FAC(LOG_LOCAL5),
	STR_FAC(LOG_LOCAL6),
	STR_FAC(LOG_LOCAL7),
};
#undef STR_FAC

/**
 * @brief Translate syslog priority
 * @details Maps syslog priority to DLog priority
 * @param[in] fac_prio Syslog facility/priority
 * @return DLog priority
 */
int translate_syslog_priority(int fac_prio)
{
	const int prio = LOG_PRI(fac_prio);

	assert(prio >= 0);
	assert(prio < NELEMS(syslog_priority_mapping));
	return syslog_priority_mapping[prio];
}

/**
 * @brief Translate syslog facility
 * @details Translates the integer representation of a syslog facility into a string
 * @param[in] fac_prio The facility/priority (as provided by syslog)
 * @return String representation of the facility name
 */
const char *translate_syslog_facility(int fac_prio)
{
	const int facility = LOG_FAC(fac_prio);

	assert(facility >= 0);

	if (facility >= NELEMS(syslog_facilities))
		return "SYSLOG-UNKNOWN";

	return syslog_facilities[facility];
}

/**
 * @}
 */
