// C
#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// POSIX
#include <dirent.h>
#include <fcntl.h>

// DLog
#include <logconfig.h>

static void conf_hash(const char *key, const char *value, void *userdata)
{
	unsigned *const hash = (unsigned *)userdata;

	while (*key)
		*hash += *(key++);
	while (*value)
		*hash += *(value++);
}

static bool fail_scandirat;
int __real_scandirat(int dirfd, const char *dirp, struct dirent ***namelist, int (*filter)(const struct dirent *), int (*compar)(const struct dirent **, const struct dirent **));
int __wrap_scandirat(int dirfd, const char *dirp, struct dirent ***namelist, int (*filter)(const struct dirent *), int (*compar)(const struct dirent **, const struct dirent **))
{
	return fail_scandirat ? -1 : __real_scandirat(dirfd, dirp, namelist, filter, compar);
}

static bool fail_openat;
int __real_openat(int dirfd, const char *pathname, int flags, mode_t mode);
int __wrap_openat(int dirfd, const char *pathname, int flags, mode_t mode)
{
	return fail_openat ? -1 : __real_openat(dirfd, pathname, flags, mode);
}

static bool fail_fdopen;
FILE *__real_fdopen(int fd, const char *mode);
FILE *__wrap_fdopen(int fd, const char *mode)
{
	return fail_fdopen ? NULL : __real_fdopen(fd, mode);
}

static bool fail_open;
int __real_open(const char *pathname, int flags, mode_t mode);
int __wrap_open(const char *pathname, int flags, mode_t mode)
{
	return fail_open ? -1 : __real_open(pathname, flags, mode);
}

static bool fail_calloc;
void *__real_calloc(size_t nmemb, size_t size);
void *__wrap_calloc(size_t nmemb, size_t size)
{
	return fail_calloc ? NULL : __real_calloc(nmemb, size);
}

int main()
{
	const char * get;
	struct log_config config;

	setenv("DLOG_CONFIG_PATH", "./non_existent_file", 1);
	assert(log_config_read(&config) == -ENOENT);

	setenv("DLOG_CONFIG_PATH", "src/tests/test.conf", 1);

#define FAIL_EXTRA_CONF_CALL(callname) \
	fail_##callname = true; \
	assert(!log_config_read(&config)); \
	assert(log_config_get(&config, "foo")); \
	assert(!log_config_get(&config, "valid_extra_conf")); \
	log_config_free(&config); \
	fail_##callname = false

	FAIL_EXTRA_CONF_CALL(scandirat);
	FAIL_EXTRA_CONF_CALL(openat);
	FAIL_EXTRA_CONF_CALL(fdopen);
	FAIL_EXTRA_CONF_CALL(open);

#undef FAIL_EXTRA_CONF_CALL

	assert(!log_config_read(&config));

	/* test if key retrieval works first (needed for set tests) */
	get = log_config_get(&config, "foo");
	assert(get);
	assert(!strncmp(get, "bar", 4));
	assert(!log_config_get(&config, "baz"));
	assert(!log_config_get(&config, "tony"));

	assert(!log_config_get(&config, "nitz"));

	get = log_config_get(&config, "valid_extra_conf");
	assert(get);
	assert(!strncmp(get, "123", 4));
	assert(log_config_get_int(&config, "valid_extra_conf", 567) == 123);

	get = log_config_get(&config, "invalid_extra_conf");
	assert(!get);
	assert(log_config_get_int(&config, "invalid_extra_conf", 678) == 678);

	/* test key length limits */
	get = log_config_get(&config, "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa");
	assert(get);
	assert(!strcmp(get, "foo"));

	get = log_config_get(&config, "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa");
	assert(!get);

	/* test existing key modification */
	log_config_set(&config, "foo", "quux");
	get = log_config_get(&config, "foo");
	assert(get);
	assert(strncmp(get, "bar", 4));
	assert(!strncmp(get, "quux", 5));

	/* test new key addition */
	log_config_set(&config, "test_add_key", "ok");
	get = log_config_get(&config, "test_add_key");
	assert(get);
	assert(!strncmp(get, "ok", 3));

	log_config_set(&config, "test_enormous_numbers", "999999999999999999999999999999999999999999999999999999");
	assert(log_config_get_int(&config, "test_enormous_numbers", 42) == 42);

	/* test boolean conversion */
	log_config_set(&config, "b_true" , "1");
	log_config_set(&config, "b_false", "0");
	log_config_set(&config, "b_none" , "foo");

	assert(log_config_get_boolean(&config, "b_true" , true));
	assert(log_config_get_boolean(&config, "b_true" , false));
	assert(!log_config_get_boolean(&config, "b_false", true));
	assert(!log_config_get_boolean(&config, "b_false", false));
	assert(log_config_get_boolean(&config, "b_none" , true));
	assert(!log_config_get_boolean(&config, "b_none" , false));

	unsigned hash_original = 0U;
	log_config_foreach(&config, conf_hash, &hash_original);

	struct log_config copy = {NULL, NULL};
	log_config_copy(&copy, &config);

	unsigned hash_copy = 0U;
	log_config_foreach(&copy, conf_hash, &hash_copy);

	assert(hash_copy == hash_original);

	assert(log_config_get_boolean(&copy, "b_true", false));
	assert(log_config_remove(&copy, "b_true"));
	assert(!log_config_remove(&copy, "b_true"));
	assert(!log_config_get_boolean(&copy, "b_true", false));
	assert(!log_config_remove(&copy, "we wtorki ser jest zielony"));
	assert(!log_config_remove(&copy, "we wtorki ser jest zielony"));

	log_config_free(&copy);
	assert(!copy.begin);
	assert(!copy.last);

	assert(!log_config_remove(&copy, "abc"));
	assert(!copy.begin);
	assert(!copy.last);

	log_config_set(&copy, "abc", "123");
	assert(copy.begin);
	assert(copy.begin == copy.last);

	fail_calloc = true;
	log_config_set(&copy, "fail", "lol");
	assert(copy.begin);
	assert(copy.begin == copy.last);
	fail_calloc = false;

	log_config_set(&copy, "def", "456");
	assert(copy.last);
	assert(copy.begin);
	assert(copy.begin != copy.last);

	assert(log_config_remove(&copy, "def"));
	assert(copy.begin);
	assert(copy.begin == copy.last);

	assert(log_config_remove(&copy, "abc"));
	assert(!copy.begin);
	assert(!copy.last);

	return 0;
}
