// C
#include <assert.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// POSIX
#include <limits.h>
#include <pthread.h>
#include <unistd.h>
#include <sys/inotify.h>

// DLog
#include <dynamic_config.h>

pthread_rwlock_t log_limiter_lock = PTHREAD_RWLOCK_INITIALIZER;
bool limiter;

static const int INOTIFY_FD = 23;
static const int INOTIFY_WD = 76;

static bool critical_failure_detected = false;
static char *expected_string;
static int expected_errno;
void __wrap_syslog_critical_failure(const char *message, ...)
{
	va_list va;
	va_start(va, message);

	/* In theory, each error message can have a unique set of values,
	 * but at the moment the only cases are either (errno) or (%s, errno). */
	if (expected_string) {
		assert(!strcmp(expected_string, va_arg(va, const char *)));
		free(expected_string);
		expected_string = NULL;
	}

	const int err = va_arg(va, int);
	va_end(va);

	assert(!expected_errno || (err == expected_errno && err == errno));
	critical_failure_detected = true;
}

static bool fail_inotify_init = false;
int __wrap_inotify_init1(int flags)
{
	if (fail_inotify_init) {
		expected_errno = errno = 14;
		return -1;
	}

	return INOTIFY_FD;
}

static bool fail_inotify_add_watch = false;
int __wrap_inotify_add_watch(int fd, const char *pathname, uint32_t mask)
{
	assert(fd == INOTIFY_FD);

	if (fail_inotify_add_watch) {
		expected_errno = errno = 42;
		return -1;
	}

	return INOTIFY_WD;
}

static bool fail_read;
static bool empty_read;
ssize_t __wrap_read(int fd, void *buf, size_t count)
{
	if (fail_read)
		return -1;

	if (empty_read)
		return 0;

	struct ievent_with_filename {
		struct inotify_event ie;
		char filename[NAME_MAX + 1];
	} *const iwf = (struct ievent_with_filename *)buf;

	assert(count >= sizeof *iwf);
	iwf->ie.wd = INOTIFY_WD;
	return sizeof *iwf;
}

static bool update_plog_called = false;
void __update_plog(const struct log_config *conf)
{
	update_plog_called = true;
}

static bool limiter_update_called = false;
void __log_limiter_update(struct log_config *config)
{
	limiter_update_called = true;
}

static int lcrf_retval;
int __wrap_log_config_read_file(struct log_config *config, const char *filename)
{
	return lcrf_retval;
}

static bool fail_asprintf;
int __wrap_asprintf(char **strp, const char *fmt, ...)
{
	if (fail_asprintf)
		return -1;

	va_list va;
	va_start(va, fmt);
	const int ret = vasprintf(strp, fmt, va);
	va_end(va);
	return ret;
}

static bool pretend_the_mutex_is_locked;
int __real_pthread_mutex_trylock(pthread_mutex_t *mutex);
int __wrap_pthread_mutex_trylock(pthread_mutex_t *mutex)
{
	return pretend_the_mutex_is_locked ? -1 : __real_pthread_mutex_trylock(mutex);
}

int main()
{
	// check whether these blow up when called before init
	__dynamic_config_update();
	__dynamic_config_destroy();
	__dynamic_config_update();

	struct log_config conf = {NULL, NULL};
	assert(!__dynamic_config_create(&conf));

	log_config_set(&conf, DYNAMIC_CONFIG_CONF_KEY, "relative/path");
	assert(!__dynamic_config_create(&conf));

	log_config_set(&conf, DYNAMIC_CONFIG_CONF_KEY, "/");

	fail_asprintf = true;
	assert(!__dynamic_config_create(&conf));
	assert(critical_failure_detected);
	critical_failure_detected = false;
	fail_asprintf = false;

	fail_inotify_init = true;
	assert(!__dynamic_config_create(&conf));
	assert(critical_failure_detected);
	critical_failure_detected = false;
	fail_inotify_init = false;

	fail_inotify_add_watch = true;
	assert(!__dynamic_config_create(&conf));
	assert(critical_failure_detected);
	critical_failure_detected = false;
	fail_inotify_add_watch = false;

	assert(__dynamic_config_create(&conf));
	assert(!critical_failure_detected);

	expected_errno = 0;
	lcrf_retval = -EINVAL;
	__dynamic_config_update();
	assert(critical_failure_detected);
	assert(!update_plog_called);
	assert(!limiter_update_called);

	limiter = false;
	lcrf_retval = 0;
	__dynamic_config_update();
	assert(update_plog_called);
	assert(!limiter_update_called);
	update_plog_called = false;

	limiter = true;
	lcrf_retval = -ENOENT;
	__dynamic_config_update();
	assert(update_plog_called);
	assert(limiter_update_called);
	update_plog_called = false;
	limiter_update_called = false;

	fail_read = true;
	__dynamic_config_update();
	assert(!update_plog_called);
	assert(!limiter_update_called);
	fail_read = false;

	empty_read = true;
	__dynamic_config_update();
	assert(!update_plog_called);
	assert(!limiter_update_called);
	empty_read = false;

	pretend_the_mutex_is_locked = true;
	__dynamic_config_update();
	assert(!update_plog_called);
	assert(!limiter_update_called);
	pretend_the_mutex_is_locked = false;

	__dynamic_config_destroy();
}
