#include <assert.h>
#include <logprint.h>
#include <stdio.h>
#include <stdlib.h>

enum {
	verbose = (1 << 0),
	debug   = (1 << 1),
	info    = (1 << 2),
	warn    = (1 << 3),
	error   = (1 << 4),
	fatal   = (1 << 5),
};

int pid_tid_filter_list[] = {
	1234,
	2345,
	34567,
	98273545,
	1,
};

/**
 * @brief Checks if a log message with a tid is printed as expected
 * @details Verify the log_should_print_line function returns the expected value
 * @param[in] line The number of source code line in this file, printed in debug messages
 * @param[in] p_filter The filter to work with
 * @param[in] tid The tid to filter with
 * @param[in] expected The expected result returned by log_should_print_line function
 * @return 0 if the log_should_print_line returns the expected result, 1 otherwise
 */
int check_print_line_tid_single(int line, log_filter *p_filter, pthread_t tid, bool expected)
{
	struct logger_entry_with_msg entry;

	entry.msg[0] = '\0'; /* empty tag */
	entry.msg[1] = '\0'; /* empty msg */
	entry.header.tag_len = 0;
	entry.header.pid = 0;
	entry.header.tid = tid;
	entry.header.priority = DLOG_DEBUG; // priority, for unmatched entries that have to face the global filter
	entry.header.len = sizeof(struct logger_entry) + 2;

	if (expected != log_should_print_line(p_filter, &entry.header)) {
		printf("Check in line %d for tid %lu failed to return %d \n", line, tid,
			expected);
		return 1;
	}
	return 0;
}

/**
 * @brief Checks if a log message if filtered properly for a group of tids
 * @details Verify the log_should_print_line function returns the expected values
 * @param[in] line The number of line in test, printed in debug messages
 * @param[in] p_filter The filter to work with
 * @param[in] tid The tid to filter with
 * @return Number of tids for which log_should_print_line returns wrong answer
 */

int check_print_line_tid(int line, log_filter *p_filter, pthread_t tid)
{
	int result = 0;
	for (int tid_no = 0; tid_no < NELEMS(pid_tid_filter_list); tid_no++)
		result += check_print_line_tid_single(line, p_filter, pid_tid_filter_list[tid_no], (pid_tid_filter_list[tid_no] == tid));
	return result;
}

/**
 * @brief Checks if a log message with a pid is printed as expected
 * @details Verify the log_should_print_line function returns the expected value
 * @param[in] line The number of source code line in this file, printed in debug messages
 * @param[in] p_filter The filter to work with
 * @param[in] pid The pid to filter with
 * @param[in] expected The expected result returned by log_should_print_line function
 * @return 0 if the log_should_print_line returns the expected result, 1 otherwise
 */
int check_print_line_pid_single(int line, log_filter *p_filter, pid_t pid, bool expected)
{
	struct logger_entry_with_msg entry;

	entry.msg[0] = '\0'; /* empty tag */
	entry.msg[1] = '\0'; /* empty msg */
	entry.header.tag_len = 0;
	entry.header.tid = 0;
	entry.header.pid = pid;
	entry.header.priority = DLOG_DEBUG; // priority, for unmatched entries that have to face the global filter
	entry.header.len = sizeof(struct logger_entry) + 2;

	if (expected != log_should_print_line(p_filter, &entry.header)) {
		printf("Check in line %d for pid %d failed to return %d \n", line, pid,
			expected);
		return 1;
	}
	return 0;
}

/**
 * @brief Checks if a log message if filtered properly for a group of pids
 * @details Verify the log_should_print_line function returns the expected values
 * @param[in] line The number of line in test, printed in debug messages
 * @param[in] p_filter The filter to work with
 * @param[in] pid The pid to filter with
 * @return Number of pids for which log_should_print_line returns wrong answer
 */

int check_print_line_pid(int line, log_filter *p_filter, pid_t pid)
{
	int result = 0;
	for (int pid_no = 0; pid_no < NELEMS(pid_tid_filter_list); pid_no++)
		result += check_print_line_pid_single(line, p_filter, pid_tid_filter_list[pid_no], (pid_tid_filter_list[pid_no] == pid));
	return result;
}

/**
 * @brief Checks if a log message with a tag and priority is printed as expected
 * @details Verify the log_should_print_line function returns the expected value
 * @param[in] line The number of source code line in this file, printed in debug messages
 * @param[in] p_filter The filter to work with
 * @param[in] tag The message's tag
 * @param[in] priority The message's priority
 * @param[in] expected The expected result returned by log_should_print_line function
 * @return 0 if the log_should_print_line returns the expected result, 1 otherwise
 */
int check_print_line_tag_prior(int line, log_filter *p_filter, char *tag, int priority, bool expected)
{
	struct logger_entry_with_msg entry;

	entry.header.tid = 0;
	entry.header.pid = 0;
	entry.header.priority = priority;
	strncpy(entry.msg, tag, strlen(tag) + 1);
	entry.msg[strlen(tag) + 1] = '\0'; // second NULL terminator: empty message
	entry.header.len = sizeof(struct logger_entry) + strlen(tag) + 2;

	if (expected != log_should_print_line(p_filter, &entry.header)) {
		printf("Check in line %d for tag %s and priority %c failed to return %d \n", line, tag,
			filter_pri_to_char(priority), expected);
		return 1;
	}
	return 0;
}

/**
 * @brief Checks if a log messages with a tag is printed ONLY for specified priorities
 * @details Verify the log_should_print_line function returns the expected values
 * @param[in] line The number of line in test, printed in debug messages
 * @param[in] p_filter The filter to work with
 * @param[in] tag The message's tag
 * @param[in] priority The message's priorities which should be printed
 * @return Number of priorities for which log_should_print_line returns wrong answer
 */

int check_print_line_tag(int line, log_filter *p_filter, char *tag, int priors)
{
	int result = 0;
	result += check_print_line_tag_prior(line, p_filter, tag, DLOG_VERBOSE, (priors & verbose) > 0);
	result += check_print_line_tag_prior(line, p_filter, tag, DLOG_DEBUG, (priors & debug) > 0);
	result += check_print_line_tag_prior(line, p_filter, tag, DLOG_INFO, (priors & info) > 0);
	result += check_print_line_tag_prior(line, p_filter, tag, DLOG_WARN, (priors & warn) > 0);
	result += check_print_line_tag_prior(line, p_filter, tag, DLOG_ERROR, (priors & error) > 0);
	result += check_print_line_tag_prior(line, p_filter, tag, DLOG_FATAL, (priors & fatal) > 0);
	return result;
}

int main()
{
	int err;
	int result = 0;
	char *tag, *tag_short, *tag_middle, *tag_long;
	log_filter *p_filter;
	struct logger_entry_with_msg entry;

	entry.header.tid = 0;
	entry.header.pid = 0;

	p_filter = log_filter_new();

	tag = "random";
	tag_short = "ab";
	tag_middle = "abc";
	tag_long = "abcde";


	// Old tests - rewritten and kept for backward compatibility (before introduction '*' and '='  in filters

	log_add_filter_rule(p_filter, "*:i"); //set global to info
	result += check_print_line_tag(__LINE__, p_filter, tag, info | warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, tag_short, info | warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, tag_middle, info | warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, tag_long, info | warn | error | fatal);

	log_add_filter_rule(p_filter, "*"); //set global to debug
	result += check_print_line_tag(__LINE__, p_filter, tag, debug | info | warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, tag_short, debug | info | warn | error | fatal);

	log_add_filter_rule(p_filter, "*:v"); //set global to verbose
	result += check_print_line_tag(__LINE__, p_filter, tag, verbose | debug | info | warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, tag_short, verbose | debug | info | warn | error | fatal);

	log_add_filter_rule(p_filter, "*:*"); //debug is default global level
	result += check_print_line_tag(__LINE__, p_filter, tag, debug | info | warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, tag_short, debug | info | warn | error | fatal);

	log_add_filter_rule(p_filter, "*:i");
	result += check_print_line_tag(__LINE__, p_filter, tag, info | warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, tag_short, info | warn | error | fatal);

	log_add_filter_rule(p_filter, "random");
	result += check_print_line_tag(__LINE__, p_filter, tag, verbose | debug | info | warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, tag_short, info | warn | error | fatal);

	log_add_filter_rule(p_filter, "random:v");
	result += check_print_line_tag(__LINE__, p_filter, tag, verbose | debug | info | warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, tag_short, info | warn | error | fatal);


	log_add_filter_rule(p_filter, "random:d");
	result += check_print_line_tag(__LINE__, p_filter, tag, verbose | debug | info | warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, tag_short, info | warn | error | fatal);

	log_add_filter_rule(p_filter, "random:w");
	result += check_print_line_tag(__LINE__, p_filter, tag, verbose | debug | info | warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, tag_short, info | warn | error | fatal);


	log_add_filter_rule(p_filter, "crap:*");

	static const size_t taglen = sizeof "crap" - 1;

	entry.header.priority = DLOG_VERBOSE;
	strncpy(entry.msg, "crap", taglen);
	entry.msg[taglen] = '\0'; /* tag delimiter */
	entry.msg[taglen + 1] = '\0'; /* empty msg */
	entry.header.len = sizeof(struct logger_entry) + taglen + 2;

	assert(log_should_print_line(p_filter, &entry.header));
	result += check_print_line_tag(__LINE__, p_filter, "crap", verbose | debug | info | warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, tag, verbose | debug | info | warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, tag_short, info | warn | error | fatal);


	//invalid expression
	err = log_add_filter_rule(p_filter, "random:z");
	assert(err < 0);
	result += check_print_line_tag(__LINE__, p_filter, tag, verbose | debug | info | warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, tag_short, info | warn | error | fatal);

	// Adding empty filter string is possible and does not erase existing
	log_add_filter_rule(p_filter, "issue550946:w");
	err = log_add_filter_string(p_filter, " ");
	assert(err == 0);
	result += check_print_line_tag(__LINE__, p_filter, "issue550946", warn | error | fatal);

	// Adding a filter with wrong priority fails also when the tag was used before
	err = log_add_filter_string(p_filter, "*:s existing:d ");
	assert(err == 0);
	result += check_print_line_tag(__LINE__, p_filter, "existing", debug | info | warn | error | fatal);
	err = log_add_filter_string(p_filter, "*:s existing:z");
	assert(err < 0);

	log_filter_free(p_filter);

	// Tests for '=' in filters

	p_filter = log_filter_new();

	log_add_filter_rule(p_filter, "pine:=i");
	result += check_print_line_tag(__LINE__, p_filter, "pine", info);
	result += check_print_line_tag(__LINE__, p_filter, tag, 0); //default policy for tag (slient) shall be used

	log_add_filter_rule(p_filter, "pine:=e");
	result += check_print_line_tag(__LINE__, p_filter, "pine", info | error);
	result += check_print_line_tag(__LINE__, p_filter, tag, 0);

	log_add_filter_rule(p_filter, "pine:i");
	result += check_print_line_tag(__LINE__, p_filter, "pine", info | warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, tag, 0);

	log_add_filter_rule(p_filter, "oak:=i");
	err = log_add_filter_string(p_filter,  "alder:=v alder:=f");
	assert(err == 0);
	result += check_print_line_tag(__LINE__, p_filter, "pine", info | warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, "alder", verbose | fatal);
	result += check_print_line_tag(__LINE__, p_filter, tag, 0);

	log_add_filter_rule(p_filter, "*:=v");
	result += check_print_line_tag(__LINE__, p_filter, "pine", info | warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, "alder", verbose | fatal);
	result += check_print_line_tag(__LINE__, p_filter, tag, verbose);

	log_filter_free(p_filter);

	// Tests for '*' in the filters

	p_filter = log_filter_new();

	log_add_filter_rule(p_filter, "a:=v");
	log_add_filter_rule(p_filter, "a*:w");
	result += check_print_line_tag(__LINE__, p_filter, "a", verbose | warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, "aa", warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, "aaa", warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, "aaab", warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, "aaaaaaaaba",  warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, "b", 0);
	result += check_print_line_tag(__LINE__, p_filter, "ba", 0);

	log_add_filter_rule(p_filter, "c?:w");
	result += check_print_line_tag(__LINE__, p_filter, "c", 0);
	result += check_print_line_tag(__LINE__, p_filter, "cx", 0);
	result += check_print_line_tag(__LINE__, p_filter, "c?", warn | error | fatal);

	log_add_filter_rule(p_filter, "zwxaaa*:e");
	result += check_print_line_tag(__LINE__, p_filter, "z", 0);
	result += check_print_line_tag(__LINE__, p_filter, "zx?*", 0);
	result += check_print_line_tag(__LINE__, p_filter, "zwxaa", 0);
	result += check_print_line_tag(__LINE__, p_filter, "zwxaab", 0);
	result += check_print_line_tag(__LINE__, p_filter, "zwxaa?", 0);
	result += check_print_line_tag(__LINE__, p_filter, "zwxaa*", 0);
	result += check_print_line_tag(__LINE__, p_filter, "zwxaa?*", 0);
	result += check_print_line_tag(__LINE__, p_filter, "zwxaaa", error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, "zwxaaaa", error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, "zwxaaab", error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, "zwxaaa*", error | fatal);

	log_filter_free(p_filter);

	// Tests for '*' and '=' in filters

	p_filter = log_filter_new();
	err = log_add_filter_string(p_filter, "*:=v TORUN:=f TOR*:=d T*:=i torun:s *:i");
	assert(err == 0);
	result += check_print_line_tag(__LINE__, p_filter, "something", info | warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, "t", info | warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, "T", info);
	result += check_print_line_tag(__LINE__, p_filter, "tor", info | warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, "TOR", debug | info);
	result += check_print_line_tag(__LINE__, p_filter, "TORUN", debug | info | fatal);
	result += check_print_line_tag(__LINE__, p_filter, "something", info | warn | error | fatal);
	result += check_print_line_tag(__LINE__, p_filter, "torun", 0);
	log_filter_free(p_filter);

	//Only *

	p_filter = log_filter_new();
	log_add_filter_rule(p_filter, "*");
	result += check_print_line_tag(__LINE__, p_filter, tag, debug | info | warn | error | fatal);
	log_filter_free(p_filter);

	//No rules - no print lines

	p_filter = log_filter_new();
	result += check_print_line_tag(__LINE__, p_filter, tag, 0);
	err = log_add_filter_string(p_filter, "");
	assert(err == 0);
	result += check_print_line_tag(__LINE__, p_filter, tag, 0);
	log_filter_free(p_filter);

	//pid filtering rules
	p_filter = log_filter_new();
	result += check_print_line_pid(__LINE__, p_filter, 0);
	err = log_add_filter_pid(p_filter, 9876);
	assert(err == 0);
	result += check_print_line_pid(__LINE__, p_filter, 9876);
	err = log_add_filter_pid(p_filter, pid_tid_filter_list[0]);
	assert(err == 0);
	result += check_print_line_pid(__LINE__, p_filter, pid_tid_filter_list[0]);
	log_filter_free(p_filter);

	//tid filtering rules
	p_filter = log_filter_new();
	result += check_print_line_tid(__LINE__, p_filter, 0);
	err = log_add_filter_tid(p_filter, 9876);
	assert(err == 0);
	result += check_print_line_tid(__LINE__, p_filter, 9876);
	err = log_add_filter_tid(p_filter, pid_tid_filter_list[0]);
	assert(err == 0);
	result += check_print_line_tid(__LINE__, p_filter, pid_tid_filter_list[0]);
	log_filter_free(p_filter);

	if (result > 0)
		exit(-1);
	return 0;
}
