// C
#include <assert.h>
#include <stdio.h>

// POSIX
#include <time.h>

// DLog
#include <dlog.h>
#include <loglimiter.h>
#include <loglimiter-internal.h>

static size_t get_rulecount()
{
	size_t cnt = 0;
	struct rule *r = NULL;
	do {
		char buf[128];
		assert(!__log_limiter_dump_rule(&r, buf, sizeof buf));
		++cnt;
	} while (r);
	return cnt;
}

static int fail_snprintf;
int __wrap_snprintf(char *str, size_t size, const char *format, ...)
{
	if (fail_snprintf && !--fail_snprintf)
		return -1;

	va_list va;
	va_start(va, format);
	const int ret = vsnprintf(str, size, format, va);
	va_end(va);
	return ret;
}

static int fail_malloc;
void *__real_malloc(size_t size);
void *__wrap_malloc(size_t size)
{
	return fail_malloc == 1 && size == sizeof(struct rule)
	    || fail_malloc == 2 && size != sizeof(struct rule)
		? NULL
		: __real_malloc(size)
	;
}

static bool fail_time;
static bool future_time;
time_t __real_time(time_t *t);
time_t __wrap_time(time_t *t)
{
	if (fail_time)
		return -1;

	return __real_time(t) + (future_time ? 40000 : 0);
}

int main()
{
	__log_limiter_destroy(); // check whether it explodes if called before init

	struct log_config conf = {NULL, NULL};
	assert(!__log_limiter_create(&conf));

	struct rule *r = NULL;
	char buffer[10];
	assert(!__log_limiter_dump_rule(&r, buffer, sizeof buffer));
	assert(r == NULL);

	log_config_set(&conf, "limiter|*|*"  , "allow");
	log_config_set(&conf, "limiter|FOO|*", "deny");
	log_config_set(&conf, "limiter|*|E"  , "deny");
	log_config_set(&conf, "limiter|FOO|E", "7");
	assert(__log_limiter_create(&conf));
	__log_limiter_update(&conf);

	for (int i = 0; i < 100; ++i) {
		assert(0 == __log_limiter_pass_log("FOO", 'F'));
		assert(1 == __log_limiter_pass_log("BAR", 'F'));
		assert(0 == __log_limiter_pass_log("BAR", 'E'));
	}

	fail_malloc = 1;
	__log_limiter_update(&conf);
	for (int i = 0; i < 100; ++i) {
		assert(1 == __log_limiter_pass_log("FOO", 'F'));
		assert(1 == __log_limiter_pass_log("BAR", 'F'));
		assert(1 == __log_limiter_pass_log("BAR", 'E'));
	}

	fail_malloc = 2;
	__log_limiter_update(&conf);
	for (int i = 0; i < 100; ++i) {
		assert(1 == __log_limiter_pass_log("FOO", 'F'));
		assert(1 == __log_limiter_pass_log("BAR", 'F'));
		assert(1 == __log_limiter_pass_log("BAR", 'E'));
	}

	fail_malloc = 0;
	__log_limiter_destroy();
	assert(__log_limiter_create(&conf));

	for (int i = 0; i < 7; ++i)
		assert(1 == __log_limiter_pass_log("FOO", 'E'));
	assert(-1 == __log_limiter_pass_log("FOO", 'E'));
	for (int i = 0; i < 2; ++i)
		assert(0 == __log_limiter_pass_log("FOO", 'E'));

	fail_time = true;
	assert(1 == __log_limiter_pass_log("FOO", 'E'));
	fail_time = false;
	assert(0 == __log_limiter_pass_log("FOO", 'E'));
	future_time = true;
	assert(1 == __log_limiter_pass_log("FOO", 'E'));
	future_time = false;
	assert(1 == __log_limiter_pass_log("FOO", 'E'));

	int rulecount = get_rulecount();
	assert(rulecount == 4);

	const struct limiter_limits limits = __log_limiter_get_limits("FOO", 'E');
	assert(limits.tag_and_prio == 7);
	assert(limits.tag          == 0);
	assert(limits.prio         == 0);
	assert(limits.global       == __LOG_LIMITER_LIMIT_MAX + 1);

	const struct limiter_limits lim_lowercase = __log_limiter_get_limits("FOO", 'e');
	const struct limiter_limits lim_enum      = __log_limiter_get_limits("FOO", DLOG_ERROR);
	const struct limiter_limits lim_numerical = __log_limiter_get_limits("FOO", '6');
	assert(!memcmp(&limits, &lim_lowercase, sizeof limits));
	assert(!memcmp(&limits, &lim_enum     , sizeof limits));
	assert(!memcmp(&limits, &lim_numerical, sizeof limits));

	log_config_set(&conf, "limiter|FOO|E", "12");
	__log_limiter_update(&conf);

	const struct limiter_limits limits_updated = __log_limiter_get_limits("FOO", 'E');
	assert(limits_updated.tag_and_prio == 12);
	assert(get_rulecount() == rulecount);

	log_config_set(&conf, "irrelevant", "entry");
	log_config_set(&conf, "limiter", "1");
	log_config_set(&conf, "limiter||E", "2");
	log_config_set(&conf, "limiter|QUUX|", "3");
	log_config_set(&conf, "limiter|", "4");
	log_config_set(&conf, "limiter||", "5");
	__log_limiter_update(&conf);
	assert(get_rulecount() == rulecount);

	log_config_set(&conf, "limiter|ABC|?", "13");
	__log_limiter_update(&conf);

	const struct limiter_limits lim_unknown = __log_limiter_get_limits("ABC", DLOG_UNKNOWN);
	assert(lim_unknown.tag_and_prio == 13);

	// bulk tests to test hashing and buckets
#define S(T) \
	log_config_set(&conf, "limiter|" #T "|D", "73"); \
	log_config_set(&conf, "limiter|" #T "|E", "74"); \
	log_config_set(&conf, "limiter|" #T "|F", "75")

	S(ABC);
	S(BCA);
	S(CAB);
	S(CBA);
	S(BAC);
	S(ACB);
#undef S
	__log_limiter_update(&conf);

#define A(T) \
	assert(__log_limiter_get_limits(#T, 'D').tag_and_prio == 73); \
	assert(__log_limiter_get_limits(#T, 'E').tag_and_prio == 74); \
	assert(__log_limiter_get_limits(#T, 'F').tag_and_prio == 75)

	A(ABC);
	A(BCA);
	A(CAB);
	A(CBA);
	A(BAC);
	A(ACB);

#undef A

	for (int i = 1; i <= 2; ++i) {
		struct rule *r = NULL;
		char buf[128];
		fail_snprintf = i;
		assert(__log_limiter_dump_rule(&r, buf, sizeof buf) == -1);
	}
	fail_snprintf = 0;

	struct rule r1 = {
		.prio = 'E',
		.hash = util_hash_key("*", '*'),
		.tag = "QWE",
		.limit = 0,
	}, r2 = {
		.prev = &r1,
		.prio = 'E',
		.hash = util_hash_key("*", '*'),
		.tag = "QWE",
		.limit = 0,
	};
	__log_limiter_destroy();
	__log_limiter_initialize(&r1);

	assert(1 == __log_limiter_pass_log("FOO", 'E'));

	/* Searching a hashmap kinda doesn't work if you
	 * fake hashes (esp. identical ones) so this just
	 * checks whether stuff explodes on creation. */
	r1.hash = r2.hash = 42;
	struct rule r3 = {
		.prev = &r2,
		.prio = 'E',
		.hash = 42,
		.tag = "ASD",
		.limit = 0,
	}, r4 = {
		.prev = &r3,
		.prio = 'F',
		.hash = 42,
		.tag = "ZXC",
		.limit = 0,
	};

	__log_limiter_destroy();
	__log_limiter_initialize(&r3);

	struct rule r5 = {
		.prev = &r4,
		.prio = 'E',
		.hash = 69,
		.tag = "ZAQ",
		.limit = 0,
	}, r6 = {
		.prev = &r5,
		.prio = 'D',
		.hash = 69,
		.tag = "WSX",
		.limit = 0,
	};

	__log_limiter_destroy();
	__log_limiter_initialize(&r5);

	// empty tag never gets blocked
	struct rule block_all = {
		.prio = '*',
		.hash = util_hash_key("*", '*'),
		.tag = "*",
		.limit = 0,
	};
	__log_limiter_destroy();
	__log_limiter_initialize(&block_all);
	assert(__log_limiter_pass_log("tag", 'W') == 0);
	assert(__log_limiter_pass_log("", 'W') == 1);

	log_config_free(&conf);
	__log_limiter_destroy();
}
