// C
#include <assert.h>

// DLog
#include <log_file.h>

static bool fail_strdup;
static char *strdup_ret;
char *__real_strdup(const char *s);
char *__wrap_strdup(const char *s)
{
	if (!fail_strdup)
		return __real_strdup(s);

	errno = EINPROGRESS;
	return strdup_ret;
}

static void *expect_free;
void __real_free(void *ptr);
void __wrap_free(void *ptr)
{
	if (!expect_free)
		__real_free(ptr);
	else
		assert(ptr == expect_free);
}

static size_t rename_calls;
int __wrap_rename(const char *oldpath, const char *newpath)
{
	++rename_calls;
	errno = EDEADLOCK;
	return -1;
}

void __wrap_extract_timestamp(const struct logger_entry *le, enum sorting_order stamp_type, struct timespec *ts)
{
	assert(stamp_type == SORT_SENT_MONO);
	assert(le == (struct logger_entry *) 0xBA5EBALL);
	ts->tv_sec = 3333;
	ts->tv_nsec = 9999;
}

char *__wrap_log_buffer_get_tag(const struct logger_entry *entry)
{
	assert(entry == (struct logger_entry *) 0xBA5EBALL);
	return "";
}

static bool custom_memcpy;
void *__real_memcpy(void *dest, const void *src, size_t n);
void *__wrap_memcpy(void *dest, const void *src, size_t n)
{
	if (!custom_memcpy)
		return __real_memcpy(dest, src, n);

	assert(src == (void *) 0xBA5EBALL);
	assert(n == sizeof(struct logger_entry));
	return dest;
}

static bool fail_snprintf;
int __wrap_snprintf(char *str, size_t size, const char *format, ...)
{
	return -fail_snprintf;
}

static int log_print_log_line_ret;
int __wrap_log_print_log_line(struct log_format p_format, int fd, const struct logger_entry *entry)
{
	assert(fd == 0xFD);
	return log_print_log_line_ret;
}

static bool fake_open;
static const char *open_pathname;
static int open_errno;
static size_t open_calls;
int __real_open(const char *pathname, int flags, mode_t mode);
int __wrap_open(const char *pathname, int flags, mode_t mode)
{
	if (!fake_open)
		return __real_open(pathname, flags, mode);

	assert(pathname == open_pathname);
	++open_calls;

	if (!open_errno)
		return 0xFD;
	errno = open_errno;
	return -1;
}

static bool fail_fstat;
int __wrap_fstat(int fd, struct stat *buf)
{
	if (fail_fstat)
		return -1;
	buf->st_size = 7654321;
	return 0;
}

int main()
{
	struct log_file lf;
	logfile_init(&lf);
	assert(lf.fd == -1);
	assert(!lf.path);

	logfile_set_fd(&lf, -10, 0);
	assert(lf.fd == -10);
	assert(!lf.should_close);

	logfile_set_fd(&lf, 15, 3);
	assert(lf.fd == 15);
	assert(lf.should_close);

	fail_strdup = true;
	assert(-EINPROGRESS == logfile_set_path(&lf, "x"));
	assert(!lf.path);

	expect_free = lf.path = (char *) 0xDEAD50UL;
	strdup_ret = (char *) 0xF007BALL;
	assert(!logfile_set_path(&lf, "y"));
	assert(lf.path == strdup_ret);
	expect_free = strdup_ret;

	fail_strdup = false;
	assert(!logfile_set_path(&lf, "pathetic"));
	assert(!strcmp(lf.path, "pathetic"));
	expect_free = NULL;

	open_pathname = lf.path;
	fake_open = true;
	open_errno = 456;
	assert(-456 == logfile_open(&lf));

	open_errno = 0;
	fail_fstat = true;
	lf.size = 9999;
	open_calls = 0;
	assert(!logfile_open(&lf));
	assert(lf.fd == 0xFD);
	assert(lf.size == 9999);
	assert(open_calls == 1);
	fail_fstat = false;

	open_calls = 0;
	lf.rotate_size_kbytes = 0;
	assert(!logfile_open(&lf));
	assert(lf.fd == 0xFD);
	assert(lf.size == 7654321);
	assert(open_calls == 1);

	lf.rotate_size_kbytes = 10;
	open_calls = 0;
	lf.max_rotated = 3;
	assert(1 == logfile_open(&lf));
	assert(rename_calls == 3);
	assert(open_calls == 2);

	custom_memcpy = true;
	lf.prev_sec = 4444;
	lf.prev_nsec = 6666;
	log_print_log_line_ret = -1;
	assert(logfile_write_with_rotation((struct logger_entry *) 0xBA5EBALL, &lf, SORT_SENT_MONO));
	assert(lf.prev_sec == 4444);
	assert(lf.prev_nsec == 6666);

	log_print_log_line_ret = 1;
	fail_snprintf = true;
	assert(!logfile_write_with_rotation((struct logger_entry *) 0xBA5EBALL, &lf, SORT_SENT_MONO));
	assert(lf.prev_sec == 3333);
	assert(lf.prev_nsec == 9999);

	fail_snprintf = false;
	assert(!logfile_write_with_rotation((struct logger_entry *) 0xBA5EBALL, &lf, SORT_SENT_MONO));
	custom_memcpy = false;

	logfile_free(&lf);

	// some extra work is done (by lcov?) after main finishes
	fake_open = false;
	expect_free = NULL;
	open_pathname = NULL;
}
