#include <log_file.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <grp.h>
#include <pwd.h>
#include <sd-daemon.h>
#include "../logger/logger_internal.h"
#include "../logger/log_storage.h"

int parse_permissions(const char *str);
int usergr2id(const char *user, const char *group, uid_t *uid, gid_t *gid);
int reset_self_privileges();
int bind_fd_create(const char *path, int type);
int listen_fd_create(const char *path, int permissions);
int systemd_sock_get(const char *path, int type);
int dev_log_sock_get();
int dev_log_sock_create();
void init_fd_entity(struct fd_entity *fd_entity, dispatch_event_t dispatch_event, void *user_data);
void set_read_fd_entity(struct fd_entity *fd_entity, int fd);
void set_write_fd_entity(struct fd_entity *fd_entity, int fd);
int writer_create(struct writer **writer, int fd, struct log_buffer *log_buffer, service_writer_t service_writer, service_socket_t service_socket);
void dispatch_event_writer(struct logger *server, struct epoll_event *event);
int create_kmsg_writer(struct writer **writer, struct log_buffer *log_buffer);
int service_writer_kmsg(struct logger *server, struct writer *wr, struct epoll_event *event);
int create_syslog_writer(struct writer **writer, struct log_buffer *log_buffer);
int service_writer_syslog(struct logger *server, struct writer *wr, struct epoll_event *event);
int service_socket_dummy(struct logger *server, struct writer *wr, struct dlog_control_msg *msg);
int socket_initialize(struct sock_data *sock, struct log_buffer *buffer, service_socket_t service_socket, struct socket_config_data *data);
void dispatch_event_sock(struct logger *server, struct epoll_event *event);
void socket_close(struct sock_data *sock);
uint64_t reader_buffered_space(const struct reader *reader);
int reader_is_bufferable(const struct reader *reader);
int reader_ms_since(const struct reader *reader, struct timespec *ts);
int reader_should_buffer(struct reader *reader, const struct buf_params *buf_params, struct timespec now);
void reader_free(struct reader* reader);
int buffer_append(const struct logger_entry *entry, struct log_buffer *b);
int reader_init_for_pipe(struct reader *reader, const char *buf_name, struct logger *server);
int reader_init_for_logger(struct reader *reader, const char *buf_name, struct logger *server);
int print_out_logs(struct reader *reader);
int service_reader_file(struct reader* reader);
int add_fd_entity(struct logger *logger, struct fd_entity *fd_entity);
int modify_fd_entity(const struct logger *logger, struct fd_entity *fd_entity, int wanted_mask);
int remove_fd_entity(struct logger *logger, struct fd_entity *fd_entity);
void writer_close_fd(struct logger* server, struct writer* wr);
int add_buffer_reader(struct log_buffer *buffer, struct reader *reader);
void reader_notify_losing_log(const struct logger_entry *le, void *reader_);

bool fail_dlog_group_server = false;
int __wrap_getgrnam_r(const char *name, struct group *grp,
                 char *buf, size_t buflen, struct group **result)
{
	*result = NULL;

	gid_t gid = -1;

	if (!strcmp(name, "root"))
		gid = 0;

	if (!strcmp(name, "mmajewski2"))
		gid = 1000;

	if (!strcmp(name, DLOG_SERVER_GROUP) && !fail_dlog_group_server)
		gid = 7;

	if (!strcmp(name, "lvoldemort")) {
		errno = EIO; // TODO: Are we sure the function does this?
		return EIO;
	}

	if (gid == -1) {
		return 0;
	}

	// Note: we don't fill in the entire struct for simplicity.
	// This means we can completely ignore buf{,len}
	grp->gr_gid = gid;
	*result = grp;

	return 0;
}
int __wrap_getpwnam_r(const char *name, struct passwd *pwd,
                      char *buf, size_t buflen, struct passwd **result)
{
	*result = NULL;

	uid_t uid = -1;

	if (!strcmp(name, "root"))
		uid = 0;

	if (!strcmp(name, "mmajewski2"))
		uid = 1001;

	if (!strcmp(name, DLOG_SERVER_USER) && !fail_dlog_group_server)
		uid = 8;

	if (!strcmp(name, "lvoldemort")) {
		errno = EIO; // TODO: Are we sure the function does this?
		return EIO;
	}

	if (uid == -1) {
		return 0;
	}

	pwd->pw_uid = uid;
	*result = pwd;

	return 0;
}

gid_t getegid_return = 0;
gid_t __wrap_getegid(void)
{
	return getegid_return;
}

uid_t geteuid_return = 0;
uid_t __wrap_geteuid(void)
{
	return geteuid_return;
}

gid_t setgid_correct = 0;
int setgid_errno = 0;
bool setgid_called = false;
int __wrap_setgid(gid_t gid)
{
	assert(gid == setgid_correct);

	if (setgid_errno != 0) {
		errno = setgid_errno;
		return -1;
	}
	setgid_called = true;
	return 0;
}

uid_t setuid_correct = 0;
int setuid_errno = 0;
bool setuid_called = false;
int __wrap_setuid(uid_t uid)
{
	assert(uid == setuid_correct);

	if (setuid_errno != 0) {
		errno = setuid_errno;
		return -1;
	}
	setuid_called = true;
	return 0;
}

int socket_type_correct = 0;
int socket_errno = 0;
int socket_last_sock = 0;
int __wrap_socket(int domain, int type, int protocol)
{
	assert(domain == AF_UNIX);
	assert(type == socket_type_correct);
	assert(protocol == 0);

	if (socket_errno != 0) {
		errno = socket_errno;
		return -1;
	}
	return ++socket_last_sock;
}

bool unlink_called_correct = false;
bool unlink_called_dev_link = false;
int __wrap_unlink(const char *pathname)
{
	if (!strcmp("/sbin/init", pathname))
		unlink_called_correct = true;
	else if (!strcmp("/dev/log", pathname))
		unlink_called_dev_link = true;
	else if (!strcmp("/run/dlog/dev-log", pathname))
		; // TODO: Maybe we should check for that?
	else
		assert(false);
	return 0; // None of the tested functions actually care what is returned here
}

int bind_errno = 0;
bool bind_called_correct = false;
bool bind_called_dev = false;
int __wrap_bind(int sockfd, const struct sockaddr *addr, socklen_t addrlen)
{
	assert(sockfd == socket_last_sock);
	assert(addrlen == sizeof(struct sockaddr_un));
	assert(addr);
	const struct sockaddr_un *addr_un = (const struct sockaddr_un *)addr;
	assert(addr_un->sun_family == AF_UNIX);

	bool *called;
	if (!strcmp("/sbin/init", addr_un->sun_path))
		called = &bind_called_correct;
	else if (!strcmp("/run/dlog/dev-log", addr_un->sun_path))
		called = &bind_called_dev;
	else
		assert(false);

	if (bind_errno != 0) {
		errno = bind_errno;
		return -1;
	}
	*called = true;
	return 0;
}

#define MAX_FD 1000
bool closed[MAX_FD];
int __wrap_close(int fd)
{
	if (fd <= 0 || fd > MAX_FD || closed[fd - 1])
		return -1;
	closed[fd - 1] = true;
	return 0;
}

mode_t chmod_correct_mode = 0;
int chmod_errno = 0;
bool chmod_called = false;
int __wrap_chmod(const char *pathname, mode_t mode)
{
	assert(pathname);
	assert(!strcmp(pathname, "/sbin/init"));
	assert(mode == chmod_correct_mode);

	if (chmod_errno != 0) {
		errno = chmod_errno;
		return -1;
	}
	chmod_called = true;
	return 0;
}

int listen_correct_sockfd = 0;
int listen_errno = 0;
bool listen_called = false;
int __wrap_listen(int sockfd, int backlog)
{
	assert(sockfd == listen_correct_sockfd);

	if (listen_errno != 0) {
		errno = listen_errno;
		return -1;
	}
	listen_called = true;
	return 0;
}

long __wrap_sysconf(int name)
{
	return -1;
}

bool sd_fail = false;
int __wrap_sd_listen_fds(int unset_environment)
{
	return sd_fail ? -EIO : 2;
}
int __wrap_sd_is_socket_unix(int fd, int type, int listening, const char *path, size_t length)
{
	assert(listening < 0);
	assert(length == 0);
	switch (fd) {
	case 3:
		return type == SOCK_STREAM && !strcmp(path, "/boot/vmlinuz");
	case 4:
		return type == SOCK_DGRAM && !strcmp(path, "/run/dlog/dev-log");
	default:
		assert(false);
	}
}

int symlink_errno = 0;
bool symlink_called = false;
int __wrap_symlink(const char *target, const char *linkpath)
{
	assert(!strcmp(target, "/run/dlog/dev-log"));
	assert(!strcmp(linkpath, "/dev/log"));

	if (symlink_errno != 0) {
		errno = symlink_errno;
		return -1;
	}
	symlink_called = true;
	return 0;
}

bool calloc_fail = false;
void *__real_calloc(size_t nmemb, size_t size);
void *__wrap_calloc(size_t nmemb, size_t size)
{
	return calloc_fail ? NULL : __real_calloc(nmemb, size);
}

int open_errno = 0;
int __real_open(const char *pathname, int flags, mode_t mode);
int __wrap_open(const char *pathname, int flags, mode_t mode)
{
	// This is needed, because one day, we might want to open
	// /home/abuild/rpmbuild/BUILD/dlog-5.5.1/src/shared/src_tests_logger-log_file.gcda...
	if (strcmp(pathname, "/dev/kmsg"))
		return __real_open(pathname, flags, mode);

	assert(flags == (O_RDONLY | O_NONBLOCK));

	if (open_errno != 0) {
		errno = open_errno;
		return -1;
	}
	return 37;
}

int fcntl_correct = 0;
int fcntl_errno = 0;
int __real_fcntl(int fd, int cmd, ...);
int __wrap_fcntl(int fd, int cmd, ...)
{
	// We allow fcntl_correct == 0 for the same reason as in __wrap_open
	if (fcntl_correct == 0 && cmd == F_SETLKW) {
		va_list args;
		va_start(args, cmd);
		struct flock *flock = va_arg(args, struct flock *);
		va_end(args);
		return __real_fcntl(fd, cmd, flock);
	}

	assert(fd == fcntl_correct);

	if (fcntl_errno != 0) {
		errno = fcntl_errno;
		return -1;
	}

	switch (cmd) {
	case F_GETFL:
		return O_APPEND | O_NOATIME;
	case F_SETFL: {
		va_list args;
		va_start(args, cmd);
		assert(va_arg(args, int) == (O_APPEND | O_NOATIME | O_NONBLOCK));
		va_end(args);
		return 0;
	}
	default:
		assert(false);
	}
}

uint64_t __wrap_log_storage_reader_get_ready_bytes(const log_storage_reader *reader)
{
	return (intptr_t)reader; // We can't cast to uint64_t directly -- it fails on 32-bit
}

void *last_free = NULL;
void __real_free(void *ptr);
void __wrap_free(void *ptr)
{
	last_free = ptr;
	__real_free(ptr);
}

int last_logfile_free_fd = 0;
void __wrap_logfile_free(struct log_file *l_file)
{
	last_logfile_free_fd = l_file->fd;
}

bool add_new_entry_result = true;
bool __wrap_log_storage_add_new_entry(log_storage *storage, const struct logger_entry *le)
{
	assert(storage == (log_storage *)570);
	assert(le == (struct logger_entry *)0xE);
	return add_new_entry_result;
}

bool reader_released = false;
void __wrap_log_storage_release_reader(log_storage_reader *reader)
{
	assert(reader == (log_storage_reader *)13);
	reader_released = true;
}

int epoll_ctl_epfd_correct = 0;
int epoll_ctl_op_correct = 0;
int epoll_ctl_fd_correct = 0;
uint32_t epoll_ctl_events_correct = 0;
void *epoll_ctl_data_correct = NULL;
int epoll_ctl_errno = 0;
int __wrap_epoll_ctl(int epfd, int op, int fd, struct epoll_event *event)
{
	assert(epfd == epoll_ctl_epfd_correct);
	assert(op == epoll_ctl_op_correct);
	assert(fd == epoll_ctl_fd_correct);
	if (op == EPOLL_CTL_DEL)
		assert(!event);
	else {
		assert(event->events == epoll_ctl_events_correct);
		assert(event->data.ptr == epoll_ctl_data_correct);
	}

	if (epoll_ctl_errno != 0) {
		errno = epoll_ctl_errno;
		return -1;
	}
	return 0;
}

struct log_storage *new_reader_storage_correct = NULL;
struct reader *new_reader_reader_correct = NULL;
bool new_reader_fail = false;
log_storage_reader *__wrap_log_storage_new_reader(log_storage *storage,
		bool dumping, log_storage_reader_callback callback, void *user_data)
{
	assert(storage == new_reader_storage_correct);
	assert(user_data == new_reader_reader_correct);
	assert(dumping == new_reader_reader_correct->dumpcount);
	assert(callback == reader_notify_losing_log);

	return new_reader_fail ? NULL : (log_storage_reader *)0x600D;
}

int main()
{
	assert(parse_permissions(NULL) == (S_IWUSR | S_IWGRP | S_IWOTH));
	assert(parse_permissions("All the permissions") == 0);
	assert(parse_permissions("6789") == 0);
	assert(parse_permissions("0001o") == 0);

	assert(parse_permissions("0000") == 0);
	assert(parse_permissions("0001") == S_IXOTH);
	assert(parse_permissions("0002") == S_IWOTH);
	assert(parse_permissions("0004") == S_IROTH);
	assert(parse_permissions("0010") == S_IXGRP);
	assert(parse_permissions("0020") == S_IWGRP);
	assert(parse_permissions("0040") == S_IRGRP);
	assert(parse_permissions("0100") == S_IXUSR);
	assert(parse_permissions("0200") == S_IWUSR);
	assert(parse_permissions("0400") == S_IRUSR);
	assert(parse_permissions("1000") == S_ISVTX);
	assert(parse_permissions("2000") == S_ISGID);
	assert(parse_permissions("4000") == S_ISUID);

	assert(parse_permissions("1234") == (S_ISVTX | S_IWUSR | S_IXGRP | S_IWGRP | S_IROTH));
	assert(parse_permissions("1") == S_IXOTH);
	assert(parse_permissions("70001") == S_IXOTH);

	uid_t uid = -1;
	gid_t gid = -1;

	assert(usergr2id("root", "root", &uid, &gid) == 0);
	assert(uid == 0);
	assert(gid == 0);

	assert(usergr2id("mmajewski2", "mmajewski2", &uid, &gid) == 0);
	assert(uid == 1001);
	assert(gid == 1000);

	assert(usergr2id("mmajewski2", "root", &uid, &gid) == 0);
	assert(uid == 0);
	assert(gid == 1000);

	assert(usergr2id("root", "mmajewski2", &uid, &gid) == 0);
	assert(uid == 1001);
	assert(gid == 0);

	assert(usergr2id(DLOG_SERVER_USER, DLOG_SERVER_GROUP, &uid, &gid) == 0);
	assert(uid == 8);
	assert(gid == 7);

	assert(usergr2id("mmajewski2", "lvoldemort", &uid, &gid) == -EIO);
	assert(usergr2id("lvoldemort", "mmajewski2", &uid, &gid) == -EIO);

	assert(usergr2id("root", "unexistent", &uid, &gid) == -ENOENT);
	assert(usergr2id("unexistent", "root", &uid, &gid) == -ENOENT);

	fail_dlog_group_server = true;
	assert(reset_self_privileges() == -ENOENT);
	fail_dlog_group_server = false;

	getegid_return = 7;
	geteuid_return = 8;
	assert(reset_self_privileges() == 0);

	getegid_return = geteuid_return = 0;
	setgid_correct = 7;
	setgid_errno = EINVAL;
	assert(reset_self_privileges() == -EINVAL);

	setgid_errno = 0;
	setgid_called = false;
	setuid_correct = 8;
	setuid_errno = EINVAL;
	assert(reset_self_privileges() == -EINVAL);
	assert(setgid_called);

	setuid_errno = 0;
	setgid_called = setuid_called = false;
	assert(reset_self_privileges() == 0);
	assert(setgid_called);
	assert(setuid_called);

	socket_type_correct = SOCK_STREAM;
	socket_errno = ENFILE;
	assert(bind_fd_create("/sbin/init", SOCK_STREAM) == -ENFILE);

	socket_type_correct = SOCK_DGRAM;
	assert(bind_fd_create("/sbin/init", SOCK_DGRAM) == -ENFILE);

	socket_errno = 0;
	bind_errno = EINVAL;
	unlink_called_correct = false;
	assert(bind_fd_create("/sbin/init", SOCK_DGRAM) == -EINVAL);
	assert(unlink_called_correct);
	assert(closed[0]);

	bind_errno = 0;
	unlink_called_correct = bind_called_correct = false;
	assert(bind_fd_create("/sbin/init", SOCK_DGRAM) == 2);
	assert(unlink_called_correct);
	assert(bind_called_correct);
	assert(!closed[1]);

	socket_type_correct = SOCK_STREAM;
	socket_errno = EACCES;
	assert(listen_fd_create("/sbin/init", 0) == -EACCES);

	socket_errno = 0;
	chmod_correct_mode = S_IWOTH;
	chmod_errno = EIO;
	assert(listen_fd_create("/sbin/init", chmod_correct_mode) == -EIO);
	assert(closed[2]);

	listen_correct_sockfd = 4;
	listen_errno = EADDRINUSE;
	assert(listen_fd_create("/sbin/init", 0) == -EADDRINUSE);
	assert(closed[3]);

	listen_errno = 0;
	listen_correct_sockfd = 5;
	listen_called = false;
	assert(listen_fd_create("/sbin/init", 0) == 5);
	assert(listen_called);
	assert(!closed[4]);

	listen_correct_sockfd = 6;
	listen_called = chmod_called = false;
	chmod_errno = 0;
	assert(listen_fd_create("/sbin/init", chmod_correct_mode) == 6);
	assert(listen_called);
	assert(chmod_called);
	assert(!closed[5]);

	sd_fail = true;
	assert(systemd_sock_get("/root", SOCK_DGRAM) == -1);
	assert(systemd_sock_get("/boot/vmlinuz", SOCK_STREAM) == -1);
	assert(systemd_sock_get("/run/dlog/dev-log", SOCK_DGRAM) == -1);
	assert(dev_log_sock_get() == -1);
	sd_fail = false;

	assert(systemd_sock_get("/root", SOCK_DGRAM) == -1);
	assert(systemd_sock_get("/boot/vmlinuz", SOCK_DGRAM) == -1);
	assert(systemd_sock_get("/boot/vmlinuz", SOCK_STREAM) == 3);
	assert(systemd_sock_get("/run/dlog/dev-log", SOCK_DGRAM) == 4);
	assert(systemd_sock_get("/run/dlog/dev-log", SOCK_STREAM) == -1);
	assert(dev_log_sock_get() == 4);

	socket_type_correct = SOCK_DGRAM;
	bind_errno = EIO;
	unlink_called_dev_link = false;
	assert(dev_log_sock_create() == -EIO);
	assert(!unlink_called_dev_link);
	assert(closed[6]);

	bind_errno = 0;
	symlink_errno = ENOSPC;
	unlink_called_dev_link = false;
	assert(dev_log_sock_create() == -ENOSPC);
	assert(unlink_called_dev_link);
	assert(closed[7]);

	symlink_errno = 0;
	unlink_called_dev_link = false;
	assert(dev_log_sock_create() == 9);
	assert(unlink_called_dev_link);
	assert(!closed[8]);

	struct fd_entity ent;
	init_fd_entity(&ent, (dispatch_event_t)0xD159A7C8, (void *)0xDA7A);
	assert(ent.dispatch_event == (dispatch_event_t)0xD159A7C8);
	assert(ent.event.data.ptr == (void *)0xDA7A);
	assert(ent.event.events == 0);
	assert(ent.fd == -1);

	set_read_fd_entity(&ent, 1);
	assert(ent.event.events == EPOLLIN);
	assert(ent.fd == 1);

	set_write_fd_entity(&ent, 2);
	assert(ent.event.events == EPOLLHUP);
	assert(ent.fd == 2);

	struct logger logger = {
		.epollfd = 102,
		.epollcnt = 0,
	};

	epoll_ctl_epfd_correct = 102;
	epoll_ctl_op_correct = EPOLL_CTL_ADD;
	epoll_ctl_fd_correct = 2;
	epoll_ctl_events_correct = EPOLLHUP;
	epoll_ctl_data_correct = (void *)0xDA7A;
	epoll_ctl_errno = ELOOP;
	assert(add_fd_entity(&logger, &ent) == -ELOOP);
	assert(logger.epollcnt == 0);

	epoll_ctl_errno = 0;
	assert(add_fd_entity(&logger, &ent) == 0);
	assert(logger.epollcnt == 1);

	// TODO: I have no idea why this is checked here, but let's roll with it.
	ent.fd = -1;
	assert(modify_fd_entity(&logger, &ent, EPOLLMSG) == -EINVAL);
	assert(ent.event.events == EPOLLHUP);

	ent.fd = 2;
	epoll_ctl_op_correct = EPOLL_CTL_MOD;
	epoll_ctl_events_correct = EPOLLMSG;
	epoll_ctl_errno = ENOENT;
	assert(modify_fd_entity(&logger, &ent, EPOLLMSG) == -ENOENT);
	assert(ent.event.events == EPOLLHUP);

	epoll_ctl_errno = 0;
	assert(modify_fd_entity(&logger, &ent, EPOLLMSG) == 0);
	assert(ent.event.events == EPOLLMSG);

	epoll_ctl_errno = EBADF;
	assert(modify_fd_entity(&logger, &ent, EPOLLMSG) == 0);
	assert(ent.event.events == EPOLLMSG);

	epoll_ctl_op_correct = EPOLL_CTL_DEL;
	epoll_ctl_errno = EPERM;
	assert(remove_fd_entity(&logger, &ent) == -EPERM);
	assert(logger.epollcnt == 1);

	epoll_ctl_errno = 0;
	assert(remove_fd_entity(&logger, &ent) == 0);
	assert(logger.epollcnt == 0);

	struct writer *writer;
	calloc_fail = true;
	assert(writer_create(&writer, 3, (struct log_buffer *)0xBU, (service_writer_t)17, (service_socket_t)18) == -ENOMEM);

	calloc_fail = false;
	closed[2] = false;
	assert(writer_create(&writer, 3, (struct log_buffer *)0xBU, (service_writer_t)17, (service_socket_t)18) == 0);
	assert(writer->fd_entity.dispatch_event == dispatch_event_writer);
	assert(writer->fd_entity.event.data.ptr == writer);
	assert(writer->fd_entity.event.events == EPOLLIN);
	assert(writer->fd_entity.fd == 3);
	assert(writer->buf_ptr == (struct log_buffer *)0xBU);
	assert(writer->readed == 0);
	assert(writer->service_writer == (service_writer_t)17);
	assert(writer->service_socket == (service_socket_t)18);
	assert(!closed[2]);

	logger.epollcnt = 1;
	epoll_ctl_op_correct = EPOLL_CTL_DEL;
	epoll_ctl_fd_correct = 3;
	writer_close_fd(&logger, writer);
	assert(closed[2]);
	assert(logger.epollcnt == 0);
	// TODO: Check if epoll_ctl called

	open_errno = ENOENT;
	assert(create_kmsg_writer(&writer, (struct log_buffer *)0x2BU) == -ENOENT);

	open_errno = 0;
	calloc_fail = true;
	assert(create_kmsg_writer(&writer, (struct log_buffer *)0x2BU) == -ENOMEM);
	assert(closed[36]);

	calloc_fail = false;
	closed[36] = false;
	assert(create_kmsg_writer(&writer, (struct log_buffer *)0x2BU) == 0);
	assert(writer->fd_entity.dispatch_event == dispatch_event_writer);
	assert(writer->fd_entity.event.data.ptr == writer);
	assert(writer->fd_entity.event.events == EPOLLIN);
	assert(writer->fd_entity.fd == 37);
	assert(writer->buf_ptr == (struct log_buffer *)0x2BU);
	assert(writer->readed == 0);
	assert(writer->service_writer == service_writer_kmsg);
	assert(writer->service_socket == service_socket_dummy);
	assert(!closed[36]);
	free(writer);

	sd_fail = true;
	bind_errno = EFAULT;
	assert(create_syslog_writer(&writer, (struct log_buffer *)0x3BU) == -EFAULT);

	bind_errno = 0;
	fcntl_correct = 11;
	fcntl_errno = EIO;
	closed[10] = false;
	assert(create_syslog_writer(&writer, (struct log_buffer *)0x3BU) == -EIO);
	assert(closed[10]);

	sd_fail = false;
	bind_errno = EFAULT;
	fcntl_correct = 4;
	closed[3] = false;
	assert(create_syslog_writer(&writer, (struct log_buffer *)0x3BU) == -EIO);
	assert(closed[3]);

	fcntl_errno = 0;
	closed[3] = false;
	calloc_fail = true;
	assert(create_syslog_writer(&writer, (struct log_buffer *)0x3BU) == -ENOMEM);
	assert(closed[3]);

	calloc_fail = false;
	closed[3] = false;
	assert(create_syslog_writer(&writer, (struct log_buffer *)0x3BU) == 0);
	assert(writer->fd_entity.dispatch_event == dispatch_event_writer);
	assert(writer->fd_entity.event.data.ptr == writer);
	assert(writer->fd_entity.event.events == EPOLLIN);
	assert(writer->fd_entity.fd == 4);
	assert(writer->buf_ptr == (struct log_buffer *)0x3BU);
	assert(writer->readed == 0);
	assert(writer->service_writer == service_writer_syslog);
	assert(writer->service_socket == service_socket_dummy);
	assert(!closed[3]);
	free(writer);
	bind_errno = 0;

	struct sock_data sock;
	struct socket_config_data data;
	memcpy(data.path, "/sbin/init", sizeof "/sbin/init");
	data.permissions = S_IWUSR;
	socket_type_correct = SOCK_STREAM;
	socket_errno = ENOBUFS;
	assert(socket_initialize(&sock, (struct log_buffer *)0x4BU, (service_socket_t)0x5E, &data) == -ENOBUFS);

	socket_errno = 0;
	listen_correct_sockfd = 12;
	chmod_correct_mode = S_IWUSR;
	closed[11] = false;
	assert(socket_initialize(&sock, (struct log_buffer *)0x4BU, (service_socket_t)0x5E, &data) == 0);
	assert(!closed[11]);
	assert(sock.fd_entity.dispatch_event == dispatch_event_sock);
	assert(sock.fd_entity.event.data.ptr == &sock);
	assert(sock.fd_entity.event.events == EPOLLIN);
	assert(sock.fd_entity.fd == 12);
	assert(sock.buf_ptr == (struct log_buffer *)0x4BU);
	assert(sock.service_socket == (service_socket_t)0x5E);

	socket_close(&sock);
	assert(closed[11]);

	sock.fd_entity.fd = 0;
	socket_close(&sock);

	struct reader reader = {
		.log_storage_reader = (log_storage_reader *)7,
		.file = {},
		.last_read_time = {
			.tv_sec = 0,
			.tv_nsec = 100000000,
		},
	};

	assert(!reader_is_bufferable(&reader));

	reader.buf_ptr = (struct log_buffer *)0x5BU;
	assert(!reader_is_bufferable(&reader));

	reader.buf_ptr = NULL;
	reader.file.path = "/dev/null";
	assert(!reader_is_bufferable(&reader));

	reader.buf_ptr = (struct log_buffer *)0x5BU;
	assert(reader_is_bufferable(&reader));

	assert(reader_buffered_space(&reader) == 7);

	assert(reader_ms_since(&reader, &(struct timespec) {
		.tv_sec = 1,
		.tv_nsec = 0,
	}) == 900);

	struct buf_params params = {
		.bytes = 100,
		.time = 0,
	};

	reader.buf_ptr = NULL;
	assert(!reader_should_buffer(&reader, &params, (struct timespec) {
		.tv_sec = 1,
		.tv_nsec = 0,
	}));
	assert(reader.last_read_time.tv_sec == 0 && reader.last_read_time.tv_nsec == 100000000);

	reader.buf_ptr = (struct log_buffer *)0x5BU;
	assert(!reader_should_buffer(&reader, &params, (struct timespec) {
		.tv_sec = 1,
		.tv_nsec = 0,
	}));
	assert(reader.last_read_time.tv_sec == 1 && reader.last_read_time.tv_nsec == 0);

	params.time = 1;
	params.bytes = 0;
	assert(!reader_should_buffer(&reader, &params, (struct timespec) {
		.tv_sec = 1,
		.tv_nsec = 500000000,
	}));
	assert(reader.last_read_time.tv_sec == 1 && reader.last_read_time.tv_nsec == 500000000);

	params.bytes = 100;
	assert(reader_should_buffer(&reader, &params, (struct timespec) {
		.tv_sec = 2,
		.tv_nsec = 0,
	}));
	assert(reader.last_read_time.tv_sec == 1 && reader.last_read_time.tv_nsec == 500000000);

	reader_free(NULL);

	struct reader *reader2 = calloc(1, sizeof(struct reader));
	assert(reader2);
	reader2->file.fd = 1;
	reader_free(reader2);
	assert(last_free == reader2 && last_logfile_free_fd == 1);

	reader2 = calloc(1, sizeof(struct reader));
	assert(reader2);
	reader2->file.fd = 3;
	reader2->fd_entity = ent;
	assert(reader2->fd_entity.fd == 2);
	reader2->read_fd = 1;
	reader2->log_storage_reader = (log_storage_reader *)13;
	closed[0] = closed[1] = false;
	reader_free(reader2);
	assert(last_free == reader2 && last_logfile_free_fd == 3);
	assert(closed[0] && closed[1]);
	assert(reader_released);

	struct log_buffer adder = {
		.log_storage = (log_storage *)570,
	};

	add_new_entry_result = false;
	assert(buffer_append((struct logger_entry *)0xE, &adder) == -ENOMEM);

	add_new_entry_result = true;
	assert(buffer_append((struct logger_entry *)0xE, &adder) == 0);

	struct reader reader3 = {};
	struct logger server = {
		.buffers = {
			[LOG_ID_KMSG] = (struct log_buffer *)7,
			[LOG_ID_SYSLOG] = (struct log_buffer *)17,
		},
	};
	assert(reader_init_for_pipe(&reader3, "invalid", &server) == -EINVAL);
	assert(reader3.service_reader == print_out_logs);

	reader3.service_reader = NULL;
	assert(reader_init_for_pipe(&reader3, "main", &server) == -EINVAL);
	assert(reader3.service_reader == print_out_logs);
	assert(reader3.buf_id == LOG_ID_MAIN);

	reader3.service_reader = NULL;
	assert(reader_init_for_pipe(&reader3, "kmsg", &server) == 0);
	assert(reader3.service_reader == print_out_logs);
	assert(reader3.buf_id == LOG_ID_KMSG);
	assert(reader3.buf_ptr == (struct log_buffer *)7);

	reader3.service_reader = NULL;
	assert(reader_init_for_pipe(&reader3, "syslog", &server) == 0);
	assert(reader3.service_reader == print_out_logs);
	assert(reader3.buf_id == LOG_ID_SYSLOG);
	assert(reader3.buf_ptr == (struct log_buffer *)17);

	reader3.service_reader = NULL;
	reader3.buf_id = 0;
	reader3.buf_ptr = NULL;
	assert(reader_init_for_logger(&reader3, "kmsg", &server) == 0);
	assert(reader3.service_reader == print_out_logs);
	assert(reader3.buf_id == LOG_ID_KMSG);
	assert(reader3.buf_ptr == (struct log_buffer *)7);

	reader3.service_reader = NULL;
	assert(reader_init_for_logger(&reader3, "syslog", &server) == 0);
	assert(reader3.service_reader == print_out_logs);
	assert(reader3.buf_id == LOG_ID_SYSLOG);
	assert(reader3.buf_ptr == (struct log_buffer *)17);

	assert(reader_init_for_logger(&reader3, "invalid", &server) == -EINVAL);
	assert(reader3.service_reader == service_reader_file);

	struct log_buffer buffer = {
		.log_storage = (struct log_storage *)570,
	};
	struct reader reader4;

	new_reader_storage_correct = buffer.log_storage;
	new_reader_reader_correct = &reader4;
	new_reader_fail = true;
	assert(add_buffer_reader(&buffer, &reader4) == -ENOMEM);

	new_reader_fail = false;
	assert(add_buffer_reader(&buffer, &reader4) == 0);
	assert(reader4.log_storage_reader == (struct log_storage_reader *)0x600D);
	assert(list_at(buffer.readers) == &reader4);

	fcntl_correct = 0;
}
