// C
#include <assert.h>
#include <time.h>

// DLog
#include <logprint.h>
#include <ptrs_list.h>

static char buffer[10240];
static size_t buffer_count = 0;
static bool fake_write;
static int partial_write;
ssize_t __real_write(int fd, const void *buf, size_t count);
ssize_t __wrap_write(int fd, const void *buf, size_t count)
{
	if (partial_write)
		return partial_write;

	if (!fake_write)
		return __real_write(fd, buf, count);

	assert(count < sizeof buffer);
	memcpy(buffer, buf, count);
	buffer[count] = '\0';
	buffer_count = count;

	return count;
}

static int fail_localtime;
struct tm *__real_localtime(const time_t *timep);
struct tm *__wrap_localtime(const time_t *timep)
{
	struct tm *const ret = (fail_localtime & 1) ? NULL : __real_localtime(timep);
	fail_localtime >>= 1;
	return ret;
}

static int fail_alloc;
void *__real_malloc(size_t size);
void *__real_calloc(size_t nmemb, size_t size);
void *__wrap_malloc(size_t size)
{
	if (fail_alloc) {
		assert(size > 512);
		return NULL;
	}

	return __real_malloc(size);
}
void *__wrap_calloc(size_t nmemb, size_t size)
{
	return fail_alloc ? NULL : __real_calloc(nmemb, size);
}

static bool fail_strndup;
char *__real_strndup(const char *s, size_t n);
char *__wrap_strndup(const char *s, size_t n)
{
	return fail_strndup ? NULL : __real_strndup(s, n);
}

static bool fail_strdup;
char *__real_strdup(const char *s);
char *__wrap_strdup(const char *s)
{
	return fail_strdup ? NULL : __real_strdup(s);
}

static bool fail_list_add;
bool __real_list_add(list_head *head, elem_value value);
bool __wrap_list_add(list_head *head, elem_value value)
{
	return fail_list_add ? false : __real_list_add(head, value);
}

int isatty_ret = 0;
int __wrap_isatty(int fd)
{
	return isatty_ret;
}

/* Timestamps are shown in the everyday format so
 * it's good to be able to work with those numbers
 * where possible instead of the underlying single
 * integer (but see below). */

#define SECONDS 1 // the base unit of the sec timestamp
#define MINUTES (60 * SECONDS)
#define HOURS   (60 * MINUTES)
#define DAYS    (24 * HOURS)

#define JANUARY  (-1 * DAYS) // dates start from Jan 01, not Jan 00
#define FEBRUARY (JANUARY  + 31 * DAYS)
#define MARCH    (FEBRUARY + 28 * DAYS) // 1970, so no 29th

/* The nano timestamp is a bit more of a monolith,
 * but can still be meaningfully split up into the
 * milliseconds part and the remainder since all
 * timestamps seem to only show the former. */

#define  NANOSECONDS 1
#define MICROSECONDS (1000 * NANOSECONDS)
#define MILLISECONDS (1000 * MICROSECONDS)

static void standard_test_msg(struct logger_entry_with_msg *entry, const char *msg, size_t msg_len)
{
	*entry = (struct logger_entry_with_msg) {
		.header = {
			.len            = sizeof entry->header + msg_len,
			.priority       = DLOG_INFO,
			.pid            = 4228,
			.tid            = 4567,
			.sec_sent_mono  = 3205048,
			.sec_sent_real  =    MARCH + 14*DAYS + 23*HOURS + 22*MINUTES + 51*SECONDS,
			.sec_recv_mono  = 365467,
			.sec_recv_real  = FEBRUARY +  1*DAYS +  0*HOURS +  5*MINUTES + 17*SECONDS,
			.nsec_sent_mono = 235*MILLISECONDS + 123*MICROSECONDS + 213*NANOSECONDS,
			.nsec_sent_real = 451*MILLISECONDS + 435*MICROSECONDS + 135*NANOSECONDS,
			.nsec_recv_mono = 346*MILLISECONDS + 253*MICROSECONDS + 662*NANOSECONDS,
			.nsec_recv_real = 734*MILLISECONDS + 543*MICROSECONDS + 254*NANOSECONDS,
			.tag_len        = -1,
		},
	};

	memcpy(entry->msg, msg, msg_len); // strcpy won't do because msg has an embedded \0 as the tag delimiter
	for (size_t i = 0; i < msg_len; ++i)
		if (!msg[i]) {
			entry->header.tag_len = i;
			break;
		}
	assert(entry->header.tag_len >= 0);

	/* Despite the comment about the single int above,
	 * sometimes we still have to work with it. Make
	 * sure we know the other representation in each
	 * case so that we can easily identify failures. */
	assert(entry->header.sec_sent_mono == FEBRUARY +  7*DAYS +  2*HOURS + 17*MINUTES + 28*SECONDS);
	assert(entry->header.sec_sent_real == 6304971);
	assert(entry->header.sec_recv_mono ==  JANUARY +  5*DAYS +  5*HOURS + 31*MINUTES +  7*SECONDS);
	assert(entry->header.sec_recv_real == 2678717);
}

struct logprint_test_case {
	log_print_format format;
	log_priority priority;
	bool color_enabled;
	const char *result;
};

void check_logprint_testcases(const char *msg, size_t msg_len, const struct logprint_test_case *tests, size_t tests_n)
{
	struct logger_entry_with_msg entry;
	standard_test_msg(&entry, msg, msg_len);

	fake_write = true;

	for (size_t i = 0; i < tests_n; ++i) {
		entry.header.priority = tests[i].priority;
		assert(log_print_log_line((struct log_format){
			.format = tests[i].format,
			.color = tests[i].color_enabled,
		}, 0, false, &entry.header) > 0);
		assert(!strcmp(tests[i].result, buffer));
	}

	fake_write = false;
}
#define CHECK_LOGPRINT_TESTCASES(MSG, TESTS) check_logprint_testcases(MSG, sizeof MSG, TESTS, NELEMS(TESTS))

void check_log_print_log_line()
{
	static const struct logprint_test_case tests_basic[] = {
		// formats using the SENT timestamp
		{FORMAT_TIME         , DLOG_INFO, false, "03-14 23:22:51.451+0000 I/BASIC TAG( 4228): Basic message.\n"             },
		{FORMAT_LONG         , DLOG_INFO, false, "[ 03-14 23:22:51.451 I/BASIC TAG P 4228, T 4567]\nBasic message.\n\n"     },
		{FORMAT_THREADTIME   , DLOG_INFO, false, "03-14 23:22:51.451+0000 I/BASIC TAG(P 4228, T 4567): Basic message.\n"    },
		{FORMAT_KERNELTIME   , DLOG_INFO, false, "3205048.235 I/BASIC TAG(P 4228, T 4567): Basic message.\n"                },

		// formats using the RECV timestamp
		{FORMAT_RECV_REALTIME, DLOG_INFO, false, "02-01 00:05:17.734 I/BASIC TAG(P 4228, T 4567): Basic message.\n"         },
		{FORMAT_RWTIME       , DLOG_INFO, false, "02-01 00:05:17 [365467.346] I/BASIC TAG(P 4228, T 4567): Basic message.\n"},

		// formats without a timestamp
		{FORMAT_BRIEF        , DLOG_INFO, false, "I/BASIC TAG( 4228): Basic message.\n"                                     },
		{FORMAT_PROCESS      , DLOG_INFO, false, "I( 4228) Basic message.  (BASIC TAG)\n"                                   },
		{FORMAT_TAG          , DLOG_INFO, false, "I/BASIC TAG: Basic message.\n"                                            },
		{FORMAT_THREAD       , DLOG_INFO, false, "I(P 4228, T 4567) Basic message.\n"                                       },
		{FORMAT_RAW          , DLOG_INFO, false, "Basic message.\n"                                                         },
		// JSON
		{FORMAT_JSON         , DLOG_INFO, false, "{" \
			"\"priority\":\"info\"," \
			"\"pid\":4228," \
			"\"tid\":4567," \
			"\"recv_real\":\"1970-02-01T00:05:17+0000\"," \
			"\"recv_mono\":365467.346253662," \
			"\"sent_real\":\"1970-03-14T23:22:51+0000\"," \
			"\"sent_mono\":3205048.235123213," \
			"\"tag\":\"BASIC TAG\"," \
			"\"msg\":\"Basic message.\"" \
		"}\n"},
	};
	CHECK_LOGPRINT_TESTCASES("BASIC TAG\0Basic message.", tests_basic);

	// Color coded headers
	static const struct logprint_test_case tests_colors[] = {
		// Fatal
		{FORMAT_TIME         , DLOG_FATAL, true, "\033[35;1m03-14 23:22:51.451+0000 F/COLORFUL TAG( 4228): \033[0mBasic message.\n"             },
		{FORMAT_LONG         , DLOG_FATAL, true, "\033[35;1m[ 03-14 23:22:51.451 F/COLORFUL TAG P 4228, T 4567]\n\033[0mBasic message.\n\n"     },
		{FORMAT_THREADTIME   , DLOG_FATAL, true, "\033[35;1m03-14 23:22:51.451+0000 F/COLORFUL TAG(P 4228, T 4567): \033[0mBasic message.\n"    },
		{FORMAT_KERNELTIME   , DLOG_FATAL, true, "\033[35;1m3205048.235 F/COLORFUL TAG(P 4228, T 4567): \033[0mBasic message.\n"                },
		{FORMAT_RECV_REALTIME, DLOG_FATAL, true, "\033[35;1m02-01 00:05:17.734 F/COLORFUL TAG(P 4228, T 4567): \033[0mBasic message.\n"         },
		{FORMAT_RWTIME       , DLOG_FATAL, true, "\033[35;1m02-01 00:05:17 [365467.346] F/COLORFUL TAG(P 4228, T 4567): \033[0mBasic message.\n"},
		{FORMAT_BRIEF        , DLOG_FATAL, true, "\033[35;1mF/COLORFUL TAG( 4228): \033[0mBasic message.\n"                                     },
		{FORMAT_PROCESS      , DLOG_FATAL, true, "\033[35;1mF( 4228) \033[0mBasic message.\033[35;1m  (COLORFUL TAG)\n\033[0m"                  },
		{FORMAT_TAG          , DLOG_FATAL, true, "\033[35;1mF/COLORFUL TAG: \033[0mBasic message.\n"                                            },
		{FORMAT_THREAD       , DLOG_FATAL, true, "\033[35;1mF(P 4228, T 4567) \033[0mBasic message.\n"                                          },
		{FORMAT_RAW          , DLOG_FATAL, true, "Basic message.\n"                                                                             },

		// Error
		{FORMAT_TIME         , DLOG_ERROR, true, "\033[31;1m03-14 23:22:51.451+0000 E/COLORFUL TAG( 4228): \033[0mBasic message.\n"             },
		{FORMAT_LONG         , DLOG_ERROR, true, "\033[31;1m[ 03-14 23:22:51.451 E/COLORFUL TAG P 4228, T 4567]\n\033[0mBasic message.\n\n"     },
		{FORMAT_THREADTIME   , DLOG_ERROR, true, "\033[31;1m03-14 23:22:51.451+0000 E/COLORFUL TAG(P 4228, T 4567): \033[0mBasic message.\n"    },
		{FORMAT_KERNELTIME   , DLOG_ERROR, true, "\033[31;1m3205048.235 E/COLORFUL TAG(P 4228, T 4567): \033[0mBasic message.\n"                },
		{FORMAT_RECV_REALTIME, DLOG_ERROR, true, "\033[31;1m02-01 00:05:17.734 E/COLORFUL TAG(P 4228, T 4567): \033[0mBasic message.\n"         },
		{FORMAT_RWTIME       , DLOG_ERROR, true, "\033[31;1m02-01 00:05:17 [365467.346] E/COLORFUL TAG(P 4228, T 4567): \033[0mBasic message.\n"},
		{FORMAT_BRIEF        , DLOG_ERROR, true, "\033[31;1mE/COLORFUL TAG( 4228): \033[0mBasic message.\n"                                     },
		{FORMAT_PROCESS      , DLOG_ERROR, true, "\033[31;1mE( 4228) \033[0mBasic message.\033[31;1m  (COLORFUL TAG)\n\033[0m"                  },
		{FORMAT_TAG          , DLOG_ERROR, true, "\033[31;1mE/COLORFUL TAG: \033[0mBasic message.\n"                                            },
		{FORMAT_THREAD       , DLOG_ERROR, true, "\033[31;1mE(P 4228, T 4567) \033[0mBasic message.\n"                                          },
		{FORMAT_RAW          , DLOG_ERROR, true, "Basic message.\n"                                                                             },

		// Warning
		{FORMAT_TIME         , DLOG_WARN , true, "\033[33;1m03-14 23:22:51.451+0000 W/COLORFUL TAG( 4228): \033[0mBasic message.\n"             },
		{FORMAT_LONG         , DLOG_WARN , true, "\033[33;1m[ 03-14 23:22:51.451 W/COLORFUL TAG P 4228, T 4567]\n\033[0mBasic message.\n\n"     },
		{FORMAT_THREADTIME   , DLOG_WARN , true, "\033[33;1m03-14 23:22:51.451+0000 W/COLORFUL TAG(P 4228, T 4567): \033[0mBasic message.\n"    },
		{FORMAT_KERNELTIME   , DLOG_WARN , true, "\033[33;1m3205048.235 W/COLORFUL TAG(P 4228, T 4567): \033[0mBasic message.\n"                },
		{FORMAT_RECV_REALTIME, DLOG_WARN , true, "\033[33;1m02-01 00:05:17.734 W/COLORFUL TAG(P 4228, T 4567): \033[0mBasic message.\n"         },
		{FORMAT_RWTIME       , DLOG_WARN , true, "\033[33;1m02-01 00:05:17 [365467.346] W/COLORFUL TAG(P 4228, T 4567): \033[0mBasic message.\n"},
		{FORMAT_BRIEF        , DLOG_WARN , true, "\033[33;1mW/COLORFUL TAG( 4228): \033[0mBasic message.\n"                                     },
		{FORMAT_PROCESS      , DLOG_WARN , true, "\033[33;1mW( 4228) \033[0mBasic message.\033[33;1m  (COLORFUL TAG)\n\033[0m"                  },
		{FORMAT_TAG          , DLOG_WARN , true, "\033[33;1mW/COLORFUL TAG: \033[0mBasic message.\n"                                            },
		{FORMAT_THREAD       , DLOG_WARN , true, "\033[33;1mW(P 4228, T 4567) \033[0mBasic message.\n"                                          },
		{FORMAT_RAW          , DLOG_WARN , true, "Basic message.\n"                                                                             },

		// Information -- no color coding needed
		{FORMAT_TIME         , DLOG_INFO , true , "03-14 23:22:51.451+0000 I/COLORFUL TAG( 4228): Basic message.\n"             },
		{FORMAT_LONG         , DLOG_INFO , true , "[ 03-14 23:22:51.451 I/COLORFUL TAG P 4228, T 4567]\nBasic message.\n\n"     },
		{FORMAT_THREADTIME   , DLOG_INFO , true , "03-14 23:22:51.451+0000 I/COLORFUL TAG(P 4228, T 4567): Basic message.\n"    },
		{FORMAT_KERNELTIME   , DLOG_INFO , true , "3205048.235 I/COLORFUL TAG(P 4228, T 4567): Basic message.\n"                },
		{FORMAT_RECV_REALTIME, DLOG_INFO , true , "02-01 00:05:17.734 I/COLORFUL TAG(P 4228, T 4567): Basic message.\n"         },
		{FORMAT_RWTIME       , DLOG_INFO , true , "02-01 00:05:17 [365467.346] I/COLORFUL TAG(P 4228, T 4567): Basic message.\n"},
		{FORMAT_BRIEF        , DLOG_INFO , true , "I/COLORFUL TAG( 4228): Basic message.\n"                                     },
		{FORMAT_PROCESS      , DLOG_INFO , true , "I( 4228) Basic message.  (COLORFUL TAG)\n"                                   },
		{FORMAT_TAG          , DLOG_INFO , true , "I/COLORFUL TAG: Basic message.\n"                                            },
		{FORMAT_THREAD       , DLOG_INFO , true , "I(P 4228, T 4567) Basic message.\n"                                          },
		{FORMAT_RAW          , DLOG_INFO , true , "Basic message.\n"                                                            },

		// Color coding disabled
		{FORMAT_TIME         , DLOG_FATAL, false, "03-14 23:22:51.451+0000 F/COLORFUL TAG( 4228): Basic message.\n"             },
	};
	CHECK_LOGPRINT_TESTCASES("COLORFUL TAG\0Basic message.", tests_colors);

	// Empty tag has special handling and gets replaced.
	static const struct logprint_test_case tests_tagless[] = {
		// formats using the SENT timestamp
		{FORMAT_RECV_REALTIME, DLOG_INFO, false, "02-01 00:05:17.734 I/DLOG_ERROR_NOTAG(P 4228, T 4567): Tagless\n"},
		{FORMAT_BRIEF        , DLOG_INFO, false, "I/DLOG_ERROR_NOTAG( 4228): Tagless\n"                            },
		{FORMAT_PROCESS      , DLOG_INFO, false, "I( 4228) Tagless  (DLOG_ERROR_NOTAG)\n"                          },
	};
	CHECK_LOGPRINT_TESTCASES("\0Tagless", tests_tagless);

	/* Some formats insert the extra info (timestamp, tag etc) every time they
	 * find a line break. The tag itself is never broken up though. */
	static const struct logprint_test_case tests_linebreak[] = {
		{FORMAT_LONG, DLOG_INFO, false,
			"[ 03-14 23:22:51.451 I/THIS\nTAG\nTAKES\nMULTIPLE\nLINES P 4228, T 4567]\n"
			"And\nso\ndoes\nthis\nmessage.\n\n" },
		{FORMAT_KERNELTIME, DLOG_INFO, false,
			"3205048.235 I/THIS\nTAG\nTAKES\nMULTIPLE\nLINES(P 4228, T 4567): And\n"
			"3205048.235 I/THIS\nTAG\nTAKES\nMULTIPLE\nLINES(P 4228, T 4567): so\n"
			"3205048.235 I/THIS\nTAG\nTAKES\nMULTIPLE\nLINES(P 4228, T 4567): does\n"
			"3205048.235 I/THIS\nTAG\nTAKES\nMULTIPLE\nLINES(P 4228, T 4567): this\n"
			"3205048.235 I/THIS\nTAG\nTAKES\nMULTIPLE\nLINES(P 4228, T 4567): message.\n" },
		{FORMAT_RECV_REALTIME, DLOG_INFO, false,
			"02-01 00:05:17.734 I/THIS\nTAG\nTAKES\nMULTIPLE\nLINES(P 4228, T 4567): And\n"
			"02-01 00:05:17.734 I/THIS\nTAG\nTAKES\nMULTIPLE\nLINES(P 4228, T 4567): so\n"
			"02-01 00:05:17.734 I/THIS\nTAG\nTAKES\nMULTIPLE\nLINES(P 4228, T 4567): does\n"
			"02-01 00:05:17.734 I/THIS\nTAG\nTAKES\nMULTIPLE\nLINES(P 4228, T 4567): this\n"
			"02-01 00:05:17.734 I/THIS\nTAG\nTAKES\nMULTIPLE\nLINES(P 4228, T 4567): message.\n" },
		{FORMAT_PROCESS, DLOG_INFO, false,
			"I( 4228) And  (THIS\nTAG\nTAKES\nMULTIPLE\nLINES)\n"
			"I( 4228) so  (THIS\nTAG\nTAKES\nMULTIPLE\nLINES)\n"
			"I( 4228) does  (THIS\nTAG\nTAKES\nMULTIPLE\nLINES)\n"
			"I( 4228) this  (THIS\nTAG\nTAKES\nMULTIPLE\nLINES)\n"
			"I( 4228) message.  (THIS\nTAG\nTAKES\nMULTIPLE\nLINES)\n" },
	};
	CHECK_LOGPRINT_TESTCASES("THIS\nTAG\nTAKES\nMULTIPLE\nLINES\0And\nso\ndoes\nthis\nmessage.", tests_linebreak);

	/* The extras' length gets to be capped (regardless
	 * of which side of the message they are put on). */
	static const struct logprint_test_case tests_long_metadata[] = {
		// Uncolored
		{FORMAT_LONG, DLOG_INFO, false,
			"[ 03-14 23:22:51.451 I/THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS"
			"foo1\nfoo2\n\n" },
		{FORMAT_KERNELTIME, DLOG_INFO, false,
			"3205048.235 I/THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGTfoo1\n"
			"3205048.235 I/THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGTfoo2\n" },
		{FORMAT_RECV_REALTIME, DLOG_INFO, false,
			"02-01 00:05:17.734 I/THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS Ifoo1\n"
			"02-01 00:05:17.734 I/THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS Ifoo2\n" },
		{FORMAT_PROCESS, DLOG_INFO, false,
			"I( 4228) foo1  (THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TA"
			"I( 4228) foo2  (THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TA" },

		// Colored
		{FORMAT_LONG, DLOG_WARN, true,
			"\033[33;1m[ 03-14 23:22:51.451 W/THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS\033[0m"
			"foo1\nfoo2\n\n" },
		{FORMAT_KERNELTIME, DLOG_WARN, true,
			"\033[33;1m3205048.235 W/THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGT\033[0mfoo1\n"
			"\033[33;1m3205048.235 W/THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGT\033[0mfoo2\n" },
		{FORMAT_RECV_REALTIME, DLOG_WARN, true,
			"\033[33;1m02-01 00:05:17.734 W/THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS I\033[0mfoo1\n"
			"\033[33;1m02-01 00:05:17.734 W/THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS I\033[0mfoo2\n" },
		{FORMAT_PROCESS, DLOG_WARN, true,
			"\033[33;1mW( 4228) \033[0mfoo1\033[33;1m  (THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TA\033[0m"
			"\033[33;1mW( 4228) \033[0mfoo2\033[33;1m  (THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TA\033[0m" },
	};
	CHECK_LOGPRINT_TESTCASES("THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGTH ~ THIS TAG IS OVER 128 CHARS IN LENGTH\0foo1\nfoo2", tests_long_metadata);
}

char *json_escape_string(char *in);
void check_json_escape_string_testcase(char *in, char *correct)
{
	char *out = json_escape_string(in);
	assert(out);
	assert(!strcmp(out, correct));
	free(out);
}

enum ts_type {
	TS_SENT,
	TS_RECV,
};
const char *json_priority_name(log_priority prio);
struct timespec entry_get_ts(const struct logger_entry *entry, const log_print_format format, enum ts_type type);
char *log_format_json(char *default_buffer, size_t default_buffer_size, const struct logger_entry *entry, size_t *p_outLength, char *tag, char *msg);
void check_json()
{
	// TODO: Failure testing
	check_json_escape_string_testcase("abc/def", "abc/def");
	check_json_escape_string_testcase("\x01\x02\x03", "\\u0001\\u0002\\u0003");
	check_json_escape_string_testcase("\x1A\x1B\x1C", "\\u001A\\u001B\\u001C");
	check_json_escape_string_testcase("\x1D\x1E\x1F", "\\u001D\\u001E\\u001F");
	check_json_escape_string_testcase("\"\\\b\f\n\r\t", "\\\"\\\\\\b\\f\\n\\r\\t");
	check_json_escape_string_testcase("zażółć gęślą jaźń", "zażółć gęślą jaźń");
	check_json_escape_string_testcase("삼성", "삼성");
	check_json_escape_string_testcase("タイゼン", "タイゼン");
	check_json_escape_string_testcase("🏀", "🏀");

	assert(!strcmp(json_priority_name(DLOG_VERBOSE), "verbose"));
	assert(!strcmp(json_priority_name(DLOG_DEBUG),   "debug"));
	assert(!strcmp(json_priority_name(DLOG_INFO),    "info"));
	assert(!strcmp(json_priority_name(DLOG_WARN),    "warning"));
	assert(!strcmp(json_priority_name(DLOG_ERROR),   "error"));
	assert(!strcmp(json_priority_name(DLOG_FATAL),   "fatal"));
	assert(!strcmp(json_priority_name(DLOG_SILENT),  "silent"));
	assert(!strcmp(json_priority_name(DLOG_UNKNOWN), "unknown"));
	assert(!strcmp(json_priority_name(123),          "unknown"));

	char data[] = "tag\0important message";
	struct logger_entry *ent = calloc(1, sizeof(struct logger_entry) + sizeof data);
	assert(ent);
	ent->len = sizeof data;
	ent->tag_len = 3;
	strncpy(ent->msg, data, ent->len);
	ent->priority = DLOG_DEBUG;
	ent->pid = 17;
	ent->tid = 18;

	struct timespec ts;

#define TEST_EMPTY_TIMESTAMP() \
	ent->nsec_sent_mono = ent->nsec_recv_mono = ent->nsec_sent_real = ent->nsec_recv_real = -1; \
	ts = entry_get_ts(ent, CURRENT_FORMAT, CURRENT_TYPE); \
	assert(ts.tv_sec == 0); \
	assert(ts.tv_nsec == 0)
#define TEST_SINGLE_TIMESTAMP(which, to_set, expected) \
	ent->sec_##which = ent->nsec_##which = to_set; \
	ts = entry_get_ts(ent, CURRENT_FORMAT, CURRENT_TYPE); \
	assert(ts.tv_sec == expected); \
	assert(ts.tv_nsec == expected)

 #define CURRENT_FORMAT FORMAT_KERNELTIME
  #define CURRENT_TYPE TS_SENT
	TEST_EMPTY_TIMESTAMP();
	TEST_SINGLE_TIMESTAMP(recv_real, 1, 1);
	TEST_SINGLE_TIMESTAMP(sent_real, 2, 2);
	TEST_SINGLE_TIMESTAMP(recv_mono, 3, 3);
	TEST_SINGLE_TIMESTAMP(sent_mono, 4, 4);
  #undef CURRENT_TYPE

  #define CURRENT_TYPE TS_RECV
	TEST_EMPTY_TIMESTAMP();
	TEST_SINGLE_TIMESTAMP(sent_real, 1, 1);
	TEST_SINGLE_TIMESTAMP(recv_real, 2, 2);
	TEST_SINGLE_TIMESTAMP(sent_mono, 3, 3);
	TEST_SINGLE_TIMESTAMP(recv_mono, 4, 4);
  #undef CURRENT_TYPE
 #undef CURRENT_FORMAT

 #define CURRENT_FORMAT FORMAT_TIME
  #define CURRENT_TYPE TS_SENT
	TEST_EMPTY_TIMESTAMP();
	TEST_SINGLE_TIMESTAMP(recv_mono, 1, 0);
	TEST_SINGLE_TIMESTAMP(sent_mono, 2, 0);
	TEST_SINGLE_TIMESTAMP(recv_real, 3, 3);
	TEST_SINGLE_TIMESTAMP(sent_real, 4, 4);
  #undef CURRENT_TYPE

  #define CURRENT_TYPE TS_RECV
	TEST_EMPTY_TIMESTAMP();
	TEST_SINGLE_TIMESTAMP(sent_mono, 1, 0);
	TEST_SINGLE_TIMESTAMP(recv_mono, 2, 0);
	TEST_SINGLE_TIMESTAMP(sent_real, 3, 3);
	TEST_SINGLE_TIMESTAMP(recv_real, 4, 4);
  #undef CURRENT_TYPE
 #undef CURRENT_FORMAT

#undef TEST_EMPTY_TIMESTAMP
#undef TEST_SINGLE_TIMESTAMP

	char buf[512] = {};
	size_t out_len;

	ent->nsec_sent_mono = ent->nsec_recv_mono = ent->nsec_sent_real = ent->nsec_recv_real = -1;
	assert(log_format_json(buf, sizeof(buf), ent, &out_len, "tag", "message") == buf);
	assert(strlen(buf) == out_len);
	assert(!strcmp(buf, "{" \
		"\"priority\":\"debug\"," \
		"\"pid\":17," \
		"\"tid\":18," \
		"\"tag\":\"tag\"," \
		"\"msg\":\"message\"" \
	"}\n"));

	char *target;
	target = log_format_json(NULL, out_len, ent, &out_len, "tag", "message");
	assert(target);
	assert(target != buf);
	assert(strlen(target) == out_len);
	assert(!strcmp(buf, target));
	free(target);

	ent->sec_sent_mono = ent->nsec_sent_mono = 21;
	ent->sec_recv_mono = ent->nsec_recv_mono = 22;
	ent->sec_sent_real = ent->nsec_sent_real = 23242527;
	ent->sec_recv_real = ent->nsec_recv_real = 28293031;
	assert(log_format_json(buf, sizeof(buf), ent, &out_len, "tag", "message") == buf);
	assert(strlen(buf) == out_len);
	assert(!strcmp(buf, "{"\
		"\"priority\":\"debug\","\
		"\"pid\":17,"\
		"\"tid\":18,"\
		"\"recv_real\":\"1970-11-24T11:10:31+0000\","\
		"\"recv_mono\":22.000000022,"\
		"\"sent_real\":\"1970-09-27T00:15:27+0000\","\
		"\"sent_mono\":21.000000021,"\
		"\"tag\":\"tag\","\
		"\"msg\":\"message\""\
	"}\n"));

	char data2[] = "∫\0\x01\x02\x03\x14\x15\x16\"\\\b\f\n\r\t🤔";
	assert(sizeof data == sizeof data2);
	strncpy(ent->msg, data2, ent->len);
	assert(log_format_json(buf, sizeof(buf), ent, &out_len, "∫", "\x01\x02\x03\x14\x15\x16\"\\\b\f\n\r\t🤔") == buf);
	assert(strlen(buf) == out_len);
	assert(!strcmp(buf, "{"\
		"\"priority\":\"debug\","\
		"\"pid\":17,"\
		"\"tid\":18,"\
		"\"recv_real\":\"1970-11-24T11:10:31+0000\","\
		"\"recv_mono\":22.000000022,"\
		"\"sent_real\":\"1970-09-27T00:15:27+0000\","\
		"\"sent_mono\":21.000000021,"\
		"\"tag\":\"∫\","\
		"\"msg\":\"\\u0001\\u0002\\u0003\\u0014\\u0015\\u0016\\\"\\\\\\b\\f\\n\\r\\t🤔\""\
	"}\n"));

	free(ent);
}

void check_isatty_detection_testcase(bool enable_colors)
{
	struct logger_entry_with_msg entry;
	standard_test_msg(&entry, "TTY_DETECTION\0test", 19);

	fake_write = true;

	entry.header.default_send_ts_type = LOGGER_ENTRY_MONOTONIC;
	entry.header.priority = DLOG_FATAL;
	assert(log_print_log_line((struct log_format){
			.format = FORMAT_TAG,
			.color = enable_colors,
		}, 0, true, &entry.header) > 0);
	if (isatty_ret)
		assert(!strcmp(buffer, "\033[35;1mF/TTY_DETECTION: \033[0mtest\n"));
	else
		assert(!strcmp(buffer, "F/TTY_DETECTION: test\n"));

	fake_write = false;
}

void check_isatty_detection()
{
	isatty_ret = 1;
	check_isatty_detection_testcase(true);
	check_isatty_detection_testcase(false);
	isatty_ret = 0;
	check_isatty_detection_testcase(true);
	check_isatty_detection_testcase(false);
}

void check_syscall_failure_handling()
{
	log_filter *const filter = log_filter_new();
	assert(filter);
	assert(!log_add_filter_tid(filter, 123));
	assert(!log_add_filter_pid(filter, 456));
	assert(!log_add_filter_rule(filter, "FOO:E"));

	log_filter *const clone = log_filter_from_filter(filter);
	assert(clone);
	log_filter_free(clone);

	fail_alloc = true;
	assert(!log_filter_new());
	assert(!log_filter_from_filter(filter));
	assert(-ENOMEM == log_add_filter_tid(filter, 123));
	assert(-ENOMEM == log_add_filter_pid(filter, 456));
	fail_alloc = false;

	// Not really a syscall, but eh
	fail_list_add = true;
	assert(-1 == log_add_filter_rule(filter, "FOO:E"));
	fail_list_add = false;

	fail_strdup = true;
	assert(!log_filter_from_filter(filter));
	assert(-1 == log_add_filter_string(filter, "FOO:E"));
	assert(-1 == log_add_filter_rule(filter, "FOO:E"));
	fail_strdup = false;

	fail_strndup = true; // N
	assert(-1 == log_add_filter_string(filter, "FOO:E"));
	fail_strndup = false;

	/* the array size is an arbitrary value between 512, i.e. the size of the default buffer,
	 * and 4096, i.e. LOG_MAX_PAYLOAD_SIZE (higher should also work but would be pointless). */
	char msg[666] = {
		[0 ... 41] = 'A', // tag
		[42] = '\0',
		[43 ... sizeof msg - 2] = 'a',  // msg
		[sizeof msg - 1] = '\0'
	};
	struct logger_entry_with_msg entry;
	standard_test_msg(&entry, msg, sizeof msg);

	struct log_format format = {
			.format = FORMAT_TAG,
			.color = false,
	};

	fail_alloc = true;
	assert(-1 == log_print_log_line(format, 0, false, &entry.header));
	fail_alloc = false;

	fail_localtime = 1;
	assert(-1 == log_print_log_line(format, 0, false, &entry.header));
	fail_localtime = 2;
	assert(-1 == log_print_log_line(format, 0, false, &entry.header));

	partial_write = -1;
	assert(0 == log_print_log_line(format, 0, false, &entry.header));

	partial_write = 17;
	assert(17 == log_print_log_line(format, 0, false, &entry.header));

	partial_write = 0;
	log_filter_free(filter);
}

void check_invalid_input()
{
	const struct logger_entry entry = { .len = 1 };
	assert(!log_buffer_get_message(&entry));
	assert(!log_buffer_get_tag(&entry));

	log_filter *const filter = log_filter_new();
	assert(filter);
	assert(-1 == log_add_filter_rule(filter, ":D")); // >mfw tagless rule gets processed correctly
	log_filter_free(filter);
}

void check_log_format_from_string()
{
	static const struct {
		char const *string;
		log_print_format format;
	} expected_formats[] = {
		{"brief"          , FORMAT_BRIEF     },
		{"  brief"        , FORMAT_OFF       },
		{"tag"            , FORMAT_TAG       },
		{"tagging up turf", FORMAT_OFF       },
		{"process"        , FORMAT_PROCESS   },
		{"long"           , FORMAT_LONG      },
		{"thien long"     , FORMAT_OFF       },
		{"threadtime"     , FORMAT_THREADTIME},
		{"kerneltime"     , FORMAT_KERNELTIME},
		{"hammertime"     , FORMAT_OFF       },
	};

	for (size_t i = 0; i < NELEMS(expected_formats); ++i)
		assert(log_format_from_string(expected_formats[i].string) == expected_formats[i].format);
}

void check_get_format_sorting()
{
	static const struct {
		log_print_format format;
		enum sorting_order sorting;
	} expected_sorting[] = {
		{FORMAT_TAG          , SORT_DEFAULT  },
		{FORMAT_PROCESS      , SORT_DEFAULT  },
		{FORMAT_THREAD       , SORT_DEFAULT  },
		{FORMAT_RAW          , SORT_DEFAULT  },
		{FORMAT_BRIEF        , SORT_DEFAULT  },
		{FORMAT_TIME         , SORT_SENT_REAL},
		{FORMAT_THREADTIME   , SORT_SENT_REAL},
		{FORMAT_KERNELTIME   , SORT_SENT_MONO},
		{FORMAT_LONG         , SORT_SENT_REAL},
		{FORMAT_RWTIME       , SORT_RECV_REAL},
		{FORMAT_RECV_REALTIME, SORT_RECV_REAL},
	};

	for (size_t i = 0; i < NELEMS(expected_sorting); ++i) {
		assert(get_format_sorting(expected_sorting[i].format) == expected_sorting[i].sorting);
	}
}

void check_filter_pri_to_char()
{
	static const struct {
		log_priority pri;
		char c;
	} convertibles[] = {
		{DLOG_VERBOSE    , 'V' },
		{DLOG_DEBUG      , 'D' },
		{DLOG_INFO       , 'I' },
		{DLOG_WARN       , 'W' },
		{DLOG_ERROR      , 'E' },
		{DLOG_FATAL      , 'F' },
		{DLOG_SILENT     , 'S' },
	}, expected_pris[] = {
		{DLOG_UNKNOWN    , '0' },
		{DLOG_DEFAULT    , '1' },
		{DLOG_VERBOSE    , '2' },
		{DLOG_DEBUG      , '3' },
		{DLOG_INFO       , '4' },
		{DLOG_WARN       , '5' },
		{DLOG_ERROR      , '6' },
		{DLOG_FATAL      , '7' },
		{DLOG_VERBOSE    , '8' }, // makes little sense but ok
		{DLOG_VERBOSE    , '9' }, // ditto

		{DLOG_VERBOSE    , 'v' },
		{DLOG_DEBUG      , 'd' },
		{DLOG_INFO       , 'i' },
		{DLOG_WARN       , 'w' },
		{DLOG_ERROR      , 'e' },
		{DLOG_FATAL      , 'f' },
		{DLOG_SILENT     , 's' },
		{DLOG_DEFAULT    , '*' },

		{DLOG_UNKNOWN    , 'l' },
		{DLOG_UNKNOWN    , 'L' },
		{DLOG_UNKNOWN    , ' ' },
		{DLOG_UNKNOWN    , '\a'},
		{DLOG_UNKNOWN    , '\0'},
	}, expected_chars[] = {
		{DLOG_DEFAULT    , '?' },
		{DLOG_UNKNOWN    , '?' },
		{(log_priority)42, '?' },
		{(log_priority)-9, '?' },
	};

	for (size_t i = 0; i < NELEMS(convertibles); ++i) {
		assert(filter_pri_to_char(convertibles[i].pri) == convertibles[i].c);
		assert(filter_char_to_pri(convertibles[i].c) == convertibles[i].pri);
	}

	for (size_t i = 0; i < NELEMS(expected_pris); ++i)
		assert(filter_char_to_pri(expected_pris[i].c) == expected_pris[i].pri);

	for (size_t i = 0; i < NELEMS(expected_chars); ++i)
		assert(filter_pri_to_char(expected_chars[i].pri) == expected_chars[i].c);
}

int main()
{
	check_log_format_from_string();
	check_get_format_sorting();
	check_filter_pri_to_char();
	check_log_print_log_line();
	check_invalid_input();
	check_syscall_failure_handling();
	check_isatty_detection();
	check_json();
}
