#include <stdbool.h>

#include <queued_entry.h>
#include <log_file.h>
#include <dlog-internal.h>
#include <sort_vector.h>
#include <fd_info.h>

bool process_log(const struct logger_entry *e, struct timespec reference_ts, enum sorting_order stamp_type, write_callback callback, void *userdata, long timeout, int *callback_ret);
struct fd_info *find_earliest_log(struct fd_info **data_fds, int fd_count, enum sorting_order sort_by);
int put_logs_into_vector(struct fd_info **data_fds, int fd_count, int nfds, struct sort_vector *logs, write_callback callback, void *userdata, log_filter *filter);
int validate_buffers(int *enabled_buffers);

int written_logs = 0;
bool free_entries = false;
int callback_ret = 0;
int callback(const struct logger_entry *entry, void *userdata)
{
	assert(userdata == (void *)0xDDD);

	if (free_entries)
		free((void *)entry); // We shouldn't do this due to const, but there's no easy way around
	++written_logs;
	return callback_ret;
}

int __wrap_clock_gettime(clockid_t clk_id, struct timespec *tp)
{
	assert(clk_id == CLOCK_MONOTONIC);

	tp->tv_sec  = 1;
	tp->tv_nsec = 900000000;
	return 0;
}

bool __wrap_log_should_print_line(log_filter *p_filter, const struct logger_entry *entry)
{
	return true;
}

struct fdi_internal
{
	bool has_log;
	const struct logger_entry *ent;
};

bool test_has_log(struct fd_info *fdi)
{
	struct fdi_internal *ii = fdi->priv_data;
	return ii->has_log;
}

const struct logger_entry *test_peek_entry(const struct fd_info *fdi)
{
	struct fdi_internal *ii = fdi->priv_data;
	return ii->ent;
}

int fdi_push_log_ret = 0;
int __wrap_fdi_push_log(struct fd_info *fdi, struct sort_vector *logs, write_callback callback, void *userdata, log_filter *filter)
{
	assert(filter == (log_filter *)0xEEE);
	if (fdi_push_log_ret != 0)
		return fdi_push_log_ret;
	struct logger_entry *ent = calloc(1, sizeof(struct logger_entry));
	assert(ent);
	memcpy(ent, fdi->ops->peek_entry(fdi), sizeof(struct logger_entry));
	struct fdi_internal *ii = fdi->priv_data;
	ii->has_log = false;
	return sort_vector_push(logs, ent, callback, userdata);
}

int main()
{
	struct logger_entry ent = {
		.sec_sent_mono = 0,
		.nsec_sent_mono = 900000000,
		.sec_sent_real = 0,
		.nsec_sent_real = 0,
		.sec_recv_mono = 1,
		.nsec_recv_mono = 0,
		.sec_recv_real = 0,
		.nsec_recv_real = 999999999,
	};

	struct timespec ts = {
		.tv_sec = 1,
		.tv_nsec = 100000,
	};

	int callback_local_ret;
	assert(process_log(&ent, ts, SORT_SENT_MONO, callback,
		(void *)0xDDD, 1, &callback_local_ret));
	assert(written_logs == 1);
	assert(callback_local_ret == 0);
	assert(process_log(&ent, ts, SORT_SENT_REAL, callback,
		(void *)0xDDD, 1, &callback_local_ret));
	assert(written_logs == 2);
	assert(callback_local_ret == 0);
	assert(!process_log(&ent, ts, SORT_RECV_MONO, callback,
		(void *)0xDDD, 1, &callback_local_ret));
	assert(written_logs == 2);
	assert(callback_local_ret == 0);
	assert(!process_log(&ent, ts, SORT_RECV_REAL, callback,
		(void *)0xDDD, 1, &callback_local_ret));
	assert(written_logs == 2);
	assert(callback_local_ret == 0);

	callback_ret = 77;
	assert(!process_log(&ent, ts, SORT_RECV_MONO, callback,
		(void *)0xDDD, 1, &callback_local_ret));
	assert(written_logs == 2);
	assert(callback_local_ret == 0);
	assert(process_log(&ent, ts, SORT_SENT_MONO, callback,
		(void *)0xDDD, 1, &callback_local_ret));
	assert(written_logs == 3);
	assert(callback_local_ret == 77);
	callback_ret = 0;

	struct sort_vector sv;
	sort_vector_init(&sv);
	sv.sort_by = SORT_SENT_MONO;
	sv.timeout = 60;
	sv.dump = DUMP_CONTINUOUS;
	sv.size = DEFAULT_SORT_BUFFER_SIZE;
	sort_vector_finalize(&sv);
	assert(sort_vector_time_span(&sv) == 0);

	sort_vector_flush(&sv, callback, (void *)0xDDD, false);
	assert(written_logs == 3 && sort_vector_used_size(&sv) == 0);

	for (int i = 0; i < 10; ++i) {
		struct logger_entry *ent2 = calloc(1, sizeof(struct logger_entry));
		assert(ent2);
		ent2->sec_sent_mono = 1;
		ent2->nsec_sent_mono = (79 + i) * 10000000;
		sort_vector_push(&sv, ent2, callback, (void *)0xDDD);
	}

	assert(sort_vector_time_span(&sv) == 110);

	sort_vector_flush(&sv, callback, (struct log_file *)0xDDD, false);
	assert(written_logs == 9 && sort_vector_used_size(&sv) == 4);

	struct logger_entry *ent3 = calloc(1, sizeof(struct logger_entry));
	assert(ent3);
	ent3->sec_sent_mono = 0;
	ent3->nsec_sent_mono = 999999999;
	sort_vector_push(&sv, ent3, callback, (void *)0xDDD);
	assert(sort_vector_time_span(&sv) == 900);

	struct fd_ops test_ops = {
		.has_log = test_has_log,
		.peek_entry = test_peek_entry,
	};

	struct fd_info fd_stor[10] = {};
	struct fdi_internal ii_stor[10] = {};
	struct logger_entry ent_stor[10 - 3] = {};
	for (int i = 0; i < 10; ++i) {
		if (i < 3)
			ii_stor[i].has_log = false;
		else {
			ent_stor[i - 3].sec_sent_mono = i % 2;
			ent_stor[i - 3].nsec_sent_mono = i;

			ii_stor[i].has_log = true;
			ii_stor[i].ent = &ent_stor[i - 3];
		}

		int t = i * 3 % 10;
		fd_stor[i].ops = &test_ops;
		fd_stor[i].priv_data = &ii_stor[t];
	}

	struct fd_info *fds[10] = {};
	for (int i = 0; i < 10; ++i) {
		fds[i] = &fd_stor[i];
	}
	assert(!find_earliest_log(fds, 0, SORT_SENT_MONO));
	assert(find_earliest_log(fds, 10, SORT_SENT_MONO) == fds[8]);

	while (!sort_vector_empty(&sv))
		sort_vector_pop(&sv);

	struct logger_entry *ent_future = calloc(1, sizeof(struct logger_entry));
	assert(ent_future);
	ent_future->sec_sent_mono = 2;
	ent_future->nsec_sent_mono = 0;
	sort_vector_push(&sv, ent_future, callback, (void *)0xDDD);
	written_logs = 0;
	sort_vector_flush(&sv, callback, (void *)0xDDD, false);
	assert(written_logs == 0);
	sort_vector_flush(&sv, callback, (void *)0xDDD, true);
	assert(written_logs == 1);
	assert(sort_vector_used_size(&sv) == 0);

	assert(put_logs_into_vector(fds, 0, 1, &sv, callback, (void *)0xDDD, (log_filter *)0xEEE) == 0);

	fdi_push_log_ret = -EIO;
	assert(put_logs_into_vector(fds, 10, 1, &sv, callback, (void *)0xDDD, (log_filter *)0xEEE) == -EIO);

	fdi_push_log_ret = 0;
	assert(put_logs_into_vector(fds, 10, 1, &sv, callback, (void *)0xDDD, (log_filter *)0xEEE) == 0);
	assert(sort_vector_used_size(&sv) == 1);

	assert(put_logs_into_vector(fds, 10, 0, &sv, callback, (void *)0xDDD, (log_filter *)0xEEE) == 0);
	assert(sort_vector_used_size(&sv) == 7);

	int buf = 0;
	assert(validate_buffers(&buf) == 0);
	assert(buf == ((1 << LOG_ID_MAIN) | (1 << LOG_ID_SYSTEM) | (1 << LOG_ID_APPS)));
	assert(validate_buffers(&buf) == 0);

	buf = 1 << LOG_ID_MAIN;
	assert(validate_buffers(&buf) == 0);
	assert(buf == (1 << LOG_ID_MAIN));

	buf |= 1 << LOG_ID_KMSG;
	assert(validate_buffers(&buf) == -EINVAL);

	sort_vector_free(&sv);
}
