// C
#include <assert.h>
#include <stddef.h>

// POSIX
#include <time.h>

// DLog
#include <logcommon.h>
#include <logconfig.h>
#include <queued_entry.h>
#include <queued_entry_timestamp.h>

int __wrap_clock_gettime(clockid_t clk_id, struct timespec *tp)
{
	assert(clk_id == CLOCK_MONOTONIC || clk_id == CLOCK_REALTIME);

	if (clk_id == CLOCK_MONOTONIC) {
		tp->tv_sec  = 21;
		tp->tv_nsec = 43;
	} else {
		tp->tv_sec  = 65;
		tp->tv_nsec = 87;
	}
	return 0;
}

int main()
{
	assert(get_proper_clock(SORT_SENT_MONO) == CLOCK_MONOTONIC);
	assert(get_proper_clock(SORT_SENT_REAL) == CLOCK_REALTIME);
	assert(get_proper_clock(SORT_RECV_MONO) == CLOCK_MONOTONIC);
	assert(get_proper_clock(SORT_RECV_REAL) == CLOCK_REALTIME);

	struct log_config conf = {NULL, NULL};
	const struct {
		const char *const str;
		enum sorting_order order;
	} ordering[] = {
		{ "sent_real", SORT_SENT_REAL },
		{ "sent_mono", SORT_SENT_MONO },
		{ "recv_real", SORT_RECV_REAL },
		{ "recv_mono", SORT_RECV_MONO },
	};
	for (size_t i = 0; i < NELEMS(ordering); ++i) {
		log_config_set(&conf, "sort_by", ordering[i].str);
		assert(get_order_from_config(&conf) == ordering[i].order);
	}

	struct timespec ts;
	struct logger_entry le = {
		 .sec_sent_mono = 1,
		.nsec_sent_mono = 2,
		 .sec_sent_real = 3,
		.nsec_sent_real = 4,
		 .sec_recv_mono = 5,
		.nsec_recv_mono = 6,
		 .sec_recv_real = 7,
		.nsec_recv_real = 8,
	};
	extract_timestamp(&le, SORT_SENT_MONO, &ts);
	assert(ts.tv_sec  == le. sec_sent_mono);
	assert(ts.tv_nsec == le.nsec_sent_mono);
	extract_timestamp(&le, SORT_RECV_REAL, &ts);
	assert(ts.tv_sec  == le. sec_recv_real);
	assert(ts.tv_nsec == le.nsec_recv_real);
	extract_timestamp(&le, SORT_SENT_REAL, &ts);
	assert(ts.tv_sec  == le. sec_sent_real);
	assert(ts.tv_nsec == le.nsec_sent_real);
	extract_timestamp(&le, SORT_RECV_MONO, &ts);
	assert(ts.tv_sec  == le. sec_recv_mono);
	assert(ts.tv_nsec == le.nsec_recv_mono);

	copy_recv_timestamp(&le);
	assert(le. sec_sent_mono == 1);
	assert(le.nsec_sent_mono == 2);
	assert(le. sec_sent_real == 3);
	assert(le.nsec_sent_real == 4);
	assert(le. sec_recv_mono == 1);
	assert(le.nsec_recv_mono == 2);
	assert(le. sec_recv_real == 3);
	assert(le.nsec_recv_real == 4);

	add_recv_timestamp(&le);
	assert(le. sec_sent_mono == 1);
	assert(le.nsec_sent_mono == 2);
	assert(le. sec_sent_real == 3);
	assert(le.nsec_sent_real == 4);
	assert(le. sec_recv_mono == 21);
	assert(le.nsec_recv_mono == 43);
	assert(le. sec_recv_real == 65);
	assert(le.nsec_recv_real == 87);

	char buffer[999];
	struct android_logger_entry *const ale = (struct android_logger_entry *)buffer;
	ale->sec_sent = 5;
	ale->nsec_sent = 8;
	ale->pid = 13;
	ale->tid = 21;
	assert(snprintf(ale->msg, sizeof buffer - offsetof(typeof(*ale), msg), "%c%s%c%s%c"
		, (char)DLOG_ERROR
		, "TAG", '\0'
		, "MSG", '\0'
	) > 0);

	struct logger_entry_with_msg lewm;
	parse_androidlogger_message(ale, &lewm.header, 9 + sizeof *ale);
	assert(lewm.header.len == sizeof lewm.header + 8); // the missing 1 is priority which went to the header
	assert(lewm.header.priority == DLOG_ERROR);
	assert(lewm.header.pid == 13);
	assert(lewm.header.tid == 21);
	assert(lewm.header .sec_sent_real == 5);
	assert(lewm.header.nsec_sent_mono == -1);
	assert(lewm.header.nsec_sent_real == 8);
	assert(lewm.header.tag_len == 3);
	assert(lewm.header.default_send_ts_type == LOGGER_ENTRY_REALTIME);
	assert(!strcmp("MSG", lewm.msg + lewm.header.tag_len + 1));

#ifdef USE_ANDROID_MONOTONIC
	struct android_logger_footer alf = {
		.magic = ANDROID_LOGGER_FOOTER_MAGIC,
		.sec_sent_mono = 4,
		.nsec_sent_mono = 7,
	};
	memcpy(buffer + 9 + sizeof *ale, &alf, sizeof alf);

	parse_androidlogger_message(ale, &lewm.header, 9 + sizeof *ale + sizeof alf);
	assert(lewm.header.len == sizeof lewm.header + 8);
	assert(lewm.header.priority == DLOG_ERROR);
	assert(lewm.header.pid == 13);
	assert(lewm.header.tid == 21);
	assert(lewm.header .sec_sent_mono == 4);
	assert(lewm.header .sec_sent_real == 5);
	assert(lewm.header.nsec_sent_mono == 7);
	assert(lewm.header.nsec_sent_real == 8);
	assert(lewm.header.tag_len == 3);
	assert(lewm.header.default_send_ts_type == LOGGER_ENTRY_REALTIME);
	assert(!strcmp("MSG", lewm.msg + lewm.header.tag_len + 1));

	// The case in which there is enough space for the footer, but the magic does not match
	const char *longer_message = "You've reached the website for Arch Linux, a lightweight and flexible Linux(R) distribution that tries to Keep It Simple.";
	int added_len;
	assert((added_len = snprintf(ale->msg, sizeof buffer - offsetof(typeof(*ale), msg), "%c%s%c%s%c"
		, (char)DLOG_ERROR
		, "TAG", '\0'
		, longer_message, '\0'
	)) > 0);

	parse_androidlogger_message(ale, &lewm.header, added_len + sizeof *ale);

	assert(lewm.header.len == sizeof lewm.header + added_len - 1);
	assert(lewm.header.priority == DLOG_ERROR);
	assert(lewm.header.pid == 13);
	assert(lewm.header.tid == 21);
	assert(lewm.header .sec_sent_real == 5);
	assert(lewm.header.nsec_sent_mono == -1);
	assert(lewm.header.nsec_sent_real == 8);
	assert(lewm.header.tag_len == 3);
	assert(lewm.header.default_send_ts_type == LOGGER_ENTRY_REALTIME);
	assert(!strcmp(longer_message, lewm.msg + lewm.header.tag_len + 1));

#endif /* USE_ANDROID_MONOTONIC */
}
