// C
#include <assert.h>

// DLog
#include <log_file.h>
#include <sort_vector.h>

size_t total_flushed;
int callback_ret = 0;
int callback(const struct logger_entry *e, void *userdata)
{
	assert(userdata == (void *)0xDA7A);
	total_flushed += (size_t) e;
	return callback_ret;
}

bool __wrap_log_entry_is_earlier(const enum sorting_order sort_by, const struct logger_entry *lhs, const struct logger_entry *rhs)
{
	return lhs < rhs;
}

int __wrap_log_config_get_int(const struct log_config *config, const char *key, int default_val)
{
	assert(config == (struct log_config *) 0xC0B0L);
	assert(key);
	assert(!strncmp(key, "util_sorting_", sizeof "util_sorting" - 1));

	return default_val;
}

static bool use_real_free;
static size_t free_total;
void __real_free(void *ptr);
void __wrap_free(void *ptr)
{
	if (use_real_free) // lcov might want to allocate stuff after main()
		__real_free(ptr);
	else
		free_total += (size_t) ptr;
}

int main()
{
	struct sort_vector sv;
	sort_vector_init(&sv);

	sort_vector_apply_config(&sv, (struct log_config *) 0xC0B0L);

	// size == 0 and no dumping shouldn't allocate any buffer at all
	sv.size = 0;
	sv.dump = 0;
	sort_vector_finalize(&sv);
	assert(!sv.data);
	assert(!sv.size);

	/* the buffer should be just large enough to fit the dumpees,
	 * i.e. #dumpees + 1 extra slot for operations, regardless of
	 * how large the requested size was */
	sv.size = 99;
	sv.dump = 6;
	sort_vector_finalize(&sv);
	assert(sv.size == 7);
	assert(sv.data);

	free(sv.data);
	sv.data = NULL;
	sv.size = 3;
	sv.dump = 6;
	sort_vector_finalize(&sv);
	assert(sv.size == 7);
	assert(sv.data);

	// sorting correctness checks
	sv.dump = 0;
	struct log_file lf;
	memset(&lf, 0, sizeof(struct log_file));

#define ADD(x) assert(sort_vector_push(&sv, (struct logger_entry *) x, callback, (void *)0xDA7A) == callback_ret)
#define POP(x) \
	free_total = 0; \
	sort_vector_pop(&sv); \
	assert(free_total == x)

	/* ┌──┬──┬──┬──┬──┬──┬──┐
	 * │  │  │  │  │  │  │  │
	 * └──┴──┴──┴──┴──┴──┴──┘ */
	assert(sort_vector_empty(&sv));
	assert(!sort_vector_full(&sv));
	assert(NULL == sort_vector_back(&sv));
	assert(0 == sort_vector_used_size(&sv));

	ADD(17);
	ADD(33);
	ADD(12);
	ADD(40);
	ADD(20);
	/* ┌──┬──┬──┬──┬──┬──┬──┐
	 * │12│17│20│33│40│  │  │
	 * └──┴──┴──┴──┴──┴──┴──┘ */
	assert(!sort_vector_empty(&sv));
	assert(!sort_vector_full(&sv));
	assert((struct logger_entry *) 12 == sort_vector_back(&sv));
	assert(5 == sort_vector_used_size(&sv));

	POP(12);
	ADD(25);
	ADD(7);
	/* ┌──┬──┬──┬──┬──┬──┬──┐
	 * │  │ 7│17│20│25│33│40│
	 * └──┴──┴──┴──┴──┴──┴──┘ */
	assert(!sort_vector_empty(&sv));
	assert(sort_vector_full(&sv)); // the empty slot is reserved for operations
	assert((struct logger_entry *) 7 == sort_vector_back(&sv));
	assert(6 == sort_vector_used_size(&sv));

	ADD(24);
	assert(7 == total_flushed);
	total_flushed = 0;
	/* ┌──┬──┬──┬──┬──┬──┬──┐
	 * │40│  │17│20│24│25│33│
	 * └──┴──┴──┴──┴──┴──┴──┘ */
	assert(!sort_vector_empty(&sv));
	assert(sort_vector_full(&sv));
	assert((struct logger_entry *) 17 == sort_vector_back(&sv));
	assert(6 == sort_vector_used_size(&sv));

	POP(17);
	ADD(18);
	ADD(41);
	assert(18 == total_flushed);
	total_flushed = 0;
	/* ┌──┬──┬──┬──┬──┬──┬──┐
	 * │33│40│41│  │20│24│25│
	 * └──┴──┴──┴──┴──┴──┴──┘ */
	assert(!sort_vector_empty(&sv));
	assert(sort_vector_full(&sv));
	assert((struct logger_entry *) 20 == sort_vector_back(&sv));
	assert(6 == sort_vector_used_size(&sv));

	callback_ret = 13;
	ADD(27);
	assert(20 == total_flushed);
	total_flushed = 0;
	/* ┌──┬──┬──┬──┬──┬──┬──┐
	 * │33│40│41│  │20│24│25│
	 * └──┴──┴──┴──┴──┴──┴──┘ */
	assert(!sort_vector_empty(&sv));
	assert(sort_vector_full(&sv));
	assert((struct logger_entry *) 20 == sort_vector_back(&sv));
	assert(6 == sort_vector_used_size(&sv));
	callback_ret = 0;

#undef ADD
#undef POP

	assert(sv.data[0] == (struct logger_entry *) 33);
	assert(sv.data[1] == (struct logger_entry *) 40);
	assert(sv.data[2] == (struct logger_entry *) 41);
	assert(sv.data[3] == (struct logger_entry *)  0);
	assert(sv.data[4] == (struct logger_entry *) 20);
	assert(sv.data[5] == (struct logger_entry *) 24);
	assert(sv.data[6] == (struct logger_entry *) 25);
	assert(sv.begin == 4);
	assert(sv.end == 3);

	struct logger_entry **const data = sv.data;
	free_total = 0;
	sort_vector_free(&sv);
	assert(total_flushed == 0);
	assert(free_total == (size_t) data + 33 + 40 + 41 + 20 + 24 + 25);
	__real_free(data);

	use_real_free = true;
}
