// C
#include <assert.h>
#include <stdlib.h>

// POSIX
#include <poll.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/un.h>

// DLog
#include <backend_androidlogger.h>
#include <logcommon.h>
#include <logconfig.h>
#include <logpipe.h>

struct pipe_msg_header {
	struct msghdr hdr;
	struct iovec iov;
	char ctl[CMSG_SPACE(sizeof(int))];
};

static struct pipe_msg_header msg_cache;
static struct dlog_control_msg dcm_cache;
static char large_buffer[1024];
static size_t large_buffer_size;

static bool fail_sendmsg;
ssize_t __wrap_sendmsg(int sockfd, const struct msghdr *msg, int flags)
{
	assert(!msg->msg_name);
	assert(!msg->msg_namelen);
	assert(msg->msg_iovlen == 1);
	assert(msg->msg_iov->iov_len == sizeof dcm_cache);
	assert(msg->msg_controllen == sizeof msg_cache.ctl);

	if (fail_sendmsg) {
		errno = 52;
		return -1;
	}

	memcpy(&dcm_cache, msg->msg_iov->iov_base, msg->msg_iov->iov_len);
	memcpy(msg_cache.ctl, msg->msg_control, sizeof msg_cache.ctl);

	return 0;
}

static ssize_t recvmsg_ret;
static bool poison_recvmsg_result;
static bool poison_recvmsg_pipefd;
ssize_t __wrap_recvmsg(int sockfd, struct msghdr *msg, int flags)
{
	assert(msg->msg_iovlen == 1);
	assert(msg->msg_iov->iov_len == sizeof dcm_cache);
	assert(msg->msg_controllen == sizeof msg_cache.ctl);

	if (recvmsg_ret <= 0) {
		if (recvmsg_ret < 0)
			errno = 322;
		return recvmsg_ret;
	}

	msg->msg_name = NULL;
	msg->msg_namelen = 0;
	memcpy(msg->msg_iov->iov_base, &dcm_cache, sizeof dcm_cache);
	memcpy(msg->msg_control, msg_cache.ctl, sizeof msg_cache.ctl);

	if (poison_recvmsg_result)
		((struct dlog_control_msg *)msg->msg_iov->iov_base)->result = 42;
	if (poison_recvmsg_pipefd)
		((struct cmsghdr *)msg->msg_control)->cmsg_type = -1;

	return recvmsg_ret;
}

static bool fail_poll;
int __wrap_poll(struct pollfd *fds, nfds_t nfds, int timeout)
{
	if (fail_poll) {
		errno = 997;
		return -1;
	}

	return 0;
}

static int fail_read;
static int partial_read;
static int poison_datalen;
static bool poison_read_result;
ssize_t __wrap_read(int fd, void *buf, size_t count)
{
	if (fail_read & 1) {
		errno = 666;
		fail_read >>= 1;
		partial_read >>= 1;
		return -1;
	}
	fail_read >>= 1;

	if (partial_read & 1) {
		partial_read >>= 1;
		return count - 1;
	}
	partial_read >>= 1;

	if (count > large_buffer_size)
		count = large_buffer_size;

	memcpy(buf, large_buffer, count);
	memmove(large_buffer, large_buffer + count, large_buffer_size - count);
	large_buffer_size -= count;

	if (poison_datalen) {
		((struct dlog_control_msg *)buf)->length -= poison_datalen;
		poison_datalen = 0;
	}

	if (poison_read_result)
		((struct dlog_control_msg *)buf)->result = 42;

	return count;
}

static bool fail_writev;
ssize_t __wrap_writev(int fd, const struct iovec *iov, int iovcnt)
{
	if (fail_writev) {
		errno = 54;
		return -1;
	}

	large_buffer_size = 0U;
	for (int i = 0; i < iovcnt; ++i) {
		memcpy(large_buffer + large_buffer_size, iov[i].iov_base, iov[i].iov_len);
		large_buffer_size += iov[i].iov_len;
	}
	return large_buffer_size;
}

static int fcntl_ret;
int __wrap_fcntl(int fd, int cmd, ...)
{
	errno = EBADF;
	return fcntl_ret;
}

static bool fail_alloc;
extern void *__real_malloc(size_t size);
void *__wrap_malloc(size_t size)
{
	return fail_alloc ? NULL : __real_malloc(size);
}

extern void *__real_calloc(size_t nmemb, size_t size);
void *__wrap_calloc(size_t nmemb, size_t size)
{
	return fail_alloc ? NULL : __real_calloc(nmemb, size);
}

static bool fail_connect;
int __wrap_connect(int sockfd, const struct sockaddr *addr, socklen_t addrlen)
{
	if (fail_connect) {
		errno = 1776;
		return -1;
	}

	return 0;
}

static bool fail_socket;
int __wrap_socket(int domain, int type, int protocol)
{
	if (fail_socket) {
		errno = 1410;
		return -1;
	}

	return 1683;
}

static bool check_open_flags;
int __real_open(const char *pathname, int flags, int mode);
int __wrap_open(const char *pathname, int flags, int mode)
{
	if (check_open_flags)
		assert(flags == 0xFECAL);

	if (!strcmp(pathname, "valid23"))
		return 23;
	if (!strcmp(pathname, "valid17"))
		return 17;
	if (!strcmp(pathname, "nonexistent")) {
		errno = ENOENT;
		return -1;
	}
	if (!strcmp(pathname, "inaccessible")) {
		errno = EPERM;
		return -1;
	}

	return __real_open(pathname, flags, mode);
}

static int ioctl_errno;
int __wrap_ioctl(int d, int request, char *argp)
{
	if (!ioctl_errno)
		return 0;

	errno = ioctl_errno;
	return -1;
}

int main(int argc, char **argv)
{
	{
		const char *strings[] = {"we", "require", "more", "vespene gas"};
		char *result = create_glued_string(strings, NELEMS(strings));
		assert(result);
		assert(!strncmp(result, "we require more vespene gas", sizeof "we require more vespene gas"));
		free(result);
	}

	{
		const char *strings[] = {"you", "must", "construct", "additional", "pylons", "!"};
		char *result = create_glued_string(strings + 3, 2);
		assert(result);
		assert(!strncmp(result, "additional pylons", sizeof "additional pylons"));
		free(result);
	}

	{
		const log_id_t invalid_ids[] = {
			LOG_ID_INVALID - 7,
			LOG_ID_INVALID,
			LOG_ID_MAX,
			LOG_ID_MAX + 7,
		};
		const char *const invalid_names[] = {
			"foo",
			"mAiN",
			"radio los santos",
			"events",
		};
		const char *const valid_names[LOG_ID_MAX] = {
			"main",
			"system",
			"radio",
			"apps",
			"kmsg",
			"syslog",
		};

		for (size_t i = 0; i < NELEMS(invalid_ids); ++i)
			assert(!strcmp(log_name_by_id(invalid_ids[i]), ""));

		for (size_t i = 0; i < NELEMS(invalid_names); ++i)
			assert(LOG_ID_INVALID == log_id_by_name(invalid_names[i]));

		for (size_t i = 0; i < NELEMS(valid_names); ++i)
			assert(!strcmp(valid_names[i], log_name_by_id(log_id_by_name(valid_names[i]))));

		for (int i = 0; i < LOG_ID_MAX; ++i)
			assert((log_id_t)i == log_id_by_name(log_name_by_id((log_id_t)i)));
	}

	{
		struct {
			bool result;
			const char *ending;
		} const test_cases[] = {
			{true, ""},
			{true, "man"},
			{false, "bat"},
			{true, "batman"},
			{false, "robin"},
			{false, "batman batman"},
		};

		for (size_t i = 0; i < NELEMS(test_cases); ++i)
			assert(test_cases[i].result == str_ends_with("batman", test_cases[i].ending));
	}

	assert(log_priority_from_char(0) == DLOG_UNKNOWN);
	assert(log_priority_from_char(1) == DLOG_DEFAULT);
	assert(log_priority_from_char(2) == DLOG_VERBOSE);
	assert(log_priority_from_char(3) == DLOG_DEBUG);
	assert(log_priority_from_char(4) == DLOG_INFO);
	assert(log_priority_from_char(5) == DLOG_WARN);
	assert(log_priority_from_char(6) == DLOG_ERROR);
	assert(log_priority_from_char(7) == DLOG_FATAL);
	assert(log_priority_from_char(8) == DLOG_SILENT);

	assert(log_priority_from_char(-120) == DLOG_UNKNOWN);
	assert(log_priority_from_char(-1)   == DLOG_UNKNOWN);
	assert(log_priority_from_char(9)    == DLOG_UNKNOWN);
	assert(log_priority_from_char(75)   == DLOG_UNKNOWN);

	fail_alloc = true;
	const char *const strings[] = {"storm", "earth", "fire"};
	assert(!create_glued_string(strings, NELEMS(strings)));
	fail_alloc = false;

	fail_sendmsg = true;
	assert(send_pipe(13, 37) == -52);
	fail_sendmsg = false;

	assert(!send_pipe(13, 37));

	fail_poll = true;
	assert(recv_pipe(13, 15) == -997);
	fail_poll = false;
	assert(recv_pipe(13, 15) == -EWOULDBLOCK);

	recvmsg_ret = -1;
	assert(recv_pipe(13, 0) == -322);
	recvmsg_ret = 0;
	assert(recv_pipe(13, 0) == -EIO);
	recvmsg_ret = 1;

	poison_recvmsg_result = true;
	assert(recv_pipe(13, 0) == -EIO);
	poison_recvmsg_result = false;
	poison_recvmsg_pipefd = true;
	assert(recv_pipe(13, 0) == -EIO);
	poison_recvmsg_pipefd = false;

	fcntl_ret = -1;
	assert(recv_pipe(13, 0) == -EBADF);
	fcntl_ret = 0;
	assert(recv_pipe(13, 0) == 37);

	char data[] = "abcdefgh";

	fail_writev = true;
	assert(send_dlog_request(5, 8, data, 6) == -54);
	fail_writev = false;
	assert(!send_dlog_reply(5, 8, DLOG_REQ_RESULT_OK, data, 6));

	void *recv_data;
	int recv_data_len;

	fail_read = 1;
	assert(recv_dlog_reply(5, 8, &recv_data, &recv_data_len) == -666);
	partial_read = 1;
	assert(recv_dlog_reply(5, 8, &recv_data, &recv_data_len) == -EINVAL);
	poison_datalen = 8;
	assert(recv_dlog_reply(5, 8, &recv_data, &recv_data_len) == -EINVAL);

	poison_read_result = true;
	assert(recv_dlog_reply(5, 8, &recv_data, &recv_data_len) == -EINVAL);
	poison_read_result = false;

	poison_datalen = 6;
	assert(!send_dlog_reply(5, 8, DLOG_REQ_RESULT_OK, data, 6));
	assert(!recv_dlog_reply(5, 8, &recv_data, &recv_data_len));
	assert(large_buffer_size == 6);
	assert(!recv_data);
	assert(!recv_data_len);

	fail_read = 2;
	assert(!send_dlog_reply(5, 8, DLOG_REQ_RESULT_OK, data, 6));
	assert(recv_dlog_reply(5, 8, &recv_data, &recv_data_len) == -666);
	assert(!recv_data);

	partial_read = 2;
	assert(!send_dlog_reply(5, 8, DLOG_REQ_RESULT_OK, data, 6));
	assert(recv_dlog_reply(5, 8, &recv_data, &recv_data_len) == -EINVAL);
	assert(!recv_data);

	fail_alloc = true;
	assert(!send_dlog_reply(5, 8, DLOG_REQ_RESULT_OK, data, 6));
	assert(recv_dlog_reply(5, 8, &recv_data, &recv_data_len) == -ENOMEM);
	assert(!recv_data);
	fail_alloc = false;

	assert(!send_dlog_reply(5, 8, DLOG_REQ_RESULT_OK, data, 6));
	memcpy(data, "qwertyui", 8);
	assert(!recv_dlog_reply(5, 8, &recv_data, &recv_data_len));
	assert(!strncmp((char *)recv_data, "abcdef", 6));

#define MAX_LEN sizeof((struct sockaddr_un *)0)->sun_path
	char sockname[MAX_LEN + 1] = {[0 ... MAX_LEN - 1] = 'x'};
	assert(connect_sock(sockname) == -ENAMETOOLONG);
	sockname[MAX_LEN - 1] = '\0';
#undef MAX_LEN

	fail_socket = true;
	assert(connect_sock(sockname) == -1410);

	fail_connect = true;
	fail_socket = false;
	assert(connect_sock(sockname) == -EBADF);

	fail_connect = false;
	assert(connect_sock(sockname) == 1683);

	/* The `logger_open_buffer` family of functions doesn't actually live in `logcommon.h`,
	 * the reason it is tested here is because it used to at some point in the past. */
	int fd;
	struct log_config conf = {NULL, NULL};
	check_open_flags = true;
	assert(logger_open_buffer_from_config(LOG_ID_MAIN, &conf, 12345, &fd) == -ENOENT);

	log_config_set(&conf, "main", "nonexistent");
	assert(logger_open_buffer_from_config(LOG_ID_MAIN, &conf, 0xFECAL, &fd) == -ENOENT);

	log_config_set(&conf, "main", "nonexistent:");
	assert(!logger_open_buffer_from_config(LOG_ID_MAIN, &conf, 0xFECAL, &fd));

	log_config_set(&conf, "main", "valid17:valid23");
	assert(logger_open_buffer_from_config(LOG_ID_MAIN, &conf, 0xFECAL, &fd) == 1);
	assert(fd == 17);
	log_config_set(&conf, "main", "valid23:valid17");
	assert(logger_open_buffer_from_config(LOG_ID_MAIN, &conf, 0xFECAL, &fd) == 1);
	assert(fd == 23);

	char ret_path[32];
	assert(logger_open_buffer_get_path(LOG_ID_MAIN, "valid17", 0xFECAL, &fd, ret_path, sizeof ret_path) == 1);
	assert(!strcmp(ret_path, "valid17"));

	ioctl_errno = EINVAL;
	assert(logger_open_buffer_get_path(LOG_ID_MAIN, "valid17", 0xFECAL, &fd, ret_path, sizeof ret_path) == -ENOTSUP);
	ioctl_errno = EINTR;
	assert(logger_open_buffer_get_path(LOG_ID_MAIN, "valid17", 0xFECAL, &fd, ret_path, sizeof ret_path) == -EINTR);
	ioctl_errno = 0;

#define CHECK_RESULT_INACCESSIBLE(str) \
	assert(logger_open_buffer(LOG_ID_MAIN, str, 0xFECAL, &fd) == -EPERM)

#define CHECK_RESULT_VALID(str) \
	assert(logger_open_buffer(LOG_ID_MAIN, str, 0xFECAL, &fd) == 1); \
	assert(fd == 17); \
	fd = 42

	CHECK_RESULT_INACCESSIBLE("inaccessible");
	CHECK_RESULT_INACCESSIBLE("inaccessible:");
	CHECK_RESULT_INACCESSIBLE("inaccessible:valid17");
	CHECK_RESULT_INACCESSIBLE("inaccessible:valid17:");
	CHECK_RESULT_INACCESSIBLE("inaccessible:valid17:nonexistent");
	CHECK_RESULT_INACCESSIBLE("inaccessible:nonexistent");
	CHECK_RESULT_INACCESSIBLE("inaccessible:nonexistent:");
	CHECK_RESULT_INACCESSIBLE("inaccessible:nonexistent:valid17");
	CHECK_RESULT_INACCESSIBLE("nonexistent:inaccessible");
	CHECK_RESULT_INACCESSIBLE("nonexistent:inaccessible:");
	CHECK_RESULT_INACCESSIBLE("nonexistent:inaccessible:valid17");
	assert(logger_open_buffer(LOG_ID_MAIN, "nonexistent", 0xFECAL, &fd) == -ENOENT);
	assert(logger_open_buffer(LOG_ID_MAIN, "nonexistent:", 0xFECAL, &fd) == 0);
	assert(logger_open_buffer(LOG_ID_MAIN, "nonexistent:nonexistent", 0xFECAL, &fd) == -ENOENT);
	CHECK_RESULT_VALID("nonexistent:valid17");
	CHECK_RESULT_VALID("nonexistent:valid17:");
	CHECK_RESULT_VALID("nonexistent:valid17:inaccessible");
	CHECK_RESULT_VALID("valid17");
	CHECK_RESULT_VALID("valid17:");
	CHECK_RESULT_VALID("valid17:inaccessible");
	CHECK_RESULT_VALID("valid17:inaccessible:");
	CHECK_RESULT_VALID("valid17:inaccessible:nonexistent");
	CHECK_RESULT_VALID("valid17:nonexistent");
	CHECK_RESULT_VALID("valid17:nonexistent:");
	CHECK_RESULT_VALID("valid17:nonexistent:inaccessible");

	check_open_flags = false; // gcov wants to open stuff after program completion
}
