/*
 * Copyright (c) 2018, Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *		http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include "../src/logger/log_storage.h"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <limits.h>

#define STR1 "a"

#define LE_SM(str, tsec, tnsec) {{sizeof(struct logger_entry)+sizeof(str)+1, 0, 0, 1, 1, (tsec), (tnsec), 0U, 0U, 0U, 0U, 0U, 0U, 0U, LOGGER_ENTRY_MONOTONIC}, str}
#define LE_SR(str, tsec, tnsec) {{sizeof(struct logger_entry)+sizeof(str)+1, 0, 0, 1, 1, 0U, 0U, (tsec), (tnsec), 0U, 0U, 0U, 0U, 0U, LOGGER_ENTRY_REALTIME}, str}
#define LE_RM(str, tsec, tnsec) {{sizeof(struct logger_entry)+sizeof(str)+1, 0, 0, 1, 1, 0U, 0U, 0U, 0U, (tsec), (tnsec), 0U, 0U, 0U, LOGGER_ENTRY_MONOTONIC}, str}
#define LE_RR(str, tsec, tnsec) {{sizeof(struct logger_entry)+sizeof(str)+1, 0, 0, 1, 1, 0U, 0U, 0U, 0U, 0U, 0U, (tsec), (tnsec), 0U, LOGGER_ENTRY_REALTIME}, str}

static bool fail_malloc = false;
void *__real_malloc(size_t size);
void *__wrap_malloc(size_t size)
{
	return fail_malloc ? NULL : __real_malloc(size);
}

void dont_call_this_callback(const struct logger_entry *le, void *user_data)
{
	assert(false);
}

int callback_count = 0;

void callback_check_msg(const struct logger_entry *le, void *user_data)
{
	assert(le);
	assert(user_data);

	const char *s = user_data;
	assert(strcmp(le->msg, s) == 0);
	callback_count++;
}

int check_sec = 0;

void callback_check_sec(const struct logger_entry *le, void *user_data)
{
	assert(le);

	assert(le->sec_sent_mono == check_sec);
	callback_count++;
}

void callback_count_inc(const struct logger_entry *le, void *user_data)
{
	assert(le);

	callback_count++;
}

void test_add_logs(log_storage *s1, unsigned cnt, unsigned sec_base, bool inc, unsigned nsec, bool set_check)
{
	assert(s1);

	int incr = inc ? 1 : -1;
	for (int i = 0; i < cnt; i++) {
		// identify each log by seconds field
		unsigned sec_to_set = sec_base+incr*(i+1);
		if (set_check)
			check_sec = sec_to_set;
		switch (log_storage_get_sorting_order(s1)) {
		case SORT_RECV_REAL: {
			struct logger_entry_with_msg e1 = LE_RR(STR1, sec_to_set, nsec);
			assert(log_storage_add_new_entry(s1, &e1.header));
			break;
		}
		case SORT_RECV_MONO: {
			struct logger_entry_with_msg e1 = LE_RM(STR1, sec_to_set, nsec);
			assert(log_storage_add_new_entry(s1, &e1.header));
			break;
		}
		case SORT_SENT_REAL: {
			struct logger_entry_with_msg e1 = LE_SR(STR1, sec_to_set, nsec);
			assert(log_storage_add_new_entry(s1, &e1.header));
			break;
		}
		default: {
			struct logger_entry_with_msg e1 = LE_SM(STR1, sec_to_set, nsec);
			assert(log_storage_add_new_entry(s1, &e1.header));
			break;
		}
		}
	}
}

void test_reader_get_new_entry(log_storage_reader *r, unsigned sec_to_check)
{
	assert(r);

	const struct log_storage *s = log_storage_reader_get_storage(r);
	const struct logger_entry *le = log_storage_reader_get_new_entry(r);
	assert(le);
	assert(s);
	int32_t entry_sec;
	switch (log_storage_get_sorting_order(s)) {
	case SORT_RECV_REAL:
		entry_sec = le->sec_recv_real;
		break;
	case SORT_RECV_MONO:
		entry_sec = le->sec_recv_mono;
		break;
	case SORT_SENT_REAL:
		entry_sec = le->sec_sent_real;
		break;
	default:
		entry_sec = le->sec_sent_mono;
		break;
	}

	assert(entry_sec == sec_to_check);
}

void test_reader_get_new_entries(log_storage_reader *r, unsigned cnt, unsigned check_base)
{
	assert(r);

	for (int i = 0; i < cnt; i++)
		test_reader_get_new_entry(r, check_base+i+1);
}

int main()
{
	/************* TEST 1 ***************/
	/* Small storage, calling callbacks on removing log entries */
	// create storage so small that no log is stored,
	{
		const int CNT = 1000;

		log_storage *s1 = log_storage_create(1, SORT_SENT_MONO);
		assert(s1);
		assert(log_storage_get_capacity(s1) == 1);

		// add entry - it should be released immediately
		struct logger_entry_with_msg e1 = LE_SM(STR1, 1000U, 1000U);
		assert(log_storage_add_new_entry(s1, &e1.header));

		// add reader - no callback should be called
		log_storage_reader *r1 = log_storage_new_reader(s1, false, dont_call_this_callback, NULL);
		assert(r1);
		// check there is no log entry stored
		assert(!log_storage_reader_is_new_entry_available(r1));
		assert(NULL == log_storage_reader_get_new_entry(r1));

		// remove the reader
		log_storage_release_reader(r1);

		// add reader with callback - when a log entry is added, a callback should be called
		r1 = log_storage_new_reader(s1, false, callback_check_msg, STR1);
		assert(r1);
		for (int i = 0; i < CNT; i++) {
			assert(log_storage_add_new_entry(s1, &e1.header));
			// check that each time log was added, the callback was called
			assert(callback_count == i+1);
		}

		// add second reader - now there should be twice as much calls to the callback
		log_storage_reader *r2 = log_storage_new_reader(s1, false, callback_check_msg, STR1);
		assert(r2);
		callback_count = 0;
		for (int i = 0; i < CNT; i++) {
			assert(log_storage_add_new_entry(s1, &e1.header));
			assert(callback_count == 2*(i+1));
		}

		assert(log_storage_reader_get_new_entry(r1) == NULL);
		assert(log_storage_reader_get_new_entry(r2) == NULL);
		log_storage_clear(s1);
		assert(log_storage_reader_get_new_entry(r1) == NULL);
		assert(log_storage_reader_get_new_entry(r2) == NULL);

		// cleanup
		log_storage_release_reader(r1);
		log_storage_release_reader(r2);
		log_storage_free(s1);
	}

	/************* TEST 2 ***************/
	/* Basic test checking normal operation - add entries, read entries */
	{
		log_storage *s1 = log_storage_create(1000000, SORT_SENT_MONO);
		assert(s1);
		callback_count = 0;

		log_storage_reader *r1 = log_storage_new_reader(s1, false, dont_call_this_callback, NULL);
		log_storage_reader *r2 = log_storage_new_reader(s1, false, dont_call_this_callback, NULL);
		assert(r1);
		assert(r2);

		const int CNT = 1000;
		// add entries
		test_add_logs(s1, CNT, 0, true, 0, false);

		// read by both readers, alternately
		for (int i = 0; i < CNT; i++) {
			test_reader_get_new_entry(r1, i+1);
			test_reader_get_new_entry(r2, i+1);
		}
		assert(log_storage_reader_get_new_entry(r1) == NULL);
		assert(log_storage_reader_get_new_entry(r2) == NULL);

		log_storage_release_reader(r1);
		log_storage_release_reader(r2);

		// create two new readers
		r1 = log_storage_new_reader(s1, false, dont_call_this_callback, NULL);
		r2 = log_storage_new_reader(s1, false, dont_call_this_callback, NULL);
		assert(r1);
		assert(r2);

		// read logs by both readers, sequentially
		test_reader_get_new_entries(r1, CNT, 0);
		test_reader_get_new_entries(r2, CNT, 0);

		assert(log_storage_reader_get_new_entry(r1) == NULL);
		assert(log_storage_reader_get_new_entry(r2) == NULL);

		assert(callback_count == 0);

		// cleanup
		log_storage_release_reader(r1);
		log_storage_release_reader(r2);
		log_storage_free(s1);
	}

	/************* TEST 3 ***************/
	/* Basic test checking normal operation - add entries, read entries */
	/* Additionally check sorting */
	{
		const int CNT = 1000;
		callback_count = 0;

		for (int i = 0; i < SORT_MAX; i++) {
			log_storage *s = log_storage_create(1000000, i);
			assert(s);
			log_storage_reader *r = log_storage_new_reader(s, false, dont_call_this_callback, NULL);
			assert(r);

			// put logs in reverse order
			test_add_logs(s, CNT, CNT+1, false, 0, false);

			// check if the storage sorted the logs
			test_reader_get_new_entries(r, CNT, 0);

			assert(log_storage_reader_get_new_entry(r) == NULL);

			assert(callback_count == 0);

			log_storage_release_reader(r);
			log_storage_free(s);
		}
	}
	/************* TEST 4 ***************/
	/* Basic test checking normal operation - add entries, read entries */
	/* Additionally check calling callbacks while sorting */
	{
		const int CNT = 1000;
		callback_count = 0;
		log_storage *s1 = log_storage_create(1000000, SORT_SENT_MONO);
		assert(s1);
		log_storage_reader *r1 = log_storage_new_reader(s1, false, callback_check_sec, NULL);
		log_storage_reader *r2 = log_storage_new_reader(s1, false, dont_call_this_callback, NULL);
		assert(r1);
		assert(r2);

		// put logs in reverse order
		test_add_logs(s1, CNT, 2*CNT+1, false, 0, false);

		// check if the storage sorted the logs
		test_reader_get_new_entries(r1, CNT, CNT);

		// put more logs in reverse order
		test_add_logs(s1, CNT, CNT+1, false, 0, true);

		assert(log_storage_reader_get_new_entry(r1) == NULL);

		// check the second reader
		test_reader_get_new_entries(r2, 2*CNT, 0);

		assert(log_storage_reader_get_new_entry(r2) == NULL);

		assert(callback_count == CNT);

		log_storage_release_reader(r1);
		log_storage_release_reader(r2);
		log_storage_free(s1);
	}
	/************* TEST 5 ***************/
	/* Check if after clearing it still works */
	{
		const int CNT = 1000;
		callback_count = 0;
		log_storage *s1 = log_storage_create(1000000, SORT_SENT_MONO);
		assert(s1);
		log_storage_reader *r1 = log_storage_new_reader(s1, false, callback_count_inc, NULL);
		log_storage_reader *r2 = log_storage_new_reader(s1, false, callback_count_inc, NULL);
		log_storage_reader *r3 = log_storage_new_reader(s1, false, dont_call_this_callback, NULL);
		assert(r1);
		assert(r2);
		assert(r3);

		// add logs
		test_add_logs(s1, 2*CNT, 0, true, 0, false);

		// keep reader r1 at the start of the logs

		// get reader r2 halfway through
		test_reader_get_new_entries(r2, CNT, 0);

		// get reader r3 to the end
		test_reader_get_new_entries(r3, 2*CNT, 0);

		// clear the storage
		log_storage_clear(s1);

		assert(callback_count == 3*CNT);

		assert(log_storage_reader_get_new_entry(r1) == NULL);
		assert(log_storage_reader_get_new_entry(r2) == NULL);
		assert(log_storage_reader_get_new_entry(r3) == NULL);

		// add logs
		test_add_logs(s1, CNT, 0, true, 0, false);

		// read by all the readers, alternately
		for (int i = 0; i < CNT; i++) {
			test_reader_get_new_entry(r1, i+1);
			test_reader_get_new_entry(r2, i+1);
			test_reader_get_new_entry(r3, i+1);
		}

		assert(log_storage_reader_get_new_entry(r1) == NULL);
		assert(log_storage_reader_get_new_entry(r2) == NULL);
		assert(log_storage_reader_get_new_entry(r3) == NULL);

		log_storage_release_reader(r1);
		log_storage_release_reader(r2);
		log_storage_release_reader(r3);
		log_storage_free(s1);
	}
	/************* TEST 6 ***************/
	/* Dumping/one-shot reader */
	{
		const int CNT = 1000;
		callback_count = 0;
		log_storage *s1 = log_storage_create(1000000, SORT_SENT_MONO);
		assert(s1);
		log_storage_reader *r1 = log_storage_new_reader(s1, true, dont_call_this_callback, NULL);
		assert(r1);

		assert(!log_storage_reader_is_new_entry_available(r1));
		assert(log_storage_reader_get_new_entry(r1) == NULL);

		log_storage_release_reader(r1);

		// add logs
		test_add_logs(s1, CNT, CNT, true, 0, false);

		r1 = log_storage_new_reader(s1, true, dont_call_this_callback, NULL);
		assert(r1);

		test_reader_get_new_entries(r1, CNT, CNT);

		assert(!log_storage_reader_is_new_entry_available(r1));
		assert(log_storage_reader_get_new_entry(r1) == NULL);
		log_storage_release_reader(r1);

		// add more logs
		test_add_logs(s1, CNT, 2*CNT, true, 0, false);

		// so, now there are 2*CNT logs in the storage
		r1 = log_storage_new_reader(s1, true, callback_check_sec, NULL);
		assert(r1);

		// add even more logs
		test_add_logs(s1, CNT, 3*CNT, true, 0, false);

		// now, there are 2*CNT logs for this reader, and 1*CNT later logs
		test_reader_get_new_entries(r1, CNT, CNT);

		// the reader has read half of "his" logs

		// add even even more logs, but earlier
		// the logs should be passed to the reader by the callback
		test_add_logs(s1, CNT, 0, true, 0, true);

		// read the remaining logs
		test_reader_get_new_entries(r1, CNT, 2*CNT);

		assert(!log_storage_reader_is_new_entry_available(r1));
		assert(log_storage_reader_get_new_entry(r1) == NULL);
		assert(callback_count == CNT);

		// reader has read all his logs
		// but adding earlier logs should still give him those logs, with callback
		test_add_logs(s1, CNT, 0, true, 1, true);

		assert(!log_storage_reader_is_new_entry_available(r1));
		assert(log_storage_reader_get_new_entry(r1) == NULL);

		// adding later logs should not give him those logs
		test_add_logs(s1, CNT, 4*CNT, true, 1, true);

		assert(!log_storage_reader_is_new_entry_available(r1));
		assert(log_storage_reader_get_new_entry(r1) == NULL);
		log_storage_release_reader(r1);

		// there should be no reader actions while adding further logs, without any readers
		check_sec = 0;
		test_add_logs(s1, CNT, 0, true, 2, false);
		test_add_logs(s1, CNT, 5*CNT, true, 1, false);

		log_storage_free(s1);
	}
	/************* TEST 6 ***************/
	/* Test basic bookkeeping */
	{
		log_storage *const s1 = log_storage_create(10000, SORT_SENT_MONO);
		log_storage_reader *const r1 = log_storage_new_reader(s1, false, callback_count_inc, NULL);

		struct logger_entry_with_msg le1 = LE_RR(STR1, 123, 456);
		assert(le1.header.len * 0 == log_storage_reader_get_ready_bytes(r1));
		assert(log_storage_add_new_entry(s1, &le1.header));
		assert(le1.header.len * 1 == log_storage_reader_get_ready_bytes(r1));
		assert(log_storage_add_new_entry(s1, &le1.header));
		assert(le1.header.len * 2 == log_storage_reader_get_ready_bytes(r1));

		callback_count = 0;
		log_storage_free(s1);
		assert(callback_count == 2);
		assert(!log_storage_reader_is_new_entry_available(r1));
		assert(0 == log_storage_reader_get_ready_bytes(r1));

		log_storage_release_reader(r1);
	}
	/************* TEST 7 ***************/
	/* Test syscall failure handling */
	{
		fail_malloc = true;
		log_storage *s1 = log_storage_create(10000, SORT_SENT_MONO);
		assert(!s1);
		fail_malloc = false;

		s1 = log_storage_create(10000, SORT_SENT_MONO);
		assert(s1);

		log_storage_reader *const r1 = log_storage_new_reader(s1, false, callback_count_inc, NULL);

		struct logger_entry_with_msg le1 = LE_RR(STR1, 123, 456);
		fail_malloc = true;
		assert(!log_storage_add_new_entry(s1, &le1.header));
		fail_malloc = false;

		callback_count = 0;
		log_storage_free(s1);
		assert(callback_count == 0);

		log_storage_release_reader(r1);
	}

	return 0;
}
