#define LOG_TAG "XXX"

// C
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// POSIX
#include <fcntl.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>

// DLOG
#include <dlog.h>
#include <dynamic_config.h>
#include <logcommon.h>
#include <loglimiter.h>

#define DYNAMIC_CONF "/tmp/dlog-filters/" DYNAMIC_CONFIG_FILENAME
#define DYNAMIC_CONF_ALTER DYNAMIC_CONF "_ALTER"
#define BUF_LEN (MAX_CONF_KEY_LEN + MAX_CONF_VAL_LEN + 3) // +3 because KEY=VAL\n\0

#define FILTER_ALLOW (__LOG_LIMITER_LIMIT_MAX + 1)
#define FILTER_DENY 0

// limit values same as in loglimiter
int set_config(const char *filename, int flag, char prio, const char *tag, int limit)
{
	assert(tag);
	assert(flag == O_APPEND || flag == O_TRUNC);

	char buf[BUF_LEN];
	const int flags = O_CREAT | O_WRONLY | flag;

	const int fd = open(filename, flags, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP);
	if (fd < 0) {
		int ret = -errno;
		perror("error opening dynamic conf file");
		return ret;
	}

	int r;
	if (limit == FILTER_ALLOW)
		r = snprintf(buf, sizeof buf, "limiter|%s|%c=allow\n", tag, prio);
	else if (limit == FILTER_DENY)
		r = snprintf(buf, sizeof buf, "limiter|%s|%c=deny\n", tag, prio);
	else
		r = snprintf(buf, sizeof buf, "limiter|%s|%c=%d\n", tag, prio, limit);
	if (r < 0) {
		r = -errno;
		perror("snprintf error");
		goto finish;
	}

	const ssize_t s = write(fd, buf, r);
	if (s < 0) {
		r = -errno;
		perror("error writing config file");
	} else
		r = 0;

finish:
	assert(!close(fd));
	return r;
}

int erase_file(const char *filename)
{
	const int fd = open(filename, O_RDWR|O_TRUNC);
	if (fd < 0) {
		int ret = -errno;
		perror("error opening dynamic conf file");
		return ret;
	}

	assert(!close(fd));

	return 0;
}

int main(int argc, char **argv)
{
	// test allow all filter
	if (set_config(DYNAMIC_CONF, O_TRUNC, '*', "*", FILTER_ALLOW) < 0)
		return EXIT_FAILURE;
	for (int i = 0; i < 3; i++)
		RLOGE("[1] ALLOWED message %d\n", i);

	// test deny all filter
	if (set_config(DYNAMIC_CONF, O_TRUNC, '*', "*", FILTER_DENY) < 0)
		return EXIT_FAILURE;
	for (int i = 0; i < 3; i++)
		RLOGE("[2] DENIED message %d\n", i);

	// still deny all, but allow by tag filter
	if (set_config(DYNAMIC_CONF, O_APPEND, '*', "YYY", FILTER_ALLOW) < 0)
		return EXIT_FAILURE;
	if (set_config(DYNAMIC_CONF_ALTER, O_TRUNC, '*', "ZZZ", FILTER_ALLOW) < 0)
		return EXIT_FAILURE;

	for (int i = 0; i < 3; i++)
		RLOG(LOG_ERROR, "YYY", "[3] ALLOWED message %d\n", i);
	for (int i = 0; i < 3; i++)
		RLOG(LOG_ERROR, "ZZZ", "[4] DENIED message %d\n", i);
	for (int i = 0; i < 3; i++)
		RLOGE("[5] DENIED message %d\n", i);


	// test deny all but allow by priority filter
	if (set_config(DYNAMIC_CONF, O_TRUNC,  '*', "*", FILTER_DENY)  < 0
	||  set_config(DYNAMIC_CONF, O_APPEND, 'W', "*", FILTER_ALLOW) < 0)
		return EXIT_FAILURE;
	for (int i = 0; i < 3; i++)
		RLOG(LOG_WARN, "ZZZ", "[6] ALLOWED message %d\n", i);
	for (int i = 0; i < 3; i++)
		RLOG(LOG_DEBUG, "ZZZ", "[7] DENIED message %d\n", i);

	// test time based limiter
	static const int TIMED_LIMIT = 2;
	if (set_config(DYNAMIC_CONF, O_TRUNC, '*', "*", TIMED_LIMIT) < 0)
		return EXIT_FAILURE;
	for (int i = 0; i < 10; i++)
		RLOG(LOG_DEBUG, "AAA", "[8] %s message %d/%d\n", i < TIMED_LIMIT ? "ALLOWED" : "DENIED", i + 1, TIMED_LIMIT);

	if (erase_file(DYNAMIC_CONF_ALTER) < 0)
		return EXIT_FAILURE;
	if (erase_file(DYNAMIC_CONF) < 0)
		return EXIT_FAILURE;
	return EXIT_SUCCESS;
}
